;============================================================================         
;
;       E V A L U A T E    N U M E R I C A L   E X P R E S S I O N S
;
;============================================================================
;       This module was originally M07
;============================================================================
        .XLIST
        INCLUDE PREAMBLE
        .LIST

PUBLIC   ARC
PUBLIC   ENDSP
PUBLIC   ERROR_24
PUBLIC   ERROR_51
PUBLIC   ERROR_74
PUBLIC   ERROR_75
PUBLIC   EVAL
PUBLIC   EXPR
PUBLIC   FINDFNC
PUBLIC   FNCEND
PUBLIC   FNCERR
PUBLIC   FNCMAX
PUBLIC   FNCMIN
PUBLIC   FNCMOD
PUBLIC   FNCRD
PUBLIC   FNCSPC
PUBLIC   FNCVER
PUBLIC   FNC_PI
PUBLIC   FN_
PUBLIC   FUNCFN
PUBLIC   LEN
PUBLIC   NUM
PUBLIC   NUMBER
PUBLIC   POS
PUBLIC   SNMBR
PUBLIC   SNUMBER
PUBLIC   SPACEK
PUBLIC   SPACEKA
PUBLIC   SPACES
PUBLIC   SPACESK
PUBLIC   TERM
PUBLIC   TERM98
PUBLIC   VAL
PUBLIC   VAR
PUBLIC   VARA
PUBLIC   VARR

EXTRN    _ABCOV  :NEAR
EXTRN    ADDVAR  :NEAR
EXTRN    ALPHA   :NEAR
EXTRN    ARAYEL  :NEAR
EXTRN    ARRAY   :NEAR
EXTRN    AVAR    :NEAR
EXTRN    AVARD   :NEAR
EXTRN    BINFLT  :NEAR
EXTRN    BINFLTA :NEAR
EXTRN    BNFLTXXX:NEAR
EXTRN    CCJ     :NEAR
EXTRN    CCJB    :NEAR
EXTRN    CPUID   :NEAR
EXTRN    _CSOS   :NEAR
EXTRN    CHARCTR :NEAR
EXTRN    ERROR   :NEAR
EXTRN    ERROR_02:NEAR
EXTRN    ERROR_04:NEAR
EXTRN    ERROR_10:NEAR
EXTRN    ERROR_11:NEAR
EXTRN    ERROR_12:NEAR
EXTRN    ERROR_13:NEAR
EXTRN    ERROR_16:NEAR
EXTRN    ERROR_19:NEAR
EXTRN    ERROR_29:NEAR
EXTRN    ERROR_34:NEAR
EXTRN    ERR_02  :NEAR
EXTRN    ERR_01  :NEAR
EXTRN    FINDVAR :NEAR
EXTRN    FLTBINA :NEAR
EXTRN    GETTERM_:NEAR
EXTRN    IMMEDERR:NEAR
EXTRN    INPER_75:NEAR
EXTRN    BOOT_ID :NEAR
EXTRN    LETRDIGT:NEAR
EXTRN    MATHADD :NEAR
EXTRN    MATHDIV :NEAR
EXTRN    MATHERR :NEAR
EXTRN    MATHMAX :NEAR
EXTRN    MATHMIN :NEAR
EXTRN    MATHMOD :NEAR
EXTRN    MATHMUL :NEAR
EXTRN    MATHRD  :NEAR
EXTRN    MATHSUB :NEAR
EXTRN    MATHX2Y :NEAR
EXTRN    NUMCH   :NEAR
EXTRN    RELOP   :NEAR
EXTRN    SCNCHSS :NEAR
EXTRN    SCRATCH :NEAR
EXTRN    SQZTEXT :NEAR
EXTRN    SYSGENFG:BYTE
EXTRN    TRACEVAR:NEAR
EXTRN    TRACEVRA:NEAR
EXTRN    VARNAME :NEAR
EXTRN    _FN_LST :NEAR
EXTRN    _MPI    :NEAR

EXTRN    @@OSEND     :DWORD
EXTRN    @@VSEND     :DWORD
EXTRN    CPBASE      :DWORD
EXTRN    ENDMEMX     :DWORD
EXTRN    ENDPART     :DWORD
EXTRN    PART_       :BYTE
EXTRN    BASE        :DWORD
EXTRN    TIMESLICE   :NEAR

;------------------------------------------------------------------------------;
;                                                                              ;
;      ************************************                                    ;
;      *                                  *                                    ;
;      *  MODULE-07: EXPRESSION ANALYSIS  *                                    ;
;      *                                  *                                    ;
;      ************************************                                    ;
;                                                                              ;
;      DESCRIPTION: EXPRESSION ANALYSIS IS PERFORMED BY                        ;
;          'EXPR' AND 'TERM' ROUTINES.  WHEN AN OPERAND                        ;
;          IS ENCOUNTERED, ITS VALUE IS STORED IN THE                          ;
;          VALUE STACK (VS). AS EACH OPERATOR IS ENCOUNT-                      ;
;          ERED, IT IS COMPARED WITH THE LAST OPERATOR                         ;
;          IN THE OPERATOR STACK. IF THE LAST OPERATOR IN                      ;
;          THE STACK IS OF HIGHER PRIORITY THAN THE                            ;
;          CURRENT OPERATOR, IT IS REMOVED FROM OS AND                         ;
;          THE SPECIFIED OPERATION IS PERFORMED ON THE                         ;
;          LAST 2 VALUES IN THE VALUE STACK.  THE RESULT                       ;
;          REPLACES THE 2 VALUES IN THE VS. STACK                              ;
;          OPERATOR EXECUTION IS CONTINUED UNTIL AN                            ;
;          OPERATOR WITH LOWER PRIORITY THAN THE CURRENT                       ;
;          OPERATOR IS FOUND; AT THIS POINT THE CURRENT                        ;
;          OPERATOR IS ADDED TO THE OS AND THE SCAN OF                         ;
;          THE EXPRESSION CONTINUES.  NOTE, IF THE                             ;
;          EXPRESSION BEGINS WITH A LEADING MINUS (-),                         ;
;          THE SIGN OF NEXT VALUE IS COMPLEMENTED AFTER                        ;
;          IT IS EVALUATED.                                                    ;
;                                                                              ;
;          EXPRESSION RECURSION: BASIC EXPRESSIONS ARE                         ;
;               RECURSIVE ('EXPR' AND 'TERM' CALL EACH                         ;
;               OTHER). RECURSION IS HANDLED BY A                              ;
;               96-LEVEL HARDWARE SUBROUTINE STACK AND A                       ;
;               64-LEVEL OPERATOR STACK. 96 LEVELS IS                          ;
;               SUFFICIENT TO HANDLE EXPRESSIONS SINCE                         ;
;               THE OPERATOR STACK WILL FILL BEFORE THE                        ;
;               SUBROUTINE STACK WRAPS AROUND.                                 ;
;                                                                              ;
;------------------------------------------------------------------------------;
;                                                                              ;
;     OPERATORS (OS ATOMS) -- HIGH 4 BITS = OPERATOR                           ;
;                             PRECEDENCE                                       ;
;          A0 = UNARY (LEADING) +                                              ;
;          B0 = +          B2 = -          B3 = LEADING -                      ;
;          C0 = *          C5 = /                                              ;
;          D0 = ^                                                              ;
;------------------------------------------------------------------------------;


;     EXPR: SCAN FOR AND EVALUATE A NUMERIC EXPRESSION
;           IN TEXT.
;
;          ENTRY: TEXTPTR POINTS TO 1ST CHARACTER OF
;                 EXPRESSION

;          RETURN: RESULTANT VALUE (8-BYTE FLOATING PT.)
;                       IN VS
;                  TEXTPTR POINTS TO CHAR. AFTER EXPR.
;
;          MODIFIED REGISTERS: ALL,AUX0,1,2
;
;     This code optimizes single-character expression
;     processing.
;
;     Description: (execution phase)
;          1. 2nd character of expression is examined.
;
;          2a. If 2nd character is expression terminator
;              (CR, stmt-separator, 3A-3F, comma, or
;              right-parenthesis), it is a single
;              character expression. Goto 3.
;
;           b. If multi-character expression, normal
;              expression processing.
;
;          3. If 1st character is digit, B  NUMBERA.
;             If letter, B  VARB.
;             If atom,  normal expression processing.
;             Else, normal expression processing.
;-----------------------------------------------------------------------------;
;++
;   = 01 means  delimeters
;   = 02 means  letters
;   = 03 means  digits
;   = 05 means  variable name
;   = 06 means  numeric constant

;            0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
MY_TAB  DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,01H,00H,00H ;0X
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;1X
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,01H,0C0H,0B0H,00H,0B2H,00H,0C5H ;2X
        DB  03H,03H,03H,03H,03H,03H,03H,03H,03H,03H,01H,01H,01H,00H,01H,01H ;3X
        DB  00H,02H,02H,02H,02H,02H,02H,02H,02H,02H,02H,02H,02H,02H,02H,02H ;4X
        DB  02H,02H,02H,02H,02H,02H,02H,02H,02H,02H,02H,00H,00H,00H,0D0H,00H ;5X
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;6X
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,06H,05H ;7X
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;8X
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;9X
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;AX
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;BX
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;CX
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;DX
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H ;EX
        DB  00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,00H,01H,00H,00H ;FX
;            0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
EXPR:   TEST    SL,2
        JZ      EXPR1                   ;BRANCH IF NOT EXECUTION PHASE
        MOV     CH,BYTE PTR [TEXTPTR]   ;CH = IST CHARACTER OF EXPRESSION
        CMP     CH,7FH
        JNE     EXPR80P
        MOV     CL,BYTE PTR [TEXTPTR+7] ;CL = 2ND CHARACTER OF EXPRESSION
        JMP     EXPR82
EXPR80P:CMP     CH,7DH
        JNE     EXPR80
        MOV     CL,BYTE PTR [TEXTPTR+7] ;CL = 2ND CHARACTER OF EXPRESSION
        JMP     EXPR82
EXPR80: CMP     CH,7EH
        JNE     EXPR81
        MOV     CL,BYTE PTR [TEXTPTR+9] ;CL = 2ND CHARACTER OF EXPRESSION
        JMP     EXPR82
EXPR81: MOV     CL,BYTE PTR [TEXTPTR+1] ;CL = 2ND CHARACTER OF EXPRESSION
EXPR82: MOV     AL,CL
        MOV     EBX,OFFSET MY_TAB
        XLATB
        CMP     AL,1
        JE      EXPR90                  ;BRANCH IF DELIMITER
        MOV     AL,CH
        XLATB
        TEST    AL,20H
        JZ      EXPR83                  ;BRANCH IF '+' OR '-'
        INC     TEXTPTR                 ;INCREMENT TEXTPTR IF CHAR. '+' OR '-'

EXPR83: MOV     EDX,OSPTR
        INC     EDX
        CMP     EDX,DWORD PTR [@@OSEND]
        JE      ERROR_04                ;ERROR IF OS STACK FULL
        MOV     BYTE PTR [EDX],@UPLUS   ;OS ATOM = UNARY +
        MOV     OSPTR,EDX
        CMP     CH,2DH
        JNE     EXPR84                  ;IF NOT LEADING '-' THEN BRANCH
                                        ;                   ELSE STORE '-' IN OS
        INC     EDX
        CMP     EDX,DWORD PTR [@@OSEND]
        JE      ERROR_04                ;ERROR IF OS STACK FULL
        MOV     BYTE PTR [EDX],@UMINUS  ;OS ATOM = UNARY -
        MOV     OSPTR,EDX

EXPR84: CALL    TERM                    ;EVAL A TERM
        DEC     OSPTR                   ;REMOVE LAST OS ATOM
        RET

EXPR90: CMP     CH,7EH
        JNE     EXPR91                  ;BRANCH IF NOT NUMERIC CONSTANT
        SUB     VSPTR,8                 ;PUSH CONSTANT INTO VS STACK
        MOV     EAX,VSPTR
        MOV     EBX,DWORD PTR [TEXTPTR+1]
        MOV     DWORD PTR [EAX],EBX
        MOV     EBX,DWORD PTR [TEXTPTR+5]
        MOV     DWORD PTR [EAX+4],EBX
        ADD     TEXTPTR,9
        MOV     R0,0
        RET

EXPR91: CMP     CH,7FH
        JE      EXPR91P
        CMP     CH,7DH
        JNE     EXPR83                  ;ERROR IF NOT VARIABLE SYMBOL
EXPR91P:MOV     K,0
        JMP     VAR                     ;PROCESS NUMERIC SCALAR VARIABLE

EXPR1:  MOV     CH,BYTE PTR [TEXTPTR]   ;GET CHAR
        MOV     AL,CH
        MOV     EBX,OFFSET MY_TAB
        XLATB
        TEST    AL,20H
        JZ      EXPR2                   ;BRANCH IF NOT '+' OR '-'
        INC     TEXTPTR                 ;INCREMENT TEXTPTR IF CHAR. '+' OR '-'
EXPR2:  JMP     TERM

;     TERM: SCAN AND EVALUATE A TERM FROM TEXT.
;
;          ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;          RETURN: RESULT IN VS
;                  CH = LAST OPERATOR TAKEN FROM OS
;                       (UNARY +)
;                  PC'S = ADDRESS OF LAST OPERATOR IN OS
;
;          NOTE, THIS ROUTINE IS RECURSIVE -- CALLS
;                ITSELF AND 'EXPR'
;                ALSO, 'TERM' MUST BE LOCAL TO 'EXPR'


TERM:   CMP     BYTE PTR [TEXTPTR],28H  ;CHECK CHAR '('
        JNE     TERM1                   ;BRANCH IF NOT '('
        INC     TEXTPTR
        CALL    EXPR                    ;CHAR. = '(' EVAL EXPRESSION IN PARENTHESES
        CMP     BYTE PTR [TEXTPTR],')'
        JNE     ERROR_11                ;ERROR EXPECTED ')'
        INC     TEXTPTR

TERM10: MOV     CH,BYTE PTR [TEXTPTR]   ;GET CHAR
        MOV     AL,CH
        MOV     EBX,OFFSET MY_TAB
        XLATB
        TEST    AL,0F0H
        JZ      TERM12                  ;BRANCH IF NOT +, -, *, / , ^
        INC     TEXTPTR
        TEST    SL,2
        JZ      TERM6                   ;BRANCH IF NOT EXECUTION PHASE
        MOV     EDX,OSPTR

TERM8:  MOV     CH,BYTE PTR [EDX]       ;CH = LAST OS ATOM
        MOV     AH,AL
        AND     AH,0F0H
        CMP     CH,AH
        JB      TERM7                   ;BR, IF CURRENT PPERATOR > PRIORITY THEN
                                        ;    LAST OPERATOR
        MOV     BYTE PTR [EDX],AL       ;STORE OPERATOR IN OS
        CALL    EVAL                    ;EVAL LAST OPERATION
        MOV     EDX,OSPTR
        MOV     AL,BYTE PTR [EDX]
        DEC     OSPTR
        DEC     EDX
        JMP     TERM8

TERM7:  MOV     EDX,OSPTR
        INC     EDX
        CMP     EDX,DWORD PTR [@@OSEND]
        JE      ERROR_04                ;ERROR IF OS STACK FULL
        MOV     BYTE PTR [EDX],AL
        MOV     OSPTR,EDX
TERM6:  CALL    TERM                    ;EVAL A TERM
TERM12: TEST    SL,2
        JZ      TERM9                   ;BR, IF NOT EXECUTION

TERM98: MOV     EDX,OSPTR
        MOV     CH,BYTE PTR [EDX]       ;CH = LAST OS ATOM
        DEC     OSPTR
        CMP     CH,0A0H
        JE      TERM99                  ;BRANCH IF NO MORE OPERATIONS TO OS
        CALL    EVAL                    ;PERFORM SPECIFIED OPERATION
        JMP     TERM98

TERM99: MOV     OSPTR,EDX               ;RESTORE OS ATOM
TERM9:  RET

TERM13: DEC     TEXTPTR
        JMP     TERM12

TERM1:  CMP     BYTE PTR [TEXTPTR],7FH
        JE      TERM4
        CMP     BYTE PTR [TEXTPTR],7DH
        JE      TERM4
        CMP     BYTE PTR [TEXTPTR],7EH
        JE      TERM70
        CMP     BYTE PTR [TEXTPTR],20H
        JB      TERM2
        CMP     BYTE PTR [TEXTPTR],3FH
        JA      TERM2
TERM69: CALL    NUMBER                  ;LOOK FOR NUMBER IF CHAR BETWWEEN 20H AND 3FH OR 7EH
        TEST    R0,0FH
        JZ      TERM10                  ;BRANCH IF FOUND
TERM3:  CALL    FUNCTION                ;LOOK FOR FUNCTION
        TEST    R0,0F0H
        JZ      TERM10                  ;BRANCH IF FOUND
TERM4:  MOV     K,0                     ;VAR FLAG = NONRECEIVER
        CALL    VAR                     ;LOOK FOR VARIABLE
        JMP     TERM10
TERM2:  TEST    SL,3
        JZ      TERM3                   ;BRANCH IF ENTRY PHASE
        TEST    BYTE PTR [TEXTPTR],80H
        JZ      TERM4                   ;BRANCH IF NOT TEXT ATOM (MUST BE VAR.)
        JMP     TERM3

TERM70: TEST    SL,2
        JZ      TERM71

        MOV     EDX,VSPTR               ;PUSH CONSTANT IN VS
        SUB     EDX,8
        MOV     EBX,DWORD PTR [TEXTPTR+1]
        MOV     DWORD PTR [EDX],EBX
        MOV     EBX,DWORD PTR [TEXTPTR+5]
        MOV     DWORD PTR [EDX+4],EBX
        MOV     VSPTR,EDX
        CMP     EDX,DWORD PTR [@@VSEND]
        JB      ERR_02                  ;VS STACK FULL

TERM71: ADD     TEXTPTR,9
        JMP     TERM10

ERROR_24:
        MOV     R0,24H
        JMP     ERROR


;     EVAL: PERFORM THE SPECIFIED ARITHMETIC OPERATION
;
;          ENTRY: ARGUMENTS IN VS (2ND ARGUMENT IN LAST)
;                 CH = OPERATOR ATOM: B0 = +
;                                     B2 = -
;                                     B3 = LEADING -
;                                     C0 = *
;                                     C5 = /
;                                     D0 = ^
;
;          RETURN: RESULT REPLACES 2 ARGUMENTS IN VS
;
;          EXIT: IF MATH ERROR, ERROR#60-69
;
;          MODIFIED REGISTERS: ALL,AUX0,1,2


EVAL:   CMP       CH,0B3H               ;BRANCH IF LEADING -
        JE        EVAL4
        MOV       EDX,VSPTR
        CMP       CH,0B0H
        JE        EVAL10                ;BR, IF OPERATOR = B0
        CMP       CH,0B2H
        JNE       EVAL1
        CALL      MATHSUB               ;CALL SUB IF OPERATOR = B2
        JMP       EVAL3
EVAL10: CALL      MATHADD               ;CALL ADD IF OPERATOR = B0
        JMP       EVAL3

EVAL1:  CMP       CH,0C0H
        JE        EVAL11
        CMP       CH,0C5H
        JNE       EVAL2
        CALL      MATHDIV               ;CALL DIV IF OPERATOR = C5
        JMP       EVAL3
EVAL11: CALL      MATHMUL               ;CALL MUL IF OPERATOR = C0
        JMP       EVAL3

EVAL2:  CMP       CH,0D0H
        JNE       EVAL5
        CALL      MATHX2Y               ;CALL XEY IF OPERATOR = D0

;     RETURN POINT FOR MATH PACKAGE
;          (MOD,ROUND FUNCTIONS ALSO RETURN HERE)
;

EVAL3:  MOV       VSPTR,EDX             ;SET VSPTR TO POINT TO RESULT IN VS
        CALL      MATHERR               ;CHECK FOR MATH ERRORS
EVAL5:  MOV       R0,0                  ;FLAG = FOUND
        RET

EVAL4:  MOV       EDX,VSPTR
        MOV       R0,0
        MOV       AX,[EDX+2]
        MOV       BX,[EDX+4]
        MOV       CX,[EDX+6]
        NOT       CX
        NOT       BX
        NEG       AX
        JNZ       EVAL20        ;CHANGE SIGN OF LAST VALUE IN VS IF UNARY -
        INC       BX
        JNZ       EVAL20
        INC       CX
EVAL20: MOV       [EDX+2],AX
        MOV       [EDX+4],BX
        MOV       [EDX+6],CX
        RET


;     VAR: SCAN AND PROCESS A NUMERIC VARIABLE.
;
;          VARIABLE FORMAT:
;               SCALAR = LETTER [DIGIT]
;               VECTOR = LETTER [DIGIT] (EXPRESSION)
;               ARRAY =  LETTER [DIGIT] (EXPR,EXPR)
;
;          ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER.
;                 K = TYPE OF VARIABLE LOOKING FOR
;                     BIT 0 = 1 IF SCALAR ONLY
;                     BIT 1 = 1 IF RECEIVER
;
;                    ??? MAY WANT TRACE-BIT (IE, MAKE
;                        RECEIVER BIT INDEPENDENT OF
;                        CURRENT OR DEFINED LENGTH) ???
;
;          RETURN: IF NOT A RECEIVER, VALUE IN VS (X-PHAS
;                  IF RECEIVER, ADDRESS OF VARIABLE IN VS
;                               & NUMERIC RECEIVER ATOM
;                               IN OS. (X-PHASE ONLY)
;
;                  TEXTPTR POINTS TO CHARACTER AFTER VAR.
;
;                  IF VARIABLE NOT FOUND, ERROR.
;
;          MODIFIED REGISTERS: ALL, AUX0,1,2
;
;          NOTE: THIS ROUTINE IS RECURSIVE (CALLS ITSELF
;                VIA EXPR). RECURSION IS HANDLED BY 96
;                LEVEL HARDWARE SUBROUTINE STACK. NO
;                OTHER PROVISION IS MADE FOR HANDLING
;                RECURSION.
;
;     VARR: SCAN AND PROCESS A NUMERIC RECEIVER VARIABLE.
;
;          ENTRY: TEXTPTR POINTS TO CHARACTER AFTER VAR.

VARR:   MOV     K,2         ;FLAG = RECEIVER

VAR:    CALL    VARNAME     ;LOOK FOR NAME OF VARIABLE (LETTER [DIGIT] [(])
        TEST    R0,0FH      ;BRANCH IF FOUND
        JZ      VAR90

ERR_24A:JMP     ERROR_24    ;NF, ERROR

VAR90:  TEST    R2,1        ;ERROR IF ALPHA VARIABLE
        JNZ     ERR_24A

VARA:   TEST    R2,2        ;BRANCH IF NUMERIC ARRAY
        JNZ     VAR2

VAR6:   TEST    SL,2        ;BRANCH IF EXECUTION PHASE
        JNZ     VARB

        TEST    SL,1        ;BRANCH IF RESOLUTION PHASE
        JNZ     VAR9

VAR3:   RET

VARB:   MOV     EAX,DWORD PTR [CPBASE]
        MOV     EDX,DWORD PTR [EAX].@DVTPTR
        LEA     EBX,[EAX].@DVTBL
        CMP     EDX,EBX         ;BR IF NO DUMMY VARIBALE
        JE      VAR9
        MOV     CX,WORD PTR [EDX-10]
        SUB     EDX,8
        CMP     CX,WORD PTR [R2]
        JE      VAR7

VAR9:   CALL    FINDVAR         ;LOOK FOR VARIABLE IN SYMBOL TABLE
        TEST    R0,0FH          ;br, if found
        JZ      VAR5

        CALL    ADDVAR          ;ADD VARIABLE TO SYMBOL TABLE
        MOV     EDX,AUX00       ;GET ADDR OF VARIABLE

VAR5:   TEST    SL,2
        JZ      VAR3
VAR7:   TEST    K,2
        JNZ     VARY

VAR4:   MOV     EAX,VSPTR
        SUB     EAX,8
        MOV     ECX,DWORD PTR [EDX]
        MOV     DWORD PTR [EAX],ECX
        MOV     ECX,DWORD PTR [EDX+4]
        MOV     DWORD PTR [EAX+4],ECX
        MOV     VSPTR,EAX
        CMP     EAX,DWORD PTR [@@VSEND]
        JB      ERROR_02
        RET

VARY:   MOV     EAX,VSPTR
        SUB     EAX,4
        MOV     DWORD PTR [EAX],EDX
        MOV     VSPTR,EAX
        CMP     EAX,DWORD PTR [@@VSEND]
        JB      ERROR_02

        MOV     EAX,OSPTR
        INC     EAX
        CMP     EAX,DWORD PTR [@@OSEND]
        JE      ERROR_04
        MOV     BYTE PTR [EAX],@NUMREC
        MOV     OSPTR,EAX
        JMP     TRACEVAR

;     ARRAY ELEMENT LOGIC
;

VAR2:   TEST    K,1         ;BRANCH IF ARRAY VARIABLE OK
        JZ      VAR75

ERROR_25:
        MOV     R0,25H      ;MISSING SCALAR VARIABLE
        JMP     ERROR

VAR75:  CALL    TRACEVRA    ;IF TRACE MODE, PRINT VARIABLE NAME
        CALL    ARAYEL      ;SCAN & PROCESS SUBSCRIPTS; PC'S = ADDR OF ARRAY EL.
        TEST    SL,2        ;EXIT IF NOT EXECUTION PHASE
        JZ      VAR3
        TEST    K,2         ;BRANCH IF NOT RECEIVER VARIABLE
        JZ      VAR4
        MOV     EAX,VSPTR
        SUB     EAX,4
        MOV     DWORD PTR [EAX],EDX     ;PUT VARIABLE ADDRESS IN VS
        MOV     VSPTR,EAX
        CMP     EAX,DWORD PTR [@@VSEND]
        JB      ERROR_02                ;ERROR IF VS FULL
        MOV     EAX,OSPTR
        INC     EAX
        CMP     EAX,DWORD PTR [@@OSEND]
        JE      ERROR_04                ;ERROR IF OS FULL               ^A02B
        MOV     BYTE PTR [EAX],@NUMREC  ;NUMERIC RECEIVER ATOM --> OS
        MOV     OSPTR,EAX
        RET


;     SNUMBER: GET A NUMBER (ASCII CHARACTERS) FROM TEXT
;              AND CONVERT IT TO INTERNAL FLOATING POINT
;              FORMAT. SIGN (+ OR -) MAY PRECEDE NUMBER.
;
;     NUMBER: SAME AS 'SNUMBER' BUT SIGN MAY NOT PRECEDE
;             NUMBER.
;
;          ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER.
;                 TEXLPTR POINTS TO CURRENT TEXT LINE.
;
;          RETURN: F0 = 0 IF FOUND. TEXTPTR POINTS TO 1ST
;                         CHARACTER AFTER ASCII NUMBER.
;                  F0 = 1 IF NOT FOUND. TEXTPR UNCHANGED.
;                  FL.PT. VALUE IN VALUE-STACK
;
;          EXIT: ERROR#20
;
;          SCRATCH AREA USED: 1ST 8 BYTES
;
;          MODIFIED REGISTERS: ALL,AUX0


SNUMBER:CMP     BYTE PTR [TEXTPTR],7EH
        JE      NMBROK          ;PUT OR SKIP NUMERIC CONSTANT IN RES. OR EXE. PHASE
        MOV     R7,0FFH         ;MAXIMUM NO. OF CHARACTERS IN STRING = 65K
        MOV     EDX,TEXTPTR     ;PC'S POINT TO ASCII NUMBER
        CALL    SNMBR           ;PROCESS SIGNED NUMBER
        MOV     AL,R0
        AND     AL,0FH
        CMP     AL,1            ;BRANCH IF NOT FOUND
        JE      SNMBR99
        MOV     EDX,AUX00
        MOV     TEXTPTR,EDX     ;RESET TEXTPTR TO AFTER ASCII STRING
SNMBR99:RET

;     Special logic for optimally handling single digit
;     numbers.  Special processing is invoked iff 1st
;     character is a digit and the 2nd character is not
;     a valid numeric character (ie, not space,.,digit,
;     or E).
;

NUMBER: CMP     BYTE PTR [TEXTPTR],7EH
        JE      NMBROK          ;PUT OR SKIP NUMERIC CONSTANT IN RES. OR EXE. PHASE
        MOV     EDX,TEXTPTR     ;POINT TO ASCII NUMBER
        MOV     R7,0FFH         ;MAX. NO. OF CHAR. = 65K; CH = 1ST CHARACTER
        MOV     CH,BYTE PTR [EDX]
        CMP     CH,30H
        JB      NUMBER1         ;BR, IF NOT DIGIT
        CMP     CH,39H
        JA      NUMBER1         ;BR, IF NOT DIGIT
        INC     EDX
        JMP     NUMBER2         ;BRANCH IF 1ST CHARACTER = DIGIT
NUMBER3:DEC     EDX
NUMBER1:CALL    NMBR            ;PROCESS UNSIGNED NUMBER
        MOV     AL,R0
        AND     AL,0FH
        CMP     AL,1            ;BRANCH IF NOT FOUND
        JE      SNMBR98
        MOV     EDX,AUX00
        MOV     TEXTPTR,EDX     ;RESET TEXTPTR TO AFTER ASCII STRING
SNMBR98:RET

NUMBER2:MOV     R3,CH           ;F3 = DIGIT; CH = 2ND CHARACTER
        MOV     CH,BYTE PTR [EDX]
        CMP     CH,20H
        JE      NUMBER3         ;BR, IF SPACE
        CMP     CH,2EH
        JE      NUMBER3         ;BR, IF '.'
        CMP     CH,45H
        JE      NUMBER3         ;BR, IF 'E'
        CMP     CH,30H
        JB      NUMBER4         ;BR, IF NOT DIGIT
        CMP     CH,39H
        JA      NUMBER4         ;BR, IF NOT DIGIT
        JMP     NUMBER3         ;BR, IF DIGIT
NUMBER4:MOV     TEXTPTR,EDX     ;POINT TO 2ND CHARACTER
NUMBERA:MOV     AUX00,TEXTPTR
        DEC     AUX00
        CALL    SQZTEXT         ;SQUEEZE VARIABLE NAME OUT OF TEXT LINE
        MOV     TEXTEND,EDX
        ADD     TEXTEND,8

;     MAKE SURE AT LEAST 10 BYTES BETWEEN TEXTEND & VS
;
        MOV     EAX,TEXTEND
        MOV     EDX,TEXTEND
        MOV     EBX,VSPTR
        SUB     EBX,10
        CMP     EBX,EAX
        JB      ERR_02                ;ERROR IF NOT AT LEAST 10 BYTES

;     EXPAND LINE BY 9 BYTES ATOM FOR NUMBER ATOM AND 8 BYTES VALUE
;
        MOV     TEXTPTR,EDX            ;SAVE 'FROM' ADDRESS
        MOV     EDX,AUX00
        MOV     DWORD PTR [A0],EDX     ;F1-0 = PTR TO WHERE LINE NO. WAS
        MOV     EDX,TEXTPTR
                                       ;PC'S = 'FROM' ADDRESS
        ADD     TEXTPTR,9              ;TEXTPTR = 'TO' ADDRESS ('FROM' + 9)

NMBR55: MOV     CH,BYTE PTR [EDX]
        XCHG    EDX,TEXTPTR
        DEC     TEXTPTR                ;MOVE NEXT BYTE OF TEXT LINE 3 BYTES
        MOV     BYTE PTR [EDX],CH
        XCHG    EDX,TEXTPTR
        DEC     TEXTPTR                ;TO THE RIGHT
        CMP     DWORD PTR [A0],EDX
        JBE     NMBR55                 ;BRANCH IF MORE TO MOVE
        MOV     EDX,AUX00
        MOV     BYTE PTR [EDX],7EH     ;INSERT VAR ATOM
        INC     EDX                    ;INCREMENT PC'S BY 1
        XOR     AX,AX
        MOV     AL,R3
        MOV     WORD PTR [EDX],0       ;INSERT EXPONENT
        ADD     EDX,2
        XOR     AX,AX
        MOV     AL,R3
        SUB     AL,30H
        MOV     WORD PTR [EDX],AX      ;INSERT MANTISSA WORD 1
        ADD     EDX,2
        MOV     WORD PTR [EDX],0       ;INSERT MANTISSA WORD 2
        ADD     EDX,2
        MOV     WORD PTR [EDX],0       ;INSERT MANTISSA WORD 3
        ADD     EDX,2
        MOV     AUX00,EDX
        MOV     TEXTPTR,EDX            ;TEXTPTR POINT TO CHAR.
        MOV     R0,0
        RET

NMBROK: TEST    SL,2                   ;SKIP CONSTANT IF RESOLUTION PHASE
        JZ      NMBROK1
        MOV     EDX,VSPTR              ;PUSH CONSTANT IN VS
        SUB     EDX,8
        MOV     EBX,DWORD PTR [TEXTPTR+1]
        MOV     DWORD PTR [EDX],EBX
        MOV     EBX,DWORD PTR [TEXTPTR+5]
        MOV     DWORD PTR [EDX+4],EBX
        MOV     VSPTR,EDX
        CMP     EDX,DWORD PTR [@@VSEND]
        JB      ERR_02                 ;VS STACK FULL

NMBROK1:ADD     TEXTPTR,9
        MOV     R0,0
        RET

;     SNMBR: GET AN ASCII NUMBER AND CONVERT IT TO
;            INTERNAL FLOATING POINT FORMAT.
;            ASCII NUMBER MAY START WITH SIGN (+ OR -).
;
;     NMBR: SAME AS 'SNMBR' EXCEPT LEADING SIGN NOT
;           ALLOWED
;
;          ENTRY: PC'S POINT TO ASCII CHARACTER STRING
;                 F7,6 = MAX. NO. OF CHARACTERS IN STRING
;
;          RETURN: F0 = 0 IF FOUND. AUX0 POINT TO 1ST
;                         CHARACTER AFTER ASCII NUMBER.
;                  F0 = 1 IF NOT FOUND.
;                  FL.PT. VALUE IN VALUE-STACK
;                  F7,6 = NO. OF CHARACTERS NOT PROCESSED
;                         ( F7,6 - LENGTH OF ASCII NO.)
;
;          EXIT: ERROR#75
;
;          SCRATCH AREA USED: 1ST 8 BYTES
;
;          REGISTER USE:
;               PC'S POINT TO NEXT CHARACTER IN ASCII
;                    CHARACTER STRING
;               AUX0 POINTS TO SCRATCH AREA
;               F0 = 80 IF 'E-XX' SPECIFIED (NEG. ADJUST)
;                  = 00 OTHERWISE
;               F1 = IMPLIED EXPONENT (NO. OF LEADING
;                    FRACTION ZEROES OR NO. OF
;                    INTEGER DIGITS -1)
;               F2 = NO. OF CHARACTERS LEFT IN STRING
;               F3,LOW 4-BITS = NO. OF SIGNIFICANT DIGITS
;                 ,HIGH 4-BITS= SAVE AREA FOR HIGH DIGIT
;                               OF PAIR TO STORE
;               F4,LOW 4-BITS = FOUND FLAG
;                       BIT 0 = 1 IF LEADING ZERO FOUND
;                       BIT 1 = 1 IF DEC. PT. FOUND
;                       BIT 3 = 1 IF SIGN FOUND
;                 ,HIGH 4-BITS= SIGN OF NUMBER
;                             = 0 IF MANTISSA +, EXP. +
;                             = 1 IF MANTISSA -, EXP. +
;                             = 8 IF MANTISSA +, EXP. -
;                             = 9 IF MANTISSA -, EXP. -
;               F5 = X'39'  (CONSTANT)
;               F2 = EXPONENT (BCD)
;               K = 0
;
;          MODIFIED REGISTERS: ALL,AUX0

NMBR:   CALL    NMBRINIT        ;INITIALIZE WORK AREA, COUNTS, FLAGS
        JMP     NMBR1

SNMBR:  CALL    NMBRINIT        ;INITIALIZE WORK AREA, COUNTS, FLAGS

;     PROCESS SIGN
;

        CMP     WORD PTR [R6],0 ;BRANCH IF VALUE ALL SPACES
        JE      NMBR32
        CALL    NMBRCH
        MOV     AL,CH
        AND     AL,0F0H
        CMP     AL,020H         ;BRANCH IF NOT '2X'
        JNE     NMBR3
        CMP     CH,2BH          ;BRANCH IF '+'
        JE      NMBR6
        CMP     CH,2DH          ;BRANCH IF '-'
        JNE     NMBR5
        OR      R4,10H          ;MANTISSA SIGN = '-'

NMBR6:  OR      R4,08H          ;FLAG = SIGN FOUND

;     PROCESS LEADING ZEROES
;

NMBR1:  CMP     WORD PTR [R6],0 ;BRANCH IF NO MORE CHARACTER
        JE      NMBR8
        CALL    NMBRCH          ;GET NEXT CHARACTER

NMBR3:  MOV     AL,CH           ;BRANCH IF NOT '3X'
        AND     AL,0F0H
        CMP     AL,030H
        JNE     NMBR9
        CMP     CH,30H         ;BRANCH IF NOT '0'
        JNE     NMBR7
        OR      R4,1           ;FLAG = LEADING ZERO FOUND
        JMP     NMBR1

;     PROCESS DECIMAL POINT WHEN NO SIGNIFICANT
;     INTEGER DIGITS
;

NMBR9:  MOV     AL,CH       ;BRANCH IF NOT '2X'
        AND     AL,0F0H
        CMP     AL,020H
        JNE     NMBR21

NMBR5:  CMP     CH,2EH      ;BRANCH IF NOT '.'
        JNE     NMBR30
        OR      R4,82H      ;FLAG = DECIMAL POINT FOUND; SIGN OF EXP. = '-'

;     PROCESS LEADING FRACTION ZEROES WHEN NO SIGNIFICANT
;          INTEGER DIGITS
;

NMBR13: MOV     AL,R1       ;INCREMENT IMPLIED EXPONENT (LEADING FRACTION ZERO COUNT)
        ADD     AL,1
        DAA
        MOV     R1,AL
        CMP     WORD PTR [R6],0 ;BRANCH IF NO MORE CHARACTER
        JE      NMBR8
        CALL    NMBRCH      ;GET NEXT CHAR AND INCREMENT POINTER
        MOV     AL,CH       ;BRANCH IF NOT '3X'
        AND     AL,0F0H
        CMP     AL,030H
        JNE     NMBR21
        CMP     CH,30H      ;BRANCH IF NOT '0'
        JNE     NMBR12
        OR      R4,1        ;FLAG = ZERO FOUND
        JMP     NMBR13

;     PROCESS SIGNIFICANT INTEGER DIGITS
;

NMBR7:  MOV     R1,99H      ;IMPLIED EXPONENT = -1
        JMP     NMBR4

NMBR15: MOV     AL,R1       ;INCREMENT IMPLIED EXPONENT
        ADD     AL,1
        DAA
        MOV     R1,AL
        INC     R3          ;INCREMENT SIGNIFICANT DIGIT COUNT
        MOV     AL,R3       ;ERROR IF > 13 DIGITS
        AND     AL,0FH
        CMP     AL,0EH
        JE      ERROR_75
        TEST    R3,1        ;BRANCH IF HIGH DIGIT OF PAIR
        JZ      NMBR16
        XCHG    EDX,AUX00
        MOV     AL,R3       ;STORE DIGIT PAIR IN SCRATCH AREA
        AND     AL,0F0H
        MOV     AH,CH
        AND     AH,0FH
        OR      AL,AH
        MOV     BYTE PTR [EDX],AL
        INC     EDX          ;INCREMENT POINTER TO SCRATCH AREA
        XCHG    EDX,AUX00
        JMP     NMBR39

NMBR16: AND     R3,0FH      ;SAVE DIGIT
        MOV     AL,CH
        SHL     AL,4
        OR      R3,AL

NMBR39: CMP     WORD PTR R6,[0] ;BRANCH IF NO MORE CHAR.
        JE      NMBR2
        CALL    NMBRCH          ;GET NEXT CHAR.
        MOV     AL,CH           ;BRANCH IF NOT '3X'
        AND     AL,0F0H
        CMP     AL,030H
        JNE     NMBR49

NMBR4:  CMP     CH,R5       ;BRANCH IF 0-9
        JBE     NMBR15

;     PROCESS DECIMAL POINT AFTER SIGNIFICANT INTEGER
;          DIGITS HAVE BEEN FOUND
;

NMBR49: MOV     AL,CH       ;BRANCH IF NOT '2X'
        AND     AL,0F0H
        CMP     AL,020H
        JNE     NMBR21
        CMP     CH,2EH      ;BRANCH IF NOT '.'
        JNE     NMBR30
        OR      R4,02H      ;FLAG = DECIMAL POINT FOUND
        JMP     NMBR18

;     PROCESS SIGNIFICANT FRACTION DIGITS
;

NMBR17: INC     R3          ;INCREMENT SIGNIFICANT DIGIT COUNT
        MOV     AL,R3       ;BRANCH IF > 13 DIGIT
        AND     AL,0FH
        CMP     AL,0EH
        JE      ERROR_75
        TEST    R3,1        ;BRANCH IF HIGH DIGIT OF PAIR
        JZ      NMBR19
        XCHG    EDX,AUX00
        MOV     AL,R3       ;STORE DIGIT PAIR IN SCRATCH AREA
        AND     AL,0F0H
        MOV     AH,CH
        AND     AH,0FH
        OR      AL,AH
        MOV     BYTE PTR [EDX],AL
        INC     EDX
        XCHG    EDX,AUX00    ;INCREMENT POINTER
        JMP     NMBR18

NMBR19: MOV     AL,CH       ;SAVE DIGIT
        SHL     AL,4
        AND     R3,0FH
        OR      R3,AL

NMBR18: CMP     WORD PTR [R6],0 ;BRANCH IF NO MORE CHARACTERS
        JE      NMBR2
        CALL    NMBRCH          ;GET NEXT CHAR.
        MOV     AL,CH           ;BRANCH IF NOT '3X'
        AND     AL,0F0H
        CMP     AL,030H
        JNE     NMBR21

NMBR12: CMP     CH,R5           ;BRANCH IF 0-9
        JBE     NMBR17

;     PROCESS EXPONENT ADJUST (E + OR - XX)
;

NMBR21: CMP     CH,45H      ;BRANCH IF NOT 'E'
        JNE     NMBR30
        CMP     WORD PTR [R6],0 ;BRANCH IF NOT MORE CHAR.
        JE      ERROR_75

NMBR24: CALL    NMBRCH          ;GET NEXT CHAR.
        MOV     AL,CH           ;BRANCH IF NOT '2X'
        AND     AL,0F0H
        CMP     AL,020H
        JNE     NMBR23
        CMP     CH,2BH      ;BRANCH IF '+'
        JE      NMBR25
        CMP     CH,2DH      ;BRANCH IF NOT '-'
        JNE     ERROR_75
        OR      R0,80H      ;EXPONENT ADJUST SIGN = '-'

NMBR25: CMP     WORD PTR [R6],0 ;ERROR IF NO MORE CHAR.
        JE      ERROR_75
        CALL    NMBRCH      ;GET NEXT CHAR.

NMBR23: MOV     AL,CH       ;BRANCH IF NOT '3X'
        AND     AL,0F0H
        CMP     AL,030H
        JNE     ERROR_75
        CMP     CH,R5       ;BRANCH IF NOT 0-9
        JG      ERROR_75
        MOV     R2,CH       ;SAVE EXPONENT ADJUST DIGIT
        AND     R2,0FH

NMBR29: CMP     WORD PTR [R6],0 ;BRANCH IF NO MORE CHAR.
        JE      NMBR22
        CALL    NMBRCH          ;GET NEXT CHAR.
        MOV     AL,CH           ;BRANCH IF NOT '3X'
        AND     AL,0F0H
        CMP     AL,030H         ;BRANCH IF NOT 0-9
        JNE     NMBR31
        CMP     CH,R5           ;BRANCH IF NOT 0-9
        JG      NMBR31
        MOV     AL,R2           ;SHLL  CH,F2,F2
        SHL     AL,4            ;R2 = EXPONENT ADJUST VALUE (BCD)
        MOV     AH,CH
        AND     AH,0FH
        OR      AL,AH
        MOV     R2,AL

;     ADJUST EXPONENT (ONLY IF 'E+-XX' FOUND)
;
;          ENTRY: F1 = IMPLIED EXPONENT (BCD)
;                 F2 = VALUE TO ADD OR SUBTRACT FROM EXP.
;                 F4,BIT7 = 0 IF EXP.+, 1 IF EXP.-
;                 F0,BIT7 = 0 IF EXP.ADJUST +, 1 IF -
;                 F3 = NO. OF SIGNIFICANT DIGITS
;

NMBR22: TEST    R3,0FH      ;BRANCH IF NO SIGNIFICANT DIGITS
        JZ      NMBR26
        MOV     AL,R0
        XOR     AL,R4
        MOV     R0,AL
        TEST    AL,80H      ;BRANCH IF SIGN OF EXP. <> SIGN OF ADJUST VALUE
        JNZ     NMBR11
        MOV     AL,R1       ;R1 = EXPONENT + ADJUST
        ADD     AL,R2
        DAA
        MOV     R1,AL
        JC      ERROR_75    ;BRANCH IF EXPONENT OVERFLOW
        JMP     NMBR2

NMBR31: INC     WORD PTR [R6];INCREMENT COUNT
        DEC     EDX          ;DECREMENT PC'S
        JMP     NMBR22

;     STORE LAST DIGIT IF NOT ALREADY STORED
;

NMBR30: DEC     EDX          ;DECREMENT PC'S
        INC     WORD PTR [R6];INCREMENT R7R6 SINCE LAST CHAR. ILLEGL

NMBR2:  TEST    R3,1
        JNZ     NMBR8
        XCHG    EDX,AUX00
        MOV     AL,R3       ;STORE LAST DIGIT FOUND IF NOT ALREADY STORED
        AND     AL,0F0H
        MOV     BYTE PTR [EDX],AL
        INC     EDX
        XCHG    EDX,AUX00

;     MAKE SURE MANTISSA IS LEGITIMATE--AT LEAST 1
;          DIGIT MUST HAVE BEEN FOUND. IF NO DIGITS
;          FOUND BUT SIGN OR DECIMAL POINT WAS FOUND,
;          ERROR; OTHERWISE, NOT FOUND RETURN.
;

NMBR8:  TEST    R3,0FH      ;BRANCH IF SOME SIGNIFICANT DIGIT FOUND
        JNZ     NMBR10

NMBR26: TEST    R4,1        ;BRANCH IF SOME LEADING ZEROS
        JNZ     NMBR10
        TEST    R4,0AH      ;BRANCH IF NO SIGN OR DEC. PT. FOUND
        JZ      NMBR32

ERROR_75:
        TEST    MODE,00002000H  ;ERROR - ILLEGAL NUMBER
        JZ      NMBR79
        JMP     INPER_75    ;IF INPUT MODE GOTO THAT ROUTINE TO CLEAR INPUT POINTERS

NMBR79: MOV     R0,75H      ;ELSE HANDLE THE PROBLEM HERE
        JMP     ERROR

NMBR32: MOV     R0,1        ;FLAG = NOT FOUND
        RET

;     STORE NUMBER IN VALUE STACK
;

NMBR10: CMP     WORD PTR [R6],0 ;BRANCH IF NO MORE CHAR.
        JE      NMBR33
        CALL    NMBRCH          ;BYPASS TRAILING SPACES
        DEC     EDX
        INC     WORD PTR [R6]   ;INCREMENT COUNT

NMBR33: MOV     AUX00,EDX       ;AUX0 = PTR TO NEXT CHAR. IN ASCII STRING
        MOV     EDX,OFFSET SCRATCH  ;GET 1ST 2 BYTE OF NUMBER
        MOV     CX,[EDX]        ; Checkpoint.
        XCHG    CH,CL
        TEST    R3,0FH         ;BRANCH IF NO SIGNIFICANT DIGITS
        JZ      NMBR14
        MOV     AL,R4          ;STORE EXPONENT AND SIGN IN SCRATCH AREA
        AND     AL,0F0H
        MOV     AH,R1
        AND     AH,0FH
        OR      AL,AH
        MOV     BYTE PTR [EDX],AL
        MOV     AH,R1
        AND     AH,0F0H
        MOV     BH,CL
        AND     BH,0FH
        OR      AH,BH
        MOV     BYTE PTR [EDX+1],AH

NMBR14: PUSH    EBP         ;CONVERT CONSTANT AND STORE IN TEXT
        MOV     W0,0
        MOV     W1,0
        MOV     W2,0
        MOV     W3,0
        MOV     EDX,OFFSET SCRATCH
        MOV     CX,WORD PTR [EDX]
        XCHG    CH,CL
        TEST    CL,0FH      ;br, if zero
        JZ      CC_ZERO
        MOV     AL,CL
        AND     AL,0F0H
        SHR     AL,4
        MOV     AH,0
        MOV     BL,10
        MUL     BL
        MOV     BL,CH
        AND     BL,0FH
        ADD     AL,BL
        TEST    CH,80H      ;check sign
        JZ      CC1
        NEG     AX
CC1:    MOV     WORD PTR [W0],AX
        AND     CL,0FH
        MOV     EDX,OFFSET B_T
        MOV     BYTE PTR [EDX],CL
        INC     EDX
        XCHG    EDX,DWORD PTR [W_TEMP]
        MOV     BL,1
        MOV     BH,1
        MOV     EDX,OFFSET SCRATCH
        ADD     EDX,2
        MOV     CH,6
CC6:    MOV     CL,BYTE PTR [EDX]
        TEST    CL,0F0H
        JZ      CC2
        MOV     BL,BH
        INC     BL
        INC     BH
        JMP     CC3
CC2:    INC     BH
CC3:    MOV     AL,CL
        AND     AL,0F0H
        SHR     AL,4
        XCHG    EDX,DWORD PTR [W_TEMP]
        MOV     BYTE PTR [EDX],AL
        INC     EDX
        XCHG    EDX,DWORD PTR [W_TEMP]
        TEST    CL,0FH
        JZ      CC4
        MOV     BL,BH
        INC     BL
        INC     BH
        JMP     CC5
CC4:    INC     BH
CC5:    MOV     AL,CL
        AND     AL,0FH
        XCHG    EDX,DWORD PTR [W_TEMP]
        MOV     BYTE PTR [EDX],AL
        INC     EDX
        XCHG    EDX,DWORD PTR [W_TEMP]
        INC     EDX
        DEC     CH
        CMP     CH,0
        JNZ     CC6
        MOV     BH,0
        SUB     WORD PTR [W0],BX
        INC     WORD PTR [W0]
        MOV     AX,WORD PTR [W0]
        OR      AX,AX
        JS      CC44
        ADD     AX,BX
        CMP     AX,13
        JG      CC44
        MOV     BX,AX
        MOV     WORD PTR [W0],0

CC44:   MOV     EAX,15
        XOR     ECX,ECX
        MOV     CX,BX
        SUB     EAX,ECX
        MOV     CL,6
        MUL     CL
        MOV     EDX,OFFSET _ABCOV
        ADD     EDX,EAX
        XCHG    EDX,DWORD PTR [W_TEMP]
        MOV     EDX,OFFSET B_T
CC9:    MOV     BH,BYTE PTR [EDX]
        MOV     WT1,0
        MOV     WT2,0
        MOV     WT3,0
        XCHG    EDX,DWORD PTR [W_TEMP]
        CMP     BH,0
        JZ      CC7
CC8:    MOV     ESI,EDX
;       CLD
        LODSW
        MOV     DI,AX
        LODSW
        MOV     CX,AX
        LODSW
        ADD     WT1,DI
        ADC     WT2,CX
        ADC     WT3,AX
        DEC     BH
        CMP     BH,0
        JNZ     CC8
CC7:    MOV     DI,WT1
        MOV     CX,WT2
        MOV     AX,WT3
        ADD     W1,DI
        ADC     W2,CX
        ADC     W3,AX
        DEC     BL
        ADD     DX,6
        XCHG    EDX,DWORD PTR [W_TEMP]
        INC     EDX
        CMP     BL,0
        JNZ     CC9
        MOV     EDX,OFFSET SCRATCH
        TEST    BYTE PTR [EDX],10H
        JZ      CC10
        NOT     WORD PTR [W3]
        NOT     WORD PTR [W2]
        NEG     WORD PTR [W1]
        JNZ     CC11
        INC     WORD PTR [W2]
CC11:   JNZ     CC10
        INC     WORD PTR [W3]

CC10:   POP     EBP

        TEST    SL,2                ;Special for some one
        JNZ     CCXX

        XCHG    TEXTPTR,AUX00
        CALL    SQZTEXT             ;SQUEEZE VARIABLE NAME OUT OF TEXT LINE
        MOV     TEXTEND,EDX
        ADD     TEXTEND,8

;     MAKE SURE AT LEAST 10 BYTES BETWEEN TEXTEND & VS
;
        MOV     EAX,TEXTEND
        MOV     EDX,TEXTEND
        MOV     EBX,VSPTR
        SUB     EBX,10
        CMP     EBX,EAX
        JB      ERR_02              ;ERROR IF NOT AT LEAST 10 BYTES

;     EXPAND LINE BY 9 BYTES AND INSERT NUMERIC ATOM AND CONSTANT
;
        MOV     TEXTPTR,EDX           ;SAVE 'FROM' ADDRESS
        MOV     EDX,AUX00
        MOV     DWORD PTR [A0],EDX    ;F1-0 = PTR TO WHERE LINE NO. WAS
        MOV     EDX,TEXTPTR
                                      ;PC'S = 'FROM' ADDRESS
        ADD     TEXTPTR,9             ;TEXTPTR = 'TO' ADDRESS ('FROM' + 9)

CC99:   MOV     CH,BYTE PTR [EDX]
        XCHG    EDX,TEXTPTR
        DEC     TEXTPTR               ;MOVE NEXT BYTE OF TEXT LINE 3 BYTES
        MOV     BYTE PTR [EDX],CH
        XCHG    EDX,TEXTPTR
        DEC     TEXTPTR               ;TO THE RIGHT
        CMP     DWORD PTR [A0],EDX
        JBE     CC99                  ;BRANCH IF MORE TO MOVE
        MOV     EDX,AUX00
        MOV     BYTE PTR [EDX],7EH    ;INSERT VAR ATOM
        INC     EDX                   ;INCREMENT PC'S BY 1
        MOV     AX,W0
        MOV     WORD PTR [EDX],AX     ;INSERT EXPONENT
        ADD     EDX,2
        MOV     AX,W1
        MOV     WORD PTR [EDX],AX     ;INSERT MANTISSA WORD 1
        ADD     EDX,2
        MOV     AX,W2
        MOV     WORD PTR [EDX],AX     ;INSERT MANTISSA WORD 2
        ADD     EDX,2
        MOV     AX,W3
        MOV     WORD PTR [EDX],AX     ;INSERT MANTISSA WORD 3
        ADD     EDX,2
        MOV     AUX00,EDX
        MOV     TEXTPTR,EDX
        JMP     NMBR28

CC_ZERO:MOV      W0,0
        MOV      W1,0
        MOV      W2,0
        MOV      W3,0
        JMP      CC10

CCXX:   MOV      EDX,VSPTR
        SUB      EDX,8
        MOV      AX,WORD PTR [W0]
        MOV      WORD PTR [EDX],AX
        MOV      AX,WORD PTR [W1]
        MOV      WORD PTR [EDX+2],AX
        MOV      AX,WORD PTR [W2]
        MOV      WORD PTR [EDX+4],AX
        MOV      AX,WORD PTR [W3]
        MOV      WORD PTR [EDX+6],AX
        MOV      VSPTR,EDX
        CMP      EDX,DWORD PTR [@@VSEND]
        JB       ERROR_02
        MOV      R0,0
        RET

W_TEMP  DD       ?

W0      DW       ?
W1      DW       ?
W2      DW       ?
W3      DW       ?

WT0     DW       ?
WT1     DW       ?
WT2     DW       ?
WT3     DW       ?

B_T     DB       13 DUP(?)

NMBR28: MOV     R0,0        ;FLAG = FOUND
        RET

;     ADJUST EXPONENT (CONT.)-- SIGN OF EXPONENT <>
;          SIGN OF EXPONENT ADJUST VALUE
;

NMBR11: MOV     AL,R1
        MOV     BL,AL
        SUB     AL,R2       ;R1 = EXPONENT - ADJUST VALUE
        DAS
        MOV     R1,AL
        cmp     al,0
        je      nmbr27

        CMP     BL,R2
        JAE     NMBR2       ;BR, IF EXPONENT DID NOT CHANGE SIGN

        MOV     AL,00H
        SUB     AL,R1
        DAS
        MOV     R1,AL
        XOR     R4,80H      ;COMPLEMENT SIGN OF EXPONENT
        JMP     NMBR2

NMBR27: AND     R4,7FH      ;MAKE SIGN OF EXPONENT + IF EXP. = 0
        JMP     NMBR2

;     NMBRINIT: INITIALIZE COUNTS, FLAGS, SCRATCH AREA                        ;
;               FOR 'NMBR' SUBROUTINE.                                        ;
;                                                                             ;
;          RETURN: AUX0 POINTS TO BYTE 2 OF SCRATCH AREA                      ;
;                  ZERO WORK AREA (1ST 8 BYTES OF SCRATCH                     ;
;                  MANTISSA & EXPONENT SIGNS = +                              ;
;                  NO. OF SIGNIFICANT DIGITS = 0                              ;
;                  IMPLIED EXPONENT = 0                                       ;
;                  FOUND FLAG = 0                                             ;
;                  F5 = C'9'   (CONSTANT)                                     ;
;                  K = 0                                                      ;
;                                                                             ;
;          MODIFIED REGISTERS: F0-5,K,AUX0                                    ;

NMBRINIT:
        MOV     AUX00,EDX
        MOV     EDX,OFFSET SCRATCH
        MOV     K,0         ;K = 0
        MOV     DWORD PTR [EDX],0    ;ZERO WORK AREA
        MOV     DWORD PTR [EDX+4],0  ;ZERO WORK AREA
        XCHG    AUX00,EDX            ;AUX0 POINT TO 2ND BYTE OF WORK AREA
        INC     AUX00
        MOV     WORD PTR [R0],0      ;ZERO COUNTS, FLAGS, SIGN, EXPONENT
        MOV     WORD PTR [R3],0
        MOV     R2,0
        MOV     R5,'9'
        RET

;     NMBRCH: GET NEXT CHARACTER OF NUMBER. IGNORE
;             SPACES.
;
;          ENTRY: PC'S POINT TO NEXT CHARACTER OF STRING
;
;          CALL MUST BE:     BLERX F7F6,DK,-----
;                            SB    NMBRCH
;
;          RETURN: CH = FOUND CHARACTER
;                  F2 (CHARACTER COUNT) DECREMENTED
;                  PC'S POINT TO CHARACTER AFTER FOUND
;                         CHARACTER.
;
;          MODIFIED REGISTERS: PC'S,F6-7,CH,CL

NMBRCH2:CMP     WORD PTR [R6],0   ;BRANCH IF NO MORE CHAR.
        JE      NMBRCH1

NMBRCH: DEC     WORD PTR [R6]     ;DECREMENT CHAR. COUNT
        MOV     CH,BYTE PTR [EDX] ;GET NEXT CHAR.
        INC     EDX               ;INCREMENT COUNT
        CMP     CH,20H
        JE      NMBRCH2
        RET

NMBRCH1:POP     ESI         ;CHANGE RETURN ADDR TO INSTRUCTION BEFORE
        SUB     ESI,19      ;    'SB NMBRCH'
        PUSH    ESI         ;RETURN TO INSTR. BEFORE 'SB NMBRCH
        RET


;     FUNCTION: SCAN AND EVALUATE ANY FUNCTION LEGAL IN
;               NUMERIC EXPRESSIONS.
;
;          ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER.
;                 TEXTLPTR POINTS TO CURRENT TEXT LINE.
;
;          RETURN: F0 = 0X IF FOUND, TEXTPTR POINTS TO
;                          CHARACTER AFTER FUNCTION.
;                          IN EXECUTION PHASE, RESULT IS
;                          IN VALUE STACK (VS).
;                  F0 <> 0X IF NOT FOUND, TEXTPTR POINTS
;                          TO NEXT NONBLANK TEXT CHAR.
;
;          NOTE: THIS ROUTINE IS RECURSIVE (CALLS ITSELF
;                VIA 'EXPR'). RECURSION IS HANDLED BY A
;                96-LEVEL HARDWARE SUBROUTINE RETURN
;                STACK. NO OTHER PROVISION IS MADE FOR
;                HANDLING RECURSION.
;
;          MODIFIED REGISTERS: ALL,AUX0,1,2

FUNCTION:
        MOV     R2,40H
        CALL    CCJ             ;LOOK FOR SINGLE-ARGUMENT MATH FUNCS. (SIGN,LOG)
        TEST    R0,0F0H         ;BRANCH IF FOUND
        JZ      FUNC1
        MOV     R2,0A0H

FUNC6:  CALL    CCJ             ;LOOK FOR MISC. FUNCTION (VAL,MAX,POS)
        TEST    R0,0F0H         ;BRANCH IF NOT FOUND
        JNZ     FUNC5

FUNC2:  JMP     EDX             ;BRANCH TO FOUND FUNCTION'S ROUTINE

;     ARC: PROCESS AN ARC FUNCTION
;

ARC:
        MOV     R2,80H          ;FLAG = ARC FUNCTION LIST
        CALL    CCJ             ;LOOK FOR ARC FUNCTION (SIN,COS,TAN)
        TEST    R0,0F0H         ;EXIT
        JNZ     FUNC5

FUNC1:  TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      FUNC3

        MOV     AUX00,EDX       ;SAVE ADDRESS OF FUNCTION ROUTINE
        MOV     EDX,OFFSET EVAL3
        PUSH    EDX             ;PUT RETURN ADDRESS FOR MATH PKG. IN STACKH
        MOV     EDX,AUX00
        AND     EDX,0FFFFFFFH   ;PUT ADDRESS OF FUNCTION ROUTINE IN STACK
        PUSH    EDX

FUNC3:  CALL    EXPR            ;SCAN AND EVALUATE EXPRESSION
        CMP     BYTE PTR [TEXTPTR],')'      ;LOOK FOR ")", R0 = 0 (FOUND FLAG)
        JNE     ERROR_11
        INC     TEXTPTR
        MOV     EDX,VSPTR       ;PC'S = PTR TO VALUE IN VS
        MOV     R0,0

FUNC5:  RET                     ;CALL MATH PKG. IF EXECUTION PHASE OR RETURN
                                ;FROM 'FUNCTION' IF NOT EXECUTION PHASE

FNCSPC: CMP     BYTE PTR [TEXTPTR],'S'  ;CHECK FOR 'S'
        JNE     FNCSPC1
        INC     TEXTPTR
        JMP     FNCSPACES
FNCSPC1:CMP     BYTE PTR [TEXTPTR],'K'  ;CHECK FOR 'K'
        JNE     FNCEND          ;NF, EVALUATE END FUNCTION
        INC     TEXTPTR

;     EVALUATE 'SPACEK' FUNCTION
;
        TEST    SL,2        ;EXIT IF NOT EXECUTION PHASE
        JZ      FNCFOUND    ;CALCULATE PARTITION SIZE
        CALL    SPACEK
        JMP     FNCFOUND

FNCSPACES:
        CMP     BYTE PTR [TEXTPTR],'K'  ;CHECK FOR 'K'
        JNE     SPACES          ;NF, EVALUATE END FUNCTION
        INC     TEXTPTR

;     EVALUATE 'SPACESK' FUNCTION
;
        TEST    SL,2        ;EXIT IF NOT EXECUTION PHASE
        JZ      FNCFOUND    ;CALCULATE PARTITION SIZE
        CALL    SPACESK
        JMP     FNCFOUND

;  NOTE:  SINCE THE PARTITION MUST HAVE A MULTIPLE OF
;     256 BYTES OF SPACE, THE RESULT OF SPACEK WILL ALWAYS BE
;     A MULTIPLE OF .25


;SPACEK: PARTITION SIZE (FL.PT.) ==> VALUE STACK (VS),                        ;
;        CURRENT PARTITION.                                                   ;
;                                                                             ;
;SPACEKA: CONVERT # PAGES OF MEMORY TO FL.PT.                                 ;
;                                                                             ;
;     ENTRY: F0 = NO. OF PAGES                                                ;
;                                                                             ;
;     MODIFIED REGISTERS: ALL,AUX0,1,2                                        ;
;

SPACEK:
        MOV     CH,BYTE PTR [SYSGENFG]
        TEST    CH,1            ;BRANCH IF SYSTEM ALREADY CONFIGURED
        JNZ     SPACEK1

;------------------ system not yet configured ----------------------------------
;        Since system not configured, SPACEK returns the amount of memory
;        available for partitions.  This is:
;
;  NO MEANING -->  56K ; (# banks -1) + (#pages in last bank) - 3K
;

        MOV     ECX,DWORD PTR [ENDMEMX] ;GET OFFSET OF PHYSICAL MEMORY END
        MOV     EAX,OFFSET GIEND
        SUB     ECX,EAX                 ;GET THE SIZE OF USER FREE SPACE
        INC     ECX
        CLC
        SHR     ECX,10                  ;SIZE IN NO. OF K
        JMP     SPACEK3
SPACEK1:
        MOV     EBX,CPBASE              ;GET CURRENT BASE
        MOV     ECX,DWORD PTR [EBX].@PARTEND    ;GET PARTITION END OFFSET
        SUB     ECX,EBX                 ;
        INC     ECX
        CLC                             ;
        SHR     ECX,10                  ;NO. OF K
SPACEK3:                                ;PUT VALUE IN VS
        MOV     EBX,VSPTR               ;GET VS PTR
        SUB     EBX,8                   ;
        MOV     WORD PTR [EBX],0        ;EXP = 0
        MOV     DWORD PTR [EBX+2],ECX   ;SAVE SPACE VALUE ( IN K )
        MOV     WORD PTR [EBX+6],0
        MOV     VSPTR,EBX               ;UPDATE VSPTR
        CMP     EBX,@@VSEND             ;OVER THE LIMIT?
        JB      ERROR_02                ;BR, IF ERROR
        RET

SPACEKA:
        XOR     ECX,ECX                 ;
        MOV     CL,R0                   ;R0 = NO. OF PAGES
        SHR     ECX,2                   ;ECX = NO. OF K
        JMP     SPACEK3

;SPACESK: TOTAL SYSTEM SIZE (FL.PT.) ==> VALUE STACK (VS),                    ;
;         SIZE IN K                                                                    ;
;                                                                             ;
;SPACES:  UNUSED SYSTEM SIZE (FL.PT.) ==> VALUE STACK (VS),                   ;
;         SIZE IN BYTES                                                                    ;
;                                                                             ;
;     MODIFIED REGISTERS: ALL,AUX0,1,2                                        ;
;

SPACESK:
        MOV     ECX,ENDMEMX
        ADD     ECX,200H
        SUB     ECX,40000H
        INC     ECX
        CLC                             ;
        SHR     ECX,10                  ;NO. OF K
        MOV     EBX,VSPTR               ;GET VS PTR
        SUB     EBX,8                   ;
        MOV     WORD PTR [EBX],0        ;EXP = 0
        MOV     DWORD PTR [EBX+2],ECX   ;SAVE SPACE VALUE ( IN K )
        MOV     WORD PTR [EBX+6],0
        MOV     VSPTR,EBX               ;UPDATE VSPTR
        CMP     EBX,@@VSEND             ;OVER THE LIMIT?
        JB      ERROR_02                ;BR, IF ERROR
        RET

SPACES:
        MOV     EAX,DWORD PTR [BASE]
        MOV     EBX,256*1024
        SUB     EBX,EAX
        SUB     EBX,200H
        MOV     EAX,ENDPART
        MOV     ECX,ENDMEMX
        SUB     ECX,EAX
        SUB     ECX,EBX
        INC     ECX
        MOV     EBX,VSPTR               ;GET VS PTR
        SUB     EBX,8                   ;
        MOV     WORD PTR [EBX],0        ;EXP = 0
        MOV     DWORD PTR [EBX+2],ECX   ;SAVE SPACE VALUE ( IN K )
        MOV     WORD PTR [EBX+6],0
        MOV     VSPTR,EBX               ;UPDATE VSPTR
        CMP     EBX,@@VSEND             ;OVER THE LIMIT?
        JB      ERROR_02                ;BR, IF ERROR
        MOV     R0,0
        RET

;     FNC#PI: PROCESS '#PI' FUNCTION
;
;          RETURN:  #PI IN VS (EXECUTION PHASE)
;

FNC_PI:
        TEST    SL,2                    ;BRANCH IF NOT EXECUTION PHASE
        JZ      FNCFOUND
        MOV     EBX,VSPTR               ;GET VS PTR
        SUB     EBX,8                   ;
        MOV     VSPTR,EBX               ;UPDATE VSPTR
        MOV     EAX,DWORD PTR _MPI
        MOV     DWORD PTR [EBX],EAX
        MOV     EAX,DWORD PTR _MPI+4
        MOV     DWORD PTR [EBX+4],EAX
        MOV     VSPTR,EBX
        CMP     EBX,@@VSEND             ;OVER THE LIMIT?
        JB      ERROR_02                ;BR, IF ERROR
FNCFOUND:
        MOV     R0,0                    ;FLAG = FOUND
        RET

;     FNCERR: PROCESS AN 'ERR' FUNCTION
;
;          RETURN: LAST MATH OR I/O ERROR CODE
;                     (FL.PT.) --> VS IF EXECUTION PHASE
;
;          NOTE: ERR IS SET TO ZERO BY:
;                    1. REFERENCING ERR
;                    2. CLEAR COMMAND
;                    3. RUN COMMAND
;                    4. MASTER INITIALIZATION
;
;                ERR IS SET TO AN ERROR CODE WHENEVER
;                    AN ERROR OCCURS.

FNCERR:
        TEST    SL,2                    ;BRANCH IF NOT EXECUTION PAHSE
        JZ      FNCFOUND
        MOV     EDX,CPBASE
        LEA     EDX,[EDX].@ERRF         ;
        MOV     CH,BYTE PTR [EDX]       ;
        MOV     BYTE PTR [EDX],0        ;ERR =0
        MOV     AL,CH
        AND     AL,0F0H
        SHR     AL,4
        MOV     BL,10
        MUL     BL
        XOR     BX,BX
        MOV     BL,CH                   ;CONVERT ERROR CODE TO BINARY
        AND     BL,0FH
        ADD     AX,BX
        MOV     WORD PTR [R0],AX
        CALL    BINFLTA                 ;CONVERT ERROR CODE TO FL. PT.
        JMP     FNCFOUND

;     FNCEND: PROCESS AN 'SPACE' FUNCTION
;
;          RETURN: FREE SPACE VALUE (BETWEEN TEXT & VS)
;                        IN VS IF EXECUTION PHASE

FNCEND:
        TEST    SL,2            ;EXIT IF NOT EXECUTION PHASE
        JZ      FNCFOUND
        MOV     EDX,VSPTR       ;FIND VALUE STACK SPACE
ENDSP:
        MOV     R7,0            ;DEFAULT SIGN = '+'
        MOV     EAX,EDX         ;
        MOV     EBX,0           ;DEFAULT SIGN = '+'
        SUB     EAX,MINVS+MINBUF+4
        SUB     EAX,TEXTEND     ;
        JS      ENDSP3          ;IF NEGATIVE
ENDSP4:
        MOV     EDX,VSPTR               ;PUT VALUE INTO VS
        SUB     EDX,8                   ;
        MOV     WORD PTR [EDX],0        ;EXP = 0
        MOV     DWORD PTR [EDX+2],EAX   ;SPACE VALUE
        MOV     WORD PTR [EDX+6],BX     ;FIRST TWO BYTE MANTISSA
        MOV     VSPTR,EDX               ;
        CMP     EDX,DWORD PTR [@@VSEND]
        JB      ERROR_02
        RET

ENDSP3:
        NEG     EAX             ;2' COMP
        MOV     EBX,-1          ;SIGN = '-'
        JMP     ENDSP4

;     FNCMOD:  PROCESS A MOD(X,Y) FUNCTION
;              MOD(X,Y) = X-Y;INT(X/Y)
;
;          RETURN: MOD(X,Y) --> VS (EXECUTION PHASE)

FNCMOD:
        CALL    EXPR            ;PROCESS EXPRESSION (X)
        cmp     byte ptr [textptr],','  ;LOOK FOR COMMA
        jne     error_13
        inc     textptr
        CALL    EXPR            ;PROCESS EXPRESSION (Y)
        cmp     byte ptr [textptr],')'   ;LOOK FOR ')'
        jne     error_11
        inc     textptr
        MOV     R0,0
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      FUNC5
        MOV     EDX,VSPTR       ;PC'S POINTS TO VALUES IN VS
        CALL    MATHMOD         ;CALCULATE X MOD Y
        JMP     EVAL3

;     FNCRD:  PROCESS A ROUND(X,N) FUNCTION
;
;          RETURN: ROUNDED X --> VS (EXECUTION PHASE)
;                  IF N>0, X ROUNDED TO NTH DECIMAL PLACE
;                  IF N=0, X ROUNDED TO INTEGER VALUE
;                  IF N<0, X ROUNDED TO N PLACES TO LEFT
;                          OF DECIMAL POINT

FNCRD:
        cmp     byte ptr [textptr],'('  ;LOOK FOR '('
        jne     error_10
        inc     textptr
        CALL    EXPR            ;PROCESS EXPRESSION (X)
        cmp     byte ptr [textptr],','  ;LOOK FOR COMMA
        jne     error_13
        inc     textptr
        CALL    EXPR            ;PROCESS EXPRESSION (N)
        cmp     byte ptr [textptr],')'   ;LOOK FOR ')'
        jne     error_11
        inc     textptr
        MOV     R0,0
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      FUNC5
        MOV     EDX,VSPTR       ;PC'S POINTS TO VALUES IN VS
        CALL    MATHRD          ;ROUND X
        JMP     EVAL3

;     LEN: PROCESS A 'LEN' FUNCTION
;
;          FORMAT:  LEN( alpha-variable )
;
;          ENTRY: TEXTPTR POINTS AFTER '('
;                 TEXTLPTR POINTS TO CURRENT TEXT LINE
;
;          MODIFIED REGISTERS: ALL, AUX0,1,2

LEN:
        MOV     K,4             ;FLAG = RETURN CURRENT LENGTH OF ALPHA VARIABLE
        CALL    AVAR            ;SCAN & PROCESS ALPHA VARIABLE
        TEST    R0,0FH          ;ERROR IF NOT FOUND
        JNZ     ERR_29B
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      LEN1
        mov     edx,vsptr       ;GET CURRENT LENGTH OF ALPHA VARIABLE
        mov     ax,word ptr [edx]
        sub     vsptr,2
        mov     edx,vsptr
        mov     word ptr [edx],0
        mov     word ptr [edx+2],ax
        mov     dword ptr [edx+4],0
        cmp     edx,dword ptr [@@vsend]
        jb      error_02
LEN1:
        cmp     byte ptr [textptr],')'      ;LOOK FOR ')' ; R0 = 0 ; AND EXIT
        jne     error_11
        inc     textptr
        mov     r0,0
        ret

ERR_29B:
        JMP     ERROR_29        ;MISSING ALPHA VARIABLE

;     FNCMAX: PROCESS A MAX(X,Y,...,Z) FUNCTION
;             RETURNS THE LARGEST VALUE
;
;          RETURN: MAXIMUM VALUE OF X,Y,...,Z --> VS
;                  (EXECUTION PHASE)

FNCMAX:
        CALL    MXMNARGX        ;PROCESS EXPRESSION (X) OR ARRAY
FNCMAX2:
        cmp     byte ptr [textptr],','      ;LOOK FOR COMMA
        jne     fncmax1         ;BRANCH IF NOT FOUND
        inc     textptr
        CALL    MXMNARGX        ;PROCESS NEXT EXPRESSION OR ARRAY
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      FNCMAX2
        MOV     EDX,VSPTR       ;PC'S POINT TO VALUES IN VS
        CALL    MATHMAX         ;GET MAX. VALUE OF 2 VALUES IN VS
        MOV     VSPTR,EDX       ;VSPTR POINTS TO MAX. VALUE (REMOVE THE OTHER)
        JMP     FNCMAX2
FNCMAX1:
        cmp     byte ptr [textptr],')'  ;LOOK FOR ')'
        jne     error_11
        inc     textptr
        mov     r0,0
        ret

;     FNCMIN: PROCESS A MIN(X,Y,...,Z) FUNCTION
;             RETURNS THE SMALLEST VALUE
;
;          RETURN: MINIMUM VALUE OF X,Y,...,Z --> VS
;                  (EXECUTION PHASE)

FNCMIN:
        MOV     K,1             ;FUNCTION = MIN
        CALL    MXMNARG         ;PROCESS EXPRESSION (X) OR ARRAY
FNCMIN2:
        cmp     byte ptr [textptr],','      ;LOOK FOR COMMA
        jne     fncmax1         ;BRANCH IF NOT FOUND
        inc     textptr
        MOV     K,1             ;FUNCTION = MIN
        CALL    MXMNARG         ;PROCESS NEXT EXPRESSION OR ARRAY
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      FNCMIN2
        MOV     EDX,VSPTR       ;PC'S POINT TO VALUE OF 2 VALUES IN VS
        CALL    MATHMIN         ;GET MIN. VALUE OF 2 VALUES IN VS
        MOV     VSPTR,EDX       ;VSPTR POINTS TO MIN. VALUE (REMOVE THE OTHER)
        JMP     FNCMIN2

;     MXMNARG: PROCESS MAX() OR MIN() FUNCTION ARGUMENT.
;              THE VALUE OF THE EXPRESSION OR MAX.(MIN.)
;              ELEMENT OF THE ARRAY ==> VALUE STACK.
;
;          FORMAT:  EXPRESSION
;                   NUMERIC ARRAY
;
;          ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;                 IN EXECUTION PHASE,
;                    K = 0 IF MAX. FUNCTION
;                        1 IF MIN. FUNCTION
;
;          RETURN: TEXTPTR POINTS AFTER ARGUMENT
;               IN EXECUTION PHASE,
;                  VALUE OF EXPRESSION OR LARGEST ARRAY
;                     ELEMENT ==> VALUE STACK
;
;          MODIFIED REGISTERS: ALL,AUX0,1,2
;
;     MXMNARGX: SAME EXCEPT ENTRY FLAG IS SET TO MAX.
;               FUNCTION BY THIS ENTRY.

MXMNARGX:
        MOV     K,0             ;FUNCTION = MAX
MXMNARG:
        CALL    ARRAY           ;LOOK FOR ARRAY NAME
        TEST    R0,0FH
        JZ      MXMNARGXX
        JMP     EXPR            ;SCAN &EVALUATE EXPRESSION

;Corrected in Rel 2.0 to properly check for alpha
;

MXMNARGXX:
        TEST    R5,1            ;ERROR IF ALPHA ARRAY
        JNZ     ERROR_74
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      MXMNARG1
        MOV     AX,WORD PTR R2  ;SAVE NO. OF ELEMENT
        MOV     WORD PTR R5,AX
        PUSH    EDX
        MOV     EAX,VSPTR       ;1ST ELEMENT TO VS
        SUB     EAX,8
        MOV     EBX,DWORD PTR [EDX]
        MOV     DWORD PTR [EAX],EBX
        MOV     EBX,DWORD PTR [EDX+4]
        MOV     DWORD PTR [EAX+4],EBX
        MOV     VSPTR,EAX
        CMP     EAX,DWORD PTR [@@VSEND]
        JB      ERROR_02
        JMP     MXMNARG4

MXMNARG2:
        POP     EDX
        ADD     EDX,8                   ;POINT TO NEXT ELEMENT
        PUSH    EDX
        MOV     EAX,VSPTR               ;PUT VALUE INTO VS
        SUB     EAX,8                   ;
        MOV     EBX,DWORD PTR [EDX]
        MOV     DWORD PTR [EAX],EBX
        MOV     EBX,DWORD PTR [EDX+4]
        MOV     DWORD PTR [EAX+4],EBX
        MOV     VSPTR,EAX
        CMP     EAX,@@VSEND             ;CHECK LIMIT?
        JB      ERROR_02                ;BR, IF ERROR
        MOV     EDX,EAX

        CMP     K,1                     ;BR, IF MIN() FUNCTION
        JE      MXMNARGY                ;
        CALL    MATHMAX                 ;FIND MAX 0F 2 ELEMENTS IN VALUE STACK
        JMP     MXMNARGZ

MXMNARGY:
        CALL    MATHMIN                 ;FIND MIN 0F 2 ELEMENTS IN VALUE STACK

MXMNARGZ:
        MOV     VSPTR,EDX               ;POINT TO MAX. VALUE

MXMNARG4:
        DEC     WORD PTR R5             ;DECREMENT ELEMENT COUNT
        JNZ     MXMNARG2                ;BR, IF MORE ARGUMENTS
        POP     EDX

MXMNARG1:
        RET

ERROR_74:
        MOV     R0,74H                  ;WRONG VARIABLE TYPE
        JMP     ERROR

; VAL: Process a 'VAL' function                                               ;
;                                                                             ;
; FORMAT:  VAL( alpha [,digit] )           where 0<digit<5                    ;
;                                                                             ;
; ENTRY: TEXTPTR points after '('                                             ;
;                                                                             ;
; MODIFIED REGISTERS: all, AUX0,1,2                                           ;
;-----------------------------------------------------------------------------;

;================== NOTE =======================================================

;        This new VAL routine was added in 3.0.  VAL was enhanced to support
;        3 and 4 byte values.
;        This routine was lifted from BASIC-3.  However, the number of bytes
;        handled was increased from 3 to 4 and the BINVALN routine added.
;
;===============================================================================

VAL:
        CALL    ALPHA                   ;GET ALPHA VARIABLE OR LITERAL
        TEST    R0,0FH                  ;NF, ERROR
        JNZ     ERR_29J

;======== handle length ========================================================

;for 3.4 +++
        mov     bx,1                    ;defalut length 1
        cmp     byte ptr [textptr],','
        jne     valx
        inc     textptr
        test    mode,0800h
        jnz     val_vvv                 ;fix for listv
        test    sl,03h
        jz      val_kkk
val_vvv:
        mov     ch,byte ptr [textptr]
        cmp     ch,30h
        jb      val_kkk
        cmp     ch,39h
        ja      val_kkk
        inc     textptr
        and     ch,0fh
        mov     bl,ch
        test    sl,01h
        jnz     val4
        jmp     val_yyy

val_kkk:
        call    expr
        test    sl,xphase
        jz      val4
        call    fltbina                 ;convert length to binary value
        mov     bx,word ptr [r0]
val_yyy:
        cmp     bx,01h                  ;check range for 1 to 6
        jb      error_34
        cmp     bx,06h
        ja      error_34
;for 3.4 ---

valx:   TEST    SL,XPHASE               ;EXIT IF NOT EXECUTION PHASE
        JZ      VAL4
        MOV     EDX,VSPTR               ;GET ADDR. AND LENGTH
        MOV     AX,WORD PTR [EDX]       ;ax = value length
        MOV     EDX,DWORD PTR [EDX+2]   ;EDX = value ADDRESS
        ADD     VSPTR,6
        cmp     ax,bx                   ;ERROR IF TOO SHORT
        jb      error_51

;======== read string into registers ===========================================

VALA:
        cmp     bl,1                    ;SKIP IF ONE BYTE VALUE
        je      val9
        cmp     bl,2                    ;SKIP IF TWO BYTE VALUE
        je      val8
        cmp     bl,3                    ;SKIP IF THREE BYTE VALUE
        je      val7
        cmp     bl,4                    ;SKIP IF FOUR BYTE VALUE
        je      val6
        cmp     bl,5                    ;SKIP IF FIVE BYTE VALUE
        je      val5

        mov     ax,word ptr [edx]       ;read six byte value
        xchg    ah,al
        mov     bx,word ptr [edx+2]
        xchg    bh,bl
        mov     cx,word ptr [edx+4]
        xchg    ch,cl
        mov     esi,vsptr
        sub     esi,8
        mov     word ptr [esi],0
        mov     word ptr [esi+2],cx
        mov     word ptr [esi+4],bx
        mov     word ptr [esi+6],ax
        mov     vsptr,esi
        cmp     esi,dword ptr [@@vsend]
        jb      error_02
        cmp     byte ptr [textptr],')'  ;look for ')'
        jne     error_11
        inc     textptr
        mov     r0,0
        ret
VAL5:
        mov     al,byte ptr [edx]       ;read five byte value
        mov     bx,word ptr [edx+1]
        xchg    bh,bl
        mov     cx,word ptr [edx+3]
        xchg    ch,cl
        mov     esi,vsptr
        sub     esi,8
        mov     word ptr [esi],0
        mov     dword ptr [esi+4],0
        mov     word ptr [esi+2],cx
        mov     word ptr [esi+4],bx
        mov     byte ptr [esi+6],al
        mov     vsptr,esi
        cmp     esi,dword ptr [@@vsend]
        jb      error_02
        cmp     byte ptr [textptr],')'  ;look for ')'
        jne     error_11
        inc     textptr
        mov     r0,0
        ret
VAL6:
        mov     bx,word ptr [edx]       ;read four byte value
        xchg    bh,bl
        mov     cx,word ptr [edx+2]
        xchg    ch,cl
        mov     esi,vsptr
        sub     esi,8
        mov     word ptr [esi],0
        mov     dword ptr [esi+4],0
        mov     word ptr [esi+2],cx
        mov     word ptr [esi+4],bx
        mov     vsptr,esi
        cmp     esi,dword ptr [@@vsend]
        jb      error_02
        cmp     byte ptr [textptr],')'  ;look for ')'
        jne     error_11
        inc     textptr
        mov     r0,0
        ret
VAL7:
        mov     bl,byte ptr [edx]       ;read three byte value
        mov     cx,word ptr [edx+1]
        xchg    ch,cl
        mov     esi,vsptr
        sub     esi,8
        mov     word ptr [esi],0
        mov     dword ptr [esi+4],0
        mov     word ptr [esi+2],cx
        mov     byte ptr [esi+4],bl
        mov     vsptr,esi
        cmp     esi,dword ptr [@@vsend]
        jb      error_02
        cmp     byte ptr [textptr],')'  ;look for ')'
        jne     error_11
        inc     textptr
        mov     r0,0
        ret
VAL8:
        mov     cx,word ptr [edx]       ;read two byte value
        xchg    ch,cl
        mov     esi,vsptr
        sub     esi,8
        mov     word ptr [esi],0
        mov     dword ptr [esi+4],0
        mov     word ptr [esi+2],cx
        mov     vsptr,esi
        cmp     esi,dword ptr [@@vsend]
        jb      error_02
        cmp     byte ptr [textptr],')'  ;look for ')'
        jne     error_11
        inc     textptr
        mov     r0,0
        ret

VAL9:
        mov     ch,byte ptr [edx]       ;read one byte value
        mov     esi,vsptr
        sub     esi,8
        mov     dword ptr [esi],0
        mov     dword ptr [esi+4],0
        mov     byte ptr [esi+2],ch
        mov     vsptr,esi
        cmp     esi,dword ptr [@@vsend]
        jb      error_02

VAL4:
        cmp     byte ptr [textptr],')'  ;look for ')'
        jne     error_11
        inc     textptr
        mov     r0,0
        ret

;======== errors ===============================================================

ERROR_51:
        MOV     R0,51H                  ;VALUE TOO SHORT
        JMP     ERROR

ERR_29J:
        JMP     ERROR_29                ;MISSING ALPHA VARIABLE


;     NUM: PROCESS A 'NUM' FUNCTION
;
;          FORMAT: NUM( alpha-variable )
;
;          ENTRY: TEXTPTR POINTS AFTER 'NUM('
;                 TEXTLPTR POINTS TO CURRENT TEXT LINE

NUM:
        CALL    AVARD                   ;SCAN &PROCESS ALPHA VARIABLE
        TEST    R0,0FH
        JZ      NUMX1
        JMP     ERROR_29                ;NF, ERROR

NUMX1:  TEST    SL,2                    ;BRANCH IF NOT EXECUTION PHASE
        JZ      NUM1
        MOV     EDX,VSPTR               ;GET LENGTH (R1,0) AND ADDR. (PC'S) OF ALPHA VALUE
        MOV     AX,WORD PTR [EDX]
        MOV     WORD PTR [R0],AX
        MOV     EDX,DWORD PTR [EDX+2]
        ADD     VSPTR,6
        MOV     R4,0                    ;DIGIT COUNT = 0
        MOV     R7,0FFH                 ;R7 = EXPONENT (-1)
        MOV     WORD PTR [R2],0         ;NUM COUNT = 0

;     REGISTER USE:
;          F6 = LEGAL CHARACTERS (SPACES ALWAYS LEGAL)
;            ,BIT0 =1 IF DIGIT LEGAL
;            ,BIT1 =1 IF DECIMAL POINT (.) LEGAL
;            ,BIT2 =1 IF SIGN (+,-) LEGAL
;            ,BIT3 =1 IF 'E' LEGAL
;          PC'S POINT TO NEXT CHARACTER OF VALUE
;          F3,2 = NUM COUNT (NO. OF LEGAL CHAR'S FOUND)
;          F1,0 = LENGTH OF VALUE
;          F4,LOW 4BITS = DIGIT COUNT OF MANTISSA
;            ,BIT80 = 1 IF ANY MANTISSA DIGIT FOUND
;          F7 = EXPONENT (BINARY)
;          F5 = 'NUMCH' RETURN FLAG
;             = 0 IF NO LEGAL CHARACTER FOUND
;             = 1 IF DIGIT FOUND
;             = 2 IF DECIMAL POINT FOUND
;             = 4 IF SIGN FOUND
;             = 8 IF 'E' FOUND


; PROCESS MANTISSA
;

        MOV     R6,0FH          ;LEGAL CHARS.:DIGIT,SIGN,.,E,SPACE
        JMP     NUM5
NUM7:
        MOV     AL,R5           ;MAKE FOUND CHAR ILLEGAL
        XOR     R6,AL
NUM11:
        AND     R6,0BH          ;MAKE SIGN ILLEGAL CHAR
NUM5:
        CALL    NUMCH           ;LOOK FOR LEGAL CHAR IN VALUE
        CMP     R5,1            ;BR, IF DIGIT NOT FOUND
        JNE     NUM3
;MANTISSA DIGIT
;
        OR      R4,80H          ;FLAG = DIGIT FOUND
        TEST    R4,0FH          ;BRANCH IF SIGNIFICANT DIGIT ALREADY FOUND
        JNZ     NUM12
        TEST    CH,0FH          ;BRANCH IF NOT LEADING ZERO
        JNZ     NUM12

;     LEADING ZERO
;
        TEST    R6,2            ;BRANCH IF '.' NOT FOUND
        JNZ     NUM11
        DEC     R7              ;DECREMENT EXPONENT
        TEST    R7,0F0H         ;BRANCH IF EXP > -240
        JNZ     NUM11
        JMP     NUM10

;     SIGNIFICANT DIGIT
;

NUM12:
        TEST    R6,2            ;BRANCH IF '.' FOUND
        JZ      NUMX2
        INC     R7              ;INCREMENT EXPONENT

NUMX2:  INC     R4              ;INCREMENT COUNT
        MOV     AL,R4
        AND     AL,0FH
        CMP     AL,0DH          ;BRANCH IF NO. OF DIGITS = 13
        JE      NUM7
        JMP     NUM11
NUM3:
        CMP     R5,0            ;BR, IF NO LEGAL CHAR FOUND
        JE      NUM2
        CMP     R5,8            ;BR, IF 'DEC PT' OR 'SIGN'
        JNE     NUM7

; PROCESS EXPONENT (E FOUND)
;
        CMP     R4,0            ;BRANCH IF NO MANTISSA DIGITS FOUND
        JE      NUM10
        MOV     AL,R2
        MOV     K,AL
        MOV     AL,R3
        MOV     R4,AL           ;SAVE NUM-COUNT

;EXPONENT SIGN
;
        MOV     R6,5            ;LEGAL CHARS.:SIGN,SPACE
        CALL    NUMCH           ;LOOK FOR SIGN
        CMP     R5,4            ;BR, IF NOT SIGN
        JNE     NUMX3
        CMP     CH,2DH          ;BR, IF '-'
        JE      NUM18

NUMX3:  TEST    R7,0F0H         ;BRANCH IF IMPLIED EXP >= 0
        JZ      NUM6
NUM17:
        MOV     R7,0            ;IMPLIED EXP = 0 (CANNOT CAUSE OVER/UNDERFLOW)
        JMP     NUM6
NUM18:
        TEST    R7,0F0H         ;BRANCH IF IMPLIED EXP < 0
        JNZ     NUM6
        JMP     NUM17

;1ST EXPONENT DIGIT
;
NUM6:
        CMP     R5,1            ;BR, IF DIGIT FOUND
        JE      NUM4            ;
        CMP     R5,4            ;BR, IF LAST CHAR <> SIGN
        JNE     NUM16
        MOV     R6,1            ;LEGAL CHARS: DIGIT,SPACE
        CALL    NUMCH           ;LOOK FOR LEGAL CHAR IN VALUE
        CMP     R5,1            ;BR, IF DIGIT FOUND
        JE      NUM4
NUM16:                          ;CARRY OFF
        MOV     AL,K            ;R4,0 = SAVE NUM - COUNT -1
        ADD     AL,0FFH
        MOV     R2,AL
        MOV     AL,R4           ;
        ADC     AL,0FFH
        MOV     R3,AL
        JMP     NUM15

;2ND MANTISSA DIGIT
;

NUM4:
        MOV     R4,CH           ;R4 = 1ST EXPONENT DIGIT
        AND     R4,0FH
        JMP     NUM14
NUM13:
        MOV     AL,R4
        AND     AL,0FH
        MOV     BL,0AH
        MUL     BL
        AND     AL,0FH
        OR      R4,AL
        MOV     K,CH
        OR      K,0FH
        MOV     AL,K
        ADD     R4,AL                   ;R4 = EXPONENT
NUM14:
        CALL    NUMCH                   ;LOOK FOR LEGAL CHAR IN VALUE
        MOV     R6,0                    ;LEGAL CHAR: SPACE
        TEST    R5,0FH                  ;BRANCH IF DIGIT FOUND
        JNZ     NUM13
        JMP     NUM9

;ADJUST & CHECK EXPONENT
;

NUM15:
        MOV     R4,0                   ;EXPLICIT EXPONENT = 0
NUM9:
        TEST    R7,0F0H
        JZ      NUMX4
        NEG     R7              ;(TWO'S COMPLEMENT.) IMPLIED EXP = ABS VALUE
        INC     EDX

NUMX4:  MOV     AL,R4           ;ADD EXPLICIT EXP TO IMPLIED EXP
        ADD     R7,AL
        JC      NUM10           ;BRANCH IF EXP > 255
        MOV     R4,63H
        MOV     AL,R4
        CMP     AL,R7           ;BRANCH IF EXP > 99
        JB      NUM10

; CONVERT NUM COUNT TO FL.PT.

        MOV     AX,WORD PTR [R2]        ;R1,0 = NUM - COUNT
        MOV     WORD PTR [R0],AX
NUM8:
        CALL    BINFLT                  ;CONVERT NUM-COUNT TO FL. PT.
NUM1:
        CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')', R0 = 0 , EXIT
        JNE     ERROR_11
        INC     TEXTPTR
        MOV     R0,0
        RET

NUM2:
        CMP     R4,0                    ;BRANCH IF DIGIT COUNT NOT 0
        JNE     NUM15
NUM10:
        MOV     WORD PTR [R0],0           ;NUM-COUNT = 0
        JMP     NUM8


;  VER:  PROCESS A VER FUNCTION
;
;  FORMAT:     VER ( alpha, alpha )
;    First alpha string (variable, literal) contains
;       data to be verified.
;    Second contains the format specification.
;
;  ENTRY:  TEXTPTR POINTS AFTER '('.
;          TEXTLPTR POINTS TO CURRENT TEXT LINE.
;  MODIFIED REGISTERS:  ALL, EXPR AUX'S

FNCVER:
        CALL    ALPHA           ;FIND FIRST ALPHA
        TEST    R0,0FH
        JZ      FNCVERX

VER_29: JMP     ERROR_29        ;ERROR IF NOT FOUND

FNCVERX:CMP     BYTE PTR [TEXTPTR],','
        JNE     ERROR_13
        INC     TEXTPTR
        CALL    ALPHA
        TEST    R0,0FH          ;ERROR IF NOT FOUND
        JNZ     VER_29
        TEST    SL,2            ;EXIT IF NOT X PHASE
        JZ      VER95
        MOV     EDX,VSPTR
        MOV     AX,WORD PTR [EDX]
        MOV     WORD PTR [R0],AX
        MOV     EDX,DWORD PTR [EDX+2]
        ADD     VSPTR,6
        MOV     AUX02,EDX         ;AUX 2 = ADDR OF 2ND ARGUMENT
        MOV     WORD PTR [R4],AX  ;R5R4 = LENGTH OF 2ND ARGUMENT
        MOV     EDX,VSPTR
        MOV     AX,WORD PTR [EDX]
        MOV     WORD PTR [R0],AX
        MOV     EDX,DWORD PTR [EDX+2]  ;PC'S = ADDRESS OF FIRST ARGUMENT
        ADD     VSPTR,6
        CMP     WORD PTR [R4],AX
        JBE     FNCVERY
        MOV     WORD PTR [R4],AX       ;R5R4 = MIN. OF THE TWO ARGUMENT LENGTHS

FNCVERY:MOV     WORD PTR [R0],0        ;R1R0 = ZERO ; INITIAL COUNT
        JMP     VER20

;  REGISTER USAGE:
;   AUX 2  = ADDRESS POINTER
;     CH   = FORMAT CHARACTER
;     F7   = DATA CHARACTER  <--------- NOW USE AH AS F7 8/25/88
;     F6   = SCRATCH REGISTER
;     F5F4 = TOTAL LENGTH TO CHECK
;     F1F0 = COUNT (RESULT) OF VALID BYTES

VER15:
        INC     WORD PTR [R0]     ;ADD ONE TO VALID BYTE COUNT
        MOV     AX,WORD PTR [R4]  ;EXIT IF ENTER LENGTH HAS BEEN CHECKED
        CMP     AX,WORD PTR [R0]
        JBE     VER90
VER20:
        MOV     CH,BYTE PTR [EDX]
        XCHG    EDX,AUX02
        INC     AUX02
        MOV     R7,CH              ;AH = DATA BYTE, TURN OFF CARRY
        MOV     CH,BYTE PTR [EDX]  ;CH = FORMAT CHAR.
        XCHG    EDX,AUX02
        INC     AUX02

        MOV     AL,CH
        AND     AL,0F0H
        CMP     AL,040H
        JNE     VER25
        CMP     CH,41H          ;SKIP IF FORMAT = A (41)
        JE      VER50
        CMP     CH,48H          ;SKIP IF FORMAT = H (48)
        JE      VER55
        CMP     CH,4EH          ;SKIP IF FORMAT = N (4E)
        JE      VER45
        JMP     VER80

VER25:
        CMP     AL,50H
        JNE     VER30
        CMP     CH,50H          ;SKIP IF FORMAT = P (50)
        JE      VER65
        CMP     CH,58H          ;SKIP IF FORMAT = X (58)
        JE      VER15
        JMP     VER80

VER30:
        CMP     AL,20H
        JNE     VER80
        CMP     CH,23H          ;SKIP IF FORMAT = # (23)
        JE      VER40
        CMP     CH,2BH          ;SKIP IF FORMAT <> + (2B)
        JNE     VER80

;    + - CHECK FOR SIGN (+ - or blank)

        CMP     CH,R7           ;OKAY IF CHAR = +
        JE      VER15
        MOV     AL,R7
        AND     AL,0F0H
        CMP     AL,020H
        JNE     VER90
        CMP     R7,2DH          ;OKAY IF CHAR = -
        JE      VER15
        CMP     R7,20H
        JE      VER15
        JMP     VER90

;   # - CHECK FOR NUMERIC

VER40:
        MOV     AL,R7
        AND     AL,0F0H
        CMP     AL,030H
        JNE     VER90
VER41:
        MOV     AL,R7
        ADD     AL,0C6H
VER42:
        JNC     VER15           ;OK IF 30-39
        JMP     VER90           ;EXIT, NOT A DIGIT

;   N - NUMERIC OR ALPHA

VER45:
        MOV     AL,R7           ;IF 30-3F, CHECK FOR NUMERIC
        AND     AL,0F0H         ;ELSE, CHECK ALPHABETIC
        CMP     AL,030H
        JE      VER41

;   A - ALPHABETIC

VER50:
        AND     R7,0FFH-20H     ;IGNORE 20 BIT OF CHAR.
        MOV     AL,R7
        AND     AL,0F0H
        CMP     AL,050H
        JE      VER52
        CMP     AL,40H
        JNE     VER90
        CMP     R7,40H          ;41-4F AND 61-6F ARE OKAY
        JNE     VER15
        JMP     VER90

VER52:
        MOV     AL,R7
        ADD     AL,0A5H
        JNC     VER15           ;50-5A AND 70-7A ARE OK
        JMP     VER42

;   H - HEXADECIMAL

VER55:
        MOV     AL,R7           ;IF 30-3F CHECK NUMERIC
        AND     AL,0F0H
        CMP     AL,030H
        JE      VER41
        CMP     AL,040H
        JNE     VER90
        CMP     R7,40H
        JE      VER90
        MOV     AL,R7
        ADD     AL,0B9H
        JMP     VER42

;   P - PACKED DECIMAL

VER65:
        MOV     AL,R7           ;Checkpoint.
        ADD     AL,0
        DAA
        MOV     R6,AL
        CMP     AL,R7           ;OKAY IF BOTH DIGITS ARE 0-9
        JE      VER15
        JMP     VER90

;   ELSE - CHECK SPECIFIC CHARACTER

VER80:
        CMP     CH,R7           ;OKAY IF DATA CHAR EQUALS FORMAT CHAR
        JE      VER15

;   END OF VERIFY

VER90:
        CALL    BINFLT                      ;CONVERT TO FL. PT. VALUE
VER95:
        CMP     BYTE PTR [TEXTPTR],')'      ;GET RIGHT PARAMETERS, AND EXIT
        JNE     ERROR_11
        INC     TEXTPTR
        MOV     R0,0
        RET


;     POS: PROCESS A 'POS' FUNCTION
;
;                                          literal
;          FORMAT: POS([-] alpha-var relop alpha-var )
;                                          hex hex
;
;               WHERE: relop ::=  <,<=,=,>=,>,<>
;
;          ENTRY: TEXTPTR POINTS AFTER 'POS('
;                 TEXTLPTR POINTS TO CURRENT TEXT LINE
;
;          MODIFIED REGISTERS: ALL,AUX0,1,2

POS:
        MOV     AL,1
        CMP     BYTE PTR [TEXTPTR],'-'      ;LOOK FOR '-'
        JNE     POSA
        INC     TEXTPTR
        MOV     AL,0
POSA:   TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      POSB
        SHL     AL,4            ;R0 = 0 IF '-' FOUND, ELSE = 10
        MOV     EDX,OSPTR       ;SAVE SEARCH DIRECTION IN OS
        INC     EDX
        CMP     EDX,DWORD [@@OSEND]
        JE      ERROR_04
        MOV     BYTE PTR [EDX],AL
        MOV     OSPTR,EDX

POSB:   CALL    ALPHA           ;SCAN & PROCESS ALPHA-VARIABLE
        TEST    R0,0FH          ;ERROR IF NOT FOUND
        JNZ     ERR_29B
        CALL    RELOP           ;LOOK FOR RELATION OPERATOR
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      POS1
        MOV     EDX,OSPTR
        MOV     CH,BYTE PTR [EDX]  ;CH = SEARCH DIRECTION
        MOV     AL,R1           ;SEARCH DIRECTION & RELATION OPERATOR IN OS
        OR      AL,CH
        MOV     BYTE PTR [EDX],AL
POS1:
        CALL    CHARCTR         ;SCAN & PROCESS ALPHA VAR. OR LITERAL OR HEX
                                ;PAIR (SEARCH -CHARACTER)
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      POS3
        MOV     AL,R0
        MOV     R6,AL
POS10:
        MOV     EDX,OSPTR
        MOV     CH,[EDX]        ;GET SEARCH DIRECTION & RELATION OPERATOR
        DEC     OSPTR
        MOV     R7,CH           ;R7 = SEARCH DIRECTION &RELATION OPERATOR
        MOV     EDX,VSPTR       ;GET LENGTH (R1,0) AND ADDR. (PC'S) OF SEARCH VAR
        MOV     AX,WORD PTR [EDX]
        MOV     WORD PTR [R0],AX
        MOV     EDX,DWORD PTR [EDX+2]
        ADD     VSPTR,6
        TEST    R7,0F0H         ;BRANCH IF SEARCH DIRECTION = '-' (BACKWARDS)
        JZ      POS4
        MOV     WORD PTR [R0],0   ;R3,2 = LENGTH OF VAR. TO SEARCH
        MOV     WORD PTR [R2],AX  ;R1,0 = 0 (POSITION)

;     FORWARD SEARCH (DIRECTION = + )
;
;          ENTRY: PC'S POINT TO BEGIN OF VAR. TO SEARCH
;                 F3,2 = LENGTH OF VARIABLE TO SEARCH
;                 F1,0 = 0 (POSITION)
;                 F6 = CHARACTER TO SEARCH FOR
;                 F7,LOW 4-BITS = RELATION OPERATOR
;                                 1 IF <
;                                 3 IF <=
;                                 2 IF =
;                                 6 IF >=
;                                 4 IF >
;                                 5 IF <>
;                F7,HIGH 4-BITS = 0 (DIRECTION = + )
;
POS5:
        INC     WORD PTR R0     ;INCREMENT POSITION
        MOV     CH,[EDX]        ;CH = NEXT CHAR. OF VALUE; INCREMENT PTR
        INC     EDX
        MOV     AL,R7
        AND     AL,1
        MOV     R5,AL
        CMP     CH,R6           ;BRANCH IF CHAR. < SEARCH CHAR.
        JB      POS6
        MOV     AL,R7
        AND     AL,4
        MOV     R5,AL
        CMP     CH,R6           ;BRANCH IF CHAR. > SEARCH CHAR.
        JNE     POS6
        MOV     AL,R7           ;CHAR. = SEARCH CHAR.
        AND     AL,2
        MOV     R5,AL
POS6:
        TEST    R5,0FH          ;BRANCH IF RELATION TRUE
        JNZ     POS7
        MOV     AX,WORD PTR [R0]  ;BRANCH IF NOT AT END OF VALUE
        CMP     AX,WORD PTR [R2]
        JB      POS5
        MOV     WORD PTR [R0],0   ;POSITION = 0 (SEARCH CHAR. NOT FOUND)
POS7:
        CALL    BINFLT          ;CONVERT BINARY POSITION TO FL. PT.
POS3:
        CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')', R0 = 0, EXIT
        JNE     ERROR_11
        INC     TEXTPTR
        MOV     R0,0
        RET

;     REVERSE SEARCH (DIRECTION = - )
;

POS4:
        XOR     EAX,EAX
        MOV     AX,WORD PTR R0  ;
        ADD     EDX,EAX         ;PC'S POINT TO END OF VALUE + 1
        MOV     R2,0FFH         ;R3,2 = FFFF (-1)
        DEC     EDX             ;PC'S POINT TO END OF VALUE
        MOV     AL,R2
        MOV     R3,AL

;          ENTRY: PC'S POINT TO END OF SEARCH VAR. +1
;                 F1,0 = LENGTH OF VALUE TO SEARCH
;                 F6 = CHARACTER SEARCHING FOR
;                 F7,LOW 4-BITS = RELATION OPERATOR
;                   ,HIGH 4-BITS = 1 (DIRECTION = + )
;                 F3,2 = FFFF (-1)

POS8:
        MOV     CH,[EDX]        ;CH = NEXT CHAR. OF VALUE
        DEC     EDX             ;DECREMENT PTR
        MOV     AL,R7
        AND     AL,1
        MOV     R5,AL           ;BRANCH IF CHAR. < SEARCH CHAR.
        CMP     CH,R6
        JB      POS9
        MOV     AL,R7
        AND     AL,4
        MOV     R5,AL
        CMP     CH,R6           ;BRANCH IF CHAR. > SEARCH CHAR.
        JNE     POS9
        MOV     AL,R7           ;CHAR = SEARCH CHAR
        AND     AL,2
        MOV     R5,AL
POS9:
        TEST    R5,0FH          ;BRANCH IF RELATION TRUE
        JNZ     POS7
        MOV     AX,WORD PTR [R2]  ;DECREMENT POSITION
        ADD     WORD PTR [R0],AX
        CMP     WORD PTR [R0],0   ;BRANCH IF AT BEGIN OF VALUE
        JNE     POS8
        JMP     POS7


;     FUNCFN: PROCESS A USER-DEFINED FUNCTION (FN)
;
;          FORMAT:  FN letter ( expression)
;
;          ENTRY: TEXTPTR POINTS AFTER 'FN'
;                 TEXTLPTR POINTS TO CURRENT TEXT LINE
;
;          DESCRIPTION:
;
;               TO PROCESS A USER-DEFINED FUNTION, THE
;               CORRESPONDING DEFFN STATEMENT IS FOUND.
;               AN ENTRY IS MADE TO THE DUMMY VARIABLE
;               TABLE (DVT) -- THE DUMMY VARIABLE IN THE
;               DEFFN STATEMENT IS ENTERED WITH THE VALUE
;               OF THE EXPRESSION IN THE FN FUNCTION.
;               THE EXPRESSION IN THE DEFFN STATEMENT IS
;               THEN EVALUATED USING THE VALUE OF THE
;               DUMMY VARIABLE WHEREVER THE DUMMY VARIABL
;               IS USED IN THE EXPRESSION ('VAR' LOGIC).
;               THE RESULT IS THE VALUE OF THE FN FUNC-
;               TION.  THE DUMMY VARIABLE IS THE REMOVED
;               FROM THE DUMMY VARIABLE TABLE.
;
;               NOTICE THAT DEFFN IS  RECURSIVE; A DEFFN
;               CAN REFER TO ANOTHER DEFFN (NESTING). THE
;               DUMMY VARIABLE TABLE PROVIDES FOR 5
;               LEVELS OF NESTING.  THE LAST DUMMY
;               VARIABLE ADDED TO THE DVT IS APPLICABLE
;               ONLY TO THE CURRENT DEFFN EVALUATION.  IT
;               IS REMOVED WHEN THE FUNCTION IS COMPLETED
;               AND ANOTHER IS ADDED WHEN ANOTHER FN
;               APPEARS IN THE DEFFN EVALUATION.
;
;          DUMMY VARIABLE TABLE (DVT):
;
;                    098E -- ENTRY 1  (12 BYTES) <--- SHOULD BE 10 BYTES
;                    0998 -- ENTRY 2
;                    09A2 -- ENTRY 3
;                    09AC -- ENTRY 4
;                    09B6 -- ENTRY 5
;
;               ENTRY = name letter (1 byte), name digit
;                       (high 1/2 of byte 2), value (8 b)
;
;               DVTPTR = POINTER TO NEXT AVAILABLE
;                        LOCATION IN DVT.
;                        (DVTPTR = 098E IF NO DUMMY VAR.)
;                        (DVTPTR = 09C0 IF DVT FULL)
;
;          CODING:
;               1. CODING IS DEPENDENT UPON EXACT
;                  LOCATION OF DVT.
;               2. TEXTPTR & TEXTLPTR TO FN STATEMENT
;                  ARE SAVED IN SR STACK (2 LEVELS).

FUNCFN:
        CALL    IMMEDERR        ;ERROR IF IMMEDIATE EXECUTION
        CALL    LETRDIGT        ;LOOK FOR LETTER OR DIGIT
        TEST    R0,0FH          ;ERROR IF NOT FOUND
        JNZ     ERR_16A
        MOV     DL,CH
        PUSH    EDX             ;SAVE FUNCTION LETTER OR DIGIT IN STACK
        CMP     BYTE PTR [TEXTPTR],'('      ;LOOK FOR '('
        JNE     ERROR_10
        INC     TEXTPTR
        CALL    EXPR            ;SCAN & EVALUATE FN EXPRESSION
        POP     EDX             ;GET FUNCTION LETTER OR DIGIT
        TEST    SL,3            ;BRANCH IF ENTRY PHASE
        JZ      FUNCFN2
        MOV     R7,DL           ;SAVE FUNCTION LETTER OR DIGIT
        PUSH    TEXTPTR         ;SAVE TEXTPTRS TO FN IN STACK
        MOV     EDX,TEXTLPTR
        PUSH    EDX

;     SCAN FOR REFERENCED FUNCTION DEFINITION (DEFFN)
;
;          Scan begins at 1st DEFFN in program. 'DEFFNADR
;          is a ptr to 1st user defined function; if no
;          DEFFN's, DEFFNADR =0.  DEFFNADR is set up by
;          DEFFN routine in resolution phase; set to zero
;          at begin of resolution by RUN.  If program not
;          resolved, search always starts at TEXTBEG.
;
        TEST    SL,2            ;BRANCH IF RESOLUTION PHASE
        JZ      FUNCFNX
        TEST    MODE,20H        ;SEARCH FORM BEG OF TEXT, IF PROG NOT RESLVD
        JZ      FUNCFN9

FUNCFNX:
        MOV     EDX,CPBASE      ;GET CURRENT BASE
        MOV     EDX,DWORD PTR [EDX].@IBASE      ;GET CALLING PART BASE
        LEA     EDX,[EDX].@DEFNADR              ;GET PTR TO 1ST DEFFNX
        MOV     ECX,DWORD PTR [EDX]
        CMP     ECX,0           ;BR, IF PTR HAS BEEN SETUP
        JNE     FUNCFN7

FUNCFN9:
        MOV     EDX,CPBASE      ;GET CURRENT BASE
        MOV     EDX,DWORD PTR [EDX].@IBASE      ;GET CALLING PART BASE
        LEA     EDX,[EDX].@TEXTBEG              ;EDX PT TO TEXT BEG

FUNCFN7:
        CALL    FINDFNC         ;SEARCH FOR DEFFNX
        OR      CH,CH           ;ERROR IF NOT FOUND
        JZ      ERROR_38
        TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
        JZ      FUNCFN8
        CMP     BYTE PTR [TEXTPTR],'('      ;LOOK FOR '('
        JNE     ERROR_10
        INC     TEXTPTR

        CMP     BYTE PTR [TEXTPTR],7FH
        JE      FUNCFNG
        JMP     FUNCFNY         ;BR, IF NOT A VAR

FUNCFNG:MOV     AX,WORD PTR [TEXTPTR+1]     ;LOOK FOR VAR. NAME
        MOV     WORD PTR [R2],AX
        TEST    R2,0FH          ;BRANCH IF NUMERIC SCALAR
        JZ      FUNCFNZ

FUNCFNY:
        JMP     ERROR_25        ;ERROR IF NOT NUMERIC SCALAR

FUNCFNZ:
        ADD     TEXTPTR,7
        MOV     EDX,CPBASE
        MOV     ECX,DWORD PTR [EDX].@DVTPTR     ;GET DUMMY VAR. TAB PTR
        LEA     EAX,[EDX].@DATAPTR      ;ADDR AFTER DUMMY VAR. TAB
        CMP     ECX,EAX         ;TABL FULL?
        JAE     ERROR_39        ;BR, ERROR
        MOV     EDX,ECX         ;PC'S POINT TO NEXT POSITION IN DVT
        MOV     AX,WORD PTR [R2];PUT VAR. NAME AND ATOM INTO DVT
        MOV     WORD PTR [EDX],AX
;       ADD     EDX,2
        MOV     EAX,VSPTR       ;PUT VALUE OF FN EXPRESSION INTO DVT
        MOV     EBX,DWORD PTR [EAX]
        MOV     DWORD PTR [EDX+2],EBX
        MOV     EBX,DWORD PTR [EAX+4]
        MOV     DWORD PTR [EDX+6],EBX
        ADD     VSPTR,8
;       ADD     EDX,8
        ADD     EDX,10
        MOV     EBX,CPBASE
        LEA     EBX,DWORD PTR [EBX].@DVTPTR       ;PT TO DVTPTR LOC
        MOV     DWORD PTR [EBX],EDX     ;NEW DVTPTR PT TO NEXT AVAIL ENTRY
        CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')' IN DEFFN STMT
        JNE     ERROR_11
        INC     TEXTPTR
        CMP     BYTE PTR [TEXTPTR],'='   ;LOOK FOR '='
        JNE     ERROR_12
        INC     TEXTPTR
        CALL    EXPR            ;EVALUATE DEFFN EXPRESSION

        MOV     EBX,CPBASE      ;
        LEA     EBX,DWORD PTR [EBX].@DVTPTR       ;PT TO DVTPTR LOC
        SUB     DWORD PTR [EBX],10      ;REMOVE DUMM VAR. FRM DVT (BY -10)

FUNCFN8:
        POP     TEXTLPTR        ;RESTORE TEXTPTRS TO FN
        POP     EDX
        MOV     TEXTPTR,EDX

FUNCFN2:
        CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')' IN FN, R0 = 0, EXIT
        JNE     ERROR_11
        INC     TEXTPTR
        MOV     R0,0
        RET

ERROR_39:
        MOV     R0,39H          ;ERROR = ILLEGAL FN USAGE
        JMP     FUNCFNW

ERROR_38:
        MOV     R0,38H          ;ERROR = UNDEFINED FN

FUNCFNW:
        POP     TEXTLPTR        ;RESTORE TEXTPTR TO FN
        POP     EDX
        MOV     TEXTPTR,EDX
        JMP     ERROR

ERR_16A:
        JMP     ERROR_16        ;MISSING LETTER


;FINDFNC:  find the specified DEFFNx statement in text.
;
;     The search for DEFFN now starts at DEFNADR in both
;     resolution and execution phases.
;
;     DEFNADR is set by a scan for the 1st DEFFN at the beginning of resolution
;     phase.  Previously, it was set when the 1st DEFFN was resolved.
;
;     It would be possible to use a 2-pass search for finding DEFFN's where
;     in pass 1 only the 1st statement in each line is checked.  If the DEFFN is
;     not found, a 2nd pass checks each statement on each line.  This scheme is not
;     yet implemented.
;
;     Entry:
;          phase = execution or resolution
;          PC's = ptr to thread of text line to start search from
;          F7 = function name (letter or digit); if F7=0, search stops at the
;               1st DEFFNx encountered.
;
;     Return:
;          CH = name letter or digit if found
;               Textptr points after name letter or digit in DEFFN
;
;          CH = 0 if DEFFNx not found
;
;     Modified registers: PC'S,CHCL,F0,TEXTPTR,TEXTLPTR

FINDFNC1:
        CALL    SCNCHSS         ;SCAN FOR STATEMENT SEPARATOR
        TEST    R0,0FH          ;BRANCH IF FOUND
        JZ      FINDFNC3
        MOV     EDX,TEXTLPTR    ;POINT TO THREAD OF LAST LINE

FINDFNC:
;       MOV     R0,0
        MOV     ECX,DWORD PTR [EDX]     ;ECX = THREAD TO NEXT LINE
        CMP     ECX,0                   ;BRANCH IF NOT FOUND
        JE      FINDFNC4
;       MOV     EDX,ECX
        MOV     TEXTLPTR,ECX    ;TEXTLPTR PT TO THREAD OF NEXT LINE
        ADD     ECX,6           ;SKIP THREAD AND LINE NO.
        MOV     TEXTPTR,ECX     ;TEXTPTR PT TO 1ST CHAR OF TEXT

FINDFNC3:
        CMP     BYTE PTR [TEXTPTR],@DEFFN       ;LOOK FOR 'DEFFN' ATOM
        JNE     FINDFNC1
        INC     TEXTPTR
        CMP     R7,0            ;BRANCH IF LOOKING FOR ANY DEFFNX
        JE      FINDFNC5
        MOV     CH,BYTE PTR [TEXTPTR]
        CMP     CH,R7           ;LOOK FOR FUNCTION LETTER OR DIGIT
        JNE     FINDFNC1
        INC     TEXTPTR
;       MOV     R0,0

FINDFNC4:
        RET

FINDFNC5:
        CALL    LETRDIGT        ;LOOK FOR ANY LETTER OR DIGIT
        TEST    R0,0FH          ;NF,BRACNCH
        JNZ     FINDFNC1
        RET

;FN#: PROCESS A '#----' FUNCTION
;

FN_:
        MOV     EDX,OFFSET _FN_LST      ;BASE ADDR(# FUNCTION LIST) IN DATA MEMORY
        CALL    CCJB                    ; Checkpoint. SCAN FOR LEGAL FUNCTION
        JMP     FN_TERM                 ;_TERM FUNCTION
        JMP     FN_PART                 ;_PART FUNCTION
        JMP     FN_ID                   ;_ID FUNCTION
        JMP     FN_CPU                  ;_CPU FUNCTION
        JMP     FN_VER                  ;_VER FUNCTION
        JMP     FN_TIMESLICE            ;_VER FUNCTION
        JMP     ERROR_19                ;ILLEGAL KEYWORD

; #TERM FUNCTION

FN_TERM:
        TEST    SL,2                    ;BRANCH IF NOT EXECUTION PHASE
        JZ      FNCFOUND
        CALL    GETTERM_                ;GET TERMINAL NUMBER

FN_T1:
        mov     al,r0
FN_T2:
        mov     edx,vsptr
        sub     edx,8
        mov     dword ptr [edx],0
        mov     dword ptr [edx+4],0
        mov     byte ptr [edx+2],al
        mov     vsptr,edx               ;CONVERT R0 TO FL. PT.
        cmp     edx,dword ptr [@@vsend]
        jb      error_02

        jmp     fncfound

; #PART FUNCTION

FN_PART:
        TEST    SL,2                    ;BRANCH IF NOT EXECUTION PHASE
        JZ      FNCFOUND
        MOV     al,part_                ;GET PART#
        inc     al                      ;R0 = PART# + 1
        JMP     FN_T2


;FN#ID:  process a #ID function
;
;     #ID returns the system identification number.  The identification number
;     is a random 2-byte binary value stored in the Bootstrap PROM.
;
;     Return:
;          In execution phase, a floating point value equal to the ID# is left
;          on the Value Stack.

FN_ID:
          TEST      SL,XPHASE       ;BRANCH IF NOT EXECUTION PHASE
          JZ        FN_ID1
          MOV       EDX,OFFSET BOOT_ID
          MOV       AX,WORD PTR [EDX]
          MOV       WORD PTR [R0],AX
          MOV       R2,80H          ;DISALLOW ID# HEX(8080); TRANSLATE TO ZERO
          MOV       AL,R2           ;BRANCH IF NOT HEX(8080)
          CMP       AL,R0
          JNE       FN_ID2
          CMP       AL,R1
          JNE       FN_ID2
          MOV       WORD PTR R0,0   ;R1R0 = 0
FN_ID2:   CALL      BINFLT          ;CONVERT BINARY TO FL. PT.
FN_ID1:   JMP       FNCFOUND

;FN#CPU: process a #CPU function
;
;     #CPU returns the system identification number.  The identification number
;     is a byte binary value stored in the Global area, and it is assigned by
;     system configuration.
;
;     Return:
;          In execution phase, a floating point value equal to the CPU# is left
;          on the Value Stack.

FN_CPU:
          TEST      SL,XPHASE       ;BRANCH IF NOT EXECUTION PHASE
          JZ        FN_CPU1
          MOV       EDX,OFFSET CPUID
          XOR       AX,AX
          MOV       AL,BYTE PTR [EDX]
          INC       AX
          MOV       WORD PTR [R0],AX
          CALL      BINFLT          ;CONVERT BINARY TO FL. PT.
FN_CPU1:  JMP       FNCFOUND

;FN#TIMESLICE: process a #TIMESLICE function
;
;     #TIMESLICE returns the system partition change number.  The number
;     is a byte binary value stored in the Global area, and it is init to 0
;
;     Return:
;          In execution phase, a floating point value equal to the CPU# is left
;          on the Value Stack.

FN_TIMESLICE:
          TEST      SL,XPHASE       ;BRANCH IF NOT EXECUTION PHASE
          JZ        FN_TIMESLICE1
          MOV       EDX,OFFSET TIMESLICE
          MOVZX     AX,BYTE PTR [EDX]
          INC       AX
          MOV       WORD PTR [R0],AX
          CALL      BINFLT          ;CONVERT BINARY TO FL. PT.
FN_TIMESLICE1:
          JMP       FNCFOUND

;FN#VER: process a #VER function
;
;     #VER returns the system version number.  The version number
;     is a word binary value stored in the Global area.
;
;     Return:
;          In execution phase, a floating point value equal to the VER# is left
;          on the Value Stack.

FN_VER:
          TEST      SL,XPHASE       ;BRANCH IF NOT EXECUTION PHASE
          JZ        FN_VER1
          MOV       EDX,OFFSET _CSOS
          MOV       AX,WORD PTR [EDX]
          mov       word ptr [r0],ax
          mov       edi,vsptr
          sub       edi,8
          mov       ax,-2
          stosw                     ;save exponent
          mov       ax,word ptr r0
          stosw                     ;save two byte binary value
          mov       eax,0           ;
          stosd                     ;save 4 byte zero
          sub       vsptr,8         ;vsptr pt to beg of this value

FN_VER1:  JMP       FNCFOUND

CODE      ENDS
          END
