          .XLIST
          INCLUDE PREAMBLE
          .LIST

PUBLIC    ALPHEXPR
PUBLIC    ERR_19A
PUBLIC    ERR_29G
PUBLIC    OPARITH
PUBLIC    OPERND1$
PUBLIC    OPERAND$
PUBLIC    OPLGCL

EXTRN     VAR     :NEAR
EXTRN     FLTBINA :NEAR
EXTRN     ALPHA   :NEAR
EXTRN     ALPHA1  :NEAR
EXTRN     ARECEVAL:NEAR
EXTRN     CCJB    :NEAR
EXTRN     CCL     :NEAR
EXTRN     CCNL    :NEAR
EXTRN     CHARCTR :NEAR
EXTRN     CURLEN  :NEAR
EXTRN     DATE    :NEAR
EXTRN     DIGIT   :NEAR
EXTRN     ERROR_29:NEAR
EXTRN     ERR$FNC :NEAR
EXTRN     ERROR_10:NEAR
EXTRN     ERROR_11:NEAR
EXTRN     ERROR_17:NEAR
EXTRN     ERROR_19:NEAR
EXTRN     ERROR_34:NEAR
EXTRN     EXPR    :NEAR
EXTRN     FILE_B  :NEAR
EXTRN     FILE_PTR:NEAR
EXTRN     FLTBNCCC:NEAR
EXTRN     HEXDIGIT:NEAR
EXTRN     NXTCHEOS:NEAR
EXTRN     PSTATFNC:NEAR
EXTRN     SCRATCH :NEAR
EXTRN     TIME    :NEAR
EXTRN     VERSION :NEAR
EXTRN     USERMSG :NEAR
EXTRN     _AOPLST :NEAR
EXTRN     _DATE   :NEAR
EXTRN     _ERR$   :NEAR
EXTRN     _MSG    :NEAR
EXTRN     _SLCTLST:NEAR
EXTRN     _TIME   :NEAR
EXTRN     _CSVER  :NEAR
EXTRN     CPBASE  :DWORD

;------------------------------------------------------------------------------;
;                                                                              ;
;      **********************************                                      ;
;      *                                *                                      ;
;      *  MODULE-13:  ALPHA-EXPRESSIONS *                                      ;
;      *                                *                                      ;
;      **********************************                                      ;
;                                                                              ;
;                                                                              ;
;      DESCRIPTION: LOGIC FOR PROCESSING RIGHT-HAND                            ;
;                   SIDE OF ALPHANUMERIC LET STATEMENTS                        ;
;                                                                              ;
;------------------------------------------------------------------------------;


;      ALPHEXPR: SCAN & PROCESS AN ALPHA-EXPRESSION (IN
;                 AN  ALPHA LET STATEMENT
;
;           FORMAT:
;
;             [LET] alpha-var [,alpha-var]... = alpha-expr
;
;
;             where:
;             alpha-expr = [operand] [operator operand]...
;                     (but null expression not allowed)
;                  OR    = operand [& operand] ...
;
;             operand =  alpha-variable
;                        literal
;                        BIN()
;                        ALL()
;
;             operator = AND, OR, XOR, or BOOL
;                        ADD[C], SUB[C], DAC, or DSC
;
;           ENTRY: TEXTPTR POINTS AFTER '=' IN LET STMT
;              IN EXECUTION PHASE, RECEIVER VARIABLES OF
;                ARE SET UP IN OS & VS.
;
;                  VALUE STACK (VS) CONTAINS--
;                   -------------------------------------
;                   ! LENGTH OF      RECEIVER (2 BYTES) !
;                   ! ADDRESS OF     RECEIVER (2 BYTES) !
;                   -------------------------------------
;
;                  OPERATOR STACK (OS) CONTAINS--
;                   -------------------------------------
;                   ! @ALPHREC FOR      RECEIVER        !
;                   -------------------------------------
;
;                       WHERE: @ALPHREC = ALPHA RECEIVER
;                                         ATOM
;
;           RETURN: TEXTPTR POINTS AFTER ALPHA-EXPRESSION
;                IN EXECUTION PHASE,
;                     RECEIVER = VALUE OF EXPRESSION
;                     RECEIVER INFO. CLEARED FROM VS & OS
;                     F1,0 = LENGTH OF RECEIVER
;                     PC'S = ADDRESS OF RECEIVER
;
;           DESCRIPTION:
;                THE ALPHA-EXPRESSION IS PROCESSED FROM
;                LEFT TO RIGHT ONE TERM AT A TIME. THE
;                LAST RECEIVER IN THE LET STATEMENT IS
;                OPERATED ON FOR EACH TERM IN THE EXPRES-
;                SION. THE OTHER RECEIVERS ARE THEN SET
;                EQUAL TO THE FINAL VALUE OF THE LAST
;                RECEIVER. IF THE '=' SIGN IS IMMEDIATELY
;                FOLLOWED BY AN OPERAND, THE LAST RECEIVER
;                IS FIRST SET EQUAL TO THE VALUE OF THE
;                OPERAND AND THEN THE REST OF THE EXPRES-
;                SION IS PROCESSED.
;
;           MODIFIED REGISTERS: ALL,AUX0,1,2,3,4
;                AUXF (EXECUTION PHASE ONLY)
;

ALPHEXPR: CALL    OPERND1$               ;SCAN & PROCESS ALPHA OPERAND
          TEST    R0,0FH
          JNZ     ALPHEXP3               ;B, NF
          TEST    SL,02                  ;BRANCH IF NOT EXECUTION PHASE
          JZ      ALPHEXP4               ;FALSE
;+++
        mov     ecx,0                   ;
        mov     ax,word ptr [r2]        ;get length of value
        mov     word ptr [r6],ax        ;save for CONCAT routine ......
        mov     esi,aux00               ;    addr of value
        mov     edx,vsptr               ;
        mov     bx,word ptr [edx]       ;get length of receiver
        mov     edi,dword ptr [edx+2]   ;    addr of receiver
        add     aux01,2                 ;???
        or      ax,ax                   ;br, if operand <> ALL function
        jnz     alphexp2                ;

;RECEIVER = ALL CHARACTER

        mov     cx,bx                   ;
        mov     al,r4                   ;get the character
        rep     stosb
        jmp     alphexp4_0              ;

;RECEIVER = OPERAND VALUE

alphexp2:
        cmp     ax,bx                   ;br, if receiver length > value length
        jb      al_4                    ;
        mov     cx,bx                   ;make the length = receiver length
        rep     movsb                   ;
al_0:
        jmp     alphexp4_0
al_4:
        mov     cx,ax                   ;make the length = value length
        rep     movsb
        mov     cx,bx                   ;fill the padded space
        sub     cx,ax                   ;
        mov     al,20h
        rep     stosb

alphexp4_0:
        cmp     byte ptr [textptr],0dh  ;br, if next charcter is CR
        je      al_5                    ;
        cmp     byte ptr [textptr],0fdh ;br, if next character is not EOS
        jne     alphexp1                ;
al_5:
        dec     osptr                   ;clear OS entry
        mov     edx,vsptr               ;
        mov     ax,word ptr [edx]       ;get length of result (receiver)
        mov     word ptr [r0],ax        ;
        mov     edx,dword ptr [edx+2]   ;    address
        add     vsptr,6                 ;clear VS entry
        ret

;
;       RECEIVER = OPERAND VALUE
;
;
ALPHEXP4: CALL    NXTCHEOS        ;LOOK FOR EOS CHARACTER
          TEST    R0,0FH          ;BRANCH IF NF
          JNZ     ALPHEXP1
ALPHEXP6: TEST    SL,02           ;BRANCH IF EXECUTION PHASE
          JNZ     ALPHEXP41       ;
SR@MD13:  RET                     ;RETURN
ALPHEXP41:
;         JMP     POPOV4          ;GET LENGTH & ADDRESS OF RESULT (RECEIVER)
          DEC     OSPTR           ;REMOVE LAST OS ATOM
          MOV     EDX,VSPTR
          MOV     AX,WORD PTR [EDX]
          MOV     WORD PTR [R0],AX
          MOV     ECX,DWORD PTR [EDX+2]
          ADD     VSPTR,6
          MOV     EDX,ECX
          RET

ALPHEXP1: CMP     BYTE PTR [TEXTPTR],'&'  ;
          JNE     ALPHEXP3        ;LOOK FOR '&'
          INC     TEXTPTR         ;BRANCH IF FOUND (CONCATENTION)
          JMP     CONCAT          ;

ALPHEXP3: CALL    OPRATOR$        ;LOOK FOR ALPHA OPERATOR
          CALL    OPERAND$        ;LOOK FOR ALPHA OPERAND
          TEST    R0,0FH          ;BRANH IF FOUND
          JZ      ALPHEXP42       ;
ERR_29G:  JMP     ERROR_29        ;ERROR IF NOT FOUND
ALPHEXP42:TEST    SL,02           ;BRANCH IF NOT EXECUTION PHASE
          JZ      ALPHEXP8        ;
          MOV     EDX,VSPTR
          MOV     CX,WORD PTR [EDX]   ;GET LENGTH OF RECEIVER FROM VS
          XCHG    EDX,AUX0F
          ADD     AUX0F,2
          MOV     WORD PTR [R6],DX    ;F7,6 = OPERATOR CODE
          MOV     EDX,AUX0F
          MOV     WORD PTR [R0],CX    ;LENGTH OF RECEIVER
          MOV     EDX,DWORD PTR [EDX] ;PC'S = ADDRESS OF RECEIVER
          CMP     WORD PTR [R2],0
          JNE     ALPHEXP43
          OR      R7,80H          ;IF OPERAND = ALL, SET IMMEDIATE OPERATION BIT
ALPHEXP43:MOV     AH,R6           ;
          AND     AH,0FH          ;
          CMP     AH,1            ;BRANCH IF ARITHMETIC OPERATOR
          JE      ALPHEXP44       ;
          CALL    OPLGCL          ;PERFORM LOGICAL OPERATION (AND,OR,XOR,BOOL)
          JMP     ALPHEXP8        ;

ALPHEXP44:CALL    OPARITH         ;PERFORM ARITHMETIC OPERATION (ADD,SUB,DAC,DSC)

ALPHEXP8: CALL    NXTCHEOS        ;LOOK FOR EOS CHARACTER
          TEST    R0,0FH          ;
          JZ      ALPHEXP6        ;BRANCH IF FOUND
          JMP     ALPHEXP3        ;BRANCH IF NF



;      CONTCAT: PERFORM CONCATENATION IN ALPHA LET STMT
;
;           FORMAT: [LET] alpha-var [,alpha-var]... =
;
;                     operand [& operand] ...
;
;              where: operand = alpha-variable, literal,
;                               BIN(), or ALL()
;
;           ENTRY: TEXTPTR POINTS AFTER 1ST '&'
;               IN EXECUTION PHASE,
;                  F7,6 = LENGTH OF 1ST OPERAND
;                  VALUE OF 1ST OPERAND ==> RECEIVER
;                  VS CONTAINS: LENGTH OF RECEIVER
;                               ADDRESS OF RECEIVER
;
;           REGISTER USE:
;                AUX3 = ADDRESS +1 OF LAST CHARACTER PUT
;                       INTO RECEIVER
;                AUX4 = ADDRESS +1 OF END OF RECEIVER
;
;           MODIFIED REGISTERS: ALL,AUX0-4

CONCAT:   TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
          JZ      CONCAT1         ;
          MOV     EDX,VSPTR

;
;  SET UP CONCATENTION POINTERS
;

          MOV     EAX,0
          MOV     AX,WORD PTR [EDX] ;GET LENGTH OF RECEIVER
          MOV     WORD PTR [R0],AX  ;F1,0 = LENGTH OF RECEIVER
          MOV     ECX,DWORD PTR [EDX+2]

;  EDX = ADDRESS OF RECEIVER

          MOV     EDX,ECX
          ADD     EDX,EAX
          MOV     AUX04,EDX         ;AUX4 = ADDRESS +1 OF END OF RECEIVER
          CMP     WORD PTR [R6],0
          JNE     CONCATX
          JMP     CONCAT6           ;BRANCH IF OPERAND = ALL()

CONCATX:  MOV     BX,WORD PTR [R0]
          CMP     BX,WORD PTR [R6]  ;BRANCH IF REC. LENGTH <= VALUE LENGTH
          JBE     CONCAT6
          MOV     AX,WORD PTR [R6]
          MOV     EDX,ECX
          ADD     EDX,EAX

CONCAT6:  MOV     AUX03,EDX         ;AUX3 = ADDRESS +1 OF LAST CHAR PUT INTO RECEVR


;  PROCESS NEXT OPERAND


CONCAT1:  CALL    OPERND1$        ;SCAN NEXT OPERAND
          TEST    R0,0FH          ;BRANCH IF NF
          JNZ     ERR_29G         ;
          TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
          JZ      CONCAT3         ;
          MOV     EDX,AUX04
          MOV     DWORD PTR [A6],EDX   ;F7,6 = ADDRESS +1 OF END OF RECIVER
          CMP     WORD PTR [R2],0      ;BRANCH IF OPERAND NOT ALL()
          JNE     CONCAT2

;
;      CONCATENATE ALL()
;

          MOV     EDX,AUX03            ;GET ADDRESS OF NEXT BYTE IN RECEIVER
          JMP     CONCAT4
CONCATZ:  MOV     AL,R4
          MOV     BYTE PTR [EDX],AL    ;ALL CHARACTER ==> RECEIVER
          INC     EDX

CONCAT4:  CMP     EDX,DWORD PTR [A6]   ;BRANCH IF MORE TO RECEIVER
          JNE     CONCATZ
          JMP     CONCAT7

;
;      CONCATENATE ALPHA-VAR, LITERAL, OR BIN()
;

CONCAT2:  MOV     K,0
          MOV     EDX,AUX00            ;GET ADDRESS OF VALUE
          MOV     DWORD PTR [A2],EDX   ;F3,2 = ADDRESS +1 OF END OF VALUE
          XOR     EAX,EAX
          MOV     AX,WORD PTR [R2]
          ADD     DWORD PTR [A2],EAX
CONCAT5:  MOV     CH,BYTE PTR [EDX]    ;GET NEXT CHARACTER OF VALUE
          INC     EDX
          XCHG    AUX03,EDX
          CMP     EDX,DWORD PTR [A6]
          JE      CONCAT7
          MOV     BYTE PTR [EDX],CH     ;CHARACTER OF VALUE ==> RECEIVER
          INC     EDX
          XCHG    EDX,AUX03
          CMP     EDX,DWORD PTR [A2]
          JNE     CONCAT5

;  LOOK FOR ANOTHER CONCATENTATION OPERATOR


CONCAT3:  CMP     BYTE PTR [TEXTPTR],'&'
          JNE     ALPHEXP6        ;BRANCH TO PROCESS ANY OTHER RECEIVER
          INC     TEXTPTR
          JMP     CONCAT1         ;BRANCH IF FOUND

CONCAT7:  XCHG    EDX,AUX03       ;RESTORE PTR TO LAST CHAR PUT INTO RECEIVER
          JMP     CONCAT3         ;


;      OPERAND$: SCAN & PROCESS AN ALPHA-EXPRESSION
;                OPERAND. RETURN DEFINED LENGTH.
;
;      OPERND1$: SAME EXCEPT RETURN CURRENT LENGTH.
;
;
;           FORMAT:  alpha-variable
;                    literal
;                    BIN()
;                    ALL()
;
;           ENTRY: TEXTPTR POINTS AT NEXT TEXT CHARACTER
;
;           RETURN (NOT FOUND): F0 = 1
;                TEXTPTR POINTS AT NEXT NONBLANK TEXT CHAR
;
;           RETURN (FOUND): F0 = 0
;                TEXTPTR POINTS AFTER OPERAND
;
;             IN EXECUTION PHASE,
;                IF OPERAND NOT ALL FUNCTION,
;                     F3,2 = LENGTH OF VALUE (NOT 0)
;                     AUX0 = ADDRESS OF VALUE
;                IF OPERAND IS ALL FUNCTION,
;                     F3,2 = 0 (ALL FLAG)
;                     F4 = ALL character
;
;           MODIFIED REGISTERS: ALL,AUX0,1,2

OPERAND$: CALL    ALPHA           ;LOOK FOR ALPHA-VARIABLE OR LITERAL
OPERND$3: TEST    R0,0FH          ;BRANCH IF NF
          JNZ     OPERND$1        ;
          TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
          JZ      OPERND$2        ;
          MOV     EDX,VSPTR       ;GET LENGTH OF VALUE FROM VS
          MOV     CX,WORD PTR [EDX]
          MOV     WORD PTR [R2],CX         ;F3,2 = LENGTH OF OPERAND VALUE
          MOV     EDX,DWORD PTR [EDX+2]    ;GET ADDRESS OF OPERAND VALUE
          ADD     VSPTR,6
          MOV     AUX00,EDX       ;AUX0 =ADDRESS OF OPERAND VALUE
OPERND$2: RET                     ;

OPERND$1: CALL    ALLFUNC         ;LOOK FOR 'ALL' FUNCTION
          TEST    R0,0FH          ;BRANCH IF NF
          JNZ     BINFUNC         ;
          MOV     WORD PTR [R2],0 ;F3,2 (LENGTH)=0 IMPLIES ALL FUNCTION
          RET

OPERND1$: MOV     K,4             ;FLAG = RETURN CURRENT LENGTH OF OPERAND
          CALL    ALPHA1          ;SCAN & PROCESS ALPHA VARIABLE OR LITERAL
          JMP     OPERND$3        ;


;  BINFUNC:  look for BIN function.  If not found, goto MSGFUNC.
;
;  BINFNC: evaluate a 'BIN' function
;
;  FORMAT: BIN( expression [,digit] )          where 0 < digit < 5
;
;  ENTRY: TEXTPTR points at next text character
;
;  EXIT: F0 = 0 and TEXTPTR points after BIN()
;
;             In execution phase,
;                binary value in start of scratch area
;                F3F2 = length of value
;                AUX0 = address of binary value
;
;  MODIFIED REGISTERS: all, AUX0, 1, 2
;
;  ================= NOTE =======================================================
;
;         The following logic was added in rel. 3.0.  BIN was enhanced to
;         handle 3 and 4 byte values.
;         This routine was lifted from BASIC-3.  However, the number of bytes
;         handled was increased from 3 to 4 and the BINVALEN routine added.
;
;  ==============================================================================

BINFUNC:  MOV     R0,@BIN?LP      ;?????
          CALL    CCL             ;LOOK FOR 'BIN'
          TEST    R0,0FH          ;
          JNZ     MSGFUNC         ;

BINFNC:   CALL    EXPR            ;SCAN & EVALUATE EXPERSSION
          TEST    SL,2            ;SKIP IF NOT EXECUTION PHASE
          JZ      BINFNC0         ;
          CALL    FLTBNCCC        ;
                                  ;NUMBER ==>  F7 F6 F5 F4 F3 F2
          test    k,08h
          jnz     error_34        ;negative number

;  ======= handle specified length ==============================================

BINFNC0:  MOV     DX,1            ;ESI =LENGTH
          CMP     BYTE PTR [TEXTPTR],','
          JNE     BINFNCA
          INC     TEXTPTR
          push    ax
          push    bx
          push    cx
          test    sl,03h
          jz      bin_kkk
          mov     ch,byte ptr [textptr]
          cmp     ch,30h
          jb      bin_kkk
          cmp     ch,39h
          ja      bin_kkk
          inc     textptr
          and     ch,0fh
          mov     dl,ch
          test    sl,01h
          jnz     binfnca1
          jmp     bin_yyy
bin_kkk:
          call    expr
          test    sl,xphase
          jz      binfnca1
          call    fltbina               ;convert number to binary value
          mov     dx,word ptr [r0]
bin_yyy:
          cmp     dx,01h                ;check range for 1 to 6
          jb      error_34
          cmp     dx,06h
          ja      error_34
binfnca1:
          pop     cx
          pop     bx
          pop     ax

BINFNCA:  TEST    SL,2            ;EXIT IF NOT EXECUTION PHASE
          JZ      BINFNC1

          MOV     AUX00,OFFSET SCRATCH  ;ADDRESS ==> AUX0

          CMP     DX,6            ;SKIP IF SIX BYTES VALUE
          JE      BINFNC2         ;
          CMP     AH,0            ;ERROR IF OVERFLOW
          JNE     BINFNC9         ;
          CMP     DX,5            ;SKIP IF FIVE BYTES VALUE
          JE      BINFNC3         ;
          CMP     AL,0            ;ERROR IF OVERFLOW
          JNE     BINFNC9         ;
          CMP     DX,4            ;SKIP IF FOUR BYTES VALUE
          JE      BINFNC4         ;
          CMP     BH,0            ;ERROR IF OVERFLOW
          JNE     BINFNC9         ;
          CMP     DX,3            ;SKIP IF THREE BYTES VALUE
          JE      BINFNC5         ;
          CMP     BL,0            ;ERROR IF OVERFLOW
          JNE     BINFNC9         ;
          CMP     DX,2            ;SKIP IF TWO BYTES VALUE
          JE      BINFNC6         ;
          CMP     CH,0            ;SKIP IF NO OVERFLOW
          JE      BINFNC7         ;
BINFNC9:  JMP     ERROR_34        ;ILLEGAL VALUE

;  ======= write binary value ===================================================

BINFNC2:  MOV     ESI,AUX00
          MOV     BYTE PTR [ESI],AH    ;WRITE SIX BYTE VALUE
          MOV     BYTE PTR [ESI+1],AL  ;WRITE FIVE BYTE VALUE
          MOV     BYTE PTR [ESI+2],BH  ;WRITE FOUR BYTE VALUE
          MOV     BYTE PTR [ESI+3],BL  ;WRITE THREE BYTE VALUE
          MOV     BYTE PTR [ESI+4],CH  ;WRITE TWO BYTE VALUE
          MOV     BYTE PTR [ESI+5],CL  ;WRITE ONE BYTE VALUE

          MOV     WORD PTR [R2],DX      ;LENGTH OF VALUE
          CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')'
          JNE     ERROR_11
          INC     TEXTPTR
          MOV     R0,0
          RET

BINFNC3:  MOV     ESI,AUX00
          MOV     BYTE PTR [ESI],AL    ;WRITE FIVE BYTE VALUE
          MOV     BYTE PTR [ESI+1],BH  ;WRITE FOUR BYTE VALUE
          MOV     BYTE PTR [ESI+2],BL  ;WRITE THREE BYTE VALUE
          MOV     BYTE PTR [ESI+3],CH  ;WRITE TWO BYTE VALUE
          MOV     BYTE PTR [ESI+4],CL  ;WRITE ONE BYTE VALUE

          MOV     WORD PTR [R2],DX      ;LENGTH OF VALUE
          CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')'
          JNE     ERROR_11
          INC     TEXTPTR
          MOV     R0,0
          RET

BINFNC4:  MOV     ESI,AUX00
          MOV     BYTE PTR [ESI],BH    ;WRITE FOUR BYTE VALUE
          MOV     BYTE PTR [ESI+1],BL  ;WRITE THREE BYTE VALUE
          MOV     BYTE PTR [ESI+2],CH  ;WRITE TWO BYTE VALUE
          MOV     BYTE PTR [ESI+3],CL  ;WRITE ONE BYTE VALUE

          MOV     WORD PTR [R2],DX      ;LENGTH OF VALUE
          CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')'
          JNE     ERROR_11
          INC     TEXTPTR
          MOV     R0,0
          RET

BINFNC5:  MOV     ESI,AUX00
          MOV     BYTE PTR [ESI],BL    ;WRITE THREE BYTE VALUE
          MOV     BYTE PTR [ESI+1],CH  ;WRITE TWO BYTE VALUE
          MOV     BYTE PTR [ESI+2],CL  ;WRITE ONE BYTE VALUE

          MOV     WORD PTR [R2],DX      ;LENGTH OF VALUE
          CMP     BYTE PTR [TEXTPTR],')'   ;LOOK FOR ')'
          JNE     ERROR_11
          INC     TEXTPTR
          MOV     R0,0
          RET

BINFNC6:  MOV     ESI,AUX00
          MOV     BYTE PTR [ESI],CH    ;WRITE TWO BYTE VALUE
          MOV     BYTE PTR [ESI+1],CL  ;WRITE ONE BYTE VALUE

          MOV     WORD PTR [R2],DX      ;LENGTH OF VALUE
          CMP     BYTE PTR [TEXTPTR],')'    ;LOOK FOR ')'
          JNE     ERROR_11
          INC     TEXTPTR
          MOV     R0,0
          RET

BINFNC7:  MOV     ESI,AUX00
          MOV     BYTE PTR [ESI],CL    ;WRITE ONE BYTE VALUE

          MOV     WORD PTR [R2],DX      ;LENGTH OF VALUE
          CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')'
          JNE     ERROR_11
          INC     TEXTPTR
          MOV     R0,0
          RET

BINFNC1:  CMP     BYTE PTR [TEXTPTR],')'   ;LOOK FOR ')'
          JNE     ERROR_11
          INC     TEXTPTR
          MOV     R0,0
          RET

;BINVAL_0:
;        PUSH    AX
;        PUSH    BX
;        PUSH    CX
;        DEC     TEXTPTR         ;RESTORE TEXTPTR BACK
;        MOV     K,0             ;
;        CALL    VAR             ;GET NUMERIC VARIABLE
;        TEST    SL,2            ;BR, IF NOT XPHASE
;        JZ      BINVAL_1        ;
;        CALL    FLTBINA         ;R0 = BINARY VALUE
;        CMP     R0,0            ;ERROR IF VALUE = 0
;        JZ      XERROR_34       ;
;        CMP     R0,6            ;ERROR IF VALUE > 6
;        JG      XERROR_34       ;
;BINVAL_1:
;        POP     CX
;        POP     BX
;        POP     AX
;        XOR     DH,DH
;        MOV     DL,R0           ;GET THE VALUE
;        JMP     BINFNCA
;
;XERROR_34:
;        POP     CX
;        POP     BX
;        POP     AX
;        JMP     ERROR_34


;   BINVALEN:  process the length parameter for a BIN or VAL function
;
;         Format:   [, digit]    where 0<digit<7
;
;         Entry:    TEXTPTR points after BIN( or VAL(
;
;         Return:   TEXTPTR points after length parameter
;                   F0 = length (default = 1)
;
;         Modified registers: PHPL,CHCL,F0
;
;@@ deleted by mll 5/15/89 ++

; MSGFUNC: PROCESS A '$MSG' FUNCTION
;
;      FORMAT:  $MSG
;
;      ENTRY: TEXTPTR POINTS AT NEXT TEXT CHARACTER
;
;      RETURN (NOT FOUND):  F0 = 1
;           TEXTPTR POINTS TO NEXT NONBLANK CHARACTER.
;           ERR#19 IF '$' FOUND BUT NOT 'MSG'.
;
;      RETURN (FOUND): F0 = 0
;           TEXTPTR POINTS AFTER '$MSG'
;        IN EXCUTION PHASE,
;           F3,2 = LENGTH OF $MSG
;           AUX0 = ADDRESS OF MSG
;
;      MODIFIED REGISTERS: ALL,AUX0,1,2
;

MSGFUNC:  MOV     R0,@$           ;
          CALL    CCL             ;LOOK FOR '$'
          TEST    R0,0FH          ;BRANCH IF NF
          JNZ     MSGFUNC3        ;
          MOV     EDX,OFFSET _MSG
          CALL    CCNL            ;LOOK FOR 'MSG'
          TEST    R0,0FH          ;BRANCH IF NF
          JNZ     MSGFUNC2        ;
          TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
          JZ      MSGFUNC1        ;
          MOV     EDX,OFFSET USERMSG
          MOV     AUX00,EDX       ;AUX0 = ADDRESS OF USER MESSAGE
          mov     WORD PTR [R0],50H ;MAX LENGTH = 80
          CALL    CURLEN          ;DETERMINE CURRENT LENGTH OF MSG
          MOV     BX,WORD PTR [R0]
          MOV     WORD PTR [R2],BX  ;F3,2 = CURRENT LENGTH
          MOV     R0,0              ;FLAG = FOUND
MSGFUNC1: RET

;  ------------------------------------------------------------------------------

MSGFUNC3: MOV     R0,@SELECT      ;
          CALL    CCL             ;LOOK FOR SELECT
          TEST    R0,0FH          ;BRANCH IF NF
          JNZ     MSGFUNCX
          JMP     SLCTFNC         ;BRANCH IF FOUND

MSGFUNCX: MOV     EDX,OFFSET _ERR$       ;ERR$ FUNCTION ADDED REL 3.0
          CALL    CCNL            ;LOOK FOR ERR$
          TEST    R0,0FH          ;BRANCH IF NOT FOUND
          JNZ     MSGFUNCY        ;
          JMP     ERR$FNC         ;BRANCH IF FOUND

MSGFUNCY: MOV     EDX,OFFSET _TIME
          CALL    CCNL            ;LOOK FOR TIME
          TEST    R0,0FH          ;BRANCH IF NOT FOUND
          JNZ     MSGFUNCZ        ;
          JMP     TIME            ;

MSGFUNCZ: MOV     EDX,OFFSET _DATE
          CALL    CCNL            ;LOOK FOR DATE
          TEST    R0,0FH          ;BRANCH IF NOT FOUND
          JNZ     MSGFUNCA        ;
          JMP     DATE            ;

MSGFUNCA: MOV     EDX,OFFSET _CSVER
          CALL    CCNL            ;LOOK FOR DATE
          TEST    R0,0FH          ;BRANCH IF NOT FOUND
          JNZ     MSGFUNC1        ;
          JMP     VERSION         ;

MSGFUNC2: JMP     PSTATFNC        ;


;      ALLFUNC: SCAN & PROCESS AN 'ALL' FUNCTION
;
;           FORMAT:       alpha-variable
;                    ALL( literal        )
;                         hh
;
;                where: h = hexdigit
;
;           ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;           RETURN (NOT FOUND): F0 = 1
;                TEXTPTR POINTS TO NEXT NONBLANK TEXT CHAR
;
;           RETURN (FOUND): F0 = 0
;                TEXTPTR POINTS AFTER 'ALL()'
;             IN EXECUTION PHASE,
;                F4 = ALL character
;
;           MODIFIED REGISTERS: ALL,AUX0,1,2
;

ALLFUNC:  MOV     R0,@ALL         ;
          CALL    CCL             ;LOOK FOR 'ALL'
          TEST    R0,0FH          ;BRANCH IF NF
          JNZ     SR@MD13         ;   RETURN IF NF
          CMP     BYTE PTR [TEXTPTR],'('       ;LOOK FOR '('
          JNE     ERROR_10
          INC     TEXTPTR
          CALL    CHARCTR         ;GET VALUE OF FUNCTION
          MOV     AL,R0           ;
          MOV     R4,AL           ;F4 = ALL CHARACTER
          CMP     BYTE PTR [TEXTPTR],')'      ;LOOK FOR ')'
          JNE     ERROR_11
          INC     TEXTPTR
          MOV     R0,0
          RET


;      OPRATOR$: LOOK FOR ALPHA-EXPRESSION OPERATOR IN
;                TEXT. ERROR IF NOT FOUND.
;
;           ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;           RETURN: TEXTPTR POINTS AFTER OPERATOR
;                IN EXECUTION PHASE,
;                   AUXF = OPERATOR CODE
;                          0000 = AND
;                          1000 = OR
;                          2000 = XOR
;                          4x00 = BOOLx
;
;                          0001 = ADD
;                          0101 = SUB
;                          4001 = ADDC
;                          4101 = SUBC
;                          4201 = DAC
;                          4301 = DSC
;
;                CH = OPERATOR TEXT ATOM
;
;           MODIFIED REG: PC'S,CH,CL,F0,1,2,7,AUX0,1,2
;                AUXF (EXECUTION PHASE ONLY)
;

OPRATOR$: MOV     EDX,OFFSET _AOPLST
          CALL    CCJB            ;LOOK FOR ALPHA EXPRESSION OPERATOR
          JMP     OP$AND          ;'AND'
          JMP     OP$OR           ;'OR'
          JMP     OP$XOR          ;'XOR'
          JMP     OP$ADD          ;'ADD'
          JMP     OP$SUB          ;'SUB'
          JMP     OP$DAC          ;'DAC'
          JMP     OP$DSC          ;'DSC'
          JMP     OP$BOOL         ;'BOOL'
ERR_19A:  JMP     ERROR_19        ;MISSING OPERATOR

OP$AND:   XOR     EDX,EDX         ;PC'S = 0
          JMP     OPRATR$5
OP$OR:    MOV     EDX,1000H
          JMP     OPRATR$5
OP$XOR:   MOV     EDX,2000H
          JMP     OPRATR$5
OP$ADD:   MOV     EDX,0001H
          JMP     OPRATR$7
OP$SUB:   MOV     EDX,0101H

OPRATR$7: MOV     AUX00,EDX       ;SAVE OPERATOR CODE
          MOV     AUX01,TEXTPTR
          CMP     BYTE PTR [TEXTPTR],'C'
          JNE     OPRATR$5
          INC     TEXTPTR

;
;      MAKE SURE NOT C$ OR C1$,ETC.
;

          CALL    DIGIT           ;LOOK FOR DIGIT
          TEST    R0,0FH          ;BRANCH IF FOUND
          JZ      OP$LAB1
          CMP     BYTE PTR [TEXTPTR],'$'
          JNE     OP$LAB2
          INC     TEXTPTR

OP$LAB1:  MOV     EDX,AUX01
          MOV     TEXTPTR,EDX     ;RESTORE TEXTPTR
OP$LAB2:  MOV     EDX,AUX00       ;PC'S = OPERATOR CODE
          TEST    R0,0FH          ;BRANCH IF '$' OR 'DIGIT' FOUND ('C' PART OF VARIABLE NAME)
          JZ      OPRATR$5
          OR      DH,40H          ;SET 'C BIT IN OPERATOR CODE

OPRATR$5: TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
          JZ      OPRATR$9
          MOV     AUX0F,EDX
OPRATR$9: RET

OP$DAC:   MOV     EDX,4201H
          JMP     OPRATR$5
OP$DSC:   MOV     EDX,4301H
          JMP     OPRATR$5
OP$BOOL:  CALL    HEXDIGIT        ;LOOK FOR HEXDIGIT
          TEST    R0,0F0H         ;BRANCH IF NOT FOUND
          JZ      OP$LAB3         ;
          JMP     ERROR_17        ;ERROR IF NOT FOUND
OP$LAB3:  MOV     DH,R0           ;
          OR      DH,40H          ;OPERATOR CODE = BOOLX
          MOV     DL,0            ;
          JMP     OPRATR$5        ;


;      OPLGCL: PERFORM THE SPECIFIED LOGICAL OPERATION
;              (AND, OR, XOR, BOOL) ON THE SPECIFIED
;              ALPHA STRINGS IF IT IS EXECUTION PHASE.
;
;           OPERAND1 + OPERAND2 --> OPERAND1
;
;           ENTRY (FOR EXECUTION PHASE):
;                  PC'S = ADDRESS OF OPERAND1
;                  F1,0 = LENGTH OF OPERAND1
;
;                  F7,BIT80= 0 IF STRING-TO-STRING OPERA-
;                              TION. (CORRESPONDING CHAR'S
;                              OF OPERANDS ARE OPERATED ON
;                     AUX0 = ADDRESS OF 2ND OPERAND
;                     F3,2 = LENGTH OF OPERAND2
;
;                  F7,BIT80= 1 IF IMMEDIATE OPERATION.
;                              (EACH CHARACTER OF OPERAND1
;                              IS OPERATED ON WITH THE
;                              IMMEDIATE CHARACTER
;                     F4 = IMMEDIATE CHARACTER
;
;                  F7,BITS 7F = LOGICAL OPERATION
;                               00 IF AND
;                               10 IF OR
;                               20 IF XOR
;                               4X IF BOOL WERE X = BOOL
;                                    FUNCTION CODE
;
;           MODIFIED REGISTERS: PC'S,CHCL,K,F0-6,AUX0

OPLGCL:   TEST    SL,2            ;
          JZ      OPLGCL1         ;EXIT IF NOT EXECUTION PHASE

        mov     ecx,0                   ;clear reg
        mov     cx,word ptr [r0]        ;length of oprd1
        mov     word ptr [r0],0         ;for safe
        mov     ah,r7                   ;opertion type
        test    ah,80h                  ;br, if immediate character
        jnz     oplg_0                  ;

        mov     esi,aux00               ;addr of oprd2
        mov     bx,word ptr [r2]        ;length of oprd2
        cmp     cx,bx                   ;br, if length of oprd1 <= oprd2
        jbe     oplg_5
        mov     cx,bx                   ;take the smaller length
oplg_5:
        test    ah,10h                  ;br, if OR
        jnz     oplg_11                 ;
        test    ah,20h                  ;br, if XOR
        jnz     oplg_12                 ;
        test    ah,40h                  ;br, if BOOL
        jnz     oplg_13                 ;
oplg_14:                ;operation AND
        lodsb                           ;get character of oprd2
        and     byte ptr [edx],al       ;save result
        inc     edx                     ;update ptr
        loop    oplg_14                 ;
        ret
oplg_11:                ;operation OR
        lodsb                           ;get character of oprd2
        or      byte ptr [edx],al       ;save result
        inc     edx                     ;update ptr
        loop    oplg_11                 ;
        ret
oplg_12:                ;operation XOR
        lodsb                           ;get character of oprd2
        xor     byte ptr [edx],al       ;save result
        inc     edx                     ;update ptr
        loop    oplg_12                 ;
        ret
;
;PROCESSING IMMEDIATED CHARACTER CASE
;
oplg_0:
        mov     bl,r4                   ;get immediate character
        test    ah,10h                  ;br, if OR
        jnz     oplg_1                  ;
        test    ah,20h                  ;br, if XOR
        jnz     oplg_2                  ;
        test    ah,40h                  ;br, if BOOL
        jnz     oplg_3                  ;
oplg_4:                 ;operation AND
;       mov     al,[edx]                ;get character of oprd1
;       and     al,bl                   ;
        and     byte ptr [edx],bl       ;save result
        inc     edx                     ;update ptr
        loop    oplg_4                  ;
        ret
oplg_1:                 ;operation OR
;       mov     al,[edx]                ;get character of oprd1
;       or      al,bl                   ;
        or      byte ptr [edx],bl       ;save result
        inc     edx                     ;update ptr
        loop    oplg_1                  ;
        ret
oplg_2:                 ;operation XOR
;       mov     al,[edx]                ;get character of oprd1
;       xor     al,bl                   ;
        xor     byte ptr [edx],bl       ;save result
        inc     edx                     ;update ptr
        loop    oplg_2                  ;
        ret
;
;PROCESSING BOOL FUNCTIONS........
;       register use --
;       CL : bit count (F2)     AL : result (F3)
;       BL : type (K)           AH : boolean operation code (F7)
;       BH : oprd1 char (F6)    CH : oprd2 char (F5)
;       DI : length count (F0,1)
;       ESI: ptr of oprd2       EDX: ptr of oprd1

oplg_13:
        lodsb                           ;get character of oprd2
        mov     r4,al                   ;make it unified
oplg_3:
        mov     di,cx                   ;save length count
oplg_30:
        mov     cl,0                    ;clear bit count
        mov     al,0                    ;clear result byte
        mov     bh,byte ptr [edx]       ;get character of oprd1
        mov     ch,r4                   ;get charcter of oprd2 or immediate character
oplgcl3:
        mov     bl,3                    ;k=3 if high order bit of op1 is 0
        test    bh,80h
        jz      oplgcla
        mov     bl,0ch                  ;otherwise k = c
oplgcla:
        test    ch,80h                  ;br, if high order bit of op2 is 1
        jnz     oplgcl4
        and     bl,5
oplgcl5:
        and     bl,ah
        shl     al,1                    ;cal result
        mov     [edx],al
        test    bl,0fh
        jz      oplgclb
        or      al,1                    ;low order bit of result = 1 iff k<>0
        mov     [edx],al
oplgclb:
        shl     bh,1                    ;shift op1, op2 left 1 bit
        shl     ch,1                    ;
        inc     cl                      ;
        cmp     cl,8                    ;br, if not through a byte
        jb      oplgcl3                 ;
oplgcl2:
        inc     edx                     ;update addr of op1
        dec     di                      ;br, if not complete
        jnz     oplg_31                 ;
        ret                             ;

oplg_31:
        test    ah,80h                  ;br, if immediate character
        jnz     oplg_30                 ;
        lodsb                           ;get next character of op2
        mov     r4,al                   ;
        jmp     oplg_30                 ;
oplgcl4:
        and     bl,0ah
        jmp     oplgcl5
OPLGCL1:
        RET


;      OPARITH: PERFORM THE SPECIFIED ARITHMETIC OPERATION
;               (ADD,ADDC,SUB,SUBC,DAC,DSC) ON THE SPECI-
;               FIED STRINGS IN EXECUTION PHASE
;
;           OPERAND1 + OPERAND2 --> OPERAND1
;
;           ENTRY (FOR EXECUTION PHASE):
;                PC'S = ADDRESS OF OPERAND1
;                F1,0 = LENGTH OF OPERAND1
;                F7,BIT7 = 0 IF STRING-TO-STRING OPERATION
;                     AUX0 = ADDRESS OF OPERAND2
;                     F3,2 = LENGTH OF OPERAND2 (=0 IF ALL
;                                                FUNCTION)
;
;                F7,BIT7 = 1 IF IMMEDIATE OPERATION
;                     F4 = IMMEDIATE CHARACTER
;                     F3,2 = 0 IF OPERAND 2 IS ALL FUNC.
;
;                F7,BITS0-3 = ARITHMETIC OPERATION
;                             0 IF ADD
;                             1 IF SUB
;                             2 IF DAC
;                             3 IF DSC
;
;                F7,BIT6 = 0 IF NO CARRY BETWEEN CHARACTER
;                          1 IF CARRY
;
;           MODIFIED REGISTERS: PC'S,CH,CL,F0-6,AUX0,K

;
;       New Register Use --
;       AH : operation code (F7)        AL : partial opreation code (F7 and 0FH)
;       BH : t_carry                    BL : immediate char (F1)
;       CL : immediate char* (F4)       CH : oprd1 char (F0)
;       EDX: ptr of beg of oprd1 (A2,3) ESI: ptr of end of oprd1 (edx)
;       EDI: ptr of end of oprd2 (aux0) AUX0:ptr of beg of oprd2 (A5,6)

qpar_3:
        jmp     opar_3
qpar_0:
        jmp     qpar_1
oparith:
        test    sl,2            ;br, if not xphase
        jz      opar_1          ;

        mov     bh,0            ;init temp carry
        mov     ah,r7           ;get operation code
        mov     al,ah           ;
        and     al,0fh          ;take low nibble
        mov     cl,r4           ;get the immediate char

        mov     k,0             ;k = 0 iff oprd2 length = 0
        cmp     word ptr [r2],0 ;chk oprd2 length
        jne     opar_a          ;
        mov     k,0ffh          ;k = 0ffh iff oprd2 <> 0
opar_a:
        mov     edi,0           ;init
        mov     di,word ptr [r2];get length of oprd2
        add     edi,aux00       ;add the starting pt of oprd2
        dec     edi             ;pt to the end char of oprd2
opar_2:
        mov     esi,0           ;init
        mov     si,word ptr [r0];get length of oprd1
        add     esi,edx         ;add the starting pt of oprd1
        dec     esi             ;pt to the end of oprd1
opar_4:
        cmp     edx,esi         ;more char in oprd1
        jbe     opar_b          ;br, if yes
opar_1:                         ;no, we finish
        ret
opar_b:
        mov     ch,[esi]        ;get char of oprd1
        test    ah,80h          ;immediate char operation?
        jz      opar_3          ;br, if not
        mov     bl,cl           ;get immediate char
        test    ah,40h          ;br, no carry between chars?
        jz      opar_c          ;
        and     cl,k            ;
opar_6:
        test    ah,40h          ;br, if carry between chars?
        jnz     qpar_1          ;
opar_c:
        mov     bh,0            ;tem_carry = 0
qpar_1:
        cmp     al,0            ;br, if not ADD
        jne     opar_d          ;
        add     bh,0ffh         ;be sure carry in (no matter)
        adc     ch,bl           ;operate on oprd1 & oprd2 (or imme char)
        mov     [esi],ch        ;save back
        mov     bh,0            ;clear tem carry
        rcl     bh,1            ;make sure last operation carry in tem carry
        dec     esi             ;pt to next char in oprd1
        jmp     opar_4          ;try next
opar_d:
        cmp     al,1            ;br, if not SUB
        jne     opar_e          ;
        add     bh,0ffh         ;be sure carry in (no matter)
        sbb     ch,bl           ;operate on oprd1 & oprd2 (or imme char)
        mov     [esi],ch        ;save back
        mov     bh,0            ;clear tem carry
        rcl     bh,1            ;make sure last operation carry in tem carry
        dec     esi             ;pt to next char in oprd1
        jmp     opar_4          ;try next
opar_e:
        cmp     al,2            ;br, if not DAC
        jne     opar_f          ;
        add     bh,0ffh         ;be sure carry in (no matter)
        xchg    al,ch           ;DAA needs al
        adc     al,bl           ;operate on oprd1 and oprd2 (or imme char)
        daa                     ;
        mov     [esi],al        ;save bach
        mov     al,ch           ;restore operation type
        mov     bh,0            ;clear tem carry
        rcl     bh,1            ;make sure last operation carry in tem carry
        dec     esi             ;pt to next char in oprd1
        jmp     opar_4          ;try next
opar_f:
                                ;operation DSC
        add     bh,0ffh         ;be sure carry in (no matter)
        xchg    al,ch           ;DAS need al
        sbb     al,bl           ;operate on oprd1 and oprd2 (or imme char)
        das                     ;
        mov     [esi],al        ;save back
        mov     al,ch           ;restore operation type
        mov     bh,0            ;clear tem carry
        rcl     bh,1            ;make sure last operation carry in tem carry
        dec     esi             ;pt to next char in oprd1
        jmp     opar_4          ;try next
opar_3:
        cmp     aux00,edi       ;if more char in oprd2?
        jbe     opar_5          ;br, if yes
        mov     bl,0            ;no more
        jmp     opar_6          ;
qpar_4:
        jmp     opar_4          ;
opar_5:
        mov     bl,[edi]        ;get next char of oprd2
        dec     edi             ;
        jmp     opar_6          ;

;   SLCTFNC:  process an alphanumeric SELECT function.
;             SELECT returns the device address for the specified parameter.  The
;             width for output devices is also returned.
;
;         Format:  A$ = SELECT select-parameter
;
;         Entry:  TEXTPTR points after SELECT
;
;         Return:
;                   ERROR#19 if SELECT found, but no select parameter
;
;                   F0 = 0 if found
;                          TEXTPTR points after select parameter
;                          In execution phase,
;                               F3F2 = length of value
;                               AUX0 = address of value
;
;                               Returned value is taa   (ASCII)
;                                  where: t = device type
;                                        aa = device address
;

SLCTFNC:  MOV     EDX,OFFSET _SLCTLST
          CALL    CCJB            ;LOOK FOR SELECT PARAMETER
          JMP     SLCFPRNT        ;SELECT PRINT
          JMP     SLCFCO          ;SELECT CO
          JMP     SLCFTAPE        ;SELECT TAPE
          JMP     SLCFCI          ;SELECT CI
          JMP     SLCFDISK        ;SELECT DISK
          JMP     ERROR_19        ;SELECT DRIVER
          JMP     ERROR_19        ;SELECT D
          JMP     ERROR_19        ;SELECT ROUND
          JMP     ERROR_19        ;SELECT NO
          JMP     ERROR_19        ;SELECT ERROR
          JMP     SLCFLIST        ;SELECT LIST
          JMP     ERROR_19        ;SELECT R
          JMP     SLCF_           ;SELECT #
          JMP     ERROR_19        ;SELECT G
          JMP     SLCFPLOT        ;SELECT PLOT
          JMP     ERROR_19        ;SELECT P
          JMP     ERROR_19        ;SELECT LINES
          JMP     SLCFINPT        ;SELECT INPUT
          JMP     ERROR_19        ;SELECT ON
          JMP     ERROR_19        ;SELECT OFF
          JMP     ERROR_19        ;SELECT @
          JMP     ERROR_19        ;SELECT TC
          JMP     ERROR_19        ;SELECT TERMINAL
          JMP     ERROR_19        ;SELECT NEW
          JMP     ERROR_19        ;SELECT OLD
          JMP     ERROR_19        ;SELECT TON
          JMP     ERROR_19        ;SELECT TOFF
          JMP     ERROR_19        ;SELECT HON
          JMP     ERROR_19        ;SELECT HOFF
          JMP     ERROR_19        ;SELECT KON
          JMP     ERROR_19        ;SELECT KOFF

SLCFPRNT: MOV     EDX,DWORD PTR [CPBASE]     ;GET DEVICE TYPE & ADDRESS (CHCL)
          MOV     CX,WORD PTR [EDX].@DEVPRNT
          XCHG    CH,CL                      ;CH = DEVICE TYPE , CL = DEVICE ADDRESS
          JMP     SLCF1
SLCFCO:   MOV     EDX,DWORD PTR [CPBASE]
          MOV     CX,WORD PTR [EDX].@DEVCO
          XCHG    CH,CL
          JMP     SLCF1
SLCFTAPE: MOV     EDX,DWORD PTR [CPBASE]
          MOV     CX,WORD PTR [EDX].@DEVTAPE
          XCHG    CH,CL
          JMP     SLCF1
SLCFCI:   MOV     EDX,DWORD PTR [CPBASE]
          MOV     CX,WORD PTR [EDX].@DEVCI
          XCHG    CH,CL
          JMP     SLCF1           ;
SLCFDISK: MOV     EDX,DWORD PTR [CPBASE]
          MOV     CX,WORD PTR [EDX].@DEVDISK
          XCHG    CH,CL
          JMP     SLCF1           ;
SLCFLIST: MOV     EDX,DWORD PTR [CPBASE]
          MOV     CX,WORD PTR [EDX].@DEVLIST
          XCHG    CH,CL
          JMP     SLCF1           ;
SLCFPLOT: MOV     EDX,DWORD PTR [CPBASE]
          MOV     CX,WORD PTR [EDX].@DEVPLOT
          XCHG    CH,CL
          JMP     SLCF1
SLCFINPT: MOV     EDX,DWORD PTR [CPBASE]
          MOV     CX,WORD PTR [EDX].@DEVINPT
          XCHG    CH,CL
          JMP     SLCF1

SLCF_:    CALL    FILE_B          ;GET FILE#
          TEST    SL,XPHASE       ;BRANCH IF NOT EXECUTION PHASE
          JZ      SLCF8           ;
          CALL    FILE_PTR        ;GET POINTER TO DEVICE TABLE SLOT
          SUB     EDX,2
          MOV     CX,WORD PTR [EDX]   ;GET DEVICE TYPE AND ADDRESS
          XCHG    CH,CL

SLCF1:    TEST    SL,XPHASE           ;BRANCH IF NOT EXECUTION PHASE
          JZ      SLCF8
          MOV     EDX,OFFSET SCRATCH
          MOV     AUX00,EDX           ;ADDRESS OF VALUE BUFFER
          AND     CH,0FH
          ADD     CH,30H
          CMP     CH,39H
          JBE     SLCFA
          ADD     CH,07H
SLCFA:    MOV     BYTE PTR [EDX],CH   ;DEVICE TYPE --> BUFFER
          INC     EDX
          MOV     AL,CL
          SHR     AL,4
          ADD     AL,30H
          CMP     AL,39H
          JBE     SLCFB
          ADD     AL,07H
SLCFB:    MOV     BYTE PTR [EDX],AL   ;HIGH DEVIEC ADDRESS DIGIT --> BUFFER
          INC     EDX
          AND     CL,0FH
          ADD     CL,30H
          CMP     CL,39H
          JBE     SLCFC
          ADD     CL,07H
SLCFC:    MOV     BYTE PTR [EDX],CL   ;LOW DEVICE ADDRESS DIGIT --> BUFFER
          MOV     WORD PTR [R2],3     ;LENGTH OF VALUE
SLCF8:    MOV     R0,0                ;FLAG = FOUND
SLCF9:    RET

CODE      ENDS
          END
