;===========================================================================
;
;     P R O G R A M   E N C R Y P T I O N   /  D E C R Y P T I O N
;
;===========================================================================
;           This module was originally M24
;===========================================================================
	    .XLIST
	    INCLUDE PREAMBLE
	    .LIST

SAWM    EQU     10H

PUBLIC  DCRYPT
PUBLIC  NCRYPT

EXTRN   CPBASE          :NEAR
EXTRN   SAVESEED        :BYTE

;*  RELINK:    link the scramble from the PREVIOUS sector to this one
;*                  if the wrap mode is on.
;*
;*   ENTRY:   F1F0 contain previous checksum and  contain wrap mode flag
;*                  (from SAVEA)
;*   RETURN:  about a quarter of the sector has been XOR'ed.
;*   Modified regs: PHPL, F1F0,  CHCL
;@3.0SCRA
;*   This link stuff is a new wrinkle to the scramble for 3.0
;*      It couples the sectors together to prevent someone from mixing
;*      two or more scrambled programs together, deleting sectors, or
;*      otherwise rearranging.
RELINK:
	TEST    R0,1
	JZ      RELINKR         ;exit if wrapmode not on
	OR      R0,3            ;bottom two bits=don't care
	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
	ADD     EDX,4           ;Pick a starting address such
	MOV     BL,4

RELINK3:
	ADD     R0,47H          ;change F0,F1
	ADD     R1,2FH
	MOV     CX,[EDX]
	XCHG    CH,CL
	XOR     CH,R0           ;XOR the byte in the sector image

;*          it's important here that the W1 and CH go together, and the W2 and CL
;*          so it's reversible with the same code.
	MOV     [EDX],CH
	XOR     CL,R1
	MOV     [EDX+1],CL
	ADD     EDX,8           ;service 2 of every 8 bytes
	ADD     BL,8            ;for nearly all of sector
	MOV     AL,BL
	AND     AL,0F0H
	CMP     AL,0F0H
	JNZ     RELINK3
RELINKR:
	RET


;
;*  LINK:    calculate a checksum from this sector, for use in next
;*
;*   ENTRY:
;*
;*   RETURN:  SAVEA contains a checksum of bytes 02-FF of  RIOBUFFR
;*            CHCL contain first two bytes of the buffer (unchanged)
;*   Modified regs: PHPL, F1F0,  CHCL
LINK:
	MOV     EDX,DWORD PTR [CPBASE]    ;2 byte scratch area used currently only
	LEA     EDX,[EDX].@RSAVE          ;only by SAVE/LOAD during X phase.
	MOV     CX,WORD PTR [EDX]
	XCHG    CH,CL
	MOV     WORD PTR R0,CX
	AND     R0,0FFH-03      ;F1F0 = old seed
	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
;* Calculate LINK  (checksum) from this unscrambled sector
	ADD     EDX,2
	CLC
	PUSHF

LINK4:
	MOV     CX,[EDX]
	MOV     BX,CX
	XCHG    CH,CL
	ADD     EDX,2
	MOV     AX,WORD PTR R0
	XOR     AX,CX
	POPF
	ADC     AX,AX           ;The LINK is a checksum of the unscrambled sector
	PUSHF
	MOV     WORD PTR R0,AX
	MOV     CX,BX
	MOV     AX,WORD PTR R0
	XOR     AX,CX
	POPF
	ADC     AX,AX
	PUSHF
	MOV     WORD PTR R0,AX
	CMP     EDX,DWORD PTR A5
	JNZ     LINK4
	POPF

	MOV     EDX,DWORD PTR [CPBASE]    ;2 byte scratch area used currently only
	LEA     EDX,[EDX].@RSAVE          ;only by SAVE/LOAD during X phase.
	MOV     CX,WORD PTR [EDX]
	XCHG    CH,CL
	AND     R0,0FFH-01
	OR      R0,02           ;turn on scramble flag
	TEST    CL,1
	JZ      DA24_01
	OR      R0,01           ;preserve bit 0001, wrap mode bit
DA24_01:
	MOV     AL,R1
	MOV     [EDX],AL
	MOV     AL,R0
	MOV     [EDX+1],AL      ;save link from this sector
	RET

;*
;*  DCRYPT: called from LOAD for protected program blocks
;*
;*  ENTRY:  sector image is in IOBUFFER.
;*
;*  RETURN:  CHCL = unscrambled first two bytes.
;*     PC'S point to IOBUFFER
;*  MODIFIED REGISTERS: F0,F1, PC'S, (CHCL)
;*   (Other registers are saved in the stack and restored)
;*======> wrap mode modification:  First, based on a checksum of the PREVIOUS
;*        sector, a variable byte is XOR'ed, then use the old algorithm.
;*
;* algorithm:  First, the low half of the first two bytes
;*    are combined, and exchanged with an embedded code
;*    byte in the record.  This code byte is put into F0
;*    and a zero is put in F1 to begin pass 1 of the
;*    decrypting.
;
DCRYPT:
	PUSH    DWORD PTR A6
	PUSH    DWORD PTR A4
	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
	MOV     DWORD PTR A0,EDX        ;A1A0 = START ADDRESS OF I/O BUFFER
	MOV     DWORD PTR A5,EDX
	ADD     DWORD PTR A5,100H       ;A6A5 = END ADDRESS OF I/O BUFFER
	MOV     EDX,DWORD PTR [CPBASE]    ;2 byte scratch area used currently only
	LEA     EDX,[EDX].@RSAVE          ;only by SAVE/LOAD during X phase.
	MOV     CX,WORD PTR [EDX]
	XCHG    CH,CL
	MOV     WORD PTR R0,CX
	CALL    RELINK
	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
	MOV     CX,WORD PTR [EDX]
	XCHG    CH,CL
	MOV     AH,CH
	AND     AH,0F0H
	MOV     AL,CL
	SHR     AL,4
	OR      AH,AL
	MOV     [EDX],AH

;*_HIGH HALVES OF THE FIRST TWO CODED BYTES
	MOV     AH,CH           ;SAVE THE LOW HALVES
	SHL     AH,4
	MOV     AL,CL
	AND     AL,0FH
	OR      AH,AL
	MOV     R0,AH
	AND     CL,60H
	MOV     EDX,DWORD PTR A0
	AND     ECX,0FFH
	ADD     EDX,ECX
	ADD     EDX,5          ;THE PASS 1 KEY IS IN BYTE  5,25,45, OR 65,

;*_DEPENDING ON THE TWO MIDDLE CODED BITS OF THE 2ND BYTE
	MOV     CH,[EDX]        ;READ THE PASS 1 KEY
	MOV     AL,R0           ;WRITE THE REPLACEMENT (FROM LOW HALVES)
	MOV     [EDX],AL
	MOV     EDX,DWORD PTR A0
	MOV     [EDX+1],CH      ;WRITE THE KEY IN THE SECOND BYTE
	PUSH    WORD PTR R2
	PUSH    WORD PTR R4     ;SAVE F5-F2 IN STACK

;*% DECRYPT PASS 2
	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
	ADD     EDX,100H-22H    ;POINT TO PASS 2 SEED
	MOV     CX,[EDX]        ;TERMINATOR IS SEED LOCATION, READ SEED
	XCHG    CH,CL           ;26 (2*13) IS INCREMENT
	MOV     DWORD PTR A3,EDX;PHPL IS STARTING LOCATION FOR LOOP
	MOV     R4,1AH
	ADD     EDX,1AH

DCRYPT10:
	MOV     WORD PTR R0,CX  ;ROTATE 1
	ADD     WORD PTR R0,CX  ;F2 = LOW PARTS ;IF THERE IS CARRY FLAG , ADD 1 MORE
	ADC     WORD PTR R0,0   ;F1 = HIGH PARTS
	MOV     CX,[EDX]        ;WRITE DECODED LOW BYTE
	XCHG    CH,CL           ;WRITE DECODED HIGH BYTE
				;ADD 26 (2 * 13)
	MOV     AH,R1
	SHL     AH,4            ;MODULO 254
	MOV     AL,R0           ;LOOP BACK IF NOT AT F3
	AND     AL,0FH
	OR      AH,AL
	MOV     R2,AH

	MOV     AH,R1
	AND     AH,0F0H
	MOV     AL,R0
	SHR     AL,4
	OR      AH,AL
	MOV     R1,AH


	MOV     AX,CX
	STC
	SBB     AL,R2
	MOV     [EDX+1],AL
	SBB     AH,R1
	MOV     [EDX],AH

	MOV     AL,R4
	AND     EAX,0FFH
	ADD     EDX,EAX
	CMP     EDX,DWORD PTR A5
	JB      DA24_02
	MOV     EAX,EDX
	SUB     EAX,DWORD PTR A5
	MOV     EDX,DWORD PTR A0
	ADD     EDX,EAX
	ADD     EDX,2
DA24_02:
	CMP     EDX,DWORD PTR A3
	JNZ     DCRYPT10        ;LOOP BACK IF NOT AT F3

;*% DECRYPT PASS 1
	MOV     EDX,DWORD PTR A0;POINT TO PASS 1 SEED, READ IT NEXT CYCLE
	MOV     CX,[EDX]        ;STOP WHEN RETURN TO SEED LOCATION
	XCHG    CH,CL           ;PHPL IS STARTING LOCATION FOR LOOP
	MOV     DWORD PTR A3,EDX
	ADD     EDX,22H         ;F1F0 = PASS 1 SEED
	MOV     R0,CL
	ADD     R0,CL
	MOV     R1,0
	ADC     R1,0
	MOV     CX,[EDX]
	XCHG    CH,CL
	JMP     DA24_03

DCRYPT20:
	MOV     WORD PTR R0,CX  ;ROTATE 1
	ADD     WORD PTR R0,CX  ;F2 = LOW PARTS
	MOV     CX,[EDX]        ;F1 = HIGH PARTS
	XCHG    CH,CL           ;WRITE DECODED LOW BYTE
DA24_03:                        ;WRITE DECODED HIGH BYTE
	ADC     WORD PTR R0,0   ;ADD 34 (2 * 17) MODULO 256
	MOV     AH,R1           ;LOOP BACK IF NOT AT F3
	SHL     AH,4
	MOV     AL,R0
	AND     AL,0FH
	OR      AH,AL
	MOV     R2,AH

	MOV     AH,R1
	AND     AH,0F0H
	MOV     AL,R0
	SHR     AL,4
	OR      AH,AL
	MOV     R1,AH

	MOV     AX,CX
	STC
	SBB     AL,R2
	MOV     [EDX+1],AL
	SBB     AH,R1
	MOV     [EDX],AH

	ADD     EDX,22H                 ;ADD 22 & MODULE 256
	CMP     EDX,DWORD PTR A5
	JB      DA24_10
	MOV     EAX,EDX
	SUB     EAX,DWORD PTR A5
	MOV     EDX,DWORD PTR A0
	ADD     EDX,EAX
DA24_10:
	CMP     EDX,DWORD PTR A3
	JNZ     DCRYPT20

;*% RECONSTRUCT FIRST TWO BYTES
	MOV     EDX,DWORD PTR A0;RECREATE FIRST BYTE (HIGH HALF ONLY)
	MOV     CX,[EDX]
	XCHG    CH,CL           ;RECREATE SECOND BYTE (FF OR FE)
	MOV     AH,CH           ;calculate checksum for next time
	AND     AH,0F0H         ;EXIT
	MOV     [EDX],AH
	AND     CH,01
	OR      CH,0FEH
	MOV     [EDX+1],CH
	CALL    LINK
	JMP     NCRYPT90

;         TITLE     SUBROUTINES
;*  ENCODE SUBROUTINE:
;*   F1F0 == SEED
;*   F5 = INCREMENT
;*   F4 = MODULO
;*   A4A3 = TERMINATOR ADDRESS
ENCODE:
	MOV     CX,WORD PTR R0  ;ROTATE 1
	ADD     WORD PTR R0,CX  ;F2 = LOW PARTS
	ADC     WORD PTR R0,0   ;F1 = HIGH PARTS
	MOV     CX,[EDX]        ;WRITE ENCODED LOW BYTE
	XCHG    CH,CL           ;WRITE ENCODED HIGH BYTE
				;ADD F5 ---
	MOV     AH,R1
	SHL     AH,4            ;MODULO F4
	MOV     AL,R0           ;LOOP BACK IF NOT AT F3
	AND     AL,0FH          ;RETURN
	OR      AH,AL
	MOV     R2,AH

	MOV     AH,R1
	AND     AH,0F0H
	MOV     AL,R0
	SHR     AL,4
	OR      AH,AL
	MOV     R1,AH

	STC
	ADC     CL,R2
	MOV     R0,CL
	MOV     [EDX+1],CL
	ADC     R1,CH
	MOV     CH,R1
	MOV     [EDX],CH

	MOV     AL,R5
	AND     EAX,0FFH
	ADD     EDX,EAX
	CMP     EDX,DWORD PTR A5
	JB      DA24_04

	MOV     EAX,EDX
	SUB     EAX,DWORD PTR A5
	MOV     EDX,DWORD PTR A0
	ADD     EDX,EAX
	MOV     AL,R4
	AND     EAX,0FFH
	ADD     EDX,EAX
DA24_04:
	CMP     EDX,DWORD PTR A3
	JNZ     ENCODE
ENCODER:
	RET

;         TITLE     NCRYPT
;*  NCRYPT: (called from SAVE for every sector of body of a
;*     scramble protected program, not header)
;*  ENTRY:  no parameters.
;*  RETURN:  none
;*
;*  MODIFIED REGISTERS: F0,F1, PC'S, (CHCL), AUX 0
;*     PC's, CHCL only, if header block
;*   (Other registers are saved in the stack and restored)
NCRYPT:
	PUSH    DWORD PTR  A6
	PUSH    DWORD PTR  A4
	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
	MOV     DWORD PTR A0,EDX        ;A1A0 = START ADDRESS OF I/O BUFFER
	MOV     DWORD PTR A5,EDX
	ADD     DWORD PTR A5,100H       ;A6A5 = END ADDRESS OF I/O BUFFER
	PUSH    WORD PTR R2
	PUSH    WORD PTR R4
	MOV     EDX,DWORD PTR [CPBASE]    ;2 byte scratch area used currently only
	LEA     EDX,[EDX].@RSAVE          ;only by SAVE/LOAD during X phase.
	MOV     CX,WORD PTR [EDX]
	XCHG    CH,CL
	PUSH    CX
	CALL    LINK                    ;checksum this sector & put in SAVEA
	PUSH    WORD PTR R6
	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
	MOV     CX,WORD PTR [EDX]
	XCHG    CH,CL

;*   NOTE:  FIRST BYTE HAS ZERO AS LOW ORDER DIGIT,
;*          SECOND IS EITHER FF OR FE.

	AND     CL,1            ;(00 IF FE, 01 IF FF)
	SHL     CL,4            ;F7 = ALL INFORMATION FROM FIRST TWO BYTES
	SHR     CH,4
	OR      CL,CH
	MOV     R7,CL

;*_(HIGH DIGIT DERIVED FROM SECOND BYTE, LOW FROM FIRST)
	MOV     EDX,OFFSET SAVESEED
	MOV     CX,[EDX]
	XCHG    CH,CL
	MOV     WORD PTR R2,CX  ;F3F2 =HIGH HALF OF 'SAVE SEED'

	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
	ADD     EDX,2
	MOV     CX,[EDX]
	XCHG    CH,CL
	MOV     WORD PTR R0,CX
	CALL    BRANDOM
	MOV     EDX,OFFSET SAVESEED
	MOV     DWORD PTR [EDX],01020304H
	ADD     EDX,4

	MOV     AL,R1
	AND     AL,60H
	MOV     R6,AL
	OR      R7,AL           ;COMBINE THE TWO SELECT BITS INTO F7
	MOV     EDX,DWORD PTR [CPBASE]
	LEA     EDX,[EDX].@IOBUFFR
	MOV     R1,0            ;F1F0 = SEED FOR PASS 1
	MOV     AL,R1           ;WRITE **RANDOM** SEED IN FIRST TWO BYTES
	MOV     [EDX],AL
	MOV     AL,R3
	MOV     R0,AL
	MOV     [EDX+1],AL


;*%  ENCRYPT - PASS 1
	MOV     R5,22H          ;INCREMENT IS 34 (2 * 17)
	MOV     R4,0            ;MODULO IS 256
	MOV     DWORD PTR A3,EDX

;*_TERMINATOR IS 00
	ADD     EDX,22H
	CALL    ENCODE          ;ENCODE PASS 1

;*%  ENCRYPT - PASS 2
	MOV     R5,1AH          ;INCREMENT IS 26 (2*13)
	MOV     R4,02           ;MODULO IS 254
	MOV     EDX,DWORD PTR A0
	ADD     EDX,0DEH        ;TERMINATOR = SEED LOCATION
	MOV     DWORD PTR A3,EDX;ENCODE PASS 2
	MOV     EDX,DWORD PTR A0
	ADD     EDX,100H-22H+1AH
	CALL    ENCODE

;*%  EMBED THE SEED
	MOV     EDX,DWORD PTR A0
	MOV     CX,[EDX]
	XCHG    CH,CL
	ADD     EDX,5
	MOV     AL,R6
	AND     EAX,0FFH
	ADD     EDX,EAX
	MOV     CH,[EDX]
	MOV     [EDX],CL
	MOV     EDX,DWORD PTR A0
	MOV     CL,CH           ;FIRST ENCODED BYTE (HIGH HALF UNCHANGED)
	MOV     AL,R7           ;SECOND ENCODED BYTE ( < HEX 80 )
	SHL     AL,4
	SHR     CH,4
	OR      AL,CH
	MOV     [EDX],AL
	MOV     AL,R7
	AND     AL,0F0H
	AND     CL,0FH
	OR      AL,CL
	MOV     [EDX+1],AL

	POP     WORD PTR R6
	POP     WORD PTR R0
	TEST    R6,SAWM         ;restore last sector's checksum
	JZ      NCRYPT89        ;skip if not wrap mode
	MOV     EDX,DWORD PTR [CPBASE]    ;2 byte scratch area used currently only
	LEA     EDX,[EDX].@RSAVE          ;only by SAVE/LOAD during X phase.
	MOV     CX,WORD PTR [EDX]
	XCHG    CH,CL
	OR      CL,1            ;flag wrap mode to LINK subroutine
	MOV     [EDX+1],CL

NCRYPT89:
	CALL    RELINK          ;xor a bunch of these bytes
NCRYPT90:
	POP     WORD PTR R4
	POP     WORD PTR R2
	POP     DWORD PTR A4
	POP     DWORD PTR A6
	RET

;*  BRANDOM:BINRAN: EXTRACT NEXT BINARY RANDOM NUMBER FROM
;*    SEQUENCE.
;*
;*  ENTRY:   LAST BINARY NUMBER (32 BITS) IS FOUND IN
;*     F3-F0.  (CALLED  'S' IN COMMENTS)
;*
;*  RETURN:  NEXT BINARY NUMBER (32 BITS) IS LEFT IN
;*     F3-F0.   (NEW VALUE OF 'S')
;*
;*  REGISTERS USED:  F3-F0, PC'S, AUX 0
;*
;*  MATHA0      EQU    AUX00
;*
;*   S = VALUE OF F3F2F1F0 (32 BITS)
;*
BRANDOM:
	MOV     EAX,DWORD PTR R0
	MOV     EBX,EAX
	SHL     EBX,16
	ADD     EBX,EAX                 ;EBX =  S * (2 ^ 16 + 1)
	SHL     EAX,2                   ;EAX =  S * 4
	ADD     EBX,EAX                 ;EBX =  S * (2 ^ 16 + 5)
	MOV     DWORD PTR R0,EBX
	RET

CODE    ENDS
	END
