;===========================================================================
;
;
;
;===========================================================================
;           This module was originally M16
;===========================================================================
            .XLIST
            INCLUDE PREAMBLE
            .LIST

PUBLIC    $BIN
PUBLIC    ALPHARAY
PUBLIC    ALPHIN
PUBLIC    ALPHINL
PUBLIC    ARGUMENT
PUBLIC    ARGUMNTR
PUBLIC    ARRAYN
PUBLIC    AVARB
PUBLIC    AVARB2
PUBLIC    BINEXP
PUBLIC    BINVAL
PUBLIC    DEF1STMT
PUBLIC    ERR_08A
PUBLIC    ERR_51C
PUBLIC    ERR_75P
PUBLIC    ERROR_03
PUBLIC    ERROR_15
PUBLIC    ERROR_27
PUBLIC    ERROR_28
PUBLIC    EXCHELMT
PUBLIC    HEXFIX
PUBLIC    INITELMT
PUBLIC    INITELM1
PUBLIC    IOBUFTXT
PUBLIC    MATDSPLA
PUBLIC    MOVEELMT
PUBLIC    MOVEVALF
PUBLIC    MOVEVALU
PUBLIC    NUMCH
PUBLIC    NXTSEP
PUBLIC    PUTHEX
PUBLIC    RCVASCII

EXTRN     ALPHA   :NEAR
EXTRN     ARECEVAL:NEAR
EXTRN     ARRAY   :NEAR
EXTRN     ARRAYA  :NEAR
EXTRN     AVAR    :NEAR
EXTRN     AVAR1   :NEAR
EXTRN     AVARD   :NEAR
EXTRN     AVARL2A :NEAR
EXTRN     ERROR   :NEAR
EXTRN     ERROR_08:NEAR
EXTRN     ERROR_24:NEAR
EXTRN     ERROR_29:NEAR
EXTRN     ERROR_51:NEAR
EXTRN     ERROR_52:NEAR
EXTRN     ERROR_57:NEAR
EXTRN     ERROR_75:NEAR
EXTRN     EXPR    :NEAR
EXTRN     FLTBINB :NEAR
EXTRN     LITERAL :NEAR
EXTRN     PMODE   :NEAR
EXTRN     SNMBR   :NEAR
EXTRN     VARA    :NEAR
EXTRN     VARNAME :NEAR
EXTRN     VARR    :NEAR

EXTRN     PART_   :BYTE
EXTRN     CPBASE  :BYTE

;------------------------------------------------------------------------------;
;                                                                              ;
;     *****************************                                            ;
;     *   MOD#16 SUBROUTINES      *                                            ;
;     *****************************                                            ;
;                                                                              ;
;------------------------------------------------------------------------------;

;      BINVAL: SCAN NUMERIC-EXPRESSION OR ALPHA-VARIABLE
;              AND CONVERT VALUE TO 16-BIT BINARY VALUE.
;
;           ENTRY: TEXTPTR POINTS AT NEXT TEXT CHARACTER
;
;           RETURN: TEXTPTR POINTS AFTER VAR. OR EXPR.
;             IN EXECUTION PHASE,
;                F1,0 = BINARY VALUE OF EXPRESSION OR
;                       1ST 2 BYTES OF ALPHA-VARIABLE
;
;           MODIFIED REGISTERS: ALL,AUX0,1,2
;

BINVAL:   CALL    AVARL2A               ;SCAN & PROCESS ALPHA VARIABLE
          TEST    R0,0FH                ;BRANCH IF NOT FOUND
          JNZ     BINVAL1               ;
          TEST    SL,2                  ;BRANCH IF NOT EXECUTION PHASE
          JZ      BINVAL2               ;
          MOV     EDX,VSPTR             ;GET ADDRESS OF VARIABLE
          MOV     EDX,DWORD PTR [EDX+4] ;EDX = ADDRESS OF VARIABLE
          MOVZX   ECX,WORD PTR [EDX]     ;ECX = FIRST TWO BYTE OF ALPHA VARIABLE
          XCHG    CH,CL
          MOV     DWORD PTR [A0],ECX    ;F1,0 = BINARY VALUE ( 1ST 2 BYTES OF VARIABLE)
          ADD     VSPTR,8
BINVAL2:  RET

BINVAL1:  CALL    EXPR                  ;SCAN & PROCESS EXPRESSION
          TEST    SL,2                  ;BRANCH IF NOT EXECUTION PHASE
          JZ      BINVAL2
          JMP     FLTBINB               ;CONVERT VALUE TO BINARY (ERROR IF <0 OR > 65535)

;      IOBUFTXT: PC'S & AUXC POINT TO BEGIN OF I/O BUFFER.
;                (1ST EVEN ADDRESS PAST END OF TEXT)
;                AN ERROR RESULTS IF NOT ENOUGH ROOM FOR
;                THE SPECIFIED NO. OF SECTORS BETWEEN TEXT
;                AREA AND VALUE STACK.
;
;           ENTRY: TEXTEND POINTS TO END OF TEXT
;                  F2 = NO. OF SECTORS (256 * F2 BYTES)
;
;           RETURN: PC'S & AUXC POINT TO I/O BUFFER
;                   F1 = NO. OF PAGES OF MEMORY AVAILABLE
;                        FOR BUFFERS
;
;           MODIFIED REGISTERS: PC'S,AUXC,F0,1
;

IOBUFTXT: MOV     EDX,TEXTEND     ;
          INC     EDX             ;PC'S POINT TO TEXTEND + 1
          TEST    DL,1
          JZ      IOBUFX
          INC     EDX             ;INCREMENT PC'S IF ADDRESS IS ODD

IOBUFX:   MOV     EAX,VSPTR
          MOV     AUX0C,EDX       ;AUX0C = PC'S
          ADD     EDX,16          ;MARGIN FOR VS STACK
          XOR     EBX,EBX         ;EBX = MEMORY SIZE WANTED FOR TEXT BUFFER
          MOV     BH,R2           ;
          SUB     EAX,EDX
          CMP     EAX,EBX
          JB      ERROR_03        ;ERROR IF NOT ENOUGH
          mov     r1,ah
          test    eax,0ffff0000h
          jz      iobuftxt1
          MOV     R1,0FFH         ;F1 = NO. OF PAGES OF MEMORY AVAILABLE
iobuftxt1:
          MOV     EDX,AUX0C
          RET

ERROR_03: MOV     R0,3            ;NOT ENOUGH MEMORY
          JMP     ERROR           ;


;   ALPHIN:  SETUP POINTERS FOR AN ALPHA ARRAY WITH
;      POSSIBLE ARRAY MODIFIERS, AND OPTIONALLY A LITERAL
;      STRING.
;
;   ENTRY:  TEXTPTR POINTS AT NEXT CHARACTER.
;
;   ALTERNATE ENTRY - ALPHINL:  A LITERAL STRING ALSO
;      ACCEPTABLE.
;
;   RETURN:  ADDRESS, THEN LENGTH, ARE IN THE STACK.  IF
;      AN ARRAY MODIFIER WAS USED, LENGTH ZERO IS POSSIBLE
;      OTHERWISE, THE FORMAT MATCHES THAT USED BY AVAR.
;      LOW PART OF F0=0 IFF VARIABLE WAS FOUND.
;
;   MODIFIED REGISTERS:  ALL, AUX 0,1,2
;   STACK:  LENGTH IS TOP TWO BYTES IN STACK, ADDRESS IS
;      THE NEXT TWO.  (EXECUTION PHASE ONLY)
;
;  ALTERNATE ENTRY: ALPHINL - LOOK FOR LITERAL FIRST,
;     THEN AS ABOVE.  (PERHAPS MAY PUT IN A FLAG 10 IF
;     LITERAL FOUND)
;

ALPHINL:  CALL    LITERAL         ;PROCESS LITERAL STRING
          TEST    R0,0FH          ;RETURN IF FOUND
          JZ      ALPHINR         ;
ALPHIN:   CALL    AVARD           ;LOOK FOR ALPHA ARRAY OR SCALAR
          TEST    R0,0FH          ;RETURN IF NF
          JNZ     ALPHINR         ;
ALPHMOD:  CMP     BYTE PTR [TEXTPTR],'<'
          JNE     ALPHIN98        ;GOOD RETURN IF NO MODIFIER
          INC     TEXTPTR
          CALL    $BIN            ;EVALUATE 2 BYTE ALPHA OR EXPRESSION (START)
          TEST    SL,2            ;BRANCH IF NOT XPHASE
          JZ      ALPHIN15        ;
          TEST    R2,0FH          ;BRANCH IF NOT FOUND
          JNZ     ALPHIN15        ;
          DEC     DWORD PTR [A0]   ;DECREMENT START COUNT
          MOV     EDX,VSPTR
          MOV     ECX,DWORD PTR [EDX]
          MOV     AUX00,EDX
          ADD     AUX00,4
          MOV     EAX,ECX
          SUB     EAX,DWORD PTR [A0]        ;CX = NEW LENGTH
          MOV     DWORD PTR [A2],EAX
          JC      ALPHIN41                  ;ERROR IF NEGATIVE
          MOV     DWORD PTR [EDX],EAX         ;SAVE NEW LENGTH IN VS
          MOV     EDX,AUX00
          MOV     ECX,DWORD PTR [EDX]
          MOV     EAX,DWORD PTR [A0]
          ADD     ECX,EAX                   ;ECX = NEW START ADDRESS
          MOV     DWORD PTR [EDX],ECX       ;SAVE NEW START ADDRESS IN VS

ALPHIN15: CALL    NXTSEP                    ;CHECK FOR , OR >
          TEST    R0,0FH                    ;BRANCH IF NOT FOUND
          JNZ     ERROR_15
          CALL    $BIN                      ;EVALUATE 2 BYTE ALPHA OR EXPRESSION (LENGTH)
          TEST    SL,2                      ;BRANCH IF NOT XPHASE
          JZ      ALPHIN70
          TEST    R2,0FH                    ;SKIP IF MISSING PARAMETER
          JNZ     ALPHIN70
          MOV     EDX,VSPTR
          MOV     ECX,DWORD PTR [EDX]
          CMP     ECX,DWORD PTR [A0]          ;ERROR IF NEW LENGTH > DEFINED LENGTH
          JB      ALPHIN41
          MOV     EAX,DWORD PTR [A0]
          MOV     DWORD PTR [EDX],EAX         ;WRITE NEW LENGTH
ALPHIN70: CMP     BYTE PTR [TEXTPTR],'>'
          JNE     ERROR_15
          INC     TEXTPTR
          JMP     ALPHIN98                  ;SKIP IF '>'
ERROR_15: MOV     R0,15H                    ;MISSING '>'
          JMP     ERROR
ALPHIN41: JMP     ERROR_57                  ;ILLEGAL STR ARGUMENTS


;  NXTSEP:  CHECK NEXT CHARACTER FOR ',' OR '>'
;
;  ENTRY:   AS ALL NXT ROUTINES
;  RETURN: F0>0  IF CHARACTER IS NEITHER,
;          F0=0  IF CHARACTER IS ','
;          F0=0, AND TEXTPTR STILL POINTS AT IT,
;                IF CHARACTER IS '>'
;
;   MODIFIED REGISTERS:  PHPL,CHCL,F0

NXTSEP:   MOV     R0,1
          CMP     BYTE PTR [TEXTPTR],','
          JNE     NXTSEP2X
          INC     TEXTPTR         ;BRANCH IF FOUND
          JMP     ALPHIN98        ;
NXTSEP2X:
          CMP     BYTE PTR [TEXTPTR],3EH
          JNE     ALPHINR         ;SKIP IF NOT '>' EITHER
ALPHIN98: MOV     R0,0            ;FAKE A 'FOUND', IF '>' WAS FOUND (HEX 3E)
ALPHINR:  RET


;  $BIN:  EVALUATE EXPRESSION OR TWO BYTE ALPHA VARIABLE,
;     CONVERTING TO BINARY VALUE 0000 <= X <= FFFF
;
;   ENTRY:  TEXTPTR POINTS TO VALUE.
;
;   RETURN:  IF A COMMA OR '>', F2=1, NOT FOUND. DEFAULT
;       VALUE F1F0=0001.  IF BINARY VALUE IS FOUND, F2=0,
;       AND THE VALUE IS RETURNED IN F1F0.
;
;   MODIFIED REGISTERS:  ALL, AUX 0,1,2
;

$BIN:                             ;CHECK FOR '>'
          CMP     BYTE PTR [TEXTPTR],'>'
          JNE     $BIN1X
          JMP     $BIN1           ;IF SO, MISSING PARAMETER
$BIN1X:
          CMP     BYTE PTR [TEXTPTR],','
          JNE     $BIN2           ;IF NOT, GET VALUE
$BIN1:                            ;RESTORE TEXT POINTER
          MOV     DWORD PTR [A0],1 ;0001 IS DEFAULT VALUE (MISSING)
          MOV     R2,1            ;FLAG MISSING VALUE
          RET

$BIN2:    CALL    BINVAL          ;GET 2 BYTE BINARY VALUE FROM ALPHA VAR, OR EXP
          MOV     R2,0            ;FLAG FOUND VALUE
          RET


;      ALPHARAY: LOOK FOR AN ALPHA VARIABLE IN TEXT. IF
;                EXECUTION PHASE, MAKE SURE LENGTH OF
;                VARIABLE >= 256. (THIS IMPLIES VARIABLE
;                MUST BE AN ARRAY). ERROR IF NOT FOUND.
;
;           ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;           RETURN (EXECUTION PHASE):
;                   F1,0 = LENGTH OF VARIABLE (>=256)
;                   PC'S & CH,CL = ADDRESS OF VARIABLE
;                   ALPHA ATOM REMOVED FROM OS
;
;           MODIFIED REGISTERS: ALL,AUX0,1,2
;

ALPHARAY: CALL    AVARD           ;LOOK FOR ALPHA VARIABLE
          TEST    R0,0FH          ;ERROR IF NF
          JNZ     ERROR_28        ;
          TEST    SL,2            ;BRANCH IF NOT EXECUTION PHASE
          JZ      ALPHARY1        ;
          MOV     EDX,VSPTR              ;GET ALPHA ADDRESS & LENGTH FROM VS
          MOV     ECX,DWORD PTR [EDX]      ;CX = LENGTH
          MOV     DWORD PTR [A0],ECX
          MOV     EDX,DWORD PTR [EDX+4]  ;EDX = ADDRESS
          ADD     VSPTR,8
          CMP     ECX,256                ;ERROR IF LENGTH < 256
          JB      ERR_51C
ALPHARY1: RET

ERROR_28: MOV     R0,28H                 ;MISSING ALPHA ARRAY VARIABLE
          JMP     ERROR

ERR_51C:  JMP     ERROR_51               ;VARIABLE TOO SMALL


;   HEXFIX  -  Turn a nibble into a printable ascii hex character
;
;   ENTRY:         F0=0X
;   RETURN:        F0= 30-39 or 41-4F  ('0123456789ABCDEF')
;
;   MODIFIED REGS:  only F0
;

HEXFIX:   ADD     R0,30H          ;
          CMP     R0,39H
          JBE     HEXFIX2
          ADD     R0,7            ;
HEXFIX2:  RET                     ;


;   PUTHEX:  convert nibble to ascii and write it to memory
;
;         Entry:    F2 = 0x, where x = nibble to convert
;                   PHPL = address to store ascii character at
;
;         Return:   F2 = ascii digit
;                   PHPL = PHPL+1
;
;         Modified registers:  F2,3
;

PUTHEX:   ADD     R2,30H
          CMP     R2,39H             ;CONVERT NIBBLE TO ASCII
          JBE     PUTHEX2
          ADD     R2,7
PUTHEX2:  MOV     AL,R2
          MOV     BYTE PTR [EDX],AL  ;WRITE ASCII TO MEMORY
          INC     EDX
          RET


;      MOVEELMT: MOVE ELEMENT OF A MATRIX TO ANOTHER
;                MATRIX. SPECIFICALLY, MOVE 8 BYTES
;                (STARTING ON EVEN ADDRESS) TO ANOTHER
;                LOCATION (STARTING ON AN EVEN ADDRESS).
;
;           ENTRY: PC'S = ADDRESS TO MOVE FROM (EVEN)
;                  AUX3 = ADDRESS TO MOVE TO (EVEN)
;
;           RETURN: ELEMENT MOVED
;                   PC'S = PC'S + 8
;                   AUX3 = AUX3 + 8
;
;           MODIFIED REGISTERS: PC'S,AUX3,CH,CL
;

MOVEELMT: MOV     ECX,2              ;USE DWORD MOVE
          MOV     ESI,EDX
          MOV     EDI,AUX03
;         CLD
          REP     MOVSD
          MOV     EDX,ESI
          MOV     AUX03,EDI
          RET


;      EXCHELMT: EXCHANGE 2 FLOATING POINT NUMBERS
;                (8 BYTES) IN RAM
;
;           ENTRY: PC'S = ADDRESS OF NUMBER 1
;                  AUX0 = ADDRESS OF NUMBER 2
;
;           RETURN: PC'S = PC'S + 8
;                   AUX0 = AUX0 + 8
;
;            MODIFIED REGISTERS: PC'S,CH,CL,F3-5,AUX0
;
EXCHELMT: MOV     EAX,AUX00                ;USE DWORD EXCHANGE
          MOV     ECX,DWORD PTR [EDX]
          XCHG    DWORD PTR [EAX],ECX
          MOV     DWORD PTR [EDX],ECX
          MOV     ECX,DWORD PTR [EDX+4]
          XCHG    DWORD PTR [EAX+4],ECX
          MOV     DWORD PTR [EDX+4],ECX
          ADD     EDX,8
          ADD     AUX00,8
          RET


;      INITELMT: SET 8 BYTES IN RAM = FL. PT. NUMBER
;                FROM 0 TO 9
;
;           ENTRY: F2 = NUMBER (0 - 9)
;                  PC'S = RAM ADDRESS
;
;           RETURN: PC'S INCREMENTED BY 8
;
;           MODIFIED REGISTERS: PC'S
;
;      INITELM1: SET 8 BYTES IN RAM = FL. PT. 1
;
;           ENTRY: PC'S = RAM ADDRESS
;

INITELM1: MOV     WORD PTR [EDX],0            ;NUMBER = 1
          MOV     WORD PTR [EDX+2],1
          MOV     DWORD PTR [EDX+4],0
          ADD     EDX,8
          RET

INITELMT: MOV     WORD PTR [EDX],0
          MOV     AX,0
          MOV     AL,R2
          MOV     WORD PTR [EDX+2],AX         ;NUMBER = R2
          MOV     DWORD PTR [EDX+4],0
          ADD     EDX,8
          RET


;      ARRAYN: SCAN & PROCESS NUMERIC ARRAY NAME
;
;           FORMAT: letter [digit] ()
;
;           ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;           RETURN:
;                F0 = 0 IF FOUND
;                     TEXTPTR POINTS AFTER NAME
;                     IN EXECUTION PHASE,
;                          AUX3 = ADDRESS OF ARRAY
;                          AUX4 = NO. OF ELEMENTS -1
;                          F3,2 = NO. OF ELEMENTS
;                          F7,6 = DIMENSION(S)
;                          F5,BITS0F = 2 IF VECTOR
;                                      6 IF 2-DIMENSIONAL
;
;                F0 <> 0 IF NF
;
;           EXIT: ERROR#43 IF ALPHA ARRAY NAME FOUND
;
;           MODIFIED REGISTERS: PC'S,CH,CL,F0-7,AUX0
;                IN EXECUTION PHASE, AUX3,4
;

ARRAYN:   CALL    ARRAY           ;LOOK FOR ARRAY NAME
          TEST    R0,0FH          ;BRANCH IF NF
          JNZ     ARRAYN1         ;
          TEST    R5,1            ;BRANCH IF ALPHA
          JNZ     ERROR_27        ;
          TEST    SL,2            ;BRANCH IF NOT XPHASE
          JZ      ARRAYN1         ;
          MOV     AUX03,EDX       ;SAVE PTR TO ARRAY
          MOV     EAX,DWORD PTR [A2]
          MOV     AUX04,EAX       ;SAVE NO. OF ELEMENTS -1
          DEC     AUX04           ;
ARRAYN1:  RET

ERROR_27: MOV     R0,27H          ;WRONG VARIABLE TYPE
          JMP     ERROR


;      MOVEVALU: MOVE VALUE (ONE BYTE AT A TIME FROM LEFT
;                TO RIGHT)
;
;           ENTRY: K = NO. OF CHARACTERS TO MOVE +1
;                  AUX3 = ADDRESS OF VALUE
;                  AUXF = ADDRESS TO MOVE TO
;
;           RETURN: AUX3 POINTS AFTER VALUE
;                   AUXF POINTS AFTER LAST CHARACTER MOVED
;
;           MODIFIED REGISTERS: PC'S,CH,CL,F0,AUX3,F
;
;      MOVEVALF: SAME AS 'MOVEVALU' EXCEPT ENTRY
;
;           ENTRY: F0 = NO. OF CHARACTERS TO MOVE
;                  AUX3,F SAME AS ABOVE
;

MOVEVALU: MOV     ECX,0
          MOV     CL,K
          DEC     ECX            ;K = NO. OF CHARACTERS TO MOVE + 1
          MOV     ESI,AUX03
          MOV     EDI,AUX0F
          REP     MOVSB
          MOV     AUX03,ESI
          MOV     AUX0F,EDI
          RET

MOVEVALF: MOV     ECX,0          ;R0 = NO. OF CHARACTERS TO MOVE
          MOV     CL,R0
          MOV     ESI,AUX03
          MOV     EDI,AUX0F
          REP     MOVSB
          MOV     AUX03,ESI
          MOV     AUX0F,EDI
          RET


;      BINEXP: GET EXPONENT OF FL.PT. VALUE & CONVERT TO
;              BINARY.  IF EXPONENT IS NEGATIVE, TAKE
;              2'S COMPLEMENT OF RESULT.
;
;           ENTRY: AUX3 = ADDRESS (EVEN) OF FL.PT. VALUE
;
;           RETURN: F0 = BINARY EXPONENT (2'S COMPLEMENT  !!! CHANGE F0 TO F1,0
;                        IF EXPONENT NEGATIVE)
;                   F0,BIT80 = 1 IFF EXPONENT NEGATIVE
;                   CH,CL = 1ST 2 BYTES OF FL.PT. VALUE
;
;           MODIFIED REGISTERS: PC'S,CH,CL,F0,1
;

BINEXP:   MOV     EDX,AUX03
          MOV     CX,WORD PTR [EDX]
          XCHG    CH,CL
          MOV     AL,CL
          SHR     AL,4
          MOV     BL,0AH
          MUL     BL
          MOV     R0,CH
          AND     R0,0FH
          ADD     R0,AL              ;F0 = BINARY EXPONET
          TEST    CH,80H
          JZ      BINEXP1
          NEG     BYTE PTR [R0]
          INC     EDX
BINEXP1:  RET


;      MATDSPLA: F1F0 = 8 * F0
;
;           ENTRY: F0 = VALUE
;
;           RETURN: F1F0 = PRODUCT
;
;           MODIFIED REGISTERS: F0,1
;

MATDSPLA: MOV     R1,0
          SHL     WORD PTR [R0],3  ;MULTIPLY 8
          RET


;      AVARB: SCAN & PROCESS AN ALPHA-VARIABLE.  ERROR IF
;             NOT FOUND.
;
;           ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;           RETURN (EXECUTION PHASE):
;                   AUXF = ADDRESS OF VARIABLE
;                   AUXE = DEFINED LENGTH OF VARIABLE
;

AVARB:    CALL    AVARD           ;LOOK FOR ALPHA VARIABLE
          TEST    R0,0FH
          JNZ     ERROR_29        ;ERRO IF NOT FOUND
          TEST    SL,2            ;BRANCH IF NOT XPHASE
          JZ      AVARB2          ;
          MOV     EDX,VSPTR       ;GET LENGTH & ADDRESS OF BUFFER VARIABLE
          MOV     EAX,DWORD PTR [EDX]      ;ECX = LENGTH
          MOV     DWORD PTR [A0],EAX
          MOV     EDX,DWORD PTR [EDX+4]  ;EDX = ADDRESS
          MOV     AUX0F,EDX       ;SAVE PTR TO BUFFER
          MOV     AUX0E,EAX       ;SAVE LENGTH OF BUFFER
          ADD     VSPTR,8
AVARB2:   RET


;      RCVASCII: VALUE (ASCII) ==> RECEIVER.  IF NUMERIC,
;                VALUE IS FIRST CONVERTED TO FL.PT.
;
;           ENTRY: PHASE = EXECUTION
;                F1,0 = LENGTH OF VALUE (ASCII)
;                AUXB = ADDRESS OF VALUE
;                AUX2,BITSFF00 = LENGTH OF RECEIVER
;                    ,BITS00FF = VARIABLE TYPE
;                                80 IF ALPHA
;                                00 IF NUMERIC
;                AUX1 = ADDRESS OF RECEIVER
;
;           RETURN: VALUE ==> RECEIVER
;                AUX1 POINTS AFTER RECEIVER
;
;           MODIFIED REGISTERS: F0-5,CHCL,PC'S,K
;                               AUX0,1,3

RCVASCII: MOV     EDX,AUX02
          TEST    DL,80H          ;BRANCH IF NUMERIC RECEIVER
          JZ      RCVASCI9        ;

;
;           ALPHA RECEIVER
;
          MOVZX   EAX,DH          ;F3,2 = LENGTH OF RECEIVER
          MOV     DWORD PTR [A2],EAX
          MOV     EDX,AUX01       ;
          ADD     EDX,DWORD PTR [A2];AUX1 POINT TO NEXT RECEIVER
          XCHG    EDX,AUX01       ;
          MOV     AUX00,EDX       ;AUX0 = ADDRESS OF RECEIVER
          MOV     EDX,AUX0B       ;PC'S = ADDRESS VALUE
          CALL    ARECEVAL        ;VALUE --> RECEIVER
          JMP     RCVASCI8        ;

;
;           NUMERIC RECEIVER
;

RCVASCI9: MOV     ECX,DWORD PTR [A6]
          MOV     AUX03,ECX        ;SAVE ELEMENT COUNT
          MOV     EAX,DWORD PTR [A0]
          MOV     DWORD PTR [A6],EAX ;F7,6 = LENGTH OF VALUE
          MOV     EDX,AUX0B        ;PC'S POINT TO VALUE
          CALL    SNMBR            ;CONVERT ASCII NO TO FL. PT.
          TEST    R0,0FH           ;BRANCH IF NF
          JNZ     ERR_75P          ;
          CMP     DWORD PTR [A6],0  ;ERROR IF ILLEGAL CHAR IN ASCII NO.
          JNE     ERR_75P          ;
          MOV     EDX,AUX01        ;PC'S = ADDRESS OF RECEIVER
;         CALL    POPV8            ;VALUE --> RECEIVER
          MOV     EAX,VSPTR
          MOV     EBX,DWORD PTR [EAX]
          MOV     DWORD PTR [EDX],EBX
          MOV     EBX,DWORD PTR [EAX+4]
          MOV     DWORD PTR [EDX+4],EBX
          ADD     EDX,6
          ADD     VSPTR,8
          MOV     AUX01,EDX        ;
          ADD     AUX01,2          ;POINT AFTER RECEIVER
          MOV     EDX,AUX03        ;
          MOV     DWORD PTR [A6],EDX ;RESTORE ELEMENT COUNT
RCVASCI8: RET

ERR_75P:  JMP     ERROR_75        ;ILLEGAL NUMBER

;------------------------------------------------------------------------------;
;                                                                              ;
;       MVP Subroutine                                                         ;
;                                                                              ;
;------------------------------------------------------------------------------;

;      ARGUMNTR: SCAN & PROCESS A RECEIVER ARGUMENT (OF
;                DATALOAD TYPE). EACH ELEMENT OF ALPHA-
;                ARRAY IS A SEPARATE RECEIVER.
;
;           FORMAT:  array-name
;                    alpha-variable
;                    numeric-variable
;
;           ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;           RETURN: TEXTPTR POINTS AFTER ARGUMENT
;             IN EXECUTION PHASE,
;                F7,6 = NO. OF ELEMENTS (1 IF SCALAR)
;                AUX1 = ADDRESS OF VARIABLE
;                AUX2,BITSFF00 = LENGTH OF RECEIVER (<125)
;                    ,BITS00FF = VARIABLE TYPE
;                                80 IF ALPHA
;                                00 IF NUMERIC
;
;                PC'S = LENGTH/TYPE
;
;           MODIFIED REGISTERS: ALL,AUX0,1,2
;
;      NOTE, DECODING ARRAY, ALPHA VARIABLE, NUMERIC
;           VARIABLE, OR STR() OPTIMIZED IN EXECUTION
;           PHASE.
;

ARGUMNTR: TEST    SL,2                    ;BRANCH IF NOT XPHASE
          JZ      ARGUMTR9                ;

;
;  EXECUTION PHASE
;
          CALL    VARNAME                 ;LOOK FOR VARIABLE NAME
          TEST    R0,0FH                  ;BRANCH IF NF
          JNZ     ARGUMTR8                ;
          TEST    R2,2                    ;BRANCH IF NOT ARRAY NAME
          JZ      ARGUMTR7
          CMP     BYTE PTR [TEXTPTR],')'  ;LOOK FOR ')'
          JNE     ARGUMTR7                ;BRANCH IF NOT ARRAY DESIGNATOR
          INC     TEXTPTR
;
;       ARRAY
;
          CALL    ARRAYA                  ;PROCESS ARRAY
          MOV     EAX,DWORD PTR [A2]      ;
          MOV     DWORD PTR [A6],EAX      ;F7,6 = NO. OF ELEMENTS
          MOV     AUX01,EDX               ;AUX1 = ADDRESS OF ARRAY
          MOV     EDX,00000800H
          TEST    R5,1                    ;BRANCH IF NUMERIC ARRAY
          JZ      ARGUMTR3                ;
          MOV     DH,CH                   ;PH = LEHGTH
          MOV     DL,80H                  ;PL = TYPE (ALPHA)
          JMP     ARGUMTR3                ;

ARGUMTR8: MOV     K,0                     ;RETURN DEFINED LENGTH
          CALL    AVAR                    ;PROCESS STR() FUNCTION
          TEST    R0,0FH                  ;BRANCH IF FOUND
          JZ      ARGUMTR1                ;
          JMP     ERROR_24                ;ERROR IF NOT FOUND
;
;       ALPHA VARIABLE
;

ARGUMTR7: TEST    R2,1                    ;BRANCH IF NUMERIC VARIABLE
          JZ      ARGUMTR6                ;
          MOV     K,0                     ;FLAG = RETURN DEFINED LENGTH
          CALL    AVAR1                   ;PROCESS ALPHA VARIABLE
ARGUMTR1: MOV     EDX,VSPTR               ;GET LENGTH & ADDRESS OF VARIABLE
          MOV     ECX,DWORD PTR [EDX]
          MOV     EDX,DWORD PTR [EDX+4]
          ADD     VSPTR,8
          MOV     AUX01,EDX               ;AUX01 = ADDRESS OF VARIABLE
          MOV     EDX,0000007CH
          CMP     ECX,EDX
          JA      ERROR_52                ;ERROR IF LENGTH > 124
          MOV     DH,CL                   ;PH = VARIABLE LENGTH
          MOV     DL,80H                  ;PL = TYPE (ALPHA)
          JMP     ARGUMTR5                ;
;
;       NUMERIC VARIABLE
;

ARGUMTR6: MOV     K,2                     ;FLAG = RECEIVER VARIABLE
          CALL    VARA                    ;PROCESS NUMERIC VARIABLE
          DEC     OSPTR                   ;REMOVE RECEIVER ATOM FROM OS
          MOV     EDX,VSPTR
          MOV     EDX,DWORD PTR [EDX]     ;GET ADDRESS OF VARIABLE FROM VS
          ADD     VSPTR,4                 ;
          MOV     AUX01,EDX               ;
          MOV     EDX,00000800H           ;LENGTH = 8, TYPE = NUMERIC
ARGUMTR5: MOV     DWORD PTR [A6],1        ;NO. OF ELEMENTS = 1
ARGUMTR3: MOV     AUX02,EDX               ;AUX2 = LENGTH / TYPE
ARGUMTR2: RET

;
;  ENTRY OR RESOLUTION PHASE
;

ARGUMTR9: CALL    ARRAY                   ;LOOK FOR ARRAY NAME
          TEST    R0,0FH                  ;BRANCH IF FOUND
          JZ      ARGUMTR2                ;
          CALL    AVARD                   ;LOOK FOR ALPHA VARIABLE OR STR()
          TEST    R0,0FH                  ;BRANCH IF FOUND
          JZ      ARGUMTR2                ;
          JMP     VARR                    ;LOOK FOR NUMERIC VARIABLE

;      ARGUMENT: SCAN & PROCESS AN ARGUMENT (OF
;                DATALOAD TYPE). EACH ELEMENT OF ALPHA-
;                ARRAY IS A SEPARATE RECEIVER.
;
;           FORMAT:  array-name
;                    alpha-variable
;                    expression
;                    literal
;
;           ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;           RETURN: TEXTPTR POINTS AFTER ARGUMENT
;             IN EXECUTION PHASE,
;                F7,6 = NO. OF ELEMENTS (1 IF SCALAR)
;                AUX3 = ADDRESS OF VALUE
;                AUX2,BITSFF00 = SOV
;                    ,BITS8000 = TYPE (1 ALPHA, 0 IF NUM)
;                    ,BITS7F00 = LENGTH (<125)
;                    ,BITS00FF = LENGTH
;                PC'S = AUX2
;
;           MODIFIED REGISTERS: ALL,AUX0,1,2
;                IN EXECUTION PHASE, AUX3
;

ARGUMENT: CALL    ARRAY                           ;LOOK FOR ARRAY NAME
          TEST    R0,0FH                          ;BRANCH IF NF
          JNZ     ARGUMNT1                        ;
          TEST    SL,2                            ;BRANCH IF NOT XPHASE
          JZ      ARGUMTR2                        ;
          MOV     EAX,DWORD PTR [A2]              ;
          MOV     DWORD PTR [A6],EAX              ;F7,6 = NO. OF ELEMENTS
          MOV     AUX03,EDX                       ;AUX3 = ADDRESS OF ARRAY
          MOV     EDX,00000808H                   ;
          TEST    R5,1                            ;BRANCH IF NUMERIC ARRAY
          JZ      ARGUMTR3                        ;
          MOV     DL,CH                           ;PL = LENGTH
          MOV     AL,CH                           ;
          OR      AL,80H                          ;
          MOV     DH,AL                           ;PH = SOV (TYPE/LENGTH)
          JMP     ARGUMTR3                        ;
ARGUMNT1: CALL    ALPHA                           ;LOOK FOR ALPHA
          TEST    R0,0FH                          ;BRANCH IF NF
          JNZ     ARGUMNT4                        ;
          TEST    SL,2                            ;BRANCH IF NOT XPHASE
          JZ      ARGUMTR2                        ;
          MOV     EDX,VSPTR                       ;GET LENGTH & ADDRESS OF VARIABLE
          MOV     ECX,DWORD PTR [EDX]
          MOV     EDX,DWORD PTR [EDX+4]
          ADD     VSPTR,8
          MOV     AUX03,EDX                       ;AUX3 = ADDRESS OF VARIABLE
          MOV     EDX,0000007CH                   ;
          CMP     ECX,EDX                         ;BRANCH IF LENGTH <= 124
          JA      ERROR_52                        ;VALUE TOO LONG
          MOV     DL,CL                           ;PL = VARIABLE LENGTH
          MOV     DH,CL                           ;
          OR      DH,80H                          ;PH = SOV (TYPE/LENGTH)
          JMP     ARGUMTR5                        ;
ARGUMNT4: CALL    EXPR                            ;LOOK FOR NUMERIC EXPRESSION
          TEST    SL,2                            ;BRANCH IF NOT XPHASE
          JZ      ARGUMTR2                        ;
          MOV     EDX,VSPTR                       ;
          MOV     AUX03,EDX                       ;AUX3 = ADDRESS OF VALUE
;         CALL    DELETV8A                        ;REMOVE VALUE FROM VS
          ADD     EDX,8
          XCHG    EDX,VSPTR                       ;VSPTR = PC'S + 8
          MOV     EDX,00000808H                   ;LENGTH = 8, TYPE = NUMERIC
          JMP     ARGUMTR5                        ;


;      NUMCH: GET NEXT CHARACTER FROM ALPHA-VALUE AND MAKE
;             SURE IT'S LEGAL. ALSO, INCREMENT NUM-COUNT.
;
;           ENTRY: F6 = LEGAL CHARACTERS (SPACE ALWAYS OK)
;                    ,BIT0 =1 IF DIGIT LEGAL
;                    ,BIT1 =1 IF DECIMAL POINT (.) LEGAL
;                    ,BIT2 =1 IF SIGN (+,-) LEGAL
;                    ,BIT3 =1 IF 'E' LEGAL
;                  PC'S POINT TO NEXT CHARACTER OF VALUE
;                  F7 = X'39'
;                  F3,2 = NUM-COUNT (NO. OF LEGAL CHAR'S
;                         FOUND THUS FAR)
;                  F1,0 = LENGTH OF VALUE
;
;           RETURN: F5 = 0 IF NO LEGAL CHARACTER FOUND
;                        1 IF DIGIT FOUND
;                        2 IF DECIMAL POINT FOUND
;                        4 IF SIGN FOUND
;                        8 IF 'E' FOUND
;                   PC'S POINT AFTER CHARACTER FOUND
;                   CH = CHARACTER FOUND
;                   F3,2 INCREMENTED
;
;           MODIFIED REGISTERS: PC'S,CH,CL,F2,3,5
;

NUMCH:    MOV     EAX,DWORD PTR [A2]
          CMP     DWORD PTR [A0],EAX                ;BRANCH IF NO MORE CHAR IN VALUE
          JBE     NUMCHA                          ;
          MOV     CH,BYTE PTR [EDX]               ;GET NEXT CHAR OF VALUE
          INC     EDX                             ;INCREMENT PTR
          CMP     CH,30H
          JB      NUMCHX
          CMP     CH,39H
          JA      NUMCHY
          MOV     R5,1                            ;FLAG = DIGIT
NUMCH3:   MOV     AL,R6
          AND     R5,AL                           ;R5 <> 0 IF LEGAL CHARACTER FOUND
          CMP     R5,0
          JE      NUMCHSR                         ;BRANCH IF NOT LEGAL CHARACTER FOUND
          INC     DWORD PTR [A2]
NUMCHSR:  RET

NUMCHX:   CMP     CH,20H
          JE      NUMCHS
          CMP     CH,'+'
          JE      NUMCHZ
          CMP     CH,'-'
          JE      NUMCHZ
          CMP     CH,'.'
          JNE     NUMCHA
          MOV     R5,2                            ;FLAG = DECIMAL POINT
          JMP     NUMCH3
NUMCHY:   CMP     CH,'E'
          JNE     NUMCHA
          MOV     R5,8                            ;FLAG = 'E'
          JMP     NUMCH3
NUMCHZ:   MOV     R5,4                            ;FLAG = SIGN
          JMP     NUMCH3
NUMCHS:   INC     DWORD PTR [A2]                  ;SPACE
          JMP     NUMCH
NUMCHA:   MOV     R5,0
          RET


;      DEF1STMT: MAKE SURE DEFFN' IS 1ST STMT ON LINE
;
;           ENTRY: TEXTPTR POINTS AFTER DEFFN' ATOM
;

DEF1STMT: MOV     EDX,TEXTLPTR                    ;
          MOV     DWORD PTR [A0],EDX              ;SAVE BEGIN OF LINE PTR
          INC     DWORD PTR [A0]                  ;
          MOV     EDX,TEXTPTR                     ;
DEF1STMX: MOV     CH,BYTE PTR [EDX]               ;READ PREVIOUS CHAR
          DEC     EDX                             ;
          CMP     CH,0CEH                         ;BRANCH IF NOT 'DEFFN' ATOM
          JNE     DEF1STMX                        ;
DEF1STM6: MOV     CH,BYTE PTR [EDX]               ;READ PREVIOUS CHAR
          DEC     EDX                             ;
          CMP     CH,0FDH                         ;ERROR IF CHAR = STMT SEPARATOR
          JE      ERR_08A                         ;
          CMP     DWORD PTR [A0],EDX              ;BRANCH IF NOT AT BEGIN OF LINE
          JB      DEF1STM6                        ;
          RET

ERR_08A:  JMP     ERROR_08                        ;STMT NOT LEGAL HERE

CODE      ENDS
          END
