;===========================================================================
;
;
;
;===========================================================================
;           This module was originally M30
;===========================================================================
            .XLIST
            INCLUDE PREAMBLE
            .LIST

PUBLIC    ARAYELRP
PUBLIC    BITF0
PUBLIC    BITTERM_
PUBLIC    CCNLPART
PUBLIC    COMPARE
PUBLIC    ERR34
PUBLIC    ERR_28S
PUBLIC    ERR_29I
PUBLIC    ERR_50S
PUBLIC    ERR_51S
PUBLIC    ERR_55S
PUBLIC    ERROR_26
PUBLIC    FIELD
PUBLIC    MATMOVE
PUBLIC    MATCOPY
PUBLIC    MERGE
PUBLIC    NAME_
PUBLIC    NAME2FR
PUBLIC    NAMEA
PUBLIC    POSARR1
PUBLIC    $POSARR1
PUBLIC    SEARCH
PUBLIC    SORT
PUBLIC    SRCHALPH
PUBLIC    NEWTOOLD
PUBLIC    OLDTONEW

EXTRN     _ABCOV  :NEAR
EXTRN     ALPHA1  :NEAR
EXTRN     ALPHIN  :NEAR
EXTRN     ALPHINL :NEAR
EXTRN     ARAYEL  :NEAR
EXTRN     ARAYELP :NEAR
EXTRN     ARECEVAL:NEAR
EXTRN     ARRAY   :NEAR
EXTRN     AVARD   :NEAR
EXTRN     $BINFLT :NEAR
EXTRN     BINFLT  :NEAR
EXTRN     BREAKPTQ:NEAR
EXTRN     BREAKPTT:NEAR
EXTRN     CCL     :NEAR
EXTRN     CCLTO   :NEAR
EXTRN     CCNL    :NEAR
EXTRN     CURLEN  :NEAR
EXTRN     ENDSTMTZ:NEAR
EXTRN     ERR_02  :NEAR
EXTRN     ERROR   :NEAR
EXTRN     ERROR_02:NEAR
EXTRN     ERROR_11:NEAR
EXTRN     ERROR_13:NEAR
EXTRN     ERROR_19:NEAR
EXTRN     ERROR_28:NEAR
EXTRN     ERROR_29:NEAR
EXTRN     ERROR_34:NEAR
EXTRN     ERROR_50:NEAR
EXTRN     ERROR_51:NEAR
EXTRN     ERROR_52:NEAR
EXTRN     ERROR_55:NEAR
EXTRN     EXPR    :NEAR
EXTRN     FINDVAR :NEAR
EXTRN     FLTBIN  :NEAR
EXTRN     FLTBINA :NEAR
EXTRN     FLTBINB :NEAR
EXTRN     GETTERM_:NEAR
EXTRN     MATHPAGE:NEAR
EXTRN     RELALPHC:NEAR
EXTRN     RELA_E:NEAR
EXTRN     RELOP   :NEAR
EXTRN     REM     :NEAR
EXTRN     VAR     :NEAR
EXTRN     VARNAME :NEAR
EXTRN     _PART   :NEAR
EXTRN     WORK_BUFFER:NEAR

EXTRN     SORT_S   :NEAR
EXTRN     MERGE_S  :NEAR
EXTRN     MATMOVE_S:NEAR
EXTRN     SEARCH_S :NEAR

EXTRN     @@VSEND :DWORD
EXTRN     CPBASE  :DWORD

;------------------------------------------------------------------------------;
;                                                                              ;
;        MODULE    MOD-30 MAT SORT                                             ;
;                                                                              ;
;------------------------------------------------------------------------------;


;        MAT SORT  -  ARRANGE AN ARRAY OF N ELEMENTS
;                       INTO ASCENDING ORDER BY MEANS
;                         OF A SHELL SORT.
;                     1) ON EACH PASS THRU THE DATA,
;                        'D' LISTS OF ELEMENTS ARE SORTED
;                        BY BUBBLING.  ELEMENTS IN EACH
;                        LIST ARE A DISTANCE 'D' APART.
;                             D(1)=(N+1)/2
;                             D(I)=(D(I-1)+1)/2
;                     2) ALL LISTS ARE BUBBLED UNTIL
;                        EVERY LIST IS SORTED.
;                     3) THE SHELL SORT CONTINUES UNTIL
;                        THE FINAL LIST (D=1) IS SORTED.
;
;
;               AN EXAMPLE OF BUBBLE SORTING:
;
;               3          1          1          1
;               1          3          3          2
;               4   --->   4   --->   2   --->   3
;               5          2          4          4
;               2          5          5          5
;
;               ON EACH BUBBLE, NOTE THAT THE FIRST
;               ELEMENT IS COMPARED TO THE SECOND AND
;               THE TWO ELEMENTS ARE EXCHANGED IF
;               NECESSARY TO PUT THEM IN THE CORRECT
;               ORDER.  THEN THE SECOND ELEMENT (WHICH
;               MAY HAVE STARTED OUT BEING THE FIRST
;               ELEMENT) IS COMPARED TO THE THIRD, AND
;               SO ON.  THIS IS CONTINUED UNTIL THE LAST
;               TWO ELEMENTS ARE COMPARED, THE WHOLE
;               PROCESS REQUIRING N-1 COMPARISONS AND
;               POSSIBLE EXCHANGES.  THIS ENTIRE ROUTINE
;               IS REPEATED UNTIL THE ELEMENTS ARE ALL
;               SORTED.  NOTE THAT ON EACH BUBBLE THRU
;               THE DATA, THE LARGEST ELEMENT 'SINKS' TO
;               THE BOTTOM SO THAT ONE LESS COMPARISON
;               CAN BE MADE ON EACH SUCCESSIVE PASS.
;
;               THE SHELL SORT USES BUBBLING TO SORT EACH
;               LIST OF DATA, BUT BY BREAKING THE DATA
;               INTO SUCCESSIVELY LARGER LISTS, ELEMENTS
;               WHICH ARE FAR OUT OF PLACE ARE SORTED
;               MORE QUICKLY.
;
;        GENERAL FORM:
;
;        *******************************************
;        *  MAT SORT (source) TO (work),(locator)  *
;        *******************************************
;
;               WHERE:  (source) = ARRAY TO BE SORTED
;                       (work) = WORK STRING FOR STORING
;                                 ELEMENT ADDRESSES
;                       (locator) = SUBSCRIPT ARRAY
;
;        MODIFIED REGISTERS:  ALL, AUX 0,1,4-6,8-F
;                      (AUX 0,1,2,4,5 IN ENTRY PHASE)
;
;        EXPR BETTER NOT USE: AUX 4,5,7,8,9,A,D,E
;

SORT:     cmp       byte ptr [textptr],'!'
          jne       sort111
          inc       textptr
          jmp       sort_s
sort111:
          CALL      ARRAY               ;SCAN SOURCE ARRAY
          TEST      R0,0FH
          JNZ       ERR_28S             ;ERROR IF NOT FOUND
          TEST      R5,1
          JZ        ERR_28S             ;ERROR IF NOT ALPHA
          TEST      SL,2
          JZ        SORT0               ;BRANCH IF NOT X PHASE
          MOV       AUX0A,EDX           ;SAVE SOURCE ADDRESS
          MOVZX     EDX,R5
          AND       DL,04H
          SHL       DL,2                ;PL = 00 IF VECTOR, 10 IF MATRIX
          MOV       AUX0F,EDX           ;SAVE FLAG VECTOR/MATRIX
          MOV       EDX,DWORD PTR [A2]
          MOV       AUX0D,EDX           ;SAVE NUMBER OF ELEMENTS
          MOV       EDX,DWORD PTR [A6]
          MOV       AUX0E,EDX           ;SAVE DIMENSIONS
          MOVZX     EDX,CH
          MOV       AUX09,EDX           ;SAVE ELEMENT LENGTH
          MOV       AUX08,EDX

SORT0:    CALL      FIELD               ;PROCESS A FIELD SPECIFICATION INTO F5F4
          CMP       A4,0
          JE        SORT1               ;BRANCH IF NOT SPECIFIED
          TEST      SL,2
          JZ        SORT1               ;BRANCH IF NOT X PHASE
          MOV       A6,0
          MOV       EDX,AUX0A           ;GET SOURCE ADDRESS
          MOV       EAX,DWORD PTR [A5]
          ADD       EDX,EAX
          MOV       AUX0A,EDX           ;SAVE ADDRESS AT START OF FIELD
          MOVZX     EDX,A4
          MOV       AUX09,EDX           ;SAVE FIELD LENGTH

SORT1:    CALL      CCLTO               ;MUST FIND "TO"
          CALL      AVARD               ;SCAN WORK STRING
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        SORT2               ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX0D           ;GET NUMBER OF SOURCE ELEMENTS
          MOV       DWORD PTR [A0],EDX
          SHL       DWORD PTR [A0],1    ;NO. * 2 = REQUIRED NO. OF WORK BYTES
          JC        ERR_51S             ;ERROR IF OVERFLOW
          MOV       EDX,VSPTR
          MOV       ECX,DWORD PTR [EDX]
          ADD       VSPTR,4             ;GET WORK STRING LENGTH
          CMP       ECX,DWORD PTR [A0]
          JB        ERR_51S             ;ERROR IF TOO SHORT
          MOV       BYTE PTR TTT,0
          MOV       EAX,DWORD PTR [A0]
          SHL       EAX,1
          JC        ERR_51S
          CMP       ECX,EAX
          JB        SORTQQQ
          MOV       BYTE PTR TTT,1
SORTQQQ:
          MOV       EDX,VSPTR           ;GET WORK ADDRESS
          MOV       EDX,DWORD PTR [EDX]
          ADD       VSPTR,4
          MOV       AUX0B,EDX           ;SAVE WORK ADDRESS
SORT2:    CMP       BYTE PTR [TEXTPTR],','   ;MUST FIND A COMMA
          JNE       ERROR_13
          INC       TEXTPTR
          CALL      ARRAY
          TEST      R0,0FH
          JNZ       ERR_28S             ;ERROR IF NOT FOUND
          TEST      R5,1
          JZ        ERR_28S             ;ERROR IF NOT ALPHA
          TEST      SL,2
          JZ        SORTH               ;BRANCH IF NOT X PHASE
          MOV       AUX0C,EDX           ;SAVE LOCATOR ADDRESS
          CMP       CH,2
          JNE       ERR_50S             ;ERROR IF NOT 2 BYTE ELEMENTS

; <<<< SHOULD WE REQUIRE TWO BYTE ELEMENTS? >>>>>>
;

          MOV       EDX,AUX0D           ;GET REQUIRED NO. OF ELEMENTS
          CMP       DWORD PTR [A2],EDX
          JB        ERR_51S             ;ERROR IF LOCATOR SIZE TOO SMALL
          MOV       R0,1                ;SET SIZE FLAG DEFAULT =
          CMP       DWORD PTR [A2],EDX
          JBE       SORTX               ;BRANCH IF SIZE = SOURCE
          MOV       R0,0                ;SET SIZE FLAG >
SORTX:    MOV       EDX,AUX0F           ;GET FLAG VECTOR/MATRIX
          MOV       AL,DL
          OR        AL,R0
          MOV       K,AL                ;SET K = FLAGS
          MOV       EDX,AUX0E           ;GET SOURCE DIMENSIONS
          MOV       DWORD PTR [A4],EDX  ;SAVE DIMENSIONS
          MOV       EDX,AUX08           ;GET SOURCE ELEMENT LENGTH
          MOV       DWORD PTR [A6],EDX  ;SAVE IT


;        REGISTERS NOW SET UP:
;               AUX 9 - FIELD LENGTH (OR ELEMENT LENGTH)
;               AUX A - ADDRESS OF SOURCE ARRAY (+FIELD)
;               AUX B - ADDRESS OF WORK STRING
;               AUX C - ADDRESS OF LOCATOR ARRAY
;               AUX D - NO. OF ELEMENTS IN SOURCE ARRAY
;
;               F5,F4 - DIMENSIONS OF SOURCE ARRAY
;               F7,F6 - ELEMENT LENGTH (F7=00)
;               K (LOW) - SIZE FLAG SET TO ZER0 IF LOC
;                   ARRAY IS > SORT ARRAY BY AT LEAST 2
;               K (HIGH) - SORT ARRAY TYPE ( VECTOR = 0,
;                          MATRIX = 1 )
;
;
;         BEGIN BY PUTTING SEQUENTIAL SUBSCRIPTS IN THE
;            LOCATOR ARRAY AND THE CORRESPONDING ELEMENT
;               ADDRESSES IN THE WORK ARRAY.
;

          cmp       byte ptr ttt,1
          je        sortwww

; added for emulate work space under sort processing
;
          mov       edx,vsptr
          mov       eax,aux0d
          shl       eax,2
          sub       edx,eax
          cmp       edx,dword ptr [@@vsend]
          jb        error_02
          mov       aux0b,edx           ;put emulate space address to work string address

SORTWWW:
          MOVZX     AX,K
          SHR       AX,4
          MOV       A1,AX               ;SET INITIAL SUBSCRIPT TO 0000 IF VECTOR,

; 0100 IF ARRAY.
;

          MOV       EDX,AUX0A           ;GET SORT ARRAY ADDRESS
          MOV       DWORD PTR [A2],EDX  ;SAVE IT
          MOV       EDX,AUX0B           ;GET WORK STRING ADDRESS
          MOV       AUX08,EDX           ;SAVE IT
          MOV       EDX,AUX0C           ;GET LOCATOR ARRAY ADDRESS
          JMP       SORTA2

SORTA0:   INC       A1                  ;INCREMENT ROW SUBSCRIPT
SORTA2:   MOV       A0,0                ;INITIALIZE COLUMN SUBSCRIPT
SORTA1:   INC       A0
SORTA:    TEST      K,0F0H
          JZ        SORTAX              ;VECTOR
          MOV       AX,A1
          MOV       BYTE PTR [EDX],AL
          INC       EDX                 ;PUT SUBSCRIPT IN LOCATOR ARRAY
          MOV       AX,A0
          MOV       BYTE PTR [EDX],AL
          XCHG      EDX,AUX08           ;SAVE LOCATOR ADDRESS
          INC       AUX08
          JMP       SORTAY

SORTAX:   MOV       EAX,DWORD PTR [A0]
          MOV       BYTE PTR [EDX],AH
          INC       EDX                 ;PUT SUBSCRIPT IN LOCATOR ARRAY
          MOV       BYTE PTR [EDX],AL
          XCHG      EDX,AUX08           ;SAVE LOCATOR ADDRESS
          INC       AUX08
SORTAY:   MOV       ECX,DWORD PTR [A2]
          MOV       DWORD PTR [EDX],ECX ;PUT ELEMENT ADDRESS IN WORK STRING
          ADD       EDX,4
          XCHG      EDX,AUX08           ;SAVE WORK ADDRESS
          MOV       EAX,DWORD PTR [A6]
          ADD       DWORD PTR [A2],EAX  ;ADDRESS OF NEXT SORT ELEMENT
          TEST      K,0F0H
          JZ        SORTBA              ;BRANCH IF VECTOR
          MOV       AX,A0
          CMP       AX,A4
          JB        SORTA1              ;CHECK COLUMN SUBSCRIPT
          MOV       AX,A1
          CMP       AX,A5
          JB        SORTA0              ;CHECK ROW SUBSCRIPT

SORTAA:   TEST      K,0FH
          JNZ       SORTB               ;BRANCH IF LOCATOR SIZE=SORT SIZE
          MOV       WORD PTR [EDX],0    ;WRITE '0000' INTO LOCATOR
          ADD       EDX,2
SORTB:    MOV       EDX,AUX09           ;GET FIELD LENGTH
          MOV       A7,DX               ;SAVE IT
          MOV       DWORD PTR [A0],1
          MOV       EDX,AUX0D           ;GET NUMBER OF ELEMENTS
          CMP       DWORD PTR [A0],EDX
          JB        SORTC               ;BRANCH IF MORE THAN ONE ELEMENT
          JMP       ENDSTMTZ

SORTBA:   INC       DWORD PTR [A0]      ;INCREMENT SUBSCRIPT
          MOV       EAX,DWORD PTR [A0]
          CMP       EAX,DWORD PTR [A4]
          JBE       SORTA               ;BRANCH IF NOT DONE
          JMP       SORTAA

;
;        BEGINNING OF NEW PASS.  COMPUTE NEW DISTANCE 'D'
;             D(1)=(N+1)/2    D(I)=(D(I-1)+1)/2
;

SORTC:    INC       EDX                  ;ADD 1, BEFORE DIVIDING BY TWO
          SHR       EDX,1                ;PHPL= (D+1)/2, NEW DISTANCE
          MOV       AUX09,EDX            ;SAVE DISTANCE 'D'
          MOV       DWORD PTR [A2],EDX
          SHL       DWORD PTR [A2],2     ;DOUBLE TO GET ADDRESS INCREMENT (F3,F2)
          MOV       EDX,AUX09            ;GET CURRENT DISTANCE 'D'
          MOV       DWORD PTR [A0],EDX   ;SUBTRACT DISTANCE BETWEEN ELEMENTS FROM
          NOT       DWORD PTR [A0]

;   THE TOTAL NUMBER OF ELEMENTS TO FIND THE NUMBER
;
          MOV       EDX,AUX0D           ;OF COMPARISONS TO BE MADE
          STC
          ADC       EDX,DWORD PTR [A0]
          MOV       AUX0F,EDX           ;SAVE COMPARISON COUNT IN AUX 7
          MOV       EDX,DWORD PTR [A0]    ;SAVE NEGATIVE OF DISTANCE ( -D )
          MOV       AUX06,EDX

;
;        NOW BUBBLE 'D' LISTS OF ELEMENTS EACH A DISTANCE
;           'D' APART IN THE ORIGINAL ARRAY.
;

SORTD:    MOV       EDX,AUX0B           ;POINT TO START OF WORK STRING
          MOV       AUX0E,EDX           ;SAVE IT
          MOV       EDX,AUX0C           ;POINT TO START OF LOCATOR ARRAY
          MOV       AUX08,EDX           ;SAVE IT
          MOV       DWORD PTR [A5],0    ;CLEAR COMPARISON COUNT
          MOV       R4,0                ;CLEAR EXCHANGE FLAG

;        DO ONE BUBBLE OF EACH LIST AND THEN CHECK IF
;          ALL LISTS ARE SORTED (EXCHANGE FLAG F4 = 0)
;

SORTE:    MOV       EDX,AUX0E           ;POINT TO WORK STRING
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX             ;POINT TO ELEMENT A DISTANCE 'D' AWAY
          MOV       EDX,DWORD PTR [EDX] ;READ
                                        ;ADDRESS OF ELEMENT 2 IN PH,PL
          MOV       AUX00,EDX           ;SAVE ADDRESS
          MOVZX     EAX,A7
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,AUX0E           ;POINT TO WORK STRING
          MOV       EDX,DWORD PTR [EDX] ;GET ADDRESS OF ELEMENT 1
          CALL      COMPARE             ;CHECK IF ELEMENT 1 < ELEMENT 2
          CMP       R0,1
          JNE       SORTF               ;BRANCH IF IT IS

;        THE FOLLOWING CODE EXCHANGES THE VALUES IN THE WORK
;          BUFFER AND THOSE IN THE LOCATOR ARRAY THAT ARE
;          A DISTANCE 'D' APART (GIVEN BY THE INCREMENT IN
;          F3,F2).
;

          MOV       EDX,AUX0E           ;POINT TO WORK STRING
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX             ;FIND SECOND ADDRESS
          MOV       CX,WORD PTR [EDX]
          XCHG      EDX,AUX0E           ;READ SECOND VALUE
          MOV       AX,CX
          MOV       CX,WORD PTR [EDX]   ;READ FIRST VALUE
          MOV       WORD PTR [EDX],AX
          XCHG      EDX,AUX0E           ;WRITE SECOND INTO FIRST
          ADD       AUX0E,2
          MOV       WORD PTR [EDX],CX
          ADD       EDX,2               ;WRITE FIRST INTO SECOND
          MOV       CX,WORD PTR [EDX]
          XCHG      EDX,AUX0E           ;READ SECOND VALUE
          MOV       AX,CX
          MOV       CX,WORD PTR [EDX]   ;READ FIRST VALUE
          MOV       WORD PTR [EDX],AX
          XCHG      EDX,AUX0E           ;WRITE 2ND INTO 1ST/ RESTORE WORK PTR
          SUB       AUX0E,2
          MOV       WORD PTR [EDX],CX   ;WRITE FIRST INTO SECOND

          MOV       EDX,AUX08           ;POINT TO LOCATOR ARRAY
          MOV       EAX,DWORD PTR [A2]
          SHR       EAX,1
          ADD       EDX,EAX             ;FIND SECOND ADDRESS
          MOV       CH,BYTE PTR [EDX]
          XCHG      EDX,AUX08           ;READ SECOND VALUE
          MOV       AL,CH
          MOV       CH,BYTE PTR [EDX]   ;READ FIRST VALUE
          MOV       BYTE PTR [EDX],AL
          XCHG      EDX,AUX08           ;WRITE SECOND INTO FIRST
          INC       AUX08
          MOV       BYTE PTR [EDX],CH
          INC       EDX                 ;WRITE FIRST INTO SECOND
          MOV       CH,BYTE PTR [EDX]
          XCHG      EDX,AUX08           ;READ SECOND VALUE
          MOV       AL,CH
          MOV       CH,BYTE PTR [EDX]   ;READ FIRST VALUE
          MOV       BYTE PTR [EDX],AL
          XCHG      EDX,AUX08           ;WRITE 2ND INTO 1ST/ RESTORE LOCATOR PTR
          DEC       AUX08
          MOV       BYTE PTR [EDX],CH   ;WRITE FIRST INTO SECOND
          MOV       R4,1                ;SET EXCHANGE FLAG

SORTF:    INC       DWORD PTR [A5]      ;INCREMENT COMPARISON COUNT
          MOV       EDX,AUX0F           ;GET TOTAL NUMBER OF COMPARISONS TO BE MADE
          CMP       DWORD PTR [A5],EDX
          JB        SORTY               ;BRANCH IF NOT DONE WITH BUBBLE
          JMP       SORTG               ;END OF BUBBLE
SORTY:    MOV       EDX,AUX0E           ;POINT TO WORK STRING
          MOV       AUX0E,EDX           ;INCREMENT TO NEXT LIST
          ADD       AUX0E,4
          MOV       EDX,AUX08           ;POINT TO LOCATOR ARRAY
          MOV       AUX08,EDX           ;INCREMENT TO NEXT LIST
          ADD       AUX08,2
          JMP       SORTE               ;CONTINUE WITH THE BUBBLE

SORTG:    TEST      R4,0FH
          JZ        SORTGA              ;BRANCH IF ALL LISTS ARE SORTED
          MOV       EDX,AUX06           ;NUMBER OF COMPARISONS FOR NEXT BUBBLE IS 1 LESS
          STC
          ADC       EDX,DWORD PTR [A5]  ;FOR EACH LIST OR (COMPARISONS - D)

; SINCE THERE ARE D LISTS.
;

          JNC       SORTGA              ;DONE IF RESULT IS <= 0
          CMP       EDX,0
          JNE       SORTI               ;BRANCH IF NOT ZERO
SORTGA:   MOV       DWORD PTR [A0],1
          MOV       EDX,AUX09           ;GET NUMBER OF LISTS = 'D'
          CMP       DWORD PTR [A0],EDX
          JB        SORTC               ;BRANCH IF MORE THAN ONE LIST
SORTH:    JMP       ENDSTMTZ            ;DONE WHEN ONLY 1 LIST ALL SORTED
SORTI:    test      byte ptr fs:[cpust],10h
          jz        sortd               ;BRANCH IF < 30MS. ELAPSED
          MOV       EDX,DWORD PTR [A2]    ;SAVE IN AN AUX FOR SWITCH PARTITIONS
          MOV       AUX0A,EDX
          MOV       EAX,EDX
          MOVZX     EDX,A7               ;SAVE IN AN AUX FOR SWITCH PARTITONS
          SHL       EDX,16
          MOV       DX,AX
          MOV       AUX0E,EDX
          MOV       R0,SAV06            ;SAVE AUX 06-1F (NEED AUX 6,9,A-F)
          CALL      BREAKPTQ            ;TRANSFER CONTROL TO NEXT USER
          MOV       EDX,AUX0A
          MOV       DWORD PTR [A2],EDX  ;RESTORE
          MOV       EDX,AUX0E
          MOV       EAX,EDX
          SHR       EAX,16
          MOV       A7,AX               ;RESTORE
          JMP       SORTD

;        MAT COPY  -  COPY A DESIGNATED CHAR STRING TO
;                        ANOTHER VARIABLE ONE BYTE AT
;                           A TIME.
;
;        GENERAL FORM:
;        **********************************
;        *  MAT COPY (source) TO (output) *
;        **********************************
;
;               WHERE:  (source) = ORIGINAL CHAR STRING
;                       (output) = NEW VARIABLE STRING
;
;        MODIFIED REGISTERS:   ALL, AUX A-C,E
;                          (AUX 0,1,2 IN ENTRY PHASE)
;
;        EXPR BETTER NOT USE:   AUX A,B,E
;
;        REGISTERS SET UP AS FOLLOWS:
;               AUX A - ADDRESS OF SOURCE VARIABLE
;               AUX B - COUNT OF SOURCE STRING
;               AUX C - ADDRESS OF OUTPUT VARIABLE
;
;               F7(LOW) - DIRECTION FLAG
;                              ++ = 3
;                              +- = 2
;                              -+ = 1
;                              -- = 0
;
;               F6(LOW) - PAD (1) OR NO PAD (0) W/ SPACES
;               F5,F4 - END ADDRESS OF COPY SEGMENT
;               F3,F2 - END ADDRESS OF PAD SEGMENT
;
;

MATCOPY:  CMP       BYTE PTR [TEXTPTR],'-' ;LOOK FOR A '-'
          MOV       DL,2
          JNE       MATCOPYX
          INC       TEXTPTR
          MOV       DL,0                ;FLAG = 03 FOR +, 01 FOR -
MATCOPYX: TEST      SL,2
          JZ        JLCOPY0             ;BRANCH IF NOT X PHASE
          MOV       AUX0E,EDX           ;SAVE DIRECTION FLAG
JLCOPY0:  CALL      ALPHIN              ;LOOK FOR AN ALPHA VARIABLE
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        JLCOPY1             ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET ADDRESS & LENGTH OF SOURCE
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX0A,EDX           ;SAVE SOURCE ADDRESS
          MOV       EDX,DWORD PTR [A0]
          MOV       AUX0B,EDX           ;SAVE SOURCE LENGTH

JLCOPY1:  CALL      CCLTO               ;MUST FIND "TO"
          CMP       BYTE PTR [TEXTPTR],'-'  ;CHECK FOR '-'
          MOV       AL,1
          JNE       MATCOPYY
          INC       TEXTPTR
          MOV       AL,0
MATCOPYY: TEST      SL,2
          JZ        JLCOPY1A            ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX0E           ;GET DIRECTION FLAG
          XOR       DL,AL               ;SET WRITE PORTION OF FLAG
          MOV       AUX0E,EDX           ;SAVE DIRECTION FLAG

;  FLAG IS AS FOLLOWS:
;       00  =  REVERSE,  REVERSE
;       01  =  REVERSE,  FORWARD
;       02  =  FORWARD,  REVERSE
;       03  =  FORWARD,  FORWARD

JLCOPY1A: CALL      ALPHIN              ;LOOK FOR AN ALPHA VARIABLE
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        JLCOPY6             ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX0E           ;GET DIRECTION FLAG
          MOV       R7,DL               ;SAVE IT
          MOV       EDX,VSPTR           ;GET ADDRESS & LENGTH OF DESTINATION VARIABLE
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX0C,EDX           ;SAVE OUTPUT ADDRESS
          MOV       DWORD PTR [A4],EAX  ;SAVE OUTPUT COUNT
          MOV       EDX,AUX0B           ;GET SOURCE COUNT
          MOV       R6,0                ;FLAG = NO PAD
          CMP       DWORD PTR [A0],EDX
          JBE       JLCOPY2             ;BRANCH IF OUTPUT COUNT <= SOURCE COUN
          MOV       R6,1                ;FLAG = PAD WITH SPACES
          MOV       DWORD PTR [A4],EDX  ;PUT SOURCE COUNT IN FOR LAST COPY
JLCOPY2:  TEST      R7,2
          JNZ       JLCOPY3             ;BRANCH IF READ IS +
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX0A           ;GET SOURCE ADDRESS
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX             ;START AT END FOR - READ
          MOV       AUX0A,EDX           ;SAVE SOURCE ADDRESS
          DEC       AUX0A
JLCOPY3:  MOV       EDX,AUX0C           ;GET OUTPUT ADDRESS
          TEST      R7,1
          JNZ       JLCOPY4             ;BRANCH IF WRITE IS +
          MOV       EAX,DWORD PTR [A0]
          ADD       EDX,EAX             ;START AT END FOR - WRITE
          MOV       AUX0C,EDX
          DEC       AUX0C
          MOV       EDX,AUX0C
          MOV       EAX,DWORD PTR [A4]
          MOV       DWORD PTR [A4],EDX
          SUB       DWORD PTR [A4],EAX  ;SUBTRACT TO FIND LAST OUTPUT ADDRESS
          JMP       MATCOPYW
JLCOPY4:  MOV       EAX,DWORD PTR [A4]
          MOV       DWORD PTR [A4],EDX
          ADD       DWORD PTR [A4],EAX  ;LAST OUT ADDRESS FOR COPY SEGMEN
MATCOPYW: TEST      R6,0FH
          JZ        JLCOPY4A            ;BRANCH IF NO PAD INDICATED
          TEST      R7,1
          JNZ       COPY4AA             ;BRANCH IF WRITE IS +
          MOV       DWORD PTR [A2],EDX
          MOV       EAX,DWORD PTR [A0]
          SUB       DWORD PTR [A2],EAX
          JMP       JLCOPY4A
COPY4AA:  MOV       EAX,DWORD PTR [A0]
          MOV       DWORD PTR [A2],EDX
          ADD       DWORD PTR [A2],EAX  ;LAST OUT ADDRESS FOR PAD SEGMENT

;        SOURCE ADDRESS IN AUX A, OUTPUT ADDRESS IN PC'S
;

JLCOPY4A: XCHG      EDX,AUX0A
          TEST      R7,1
          JZ        JLCOPY4X            ;NOW BRANCH TO THE APPROPRIATE COPY AND
          TEST      R7,2
          JNZ       COPYPP1             ;PAD ROUTINES BASED ON THE READ AND
          JMP       COPYMP1             ;WRITE DIRECTIONS. (++, +-, -+, --)
JLCOPY4X: TEST      R7,2
          JNZ       COPYPM1
          JMP       COPYMM1

;    02   + -
;

COPYPM:   MOV       BYTE PTR [EDX],CH
          XCHG      EDX,AUX0A
          DEC       AUX0A
COPYPM1:  MOV       CH,BYTE PTR [EDX]
          XCHG      EDX,AUX0A
          INC       AUX0A               ;READ
          CMP       DWORD PTR [A4],EDX
          JB        COPYPM              ;BRANCH IF NOT DONE WITH COPY SEGMENT
COPYPADM: TEST      R6,0FH              ;BRANCH IF NO PAD
          JZ        JLCOPY6

JLCOPY5:  MOV       BYTE PTR [EDX],20H  ;WRITE IN A SPACE
          DEC       EDX
          CMP       DWORD PTR [A2],EDX
          JB        JLCOPY5             ;BRANCH IF NOT DONE WITH PAD SEGMENT
JLCOPY6:  JMP       ENDSTMTZ

;   00   - -
;

COPYMM:   MOV       BYTE PTR [EDX],CH
          XCHG      EDX,AUX0A
          DEC       AUX0A
COPYMM1:  MOV       CH,BYTE PTR [EDX]   ;READ
          XCHG      EDX,AUX0A
          DEC       AUX0A
          CMP       DWORD PTR [A4],EDX
          JB        COPYMM              ;BRANCH IF NOT DONE WITH COPY SEGMENT
          JMP       COPYPADM

;     03   + +
;

COPYPP:   MOV       BYTE PTR [EDX],CH
          XCHG      EDX,AUX0A
          INC       AUX0A
COPYPP1:  MOV       CH,BYTE PTR [EDX]   ;READ
          XCHG      EDX,AUX0A
          INC       AUX0A
          CMP       DWORD PTR [A4],EDX
          JBE       COPYPADP            ;BRANCH IF DONE WITH COPY SEGMENT
          JMP       COPYPP

COPYPADP: TEST      R6,0FH              ;BRANCH IF NO PAD
          JZ        JLCOPY6
          MOV       R0,20H              ;SET UP A SPACE CHAR
JLCOPY7:  CMP       DWORD PTR [A2],EDX
          JBE       JLCOPY6             ;BRANCH IF DONE WITH PAD SEGMEN
          MOV       AL,R0
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          JMP       JLCOPY7

;    01   - +
;

COPYMP:   MOV       BYTE PTR [EDX],CH
          XCHG      EDX,AUX0A
          INC       AUX0A
COPYMP1:  MOV       CH,BYTE PTR [EDX]   ;READ
          XCHG      EDX,AUX0A
          DEC       AUX0A
          CMP       DWORD PTR [A4],EDX
          JBE       COPYPADP
          JMP       COPYMP


;        MAT SEARCH  -  SEARCH A DESIGNATED STRING FOR
;                        CHARACTER STRINGS WHICH SATISFY
;                         A GIVEN RELATION AND PUT THE
;                          TWO BYTE POSITION INTO A
;                           LOCATOR STRING.
;
;        GENERAL FORM:
;        ***********************************************
;        * MAT SEARCH (search),(relation) (compare)    *
;        *            TO (locator) STEP (expression)   *
;        ***********************************************
;
;               WHERE:  (search) = STRING TO BE SEARCHED (no test for null string)
;                       (relation) = RELATION TESTED
;                       (compare) = COMPARISON STIRNG (no test for null string)
;                       (locator) = STRING OF POSITIONS
;                       (expression) = STEP PARAMETER
;
;        MODIFIED REGISTERS:  ALL, AUX 0-A,E,F
;
;        EXPR BETTER NOT USE:   AUX 3-A,E,F
;

SEARCH:   cmp       byte ptr [textptr],'!'
          jne       search111
          inc       textptr
          jmp       search_s
search111:
          CALL      ALPHINL             ;SCAN SEARCH VARIABLE OR LITERAL (RELEASE 1.5
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        SEARCH1             ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET SEARCH ADDRESS & LENGTH
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX03,EDX           ;SAVE INITIAL SEARCH STRING ADRESS
          MOV       AUX09,EDX           ;AUX9 = ADR(SEARCH STRING)-1
          DEC       AUX09
          ADD       EDX,EAX             ;PC's = ADR(END OF SEARCH STRING)+1
          MOV       AUX0A,EDX           ;AUXA = ADR(END OF SEARCH STRING)
          DEC       AUX0A
          MOV       EDX,DWORD PTR [A0]  ;AUX4 = INITIAL LENGTH TO SEARCH
          MOV       AUX04,EDX

SEARCH1:  CMP       BYTE PTR [TEXTPTR],','  ;MUST FIND COMMA
          JNE       ERROR_13
          INC       TEXTPTR
          CALL      RELOP               ;PUTS RELATION OPERATOR IN F1
          TEST      SL,2
          JZ        SEARCHX             ;BRANCH IF NOT X PHASE
          MOV       DL,R1
          OR        DL,0F0H             ;PL = RELOP AND FORWARD SEARCH DIRECTION
          MOV       AUX08,EDX           ;SAVE RELOP
SEARCHX:  MOV       K,4                 ;(CURRENT LENGTH))
          CALL      ALPHA1              ;SCAN ALPHA VALUE ARGUMENT
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        SEARCH2             ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET VALUE ADDRESS & LENGTH
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX05,EDX           ;SAVE ADDRESS
          MOV       EDX,DWORD PTR [A0]
          MOV       AUX06,EDX           ;SAVE LENGTH

SEARCH2:  CALL      CCLTO               ;MUST FIND "TO"
          CALL      AVARD               ;SCAN POSITION VARIABLE
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        SEARCH2B            ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET LOCATOR ADDRESS & LENGTH
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX0E,EDX           ;SAVE ADDRESS
          AND       A0,0FFFEH           ;MAKE LENGTH EVEN
          CMP       A0,0
          JNE       SEARCHY
          CMP       A1,0
          JE        ERR_51S             ;ERROR IF LENGTH WAS < 2 BYTES
SEARCHY:  ADD       EDX,DWORD PTR [A0]
          MOV       AUX0F,EDX           ;SAVE LAST ADDRESS
SEARCH2B: MOV       R0,@STEP
          CALL      CCL                 ;LOOK FOR "STEP"
          TEST      R0,0FH
          JZ        SEARCHS             ;BRANCH IF FOUND
          TEST      SL,2
          JZ        SEARCH6             ;BRANCH IF NOT X PHASE
          MOV       EDX,1               ;SET STEP VALUE TO ONE
SEARCH2A: MOV       AUX07,EDX           ;SAVE STEP VALUE

;       REGISTERS NOW SET UP AS FOLLOWS:
;         AUX3 = CURRENT ADR WITHIN SEARCH STRING
;         AUX4 = REMAINING LENGTH OF SEARCH STRING
;         AUX5 = ADR(STRING TO SEARCH FOR)
;         AUX6 = LENGTH OF STRING TO SEARCH FOR
;         AUX7 = STEP SIZE
;         AUX8 = COMPARISON OPERATOR
;         AUX9 = ADR(SEARCH STRING)-1
;         AUXA = ADR(END OF SEARCH STRING)
;         AUXE = ADR(LOCATOR STRING)
;         AUXF = LOCATOR LAST ADR

SEARCH3:  CALL      SRCHALPH            ;SEARCH FOR NEXT OCCURANCE
          TEST      R7,0FH
          JZ        SEARCHD             ;BRANCH IF NO MATCH; EXIT MAT SEARCH
          MOV       DWORD PTR [A2],EDX  ;F3F2 = ADR(MATCHED STRING)
          MOV       EDX,AUX09           ;PC's = ADR(SEARCH STRING)-1
          MOV       EAX,DWORD PTR [A2]
          SUB       EAX,EDX             ;F1F0 = OFFSET OF FOUND STRING IN SEARCH STRING
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,AUX07           ;PC's = STEP SIZE
          MOV       EBX,EDX
          MOV       EAX,DWORD PTR [A2]
          ADD       EAX,EBX             ;F5F4 = ADR(NEXT STRING TO TEST)
          MOV       DWORD PTR [A4],EAX
          JC        SEARCH4             ;BRANCH IF 16 BIT OVERFLOW
          MOV       EDX,AUX0A           ;PC's = ADR(END OF SEARCH STRING)
          CMP       DWORD PTR [A4],EDX
          JBE       SEARCH7             ;BRANCH IF STILL WITHIN SEARCH STRING

;      LAST OCCURANCE FOUND
;

SEARCH4:  CALL      POSARR1             ;PUT LAST OCCURANCE IN LOCATOR ARRAY
          CMP       R1,0
          JE        SEARCH6             ;BRANCH IF LOCATOR STRING FULL

;      SIGNAL END OF LOCATOR ARRAY
;

SEARCHD:  MOV       DWORD PTR [A0],0    ;F1F0 = ZERO
          CALL      POSARR1             ;PUT IN LOCATOR STRING
SEARCH6:  JMP       ENDSTMTZ            ;EXIT STATEMENT

;      LOOP AND SEARCH AGAIN
;

SEARCH7:  MOV       DWORD PTR [A2],EDX  ;F3F2 = ADR(END OF SEARCH STRING)+1
          INC       DWORD PTR [A2]
          MOV       EDX,DWORD PTR [A2]  ;PC's = REMAINING LENGTH TO SEARCH
          SUB       EDX,DWORD PTR [A4]
          MOV       AUX04,EDX           ;AUX4 = REMAINING LENGTH TO SEARCH
          MOV       EDX,DWORD PTR [A4]  ;AUX3 = ADR(NEXT STRING TO TEST)
          MOV       AUX03,EDX
          CALL      POSARR1             ;PUT OCCURANCE OF STRING IN LOCATOR STRING
          CMP       R1,0
          JNE       SEARCH3             ;LOOP IF LOCATOR STRING NOT FULL
          JMP       SEARCH6             ;EXIT STATEMENT IF LOCATOR STRING FULL

;
;        PROCESS STEP FUNCTION
;

SEARCHS:  CALL      EXPR                ;EVALUATE EXPRESSION
          TEST      SL,2
          JZ        SEARCH6             ;BRANCH IF NOT X PHASE
          CALL      FLTBIN              ;PUT STEP INTO F1,F0
          TEST      R2,2
          JZ        QERR34              ;ERROR IF NOT 0<X<2^32
          MOV       EDX,DWORD PTR [A0]
          JMP       SEARCH2A
QERR34:   JMP       ERR34


;        MAT MOVE  -  THERE ARE FOUR GENERAL CASES IN
;                        MOVING DATA ELEMENTS FROM ONE
;                        ARRAY TO ANOTHER:
;                         1)  ALPHA --> ALPHA
;                         2)  NUMERIC --> ALPHA
;                         3)  ALPHA --> NUMERIC
;                         4)  NUMERIC --> NUMERIC
;
;                     CASE 2 IMPLIES CONVERSION FROM 2600
;                         INTERNAL FORMAT TO SORT FORMAT.
;                     CASE 3 IMPLIES CONVERSION FROM SORT
;                         FORMAT TO 2600 INTERNAL FORMAT.
;
;        GENERAL FORM:
;    ****************************************************
;    *  MAT MOVE (move array),(locator array element),n *
;    *                    TO (receiver array element)   *
;    ****************************************************
;
;          WHERE:  (move array) = ALPHA OR NUMERIC
;                  (loc array element) = ALPHA SUBSCRIPTS
;                         IN THE ARRAY BEGINNING WITH THE
;                         ELEMENT SPECIFIED (OR BEGIN IF
;                         NO SUBSCRIPT GIVEN).
;                  n = NUMERIC SCALAR VARIABLE GIVING
;                         MAXIMUM # OF ELEMENTS TO MOVE
;                         AND RECEIVING #  OF ELEMENTS
;                         ACTUALLY MOVED.
;                  (rec array element) = DESTINATION OF
;                         ELEMENTS FROM MOVE ARRAY
;                         BEGINNING WITH THE ELEMENT
;                         SPECIFIED (OR BEGIN IF NO
;                         SUBSCRIPT GIVEN).
;
;          THE LOCATOR ARRAY IS OPTIONAL.  HENCE, THE
;             MOVE MAY BE DIRECT OR INDIRECT.
;
;          REGISTERS MODIFIED:   ALL, AUX 0-E
;                              (AUX 0,1,5 IN ENTRY PHASE)
;
;          EXPR BETTER NOT USE:   AUX 3-E


;     MOVE ARRAY
;


MATMOVE:  cmp       byte ptr [textptr],'!'
          jne       matmove111
          inc       textptr
          jmp       matmove_s
matmove111:
          CALL      ARRAY               ;SCAN MOVE ARRAY VARIABLE
          TEST      R0,0FH
          JNZ       ERROR_26            ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        JLMOVE1             ;BRANCH IF NOT X PHASE
          MOV       AUX09,EDX           ;SAVE ADDRESS OF MOVE ARRAY
          TEST      R5,1
          JZ        MATMOVEX            ;BRANCH IF NUMERIC ARRAY
          DEC       EDX                 ;DEDUCT 1 IF ALPHA
MATMOVEX: MOV       AUX06,EDX           ;SAVE ADDRESS OF DIMS
          SUB       AUX06,4
          XOR       EDX,EDX             ;CLEAR STATUS FLAG
          TEST      R5,1
          JZ        MATMOVEY            ;BRANCH IF NUMERIC
          OR        EDX,12H             ;SET STATUS FLAG TO ALPHA
MATMOVEY: TEST      R5,4
          JZ        MATMOVEZ            ;BRANCH IF VECTOR
          OR        EDX,40H             ;SET STATUS FLAG TO MATRIX
MATMOVEZ: MOV       AUX07,EDX           ;SAVE IT
          MOV       EDX,DWORD PTR [A2]
          MOV       AUX08,EDX           ;SAVE ELEMENT COUNT OF MOVE ARRAY
          MOV       AUX0E,0             ;FIELD POSITION = 0 FOR NUMERIC ARRAYS
          MOV       EDX,8                ;NUMERIC LENGTH = 8
          TEST      R5,1
          JZ        MATMOVEA            ;BRANCH IF NUMERIC
          MOVZX     EDX,CH               ;SAVE ELEMENT LENGTH IF ALPHA
MATMOVEA: MOV       AUX0A,EDX
          MOV       AUX0B,EDX
JLMOVE1:  TEST      R5,1
          JZ        JLMOVE2             ;BRANCH IF NUMERIC
          CALL      FIELD               ;PROCESS A FIELD SPECIFICATION IF ALPHA
          TEST      SL,2
          JZ        JLMOVE2             ;BRANCH IF NOT X PHASE
          MOV       A6,0
          MOV       EDX,DWORD PTR [A5]
          MOV       AUX0E,EDX           ;TEMPORARY SAVE OF FIELD POSITION
          CMP       A4,0
          JE        JLMOVE2             ;BRANCH IF FIELD IS NOT SPECIFIED
          MOV       EDX,AUX09           ;GET MOVE ADDRESS
          MOV       EAX,DWORD PTR [A5]
          ADD       EDX,EAX             ;ADDRESS AT START OF FIELD
          MOV       AUX09,EDX           ;SAVE IT
          MOVZX     EDX,A4
          MOV       AUX0B,EDX           ;SAVE FIELD LENGTH
JLMOVE2:  CMP       BYTE PTR [TEXTPTR],','  ;LOOK FOR A COMMA
          JNE       JLMOVE4             ;BRANCH IF NOT FOUND
          INC       TEXTPTR

;   LOCATOR ARRAY(ELEMENT)
;
          MOV       AUX20,TEXTPTR
          CALL      VARNAME             ;LOOK FOR A LOCATOR VARIABLE
          TEST      R0,0FH
          JNZ       JLMOVE3A            ;BRANCH IF NOT FOUND
          TEST      R2,1
          JNZ       JLMOVE2A            ;BRANCH IF ALPHA
          MOV       TEXTPTR,AUX20       ;RETORE TEXT POINTER IF NOT ALPHA
          JMP       JLMOVE3A

JLMOVE2A: TEST      R2,2
          JZ        ERROR_26            ;ERROR IF NOT AN ARRAY
          CALL      ARAYELRP            ;PROCESS ARRAY ELEMENT
          TEST      SL,2
          JZ        JLMOVE3             ;BRANCH IF NOT X PHASE
          CMP       CH,2
          JNE       ERR_50S             ;ERROR IF ELEMENT LENGTH NOT = 2
          MOV       AUX09,EDX           ;SAVE LOCATOR ADDRESS
          MOV       EDX,AUX07           ;GET STATUS FLAG
          OR        EDX,4               ;SET STATUS TO INDIRECT
          MOV       AUX07,EDX           ;SAVE IT
          MOV       EAX,DWORD PTR [A6]    ;FIND # OF ELEMENTS
          MOV       DWORD PTR [A0],EAX    ;FIND # OF ELEMENTS
          TEST      K,40H
          JZ        MATMOVEB            ;BRANCH IF VECTOR
          MOVZX     EAX,A0              ;MULTIPLY DIMENSIONS FOR MATRIX
          MOVZX     EBX,A1
          PUSH      EDX
          MUL       EBX
          POP       EDX
          MOV       DWORD PTR [A0],EAX
MATMOVEB: MOV       EDX,AUX02           ;GET POSITION # OF ELEMENT-1
          MOV       EAX,DWORD PTR [A0]
          SUB       EAX,EDX
          MOV       EDX,EAX               ;# OF REMAINING ELEMENTS IN PH,PL
          MOV       AUX08,EDX           ;SAVE IT
          MOV       EDX,AUX0E           ;GET FIELD POSITION
          MOV       AUX0A,EDX           ;SAVE IT
JLMOVE3:  CMP       BYTE PTR [TEXTPTR],',' ;LOOK FOR A COMMA
          JNE       JLMOVE4             ;BRANCH IF NOT FOUND
          INC       TEXTPTR

;   COUNT VARIABLE
;

JLMOVE3A: MOV       K,3                 ;SET = SCALAR, RECEIVER
          CALL      VAR                 ;SCAN NUMERIC VAR, ERROR IF NOT FOUND
          TEST      SL,2
          JZ        JLMOVE4             ;BRANCH IF NOT X PHASE
          DEC       OSPTR               ;REMOVE OS ATOM
          MOV       EDX,VSPTR           ;GET ADDRESS OF VAR
          MOV       EDX,DWORD PTR [EDX]
          ADD       VSPTR,4
          MOV       AUX0F,EDX           ;SAVE IT
          MOV       EAX,VSPTR           ;VALUE TO VS
          SUB       EAX,8
          MOV       ECX,DWORD PTR [EDX]
          MOV       DWORD PTR [EAX],ECX
          MOV       ECX,DWORD PTR [EDX+4]
          MOV       DWORD PTR [EAX+4],ECX
          MOV       VSPTR,EAX
          CMP       EAX,DWORD PTR [@@VSEND]
          JB        ERR_02

          CALL      FLTBIN              ;CONVERT TO BINARY IN F1,F0
          TEST      R2,02H
          JNZ       MATMOVEC            ;BRANCH IF 0<VAR<65535
          JMP       REM                 ;N SPECIFIES ZERO ELEMENTS MOVED
MATMOVEC: MOV       EDX,AUX08           ;GET MAXIMUM MOVE COUNT
          CMP       DWORD PTR [A0],EDX
          JB        MATMOVED            ;BRANCH IF N<MAX
          JMP       MATMOVEE
MATMOVED: MOV       EDX,DWORD PTR [A0]
          MOV       AUX08,EDX           ;SAVE MAXIMUM MOVE COUNT
MATMOVEE: MOV       EDX,AUX07           ;GET STATUS FLAG
          OR        EDX,8               ;SET STATUS = N SPECIFIED
          MOV       AUX07,EDX           ;SAVE STATUS FLAG
JLMOVE4:  CALL      CCLTO               ;MUST FIND "TO"

;   RECEIVER ARRAY [(ELEMENT)]
;

          CALL      VARNAME             ;LOOK FOR RECEIVER VARIABLE
          TEST      R0,0FH
          JNZ       ERROR_26            ;ERROR IF NOT FOUND
          TEST      R2,2
          JZ        ERROR_26            ;ERROR IF NOT AN ARRAY
          TEST      R2,1
          JZ        MATMOVEF            ;BRANCH IF NOT ALPHA
          MOV       R0,1                ;SET FLAG TO ALPHA
MATMOVEF: MOVZX     EDX,R0
          MOV       AUX05,EDX           ;SAVE FLAG FOR ALPHA/NUMERIC
          CALL      ARAYELRP            ;PROCESS ARRAY VARIABLE
          TEST      SL,2
          JZ        JLMOVE5             ;BRANCH IF NOT X PHASE
          MOV       AUX0C,EDX           ;SAVE RECEIVER ADDRESS
          MOV       EDX,AUX07           ;GET STATUS FLAG
          TEST      K,10H
          JZ        MATMOVEG            ;BRANCH IF NOT ALPHA
          OR        EDX,1
          MOV       AUX07,EDX           ;SAVE STATUS FLAG
MATMOVEG: MOV       EAX,DWORD PTR [A6]
          MOV       DWORD PTR [A0],EAX  ;GET DIMENSIONS TO F1,F0
          TEST      K,40H
          JZ        MATMOVEH            ;BRANCH IF VECTOR
          MOVZX     EAX,A0              ;FIND TOTAL # OF ELEMENTS
          MOVZX     EBX,A1
          PUSH      EDX
          MUL       EBX
          POP       EDX
          MOV       DWORD PTR [A0],EAX
MATMOVEH: MOV       EDX,AUX02           ;GET POSITION # OF THIS ELEMENT-1
          SUB       DWORD PTR [A0],EDX  ;REMAINING # OF ELEMENTS IN F1,F0
          MOV       EDX,AUX08           ;GET MAXIMUM MOVE COUNT
          CMP       DWORD PTR [A0],EDX
          JB        MATMOVEI            ;BRANCH IF RECEIVER < MAX
          JMP       MATMOVEJ
MATMOVEI: MOV       EDX,DWORD PTR [A0]
          MOV       AUX08,EDX           ;SAVE MAXIMUM MOVE COUNT
MATMOVEJ: MOV       EDX,8               ;NUMERIC LENGTH = 8
          TEST      K,10H
          JZ        MATMOVEK            ;BRANCH IF NOT ALPHA
          MOVZX     EDX,CH              ;SAVE ELEMENT LENGTH IF ALPHA
MATMOVEK: MOV       AUX0E,EDX           ;SAVE ELEMENT LENGTH
          MOV       AUX0D,EDX
JLMOVE5:  MOV       EDX,AUX05           ;GET FLAG ALPHA/NUMERIC
          MOV       AL,DL
          AND       AL,0FH
          CMP       AL,1
          JNE       JLMOVE6             ;BRANCH IF NOT ALPHA
          MOV       EDX,AUX0D
          CALL      FIELD               ;PROCESS A FIELD SPECIFICATION IF ALPHA
          CMP       A4,0
          JE        JLMOVE6             ;BRANCH IF NOT SPECIFIED135
          TEST      SL,2
          JZ        JLMOVED             ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          MOVZX     EAX,A5
          ADD       EDX,EAX             ;ADDRESS AT START OF FIELD
          MOV       AUX0C,EDX           ;SAVE IT
          MOVZX     EDX,A4
          MOV       AUX0E,EDX           ;SAVE FIELD LENGTH
JLMOVE6:  TEST      SL,2
          JZ        JLMOVED             ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX08           ;GET MOVE COUNT
          MOV       AUX08,EDX           ;DECREMENT FOR START OF LOOP
          DEC       AUX08
          XCHG      EDX,AUX07           ;SET # MOVED
          MOV       K,DL                ;SAVE STATUS FLAG
          TEST      K,1
          JNZ       JLMOVE6A
          TEST      K,2
          JZ        JLMOVE6A            ;BRANCH IF NOT ALPHA ==> NUMERIC
          MOV       EDX,AUX0B
          DEC       EDX
          CMP       EDX,0
          JNE       JLMOVE6A
          JMP       ERROR_51            ;ERROR IF LENGTH OF MOVE ELEMENT = 1

;        REGISTER USE IS NOW AS FOLLOWS:
;          AUX 6 - POINTS TO DIMS OF MOVE ARRAY
;          AUX 7 - # 0F ELEMENTS MOVED
;          AUX 8 - MAXIMUM # OF ELEMENTS TO BE MOVED
;                     = SMALLEST OF # OF MOVE ELEMENTS,
;                       RECEIVER ELEMENTS, OR N FOR
;                       DIRECT MOVES
;                     = SMALLEST OF # OF LOCATOR ELEMENTS
;                       OR RECEIVER ELEMENTS OR N FOR
;                       INDIRECT MOVES
;          AUX 9 - MOVE ARRAY ADDRESS IF DIRECT MOVE OR
;                  LOCATOR ARRAY ADDRESS IF INDIRECT
;          AUX A - ELEMENT LENGTH OF MOVE ARRAY IF DIRECT
;                  MOVE (USED TO POINT TO NEXT ELEMENT)
;                  OR FIELD POSITION WITHIN MOVE ELEMENT
;                  IF INDIRECT MOVE (MAY BE = 0 IF FIELD
;                  IS NOT SPECIFIED)
;          AUX B - FIELD LENGTH OF MOVE ARRAY OR ELEMENT
;                  LENGTH IF NO FIELD SPECIFIED
;          AUX C - ADDRESS OF RECIEVER ( FIELD POSITION
;                  IS ADDED IF IT IS SPECIFIED)
;          AUX D - ELEMENT LENGTH OF RECEIVER (USED AS
;                  INCREMENT TO NEXT ELEMENT POSITIONED
;                  TO FIELD IF SPECIFIED)
;          AUX E - FIELD LENGTH OF RECEIVER OR ELEMENT
;                  LENGTH IF NO FIELD SPECIFIED
;          AUX F - ADDRESS OF NUMERIC SCALAR VARIABLE (N)
;
;          STATUS FLAG (K):
;                  BITS 1,0 - MOVE FUNCTION          (03)
;                          11 = ALPHA --> ALPHA
;                          10 = ALPHA --> NUMERIC
;                          01 = NUMERIC --> ALPHA
;                          00 = NUMERIC --> NUMERIC
;                  BIT 2 - INDIRECT MOVE = 1         (04)
;                          DIRECT MOVE = 0
;                  BIT 3 - N SPECIFIED (1=YES,0=NO)  (08)
;                  BIT 4 - SAME AS BIT 1 (FOR ARAYELP)  (10)
;                  BIT 6 - MOVE ARRAY (VECTOR=0,MATRIX=1) (40)
;                  BITS 7,5 - NOT USED
;
;

JLMOVE6A: TEST      K,4                 ;BRANCH IF INDIRECT
          JNZ       MOVEAAI

;        'DIRECT MOVE'
;

          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          JMP       MOVEAAD

MOVEAA1:  MOV       EDX,AUX0A           ;GET ADDRESS INCREMENT
          MOV       EAX,EDX
          MOV       EDX,AUX09           ;INCREMENT VALUE ADDRESS
          ADD       EDX,EAX
          MOV       AUX09,EDX           ;SAVE IT
          MOV       EDX,AUX0D           ;GET ADDRESS INCREMENT
          MOV       EAX,EDX
          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          ADD       EDX,EAX             ;INCREMENT
          MOV       AUX0C,EDX           ;SAVE
MOVEAAD:  MOV       AUX00,EDX           ;SAVE RECEIVER ADDRESS
          MOV       EDX,AUX0B           ;GET VALUE LENGTH
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX0E           ;GET RECEIVER LENGTH
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX09           ;GET VALUE ADDRESS
          MOV       AL,K
          AND       AL,03H
          CMP       AL,3                ;BRANCH IF ALPHA --> ALPHA
          JE        MOVEAAD1
          CMP       AL,1                ;BRANCH IF NUMERIC --> ALPHA
          JE        MOVECON
          CMP       AL,2                ;BRANCH IF ALPHA --> NUMERIC
          JE        MOVEDCON
MOVEAAD1: CALL      ARECEVAL            ;MOVE VALUE TO RECEIVER
          MOV       EDX,AUX08           ;GET MOVE COUNT
          MOV       AUX08,EDX           ;DECREMENT AND SAVE
          DEC       AUX08
          MOV       EAX,EDX
          DEC       EDX
          TEST      EAX,0FFH
          JNZ       MOVEAA1
          CMP       EAX,0
          JE        MOVEDONE
          test      byte ptr fs:[cpust],10h
          jz        moveaa1             ;BRANCH IF < 30MS. ELAPSED
          MOV       DL,K
          MOV       AUX05,EDX           ;SAVE K
          MOV       R0,SAV03
          CALL      BREAKPTQ            ;BREAKPOINT
          MOV       EDX,AUX05
          MOV       K,DL                ;RESTORE K
          JMP       MOVEAA1

MOVEDONE: TEST      K,8                 ;BRANCH IF N IS NOT SPECIFIED
          JZ        JLMOVED
          MOV       EDX,AUX07           ;GET # OF ELEMENTS MOVED
          MOV       DWORD PTR [A0],EDX
MOVEIDD:  CALL      $BINFLT             ;CONVERT TO FLOATING
          MOV       EDX,AUX0F           ;GET SCALAR VARIABLE ADDRESS
          MOV       EAX,VSPTR           ;PUT # MOVED INTO SCALAR VARIABLE
          MOV       ECX,DWORD PTR [EAX]
          MOV       DWORD PTR [EDX],ECX
          MOV       ECX,DWORD PTR [EAX+4]
          MOV       DWORD PTR [EDX+4],ECX
          ADD       VSPTR,8
JLMOVED:  JMP       ENDSTMTZ

;       'INDIRECT MOVE'
;

MOVEAAI0: MOV       EDX,AUX0D           ;GET ADDRESS INCREMENT
          MOV       EAX,EDX
          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          ADD       EDX,EAX
          MOV       AUX0C,EDX           ;SAVE IT
MOVEAAI:  MOV       EDX,AUX09           ;GET LOCATOR ADDRESS
          test      k,40h
          jz        moveaax
          MOVZX     CX,BYTE PTR [EDX]   ;READ SUBSCRIPTS AND PUT IN F1,F0
          INC       EDX
          MOV       A1,CX
          MOVZX     CX,BYTE PTR [EDX]
          MOV       AUX09,EDX           ;INCREMENT LOCATOR ADDRESS
          INC       AUX09
          MOV       A0,CX
          CMP       A0,0
          JNE       MOVEXY
          CMP       A1,0                ;DONE IF SUBSCRIPT IS '0000'
          JE        MOVEID
          jmp       movexy
moveaax:
          XOR       EAX,EAX
          MOV       AH,BYTE PTR [EDX]   ;READ SUBSCRIPTS AND PUT IN F1,F0
          INC       EDX
          MOV       AL,BYTE PTR [EDX]
          MOV       AUX09,EDX           ;INCREMENT LOCATOR ADDRESS
          INC       AUX09
          MOV       DWORD PTR [A0],EAX
          CMP       A0,0
          JNE       MOVEXY
          CMP       A1,0                ;DONE IF SUBSCRIPT IS '0000'
          JE        MOVEID
MOVEXY:   MOV       EDX,AUX06           ;PC'S POINT TO DIMS OF MOVE ARRAY
          CALL      ARAYELP             ;DETERMINE ADDRESS
          MOV       EBX,AUX0A           ;GET FIELD POSITION
          MOV       EAX,EBX
          ADD       EAX,EDX             ;ADD TO FIND VALUE ADDRESS
          MOV       AUX01,EAX           ;SAVE IT
          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          MOV       AUX00,EDX
          MOV       EDX,AUX0B           ;GET VALUE LENGTH
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX0E           ;GET RECEIVER LENGTH
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX01           ;GET VALUE ADDRESS
          MOV       AL,K
          AND       AL,03H
          CMP       AL,3                ;BRANCH IF ALPHA --> ALPHA
          JE        MOVEAAI1
          CMP       AL,1                ;BRANCH IF NUMERIC --> ALPHA
          JE        MOVECON
          CMP       AL,2                ;BRANCH IF ALPHA --> NUMERIC
          JE        MOVEDCON
MOVEAAI1: CALL      ARECEVAL            ;MOVE VALUE TO RECEIVER
          MOV       EDX,AUX08           ;GET MOVE COUNT
          MOV       AUX08,EDX           ;DECREMENT AND SAVE
          DEC       AUX08
          MOV       EAX,EDX
          DEC       EDX
          TEST      EAX,0FFH
          JNZ       MOVEAAI0
          CMP       EAX,0                ;BRANCH IF DONE
          JE        MOVEDONE
          test      byte ptr fs:[cpust],10h
          jz        moveaai0            ;BRANCH IF < 30MS. ELAPSED
          MOV       DL,K
          MOV       AUX05,EDX           ;SAVE K
          MOV       R0,SAV03
          CALL      BREAKPTQ            ;BREAKPOINT
          MOV       EDX,AUX05
          MOV       K,DL                ;RESTORE K
          JMP       MOVEAAI0

MOVEID:   TEST      K,8                 ;BRANCH IF N IS NOT SPECIFIED
          JZ        JLMOVED
          MOV       EDX,AUX07           ;GET # OF ELEMENTS EXPECTED WOULD BE MOVED
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX08           ;GET # OF ELEMENTS NOT YET MOVED FROM EXPECTED
          SUB       DWORD PTR [A0],EDX    ;CALCULATES # ACTUALLY MOVED
          DEC       DWORD PTR [A0]
          JMP       MOVEIDD

;        CONVERSION ROUTINES:
;
;        THIS SECTION 'CONVERTS' DATA ELEMENTS FROM
;           2600 INTERNAL FORMAT TO SORT FORMAT.
;
;
;        SET UP F7,F6,F5 SUCH THAT:
;          F7 (HIGH,LOW) = SIGN DIGIT, UNUSED
;          F6 (HIGH,LOW) = EXP (HIGH), EXP (LOW)
;          F5 (HIGH,LOW) = ZERO      , 1ST MANTISSA DIGIT
;

MOVECON:  MOV       R8,0
          CALL      NEWTOOLD            ;CONVERT NEW FORMAT TO OLD FORMAT
          MOV       CX,WORD PTR [EDX]   ;READ VALUE AND PUT INTO SCRATCH AREA
          XCHG      CH,CL
          MOV       AUX01,EDX           ;READ VALUE AND PUT INTO SCRATCH AREA
          ADD       AUX01,2
          MOV       R7,CH
          MOV       R5,CL
          AND       R5,0FH
          AND       CH,0FH
          AND       CL,0F0H
          OR        CL,CH
          MOV       R6,CL
          MOV       EDX,OFFSET MATHPAGE+2       ;POINT TO SCRATCH AREA
          XCHG      EDX,AUX01           ;VALUE ADDRESS TO PH,PL
          MOV       CX,WORD PTR [EDX]   ;READ DIGITS 2-5
          XCHG      CH,CL
          XCHG      EDX,AUX01
          ADD       AUX01,2
          MOV       BYTE PTR [EDX],CH
          INC       EDX
          MOV       BYTE PTR [EDX],CL
          XCHG      EDX,AUX01
          INC       AUX01
          MOV       CX,WORD PTR [EDX]   ;READ DIGITS 6-9
          XCHG      CH,CL
          XCHG      EDX,AUX01
          ADD       AUX01,2
          MOV       BYTE PTR [EDX],CH
          INC       EDX
          MOV       BYTE PTR [EDX],CL
          XCHG      EDX,AUX01
          INC       AUX01
          MOV       CX,WORD PTR [EDX]   ;READ DIGITS 10-13
          XCHG      CH,CL
          XCHG      EDX,AUX01
          ADD       AUX01,2
          MOV       BYTE PTR [EDX],CH   ;TURN OFF CARRY
          INC       EDX
          MOV       BYTE PTR [EDX],CL
          XCHG      EDX,AUX01
          INC       AUX01
          TEST      R5,0FH
          JNZ       JLMOVE7A            ;BRANCH IF NOT ZERO
          MOV       R7,80H              ;SET UP ZERO IN SORT FORMAT
JLMOVE7:  MOV       EDX,OFFSET MATHPAGE ;POINT TO SCRATCH AREA
          MOV       AL,R7               ;WRITE SIGN, EXPONENT AND FIRST DIGIT
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          MOV       AL,R6
          MOV       BYTE PTR [EDX],AL
          DEC       EDX
          CMP       R8,0
          JE        JLMOVE99
          CALL      OLDTONEW
JLMOVE99: MOV       R8,0
          TEST      K,4
          JNZ       MOVEAAI1            ;BRANCH IF INDIRECT
          JMP       MOVEAAD1


;        FIRST FIX SIGN DIGIT ( F7(HIGH) ):
;               0 --> 9
;               8 --> 8
;               9 --> 1
;               1 --> 0
;

JLMOVE7A: MOV       AL,R7
          AND       AL,0F0H
          CMP       AL,80H
          JE        JLMOVE8
          MOV       AL,R7               ;CARRY WAS TURNED OFF ABOVE
          ADD       AL,90H
          DAA
          MOV       R7,AL
          AND       AL,0F0H
          CMP       AL,80H
          JNE       JLMOVEX
          MOV       R7,10H
JLMOVEX:  TEST      R7,0F0H
          JNZ       JLMOVE9

;        TAKE 9'S COMPLEMENT OF EXPONENT IF MANTISSA
;           AND EXPONENT SIGNS DIFFER
;

JLMOVE8:  STC
          MOV       AL,0
          SBB       AL,R6
          DAS
          MOV       R6,AL
          INC       EDX
JLMOVE9:  AND       R7,0F0H
          MOV       AL,R6
          SHR       AL,4
          OR        R7,AL               ;PACK SIGN, EXPONENT, AND FIRST DIGIT INT
          SHL       R6,4
          MOV       AL,R5
          AND       AL,0FH
          OR        R6,AL               ;TWO BYTES FOR WRITE TO SCRATCH AREA.
          TEST      R7,80H
          JNZ       JLMOVE7             ;BRANCH IF MANTISSA IS POSITIVE

;
;        IF MANTISSA IS NEGATIVE, TAKE 9'S COMPLEMENT
;           OF ALL MANTISSA DIGITS.
;

MOVECOMP: STC                           ;COMPLEMENTS FIRST DIGIT
          MOV       AL,0                ;COMPLEMENTS FIRST DIGIT
          SBB       AL,R5
          DAS
          MOV       R5,AL
          INC       EDX
          AND       R6,0F0H             ;PACK INTO F6 FOR WRITE TO SCRATCH AREA
          MOV       AL,R5
          AND       AL,0FH
          OR        R6,AL
          MOV       EDX,OFFSET MATHPAGE+2        ;COMPLEMENT DIGITS 2-5
          MOV       CX,WORD PTR [EDX]
          XCHG      CH,CL
          STC
          MOV       AL,0
          SBB       AL,CH
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          STC
          MOV       AL,0
          SBB       AL,CL
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          MOV       CX,WORD PTR [EDX]   ;COMPLEMENT DIGITS 6-9
          XCHG      CH,CL
          STC
          MOV       AL,0
          SBB       AL,CH
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          STC
          MOV       AL,0
          SBB       AL,CL
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          MOV       CX,WORD PTR [EDX]   ;COMPLEMENT DIGITS 10-13
          XCHG      CH,CL
          STC
          MOV       AL,0
          SBB       AL,CH
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          STC
          MOV       AL,0
          SBB       AL,CL
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          JMP       JLMOVE7


;        THIS SECTION 'UNCONVERTS' DATA FROM SORT FORMAT
;           TO 2600 INTERNAL FORMAT.
;
;        ALPHA ELEMENTS MAY BE DEFINED TO ANY LENGTH, BUT
;           ONLY THE FIRST EIGHT BYTES WILL BE USED IN
;           'UNCONVERT'.  IF THE VALUE LENGTH IS LESS
;           THAN 8 BYTES, THE VALUE WILL BE PADDED WITH
;           ZERO'S IF THE VALUE IS POSITIVE OR NINE'S IF
;           THE VALUE IS NEGATIVE BEFORE UNCONVERSION.
;
;

MOVEDCON: MOV       R8,1
          MOV       AUX01,EDX           ;SAVE VALUE ADDRESS
          MOV       A5,08H              ;MAXIMUM OF 8 BYTES WILL BE MOVED
          MOV       AX,A0
          CMP       A5,AX
          JBE       MOVEDCONX           ;BRANCH IF LENGTH IS 8 OR MORE BYTES
          MOV       A5,AX               ;VALUE LENGTH IS LESS THAN 8
MOVEDCONX:MOV       A0,08H              ;LENGTH OF SCRATCH AREA
          MOV       EDX,OFFSET MATHPAGE ;POINTS TO SCRATCH AREA
          XCHG      EDX,AUX01           ;POINTERS READY FOR FIRST READ
          ADD       A5,DX               ;END CONDITION FOR READ/WRITE
MOVE10A:  MOV       CH,BYTE PTR [EDX]
          XCHG      EDX,AUX01
          INC       AUX01
          MOV       BYTE PTR [EDX],CH
          XCHG      EDX,AUX01
          INC       AUX01
          CMP       A5,DX
          JNE       MOVE10A             ;BRANCH IF NOT DONE
          MOV       EDX,OFFSET MATHPAGE ;POINT TO SIGN,EXPONENT, AND FIRST DIGIT
          MOV       CX,WORD PTR [EDX]
          XCHG      CH,CL
          MOV       R7,CH               ;SETUP F7F6F5 AS SHOWN ABOVE FOR'CONVERT'
          MOV       R5,CL
          AND       R5,0FH
          AND       CH,0FH
          AND       CL,0F0H
          OR        CL,CH
          MOV       R6,CL
          MOV       EDX,AUX01           ;GET NEXT POSITION IN SCRATCH AREA
          TEST      R7,80H
          JZ        MOVE10B             ;BRANCH IF NUMBER IS NEGATIVE
          JMP       MOVEDCONY
MOVEXX:   MOV       BYTE PTR [EDX],0    ;PAD SCRATCH AREA WITH ZERO'S
          INC       EDX
MOVEDCONY:CMP       EDX,OFFSET MATHPAGE+8  ;LOOP IF NOT END OF SCRATCH AREA
          JNE       MOVEXX
          TEST      R5,0FH
          JNZ       MOVE10              ;BRANCH IF VALUE IS NOT ZERO
          MOV       R7,0                ;INTERNAL FORMAT ZERO
          JMP       JLMOVE7
MOVE10B:  MOV       R4,99H              ;PAD SCRATCH AREA WITH NINE'S
          JMP       MOVEDCONZ
MOVEYY:   MOV       AL,R4
          MOV       BYTE PTR [EDX],AL
          INC       EDX
MOVEDCONZ:CMP       EDX,OFFSET MATHPAGE+8
          JNE       MOVEYY

;        FIRST FIX SIGN DIGIT ( F7(HIGH) ):
;               9 --> 0
;               8 --> 8
;               1 --> 9
;               0 --> 1
;

MOVE10:   INC       EDX                 ;TURN OFF CARRY
          MOV       AL,R7
          AND       AL,0F0H
          CMP       AL,80H
          JE        JLMOVE11
          MOV       AL,R7
          ADD       AL,10H
          DAA
          MOV       R7,AL
          AND       AL,0F0H
          CMP       AL,20H
          JNE       MOVEZZ
          MOV       R7,90H
MOVEZZ:   MOV       AL,R7
          AND       AL,0F0H
          CMP       AL,10H
          JNE       JLMOVE12

;  IF SIGNS OF MANTISSA AND EXPONENT DO NOT DIFFER, DON'T
;    COMPLEMENT EXPONENT.

JLMOVE11: STC
          MOV       AL,0
          SBB       AL,R6
          DAS
          MOV       R6,AL
          INC       EDX
JLMOVE12: AND       R7,0F0H
          MOV       AL,R6
          SHR       AL,4
          OR        R7,AL               ;PACK SIGN, EXPONENT, AND FIRST DIGIT INT
          SHL       R6,4
          MOV       AL,R5
          AND       AL,0FH
          OR        R6,AL               ;TWO BYTES FOR WRITE TO SCRATCH AREA.
          TEST      R7,10H
          JZ        JLMOVE7             ;BRANCH IF MANTISSA IS POSITIVE
          JMP       MOVECOMP

;        MAT MERGE  -  MERGE ROWS OF SORTED DATA
;                         INTO A SINGLE OUTPUT BUFFER
;                         IN ASCENDING ORDER.
;
;        GENERAL FORM:
;        ************************************************
;        *  MAT MERGE (merge array) TO (control area),  *
;        *               (work area),(locator array)    *
;        ************************************************
;
;               WHERE:   (merge array) - ROWS OF SORTED
;                              DATA WITH NOT MORE THAN
;                              254 ELEMENTS PER ROW.
;
;                        (control area) - USED TO STORE
;                              THE 1-BYTE COLUMN POSITION
;                              OF THE NEXT ELEMENT IN
;                              EACH ROW.  THE LAST BYTE
;                              IS USED AS A CONTROL FLAG
;                              SIGNALING TO THE USER IF A
;                              ROW IS EMPTY OR IF THE
;                              LOCATOR ARRAY IS FULL.
;
;                              REQUIRES: N+1 BYTES WHERE
;                                N = (no. of rows)  IN
;                                THE MERGE ARRAY.
;
;                        (work area) - USED TO STORE THE
;                              ADDRESS OF THE ELEMENT
;                              SPECIFIED BY THE COLUMN
;                              POSITION IN THE RESPECTIVE
;                              ROW OF THE CONTROL AREA.
;
;                              REQUIRES: N*2 BYTES WHERE
;                                N = (no. of rows) IN THE
;                                MERGE ARRAY.
;
;                        (locator array) - AN ARRAY WHICH
;                              RECEIVES THE 2-BYTE
;                              SUBSCRIPTS OF ELEMENTS IN
;                              THE MERGE ARRAY AS THEY
;                              ARE ARRANGED IN ASCENDING
;                              ORDER.
;
;
;        MODIFIED REGISTERS:  ALL, AUX 0-F
;                             (AUX 0,1 IN ENTRY PHASE)
;
;        EXPR BETTER NOT USE:   AUX 3,4,5,7,8,9,A,B,C,D
;

MERGE:    cmp       byte ptr [textptr],'!'
          jne       merge111
          inc       textptr
          jmp       merge_s
merge111:
          CALL      ARRAY               ;SCAN MERGE ARRAY VARIABLE
          TEST      R0,0FH
          JNZ       ERR_28S             ;ERROR IF NOT FOUND
          TEST      R5,1
          JZ        ERR_28S             ;ERROR IF NOT ALPHA
          TEST      SL,2
          JZ        MERGE0              ;BRANCH IF NOT X PHASE
          MOV       AL,R5
;         AND       AL,0FH              ;05/12/92
          AND       AL,07H
          CMP       AL,7
          JNE       ERR_50S             ;ERROR IF NOT ALPHA MATRIX
          MOV       AUX0B,EDX           ;SAVE ADDRESS OF DIMS
          SUB       AUX0B,5
          MOV       EDX,DWORD PTR [A6]
          CMP       DX,0FFH
          JAE       ERR_50S             ;ERROR IF MORE THAN 254 ELEMENTS/ROW
          MOV       AUX0C,EDX           ;SAVE MERGE DIMS
          MOVZX     EDX,CH
          MOV       AUX0D,EDX           ;SAVE ELEMENT LENGTH
          MOV       AUX03,EDX           ;SAVE AS DEFAULT FIELD LENGTH
MERGE0:   CALL      FIELD               ;SCAN AND PROCESS A FIELD SPECIFICATION
          TEST      SL,2
          JZ        MERGE1              ;BRANCH IF NOT X PHASE
          MOV       A6,0
          MOV       EDX,DWORD PTR [A5]    ;SAVE INCREMENT TO START OF FIELD
          MOV       AUX0A,EDX           ;SAVE IT
          CMP       A4,0
          JE        MERGE1              ;BRANCH IF FIELD NOT SPECIFIED
          MOVZX     EDX,A4
          MOV       AUX03,EDX           ;SAVE FIELD LENGTH
MERGE1:   CALL      CCLTO               ;MUST FIND "TO"

; CONTROL AREA
;
          CALL      AVARD               ;SCAN CONTROL AREA VARIABLE
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        MERGE2              ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET ADDRESS LENGTH OF CONTROL AREA
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX08,EDX           ;SAVE ADDRESS
          ADD       EDX,EAX
          MOV       AUX09,EDX           ;SAVE ADDRESS OF CONTROL FLAG (LAST BYTE)
          DEC       AUX09
          MOV       EDX,AUX0C           ;GET MERGE DIMS
          SHR       EDX,16              ;NO. OF ROWS PUT IN PH,PL
          MOV       AUX07,EDX           ;TEMPORARY SAVE # OF ROWS
          CMP       DWORD PTR [A0],EDX
          JBE       ERR_51S             ;ERROR IF NOT LARGE ENOUGH
MERGE2:   CMP       BYTE PTR [TEXTPTR],','  ;MUST FIND COMMA
          JNE       ERROR_13
          INC       TEXTPTR

; WORK VARIABLE
;
          CALL      AVARD               ;SCAN WORK AREA VARIALBLE
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        MERGE3              ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET ADDRESS & LENGTH OF BUFFER
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX06,EDX           ;SAVE ADDRESS
          MOV       EDX,AUX07           ;GET NO. OF ROWS IN MERGE ARRAY
          MOV       DWORD PTR [A2],EDX
          SHL       EDX,1                ;DOUBLE TO FIND REQUIRED #  OF WORK BYTES
          CMP       DWORD PTR [A0],EDX
          JB        ERR_51S             ;ERROR IF WORK AREA ISN'T LARGE ENOUGH
          MOV       BYTE PTR TTT,0
          shl       edx,1
          jc        err_51s
          cmp       dword ptr [a0],edx
          jb        merge3
          mov       byte ptr ttt,1
MERGE3:   CMP       BYTE PTR [TEXTPTR],',' ;MUST FIND A COMMA
          JNE       ERROR_13
          INC       TEXTPTR

; LOCATOR ARRAY
;
          CALL      ARRAY               ;SCAN LOCATOR ARRAY
          TEST      R0,0FH
          JNZ       ERR_28S             ;ERROR IF NOT FOUND
          TEST      R5,1
          JZ        ERR_28S             ;ERROR IF NOT ALPHA
          TEST      SL,2
          JZ        MERGE14             ;BRANCH IF NOT X PHASE
          CMP       CH,2
          JNE       ERR_50S             ;ERROR IF ELEMENT LENGTH IS NOT 2
          MOV       AUX0E,EDX           ;SAVE LOCATOR ADDRESS
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX
          ADC       EDX,EAX
          MOV       AUX0F,EDX           ;SAVE LAST ADDRESS + 2

;        REGISTER USE IS AS FOLLOWS:
;               AUX 3 - FIELD LENGTH OF MERGE ARRAY
;                       ELEMENT ( = ELEMENT LENGTH IF
;                       FIELD IS NOT SPECIFIED)
;               AUX 6 - ADDRESS OF WORK AREA
;               AUX 8 - ADDRESS OF CONTROL AREA
;               AUX 9 - ADDRESS OF CONTROL FLAG ( LAST
;                       BYTE OF CONTROL AREA)
;               AUX A - INCREMENT TO START OF FIELD
;               AUX B - ADDRESS OF DIMS OF MERGE ARRAY
;               AUX C - DIMENSIONS OF MERGE ARRAY
;               AUX D - ELEMENT LENGTH OF MERGE ARRAY
;               AUX E - ADDRESS OF LOCATOR ARRAY
;               AUX F - LAST ADDRESS OF LOCATOR ARRAY
;
;
;        READ STARTING POSITIONS FOR EACH ROW FROM
;        CONTROL AREA AND PUT CORRESPONDING ELEMENT
;        ADDRESSES INTO WORK BUFFER.  IF POSITION IS 'FF'
;        THEN IGNORE ROW BY PUTTING '0000' INTO WORK
;        BUFFER.  SAVE ROW # AND ELEMENT ADDRESS OF THE
;        FIRST ROW NOT IGNORED.
;
;        K IS USED AS A FLAG FOR ARAYELP INDICATING
;           ALPHA MATRIX ( HIGH 8 BITS = 5 ) AND AS A
;              FLAG INDICATING WHETHER ALL ROWS ARE
;                 TO BE IGNORED ( LOW 8 BITS, 0=YES,1=NO)
;

          cmp       byte ptr ttt,1
          je        mergeqqq

; added for emulate work space under merge processing
;
          mov       edx,vsptr
          mov       eax,aux07
          shl       eax,2
          sub       edx,eax
          cmp       edx,dword ptr [@@vsend]
          jb        error_02
          mov       aux06,edx           ;put emulate space address to work string address

mergeqqq:
          MOV       K,50H
          MOV       AUX05,0             ;INIT ROW POSITION (0001)
          INC       AUX05
          MOV       EDX,AUX0B
          MOV       CX,WORD PTR [EDX]
          MOVZX     EDX,CX
          DEC       EDX
          PUSH      EDX                 ;SAVE NO. OF ROWS IN MERGE ARRAY -1
          MOV       EDX,AUX06           ;SAVE WORK AREA ADDRESS
          MOV       AUX07,EDX
          MOV       EDX,AUX08           ;ADDRESS OF CONTROL AREA
MERGE4:   MOV       CH,BYTE PTR [EDX]   ;READ COLUMN POSITION IN ROW
          MOV       AUX04,EDX           ;READ COLUMN POSITION IN ROW
          INC       AUX04
          MOV       EDX,AUX05
          INC       AUX05               ;INCREMENT TO NEXT ROW AND SAVE
          CMP       CH,0FFH
          JNE       MERGE4A             ;BRANCH IF NOT IGNORE ROW
          MOV       DWORD PTR [A0],0     ;PUT '0000' INTO WORK AREA
          JMP       MERGE45

MERGE4A:  MOVZX     AX,CH               ;PUT ELEMENT SUBSCRIPT INTO F1,F0
          MOV       A0,AX
          MOV       A1,DX
          MOV       AL,K
          AND       AL,0FH
          CMP       AL,1
          JE        MERGE4B             ;BRANCH IF NOT FIRST GOOD ROW
          MOV       K,51H               ;FLAG = FIRST GOOD ROW FOUND
          MOV       EAX,EDX
          SHL       EDX,16
          MOV       DX,AX
          MOV       AUX05,EDX           ;SAVE ROW NUMBER OF FIRST GOOD ROW
          INC       AUX05
          MOV       EDX,AUX07           ;SAVE WORK AREA ADDRESS OF FIRST GOOD ROW
          MOV       AUX01,EDX
MERGE4B:  MOV       EDX,AUX0B           ;GET MERGE DIM ADDRESS
          CALL      ARAYELP             ;GET ELEMENT ADDRESS TO PC'S

MERGE4C:  MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX0A           ;GET INCREMENT TO START OF FIELD
          MOV       EAX,EDX               ;F1,0 = ADDRESS OF START OF FIELD
          ADD       DWORD PTR [A0],EAX
MERGE45:  MOV       EDX,AUX07           ;GET ADDRESS IN WORK AREA
          MOV       ECX,DWORD PTR [A0]
          MOV       DWORD PTR [EDX],ECX ;WRITE ELEMENT ADDRESS INTO WORK AREA
          ADD       EDX,4
          MOV       AUX07,EDX
          POP       EDX                 ;GET ROW COUNT
          MOV       EAX,EDX
          DEC       EDX
          CMP       EAX,0
          JE        MERGE5              ;BRANCH IF LAST ROW DONE
          PUSH      EDX                 ;DECREMENT ROW COUNT
          MOV       EDX,AUX04           ;GET ADDRESS OF CONTROL AREA
          JMP       MERGE4
MERGE5:   TEST      K,0FH
          JZ        SEARCHD             ;BRANCH IF ALL ROWS IGNORED

;        NOW PERFORM POLYPHASE MERGE:
;          1) ONE EACH PASS, COMPARE NEXT ELEMENT OF
;             EACH ROW, AS GIVEN BY ADDRESS IN WORK AREA.
;          2) PUT SUBSCRIPT OF LOWEST ELEMENT FOUND IN
;             PASS INTO THE LOCATOR ARRAY.
;          3) INCREMENT THE COLUMN SUBSCRIPT IN THE
;             CONTROL AREA, AND PUT THE ADDRESS OF THIS
;             NEXT ELEMENT INTO THE WORK AREA.
;          4) REPEAT STEPS 1 - 3 UNTIL THE END OF A ROW
;             IS REACHED OR UNTIL THE LOCATOR ARRAY IS
;             FILLED.
;
;        REGISTERS NOW SET UP AS FOLLOWS:
;               AUX 1 - WORK AREA ADDRESS OF FIRST GOOD
;                       ROW
;               AUX 5 (HIGH 8 BITS) - ROW NUMBER OF FIRST
;                       GOOD ROW.
;
;        REGISTER USE WILL BE AS FOLLOWS:
;               AUX 2 (LOW 8 BITS) - ROW POSITION OF
;                       LOWEST ELEMENT IN THE PASS
;               AUX 3 - ADDRESS OF LOWEST ELEMENT IN THE
;                       PASS
;               AUX 4 - CURRENT ADDRESS IN THE WORK AREA
;               AUX 5 (LOW 8 BITS) - CURRENT ROW POSITION
;               F7,F6 - DIMENSIONS OF THE MERGE ARRAY
;               F5,F4 - FIELD LENGTH OF THE MERGE ARRAY
;
          MOV       EDX,AUX0C           ;GET DIMENSIONS OF MERGE ARRAY
          MOV       DWORD PTR [A6],EDX
          MOV       EDX,AUX03           ;GET FIELD LENGTH OF MERGE ARRAY
          MOV       DWORD PTR [A4],EDX
MERGE6:   MOV       EDX,AUX01           ;GET FIRST GOOD WORK ADDRESS
          MOV       ECX,DWORD PTR [EDX]
          ADD       EDX,4               ;READ ELEMENT ADDRESS
          MOV       AUX04,EDX
          MOV       AUX03,ECX           ;INIT FIRST GOOD ROW AS LOWEST
          MOV       EDX,AUX05           ;GET FIRST GOOD ROW POSITION
          MOV       EAX,EDX
          SHR       EAX,16
          MOV       DX,AX
          MOV       AUX05,EDX           ;START COMPARE AT NEXT ROW
          INC       AUX05
          SHL       EDX,16
          SHR       EDX,16
          MOV       AUX02,EDX           ;INIT FIRST ROW AS LOWEST
          DEC       AUX02

MERGE7:   CMP       A7,DX
          JE        MERGE9              ;BRANCH IF LAST ROW
          MOV       EDX,AUX04           ;GET NEXT WORK ADDRESS
          MOV       ECX,DWORD PTR [EDX]
          ADD       EDX,4               ;READ ELEMENT ADDRESS
          MOV       AUX04,EDX
          CMP       ECX,0
          JE        MERGE8              ;BRANCH IF IGNORED '0000'
          MOV       AUX0C,ECX           ;SAVE ELEMENT ADDRESS
          MOV       AUX00,ECX
          MOV       EDX,AUX03           ;ADDRESS OF CURRENT LOWEST ELEMENT
          MOV       EAX,DWORD PTR [A4]
          MOV       DWORD PTR [A0],EAX  ;SET UP FIELD LENGTH
          CALL      COMPARE             ;COMPARE NEXT ELEMENT WITH LOWEST FOR PASS
          CMP       R0,1
          JNE       MERGE8              ;BRANCH IF STILL THE LOWEST
          MOV       EDX,AUX0C           ;GET ELEMENT ADDRESS OF NEW ELEMENT
          MOV       AUX03,EDX           ;SAVE AS THE NEW LOWEST FOR THE PASS
          MOV       EDX,AUX05           ;GET CURRENT ROW POSITION
          MOV       AUX05,EDX           ;INCREMENT TO THE NEXT ROW
          INC       AUX05
          SHL       EDX,16
          SHR       EDX,16
          MOV       AUX02,EDX           ;SAVE AS THE NEW LOWEST FOR THE PASS
          DEC       AUX02
          JMP       MERGE7

MERGE8:   MOV       EDX,AUX05           ;INCREMENT TO THE NEXT ROW
          MOV       AUX05,EDX
          INC       AUX05
          JMP       MERGE7

;        END OF PASS:  NOW PUT SUBSCRIPTS OF LOWEST
;              ELEMENT FOUND FOR THE PASS INTO THE
;              LOCATOR ARRAY AND TEST FOR END OF MERGE
;              CONDITIONS.
;

MERGE9:   MOV       EDX,AUX02           ;GET ROW POSITON OF LOWEST ELEMENT
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX08           ;GET ADDRESS OF CONTROL AREA
          MOV       EAX,DWORD PTR [A0]
          ADD       EDX,EAX             ;FIND ADDRESS OF ROW IN CONTROL AREA
          MOV       CH,BYTE PTR [EDX]
          MOV       AUX00,EDX           ;SAVE AND READ
          MOV       AX,A0
          INC       AX
          MOV       A1,AX               ;PUT SUBSCRIPT INTO F1,F0
          MOVZX     AX,CH
          MOV       A0,AX
          CALL      $POSARR1            ;PUT SUBSCRIPT INTO LOCATOR ARRAY
          MOV       AX,A0
          CMP       AX,A6
          JE        MERGE10             ;BRANCH IF ROW IS EXHAUSTED
          MOV       AL,R1
          MOV       K,AL                ;SAVE LOCATOR STATUS
          INC       A0                  ;INCREMENT COLUMN POSITION
          MOV       EDX,AUX00           ;GET CONTROL AREA ADDRESS
          MOV       AX,A0
          MOV       BYTE PTR [EDX],AL   ;WRITE COLUMN SUBSCRIPT INTO CONTROL AREA
          MOV       EDX,AUX03           ;ADDRESS OF LOWEST ELEMENT
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX0D           ;GET ELEMENT LENGTH OF MERGE ARRAY
          MOV       EAX,EDX
          ADD       DWORD PTR [A0],EAX  ;FIND ADDRESS OF NEXT ELEMENT IN ROW
          MOV       EDX,AUX02           ;GET ROW POSITION OF LOWEST
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX06           ;GET ADDRESS OF WORK AREA
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX             ;FIND ADDRESS OF ROW IN WORK AREA
          ADC       EDX,EAX
          ADC       EDX,EAX
          ADC       EDX,EAX
          MOV       EAX,DWORD PTR [A0]
          MOV       DWORD PTR [EDX],EAX ;WRITE IN NEXT ELEMENT ADDRESS
          ADD       EDX,4
          CMP       K,0
          JNE       MERGE6              ;BRANCH IF LOCATOR NOT FULL

;
;        THIS SECTION CODES FOR LOCATOR ARRAY FULL
;

MERGE12:  MOV       EDX,AUX09           ;GET ADDRESS OF CONTROL FLAG
MERGE13:  MOV       BYTE PTR [EDX],0    ;'00' SIGNALS LOCATOR FULL
MERGE14:  JMP       ENDSTMTZ

;
;        THIS SECTION CODES FOR ROW EMPTY
;

MERGE10:  MOV       EDX,AUX02           ;GET ROW POSITION
          MOV       A0,DX
          INC       A0
          MOV       EDX,AUX09           ;GET ADDRESS OF CONTROL FLAG
          MOV       AX,A0
          MOV       BYTE PTR [EDX],AL   ;SIGNAL THAT ROW IS EMPTY
          MOV       EDX,AUX00           ;GET CONTROL AREA ADDRESS
          MOV       BYTE PTR [EDX],0FFH ;SIGNAL TO IGNORE ROW
          TEST      R1,0FH
          JZ        MERGE14             ;BRANCH IF LOCATOR IS FULL
          MOV       EDX,AUX0E           ;GET LOCATOR ADDRESS
          MOV       BYTE PTR [EDX],0    ;WRITE '0000' AS END OF BUFFER
          INC       EDX
          JMP       MERGE13

;
;
;        CALLS TO ERROR ROUTINES
;
;

ERROR_26: MOV       R0,26H
          JMP       ERROR               ;MISSING ARRAY
ERR_51S:  JMP       ERROR_51            ;VARIABLE TOO SMALL
ERR_28S:  JMP       ERROR_28            ;MISSING ALPHA ARRAY
ERR_50S:  JMP       ERROR_50            ;ILLEGAL DIM'S OR LENGTH
ERR_29I:  JMP       ERROR_29            ;MISSING ALPHA VARIABLE
ERR_55S:  JMP       ERROR_55            ;UNDEFINED ARRAY


;     SRCHALPH: Find Postion of a string in a string satisfying a relation
;
;     ENTRY: AUX3 = ADR(STRING TO SEARCH)
;            AUX4 = LENGTH OF STRING TO SEARCH (no check for null string)
;            AUX5 = ADR(STRING TO SEARCH FOR)
;            AUX6 = LENGTH OF STRING TO SEARCH FOR (no check for null string)
;            AUX7 = STEP SIZE (16 bit binary)
;            AUX8 = RELATION
;                 AUX8,LOW 4-BITS:
;                                 1 BIT FOR <
;                                 2 BIT FOR =
;                                 4 BIT FOR >
;
;         RETURN: LOW 4 BITS OF F7 = X'0' IF NOT FOUND; ELSE IF FOUND
;                 PC's = ADDRESS OF FOUND CHARACTER (valid if F7 = indicates found)
;                 INPUT AUX's NOT MODIFIED
;
;         MODIFIED REGISTERS: ALL, AUX 0-2
;
;         NOTE: This routine can breakpoint!
;         NOTE: Value to be searched cannot start at address 0000 or end at address FFFF
;

SRCHALPH: MOV       EDX,AUX04           ;PC's = LENGTH OF SEARCH STRING
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX06           ;PC's = LENGTH OF COMPARE STRING
          CMP       DWORD PTR [A2],EDX
          JB        SRCHAL9             ;BRANCH IF SEARCH STRING SHORTER THAN COMPARE STRING
          SUB       DWORD PTR [A2],EDX    ;F3F2 = LENGTH OF STRING TO ACTUALLY SEARCH
          MOV       EDX,AUX03           ;PC's = ADR(SEARCH STRING)

;      FORWARD SEARCH
;

SRCHFWD:  MOV       AUX01,EDX           ;AUX1 = FIRST ADR TO SEARCH
          MOV       EAX,DWORD PTR [A2]
          STC
          ADC       EDX,EAX             ;PC's = LAST ADR TO SEARCH +1
          MOV       AUX02,EDX           ;AUX2 = END OF LOOP TEST VALUE
          CALL      SRCHARGX

SRCHFWD3: MOV       CH,BYTE PTR [EDX]
          MOV       AL,R7
          AND       AL,04H
          MOV       K,AL                ;SET > CONDITION, IF REQUESTED
          CMP       R6,CH
          JB        SRCHFWD1            ;BRANCH IF CHAR > SEARCH CHAR
          MOV       AL,R7
          AND       AL,02H
          MOV       K,AL                ;GET CHAR; SET = CONDITION, IF REQUESTED
          CMP       CH,R6
          JE        SRCHFWD1            ;BRANCH IF CHAR = SEARCH CHAR
          MOV       AL,R7
          AND       AL,01H
          MOV       K,AL                ;SET < CONDITION, IF REQUESTED
SRCHFWD1: TEST      K,0FH
          JNZ       SRCHAL1             ;BRANCH IF RELATION TRUE
SRCHFWD2: test      byte ptr fs:[cpust],10h
          jz        srchfwd2a
          CALL      BREAKPTT            ;IF 30MS. ELAPSED, BREAKPOINT
SRCHFWD2A:
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX             ;PC'S = ADR(NEXT BYTE)
          CMP       DWORD PTR [A0],EDX
          JBE       SRCHAL9             ;BRANCH IF END OF VALUE
          JNC       SRCHFWD3            ;REPEAT IF ADDRESS DIDN'T WRAP

;       END OF VALUE, RETURN NOT FOUND
;

SRCHAL9:  MOV       R7,0                ;SIGNAL NOT FOUND
          RET

;      1 BYTE FOUND - TEST FOR ENTIRE COMPARISON VALUE
;

SRCHAL1:  MOV       AUX01,EDX           ;AUX1=ADR(LAST BYTE TESTED)

;2/21/89 ++
        mov     ecx,aux06               ;length of compare string
        mov     dl,byte ptr aux08       ;relational operator
        and     dl,0fh                  ;
        mov     r7,dl                   ;
        mov     edi,aux05               ;addr of comparision string
        mov     esi,aux01               ;addr of last byte tested
        call    rela_e                  ;compare two equal length string

;         MOV       EDX,AUX06           ;PC's = LENGTH OF COMPARE STRING
;         MOV       WORD PTR [R0],DX    ;F1F0 = LENGTH OF COMPARE STRING
;         MOV       EDX,AUX08
;         MOV       R7,DL
;         AND       R7,0FH              ;F7 = RELATIONAL OPERATOR (without search direction flag)
;         MOV       EDX,AUX05
;         MOV       AUX00,EDX           ;AUX0 = ADDRESS OF COMPARISION STRING
;         MOV       EDX,AUX01           ;PC's = ADR(LAST BYTE TESTED)
;         MOV       R4,0                ;F4 = 0
;         CALL      RELALPHC            ;COMPARE MULTI-BYTE STRINGS
;2/21/89 --
          TEST      R7,0FH
          JNZ       SRCHAL2             ;BRANCH IF MATCH
          CALL      SRCHARGS            ;SET UP REGISTERS FOR FURTHUR 1-BYTE TESTS
          JMP       SRCHFWD2            ;BRANCH BACK TO 1 BYTE SEARCH LOOP

SRCHAL2:  MOV       EDX,AUX01           ;PC's = ADR(1st BYTE OF MATCHING STRING)
SRCHAL3:  OR        R7,08H              ;TURN ON A BIT IN F7 TO INDICATE MATCH
          RET

;       SET UP REGISTERS FOR 1 BYTE COMPARE LOOPS
;

SRCHARGS: MOV       EDX,AUX02           ;PC's = LAST ADR TO TEST

; SRCHARGS+1 ENTRY POINT (PC's = LAST ADR TO TEST)
;
SRCHARGX: MOV       DWORD PTR [A0],EDX  ;F1F0 = LAST ADR TO TEST
          MOV       EDX,AUX07           ;PC's = STEP SIZE
          MOV       DWORD PTR [A2],EDX  ;F3F2 = STEP SIZE
          MOV       EDX,AUX05           ;PC's = ADDRESS OF COMPARISION STRING
          MOV       CH,BYTE PTR [EDX]
          MOV       EDX,AUX08           ;CH = 1st BYTE OF COMPARE STRING; PC's = RELATIONAL OPERATOR
          MOV       R7,DL
          OR        R7,02H              ;F7 = RELEATIONAL OPERATOR WITH EQUAL CONDITION ALWAYS ENABLED
          MOV       R6,CH               ;F6 = 1st BYTE OF COMPARE STRING
          MOV       EDX,AUX01           ;PC's =ADR(NEXT BYTE TO TEST)
          RET


;        FIELD  -  PROCESS A FIELD SPECIFICATION WITHIN
;                     AN ARRAY ELEMENT.
;
;        ENTRY  -  LENGTH OF ELEMENT IN PH,PL
;
;        RETURN -  FOUND:
;                     F5 = FIELD POSITION - 1
;                     F4 = FIELD LENGTH
;                  NOT FOUND:
;                     F5,F4 = 0  (NOTE THAT F5 MAY ALSO
;                        EQUAL ZERO IF FOUND)
;
;        MODIFIED REGISTERS - ALL, AUX 4,5
;
;        EXPR BETTER NOT USE:   AUX 4,5
;


FIELD:    MOV       AUX05,EDX           ;SAVE ELEMENT LENGTH
          MOV       DWORD PTR [A4],0    ;CLEAR FIELD LENGTH AND FIELD POSITION
          CMP       BYTE PTR [TEXTPTR],'('  ;LOOK FOR "("
          JNE       FIELD3              ;BRANCH IF NOT FOUND
          INC       TEXTPTR
          CALL      EXPR                ;SCAN FOR FIELD POSITION
          TEST      SL,2
          JZ        FIELD1              ;BRANCH IF NOT X PHASE
          CALL      FLTBIN              ;CONVERT TO BINARY VALUE IN F1,F0
          TEST      R2,2
          JZ        ERR34               ;ERROR IF < 0 OR > 2^32
          CMP       DWORD PTR [A0],0FFFFH
          JA        ERR34
          MOVZX     EDX,A0
          MOV       AUX04,EDX           ;SAVE FIELD POSITION
          DEC       AUX04

FIELD1:   CMP       BYTE PTR [TEXTPTR],')'  ;LOOK FOR A ")"
          JNE       FIELDX              ;BR, IF NF
          INC       TEXTPTR
          JMP       FIELD4              ;BRANCH IF FOUND
FIELDX:   CMP       BYTE PTR [TEXTPTR],','  ;MUST FIND A COMMA
          JNE       ERROR_13
          INC       TEXTPTR
          CALL      EXPR                ;SCAN FOR FIELD LENGTH
          TEST      SL,2
          JZ        FIELD2              ;BRANCH IF NOT X PHASE
          CALL      FLTBIN              ;CONVERT TO BINARY IN F1,F0
          TEST      R2,2
          JZ        ERR34               ;ERROR IF <0 OR >2^32
          CMP       DWORD PTR [A0],0FFFFH
          JA        ERR34
          MOV       EDX,AUX04           ;GET FIELD POSITION
          MOV       A5,DX               ;SAVE IT
          MOV       EAX,DWORD PTR [A0]
          ADD       EAX,EDX
          MOV       DWORD PTR [A2],EAX  ;FIND END POSITION
          MOV       EDX,AUX05           ;GET ELEMENT LENGTH
          CMP       DWORD PTR [A2],EDX
          JBE       FIELDY              ;BRANCH IF LEGAL FIELD
ERR34:    JMP       ERROR_34            ;ILLEGAL VALUE
FIELDY:   MOV       AX,A0               ;SAVE FIELD LENGTH
          MOV       A4,AX
FIELD2:   CMP       BYTE PTR [TEXTPTR],')' ;MUST FIND ")"
          JNE       ERROR_11
          INC       TEXTPTR
FIELD3:   RET

FIELD4:   TEST      SL,2
          JZ        FIELD3              ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX04           ;GET FIELD POSITION
          MOV       A5,DX               ;SAVE IT
          MOV       A0,DX
          XOR       A0,0FFFFH           ;COMPLEMENT FOR SUBTRACT
          MOV       EDX,AUX05           ;GET ELEMENT LENGTH
          STC
          MOV       AX,A0
          ADC       AX,DX
          MOV       A4,AX               ;SAVE FIELD LENGTH
          JNC       ERR34               ;ERROR IF CARRY IS OFF (INVALID POSITION)
          CMP       A4,0
          JE        ERR34               ;ERROR IF ZERO (INVALID POSITION)
          RET


;        POSARR1   -  PUTS THE TWO BYTE VALUE STORED IN
;                     F1F0 INTO THE ADDRESS GIVEN BY
;                     AUX E, AND THEN INCREMENTS THE
;                     ADDRESS AND CHECKS IF DONE.
;
;        ENTRY:  AUX E = RECEIVER ADDRESS
;                AUX F = LAST ADDRESS + 2
;
;        RETURN: F1 = 0 IF DONE
;                F1 <> 0 IF NOT DONE
;
;        MODIFIED REGISTERS:  PH,PL,F0-F3,AUX E,F
;

POSARR1:  MOV       EDX,AUX0F           ;GET LAST ADDRESS
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX0E           ;GET CURRENT ADDRESS
          MOV       EAX,DWORD PTR [A0]
          XCHG      AH,AL
          MOV       WORD PTR [EDX],AX   ;WRITE VALUE IN
          ADD       EDX,2
          MOV       AUX0E,EDX           ;SAVE NEW CURRENT ADDRESS
          CMP       DWORD PTR [A2],EDX
          JBE       POSA0               ;BRANCH IF DONE
          MOV       R1,1
          RET

POSA0:    MOV       R1,0
          RET

$POSARR1: MOV       EDX,AUX0F           ;GET LAST ADDRESS
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX0E           ;GET CURRENT ADDRESS
          MOV       AX,A1
          MOV       BYTE PTR [EDX],AL
          MOV       AX,A0
          MOV       BYTE PTR [EDX+1],AL ;WRITE VALUE IN
          ADD       EDX,2
          MOV       AUX0E,EDX           ;SAVE NEW CURRENT ADDRESS
          CMP       DWORD PTR [A2],EDX
          JBE       $POSA0               ;BRANCH IF DONE
          MOV       R1,1
          RET

$POSA0:   MOV       R1,0
          RET


;        COMPARE - COMPARE 2 VALUES OF EQUAL LENGTH
;
;        ENTRY:  AUX 0 = ADDRESS OF ELEMENT 2
;                PH,PL = ADDRESS OF ELEMENT 1
;                F1,F0 = ELEMENT LENGTH
;
;        RETURN:    F0 = 0   ELEMENT 1 <= ELEMENT 2
;                   F0 = 1   ELEMENT 1 > ELEMENT 2
;
;        MODIFIED REGISTERS: CH,CL,K,F0,F1,PH,PL,AUX 0
;

COMPARE:  ADD       DWORD PTR [A0],EDX  ;FIND LAST ADDRESS IN ELEMENT 1

COMP:     CMP       DWORD PTR [A0],EDX  ;BRANCH IF DONE
          JBE       COMP0
          MOV       CH,BYTE PTR [EDX]   ;READ CHARACTER OF ELEMENT 1
          XCHG      EDX,AUX00
          INC       AUX00
          MOV       K,CH
          MOV       CH,BYTE PTR [EDX]   ;READ CHARACTER OF ELEMENT 2
          XCHG      EDX,AUX00
          INC       AUX00
          CMP       CH,K
          JE        COMP                ;LOOP IF EQUAL
          JB        COMP1               ;BRANCH IF ELEMENT 1 > ELEMENT 2
COMP0:    MOV       R0,0
          RET

COMP1:    MOV       R0,1
          RET


;  ARAYELRP:  LOOK FOR ARRAY ELEMENT, OR JUST A RIGHT
;    PARENTHESIS, WHICH DEFAULTS TO FIRST ELEMENT.
;

ARAYELRP: CMP       BYTE PTR [TEXTPTR],')'  ;CHECK FOR RP
          JNE       ARAYEL              ;NOT FOUND, GET ELEMENT
          INC       TEXTPTR
          TEST      SL,3                ;RETURN IF ENTRY PHASE
          JZ        ARSR
          DEC       TEXTPTR
          CALL      FINDVAR             ;FIND/DEFINE VARIABLE
          INC       TEXTPTR
          TEST      SL,2
          JZ        ARSR                ;RETURN IF NOT X PHASE
          TEST      R0,0FH
          JNZ       ERR_55I             ;NF, ERROR
          MOV       AL,K
          MOV       R0,AL

          SHR       AL,4
          MOV       AH,CL
          SHL       AH,4
          OR        AL,AH
          MOV       K,AL

          MOV       CX,WORD PTR [EDX]   ;AUX 0 POINTS AFTER DIMS
          shl       ecx,16
          mov       cx,word ptr [edx+2]
          MOV       AUX00,EDX
          ADD       AUX00,4
          MOV       DWORD PTR [A6],ECX    ;F7F6 HAS DIMS
          MOV       AUX02,0               ;AUX 2 HAS ELEMENT # -1  (=0)
          MOV       EDX,AUX00
          TEST      K,10H
          JZ        ARSR
          MOV       CH,BYTE PTR [EDX]   ;IF ALPHA, CH = ELEMENT LENGTH
          INC       EDX
ARSR:     RET                           ;PC'S POINT AT VALUE


;      ERROR INDIRECTION
;

ERR_55I:  JMP       ERROR_55            ;UNDEFINED ARRAY


; NAME: LOOK FOR NAME IN TEXT. ERROR IF NOT FOUND.
;
;     FORMAT: name = literal or alpha-variable
;
;     RETURN (EXECUTION PHASE):
;             F0 = LENGTH OF NAME (1-8)
;             PC'S = ADDRESS OF NAME
;
;     MODIFIED REGISTERS: ALL,AUX0-2
;
; NAMEA: LOOK FOR NAME IN TEXT
;
;     RETURN: F0 = FF IF NOT FOUND
;             F0 = 0X IF FOUND
;               IN EXECUTION PHASE,
;                    F0 = LENGTH OF NAME (1-8)
;                    PC'S = ADDRESS OF NAME
;

NAMEA:    MOV       K,4                 ;GET CURRENT LENGTH
          CALL      ALPHA1              ;LOOK FOR ALPHA VARIABLE OR LITERAL
          TEST      R0,0FH
          JZ        NAME2               ;F, BRANCH
          MOV       R0,0FFH             ;FLAG = NF
NAME1:    RET

NAME_:    MOV       K,4                 ;GET CURRENT LENGTH
          CALL      ALPHA1              ;LOOK FOR ALPHA VARIABLE OR LITERAL
          TEST      R0,0FH
          JNZ       ERROR_29            ;ERROR IF NOT FOUND
NAME2:    TEST      SL,2                ;BRANCH IF NOT EXECUTION PHASE
          JZ        NAME1
          MOV       EDX,VSPTR           ;GET ADDRESS & LENGTH OF NAME
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          CALL      CURLEN              ;CALCULATE CURRENT LENGTH OF VALUE
          MOV       EDX,DWORD PTR [A2]  ;PC'S = ADDRESS OF VALUE
          CMP       DWORD PTR [A0],8
          JBE       NAMEX
          JMP       ERROR_52            ;ERROR IF LENGTH > 8
namex:    mov       ax,word ptr [a0]
          mov       word ptr [r0],ax
          jmp       name1

; BITTERM#: CONVERT CURRENT BINARY TERM# TO A BIT POSITION IN
;          A 2-BYTE VALUE
;
;          BYTE = 8000 IF TERM# 15
;                 4000 IF TERM# 14
;                 .
;                 .
;                 .
;                 0001 IF TERM# 00
;
;     RETURN: F2F1 = BIT POSITION OF TERM#
;
;     MODIFIED REGISTERS: PC'S,CH,CL,F0-2
;
; BITF0: CONVERT BINARY VALUE IN F0 TO BIT POSITION
;
;     MODIFIED REGISTERS: F0-1
;

BITTERM_: CALL      GETTERM_            ;F0 = TERM#

BITF0:    MOV       WORD PTR [R1],01H   ;INITIALIZE BIT POSITION TO VALUE FOR 01
          JMP       BITTERM2

BITTERM1: SHL       WORD PTR [R1],1     ;LEFT SHIFT BIT POSITION 1 PLACE
BITTERM2: DEC       R0                  ;DECREMENT BINARY COUNT
          TEST      R0,0FH
          JNZ       BITTERM1            ;BRANCH IF VALUE NOT 0
          RET

; NAME2FR: MOVE NAME FROM MEMORY TO FILE REGISTERS
;
;     ENTRY: F0 = LENGTH OF NAME (1-8)
;            PC'S = ADDRESS OF NAME
;
;     RETURN: F7-0 = NAME (PADDED WITH TRAILING SPACES)
;
;     MODIFIED REGISTERS: F0-7,PC'S,CH,CL
;

NAME2FR:  MOV       CH,BYTE PTR [EDX]   ;GET 1ST BYTE OF NAME
          INC       EDX
          MOV       R7,CH               ;F7 = 1ST BYTE OF NAME
          MOV       CH,BYTE PTR [EDX]
          INC       EDX
          DEC       A0                  ;DECREMENT LENGTH OF NAME
          JZ        NAME2FR6            ;BRANCH IF NO MORE NAME BYTES
          MOV       R6,CH
          MOV       CH,BYTE PTR [EDX]   ;F6 = 2ND BYTE OF NAME
          INC       EDX
          DEC       A0
          JZ        NAME2FR5
          MOV       R5,CH
          MOV       CH,BYTE PTR [EDX]
          INC       EDX
          DEC       A0
          JZ        NAME2FR4
          MOV       R4,CH
          MOV       CH,BYTE PTR [EDX]
          INC       EDX
          DEC       A0
          JZ        NAME2FR3
          MOV       R3,CH
          MOV       CH,BYTE PTR [EDX]
          INC       EDX
          DEC       A0
          JZ        NAME2FR2
          MOV       R2,CH
          MOV       CH,BYTE PTR [EDX]
          INC       EDX
          DEC       A0
          JZ        NAME2FR1
          MOV       R1,CH
          MOV       CH,BYTE PTR [EDX]
          INC       EDX
          DEC       A0
          JZ        NAME2FR0
          MOV       R0,CH
          MOV       CH,BYTE PTR [EDX]
          INC       EDX
          RET

NAME2FR6: MOV       R6,20H              ;PAD WITH TRAILING SPACES
NAME2FR5: MOV       R5,20H
NAME2FR4: MOV       R4,20H
NAME2FR3: MOV       R3,20H
NAME2FR2: MOV       R2,20H
NAME2FR1: MOV       R1,20H
NAME2FR0: MOV       R0,20H

          RET

; CCNLPART: LOOK FOR 'PART' IN TEXT. ERROR IF NOT FOUND.
;
;     ENTRY: TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;     MODIFIED REGISTERS;
;

CCNLPART: MOV       EDX,OFFSET _PART
          CALL      CCNL                   ;LOOK FOR 'PART'
          TEST      R0,0FH
          JNZ       ERROR_19               ;ERROR IF NOT FOUND
          RET

; newtoold: convert new floating point format to old floating format
; entry : edx point to 8 byte new floating point
; return: edx point to 8 byte old floating point

newtoold: push      textptr                  ;save textptr
          mov       ax,word ptr [edx]        ;get exponent
          add       ax,13
          mov       word ptr [a8],ax
          mov       ebp,dword ptr [edx+2]    ;get mantissa 1
          mov       bx,word ptr [edx+6]      ;get mantissa 2
          mov       edi,offset work_buffer   ;get output buffer pointer
          mov       dword ptr [edi],0        ;set sign to plus and mantissa to 0
          mov       dword ptr [edi+4],0
          test      bx,8000h                 ;test mantissa sign
          jz        new1                     ;branch if plus
          or        byte ptr [edi],10h       ;set mantissa sign to minus
          not       bx                       ;complement mantissa
          neg       ebp
          jnz       new1
          inc       bx
new1:     test      bx,0ffffh                ;check zero mantissa
          jnz       new2
          test      ebp,0ffffffffh
          jz        new_zero
new2:     mov       esi,offset _abcov+6
new3:     lodsd
          mov       ecx,eax
          lodsw
          mov       dl,2fh
new4:     inc       dl
          sub       ebp,ecx
          sbb       bx,ax
          jnb       new4
          add       ebp,ecx
          adc       bx,ax
          cmp       dl,30h
          jne       new5
          dec       word ptr [a8]
          jmp       new3
new5:     and       dl,0fh
          or        byte ptr [edi+1],dl
          test      word ptr [a8],8000h      ;check exponent sign
          jz        new6
          neg       word ptr [a8]            ;neg exponent value
          or        byte ptr [edi],80h       ;set exponent sign
new6:     mov       ax,word ptr [a8]
          test      ax,0ffffh
          jz        new10
          xor       dx,dx
          mov       cx,10
          div       cx
          or        byte ptr [edi],dl        ;set exponent bcd high
          shl       al,4
          or        byte ptr [edi+1],al      ;set exponent bcd low

new10:    add       edi,2
          xor       dh,dh
new7:     lodsd
          test      eax,0ffffffffh
          jz        newend
          mov       ecx,eax
          lodsw
          mov       dl,2fh
new8:     inc       dl
          sub       ebp,ecx
          sbb       bx,ax
          jnb       new8
          add       ebp,ecx
          adc       bx,ax
          test      dh,0ffh
          jz        new9
          and       dl,0fh
          xor       dh,dh
          or        byte ptr [edi],dl
          inc       edi
          jmp       new7

new9:     shl       dl,4
          mov       dh,1
          or        byte ptr [edi],dl
          jmp       new7

newend:   mov       edx,offset work_buffer
          pop       textptr
          ret

new_zero: mov       edx,edi
          pop       textptr
          ret

oldtonew: push      textptr                  ;save textptr
          mov       edi,offset work_buffer   ;set output buffer pointer
          mov       ebp,offset mathpage      ;set input buffer pointer
          mov       cx,word ptr [ebp]
          xchg      ch,cl
          test      cl,0fh                   ;check zero number
          jz        old_zero
          xor       ax,ax
          test      cl,0f0h
          jz        old1
          mov       al,cl
          shr       al,4
          mov       bl,10
          mul       bl
old1:     xor       bx,bx
          mov       bl,ch
          and       bl,0fh
          add       ax,bx
          test      ch,80h                   ;check sign
          jz        old2
          neg       ax
old2:     mov       bx,12
          mov       cx,word ptr [ebp+6]
          test      cx,0ffffh
          jnz       old4
          sub       bx,4
          jmp       old5
old4:     test      cx,0f00h
          jnz       old3
          dec       bx
          test      cx,0f000h
          jnz       old3
          dec       bx
          test      cx,000fh
          jnz       old3
          dec       bx
          test      cx,00f0h
          jnz       old3
          dec       bx
old5:     mov       cx,word ptr [ebp+4]
          test      cx,0ffffh
          jnz       old6
          sub       bx,4
          jmp       old7
old6:     test      cx,0f00h
          jnz       old3
          dec       bx
          test      cx,0f000h
          jnz       old3
          dec       bx
          test      cx,000fh
          jnz       old3
          dec       bx
          test      cx,00f0h
          jnz       old3
          dec       bx
old7:     mov       cx,word ptr [ebp+2]
          test      cx,0ffffh
          jnz       old8
          sub       bx,4
          jmp       old3
old8:     test      cx,0f00h
          jnz       old3
          dec       bx
          test      cx,0f000h
          jnz       old3
          dec       bx
          test      cx,000fh
          jnz       old3
          dec       bx
          test      cx,00f0h
          jnz       old3
          dec       bx

old3:     sub       ax,bx
          mov       word ptr [edi],ax
          mov       word ptr [edi+2],0
          mov       dword ptr [edi+4],0
          xor       eax,eax
          mov       ax,15
          inc       bx
          sub       ax,bx
          mov       cx,ax
          shl       ax,2
          add       ax,cx
          add       ax,cx

          mov       esi,offset _abcov
          add       esi,eax

          inc       ebp
          mov       cl,1

old15:    mov       ch,byte ptr [ebp]
          cmp       cl,1
          je        old11
          shr       ch,4
          mov       cl,1
          jmp       old12
old11:    and       ch,0fh
          inc       ebp
          xor       cl,cl
old12:    cmp       ch,0
          jnz       old14
          add       esi,6
          dec       bx
          jmp       old15
old14:    mov       aux20,0
          mov       word ptr [a8],0
old13:    lodsd
          mov       edx,eax
          lodsw
OLD17:    add       aux20,edx
          adc       word ptr [a8],ax
          dec       ch
          jnz       old17
          mov       edx,aux20
          mov       ax,word ptr [a8]
          add       dword ptr [edi+2],edx
          adc       word ptr [edi+6],ax
          dec       bx
          jnz       old15
          mov       ebp,offset mathpage
          test      byte ptr [ebp],10h
          jz        old16
          not       word ptr [edi+6]
          neg       dword ptr [edi+2]
          jnz       old16
          inc       word ptr [edi+6]
old16:    mov       edx,edi
          pop       textptr
          ret

old_zero: mov       edx,offset work_buffer
          mov       dword ptr [edx],0
          mov       dword ptr [edx+4],0
          pop       textptr
          ret

ttt       db        0

CODE      ENDS
          END
