;===========================================================================
;
;
;
;===========================================================================
;           This module was originally M31
;===========================================================================
            .XLIST
            INCLUDE PREAMBLE
            .LIST

PUBLIC    MATMOVE_S
PUBLIC    MERGE_S
PUBLIC    SORT_S
PUBLIC    SEARCH_S

EXTRN     ALPHINL :NEAR
EXTRN     RELOP   :NEAR
EXTRN     ALPHA1  :NEAR
EXTRN     CCL     :NEAR
EXTRN     EXPR    :NEAR
EXTRN     ERR34   :NEAR
EXTRN     SRCHALPH:NEAR

EXTRN     $BINFLT :NEAR
EXTRN     ARAYELP :NEAR
EXTRN     ARAYELRP:NEAR
EXTRN     ARECEVAL:NEAR
EXTRN     ARRAY   :NEAR
EXTRN     AVARD   :NEAR
EXTRN     BREAKPTQ:NEAR
EXTRN     CCLTO   :NEAR
EXTRN     COMPARE :NEAR
EXTRN     ENDSTMTZ:NEAR
EXTRN     ERROR   :NEAR
EXTRN     ERROR_13:NEAR
EXTRN     ERROR_28:NEAR
EXTRN     ERROR_29:NEAR
EXTRN     ERROR_50:NEAR
EXTRN     ERROR_51:NEAR
EXTRN     ERROR_55:NEAR
EXTRN     ERR_02  :NEAR
EXTRN     FIELD   :NEAR
EXTRN     FLTBIN  :NEAR
EXTRN     MATHPAGE:NEAR
EXTRN     NEWTOOLD:NEAR
EXTRN     OLDTONEW:NEAR
EXTRN     REM     :NEAR
EXTRN     VAR     :NEAR
EXTRN     VARNAME :NEAR

EXTRN     @@VSEND :DWORD

;------------------------------------------------------------------------------;
;                                                                              ;
;        MODULE    MOD-30 MAT SORT                                             ;
;                                                                              ;
;------------------------------------------------------------------------------;


;        MAT SORT  -  ARRANGE AN ARRAY OF N ELEMENTS
;                       INTO ASCENDING ORDER BY MEANS
;                         OF A SHELL SORT.
;                     1) ON EACH PASS THRU THE DATA,
;                        'D' LISTS OF ELEMENTS ARE SORTED
;                        BY BUBBLING.  ELEMENTS IN EACH
;                        LIST ARE A DISTANCE 'D' APART.
;                             D(1)=(N+1)/2
;                             D(I)=(D(I-1)+1)/2
;                     2) ALL LISTS ARE BUBBLED UNTIL
;                        EVERY LIST IS SORTED.
;                     3) THE SHELL SORT CONTINUES UNTIL
;                        THE FINAL LIST (D=1) IS SORTED.
;
;
;               AN EXAMPLE OF BUBBLE SORTING:
;
;               3          1          1          1
;               1          3          3          2
;               4   --->   4   --->   2   --->   3
;               5          2          4          4
;               2          5          5          5
;
;               ON EACH BUBBLE, NOTE THAT THE FIRST
;               ELEMENT IS COMPARED TO THE SECOND AND
;               THE TWO ELEMENTS ARE EXCHANGED IF
;               NECESSARY TO PUT THEM IN THE CORRECT
;               ORDER.  THEN THE SECOND ELEMENT (WHICH
;               MAY HAVE STARTED OUT BEING THE FIRST
;               ELEMENT) IS COMPARED TO THE THIRD, AND
;               SO ON.  THIS IS CONTINUED UNTIL THE LAST
;               TWO ELEMENTS ARE COMPARED, THE WHOLE
;               PROCESS REQUIRING N-1 COMPARISONS AND
;               POSSIBLE EXCHANGES.  THIS ENTIRE ROUTINE
;               IS REPEATED UNTIL THE ELEMENTS ARE ALL
;               SORTED.  NOTE THAT ON EACH BUBBLE THRU
;               THE DATA, THE LARGEST ELEMENT 'SINKS' TO
;               THE BOTTOM SO THAT ONE LESS COMPARISON
;               CAN BE MADE ON EACH SUCCESSIVE PASS.
;
;               THE SHELL SORT USES BUBBLING TO SORT EACH
;               LIST OF DATA, BUT BY BREAKING THE DATA
;               INTO SUCCESSIVELY LARGER LISTS, ELEMENTS
;               WHICH ARE FAR OUT OF PLACE ARE SORTED
;               MORE QUICKLY.
;
;        GENERAL FORM:
;
;        *******************************************
;        *  MAT SORT (source) TO (work),(locator)  *
;        *******************************************
;
;               WHERE:  (source) = ARRAY TO BE SORTED
;                       (work) = WORK STRING FOR STORING
;                                 ELEMENT ADDRESSES
;                       (locator) = SUBSCRIPT ARRAY
;
;        MODIFIED REGISTERS:  ALL, AUX 0,1,4-6,8-F
;                      (AUX 0,1,2,4,5 IN ENTRY PHASE)
;
;        EXPR BETTER NOT USE: AUX 4,5,7,8,9,A,D,E
;

SORT_S:   CALL      ARRAY               ;SCAN SOURCE ARRAY
          TEST      R0,0FH
          JNZ       ERR_28S             ;ERROR IF NOT FOUND
          TEST      R5,1
          JZ        ERR_28S             ;ERROR IF NOT ALPHA
          TEST      SL,2
          JZ        SORT0               ;BRANCH IF NOT X PHASE
          MOV       AUX0A,EDX           ;SAVE SOURCE ADDRESS
          MOVZX     EDX,R5
          AND       DL,04H
          SHL       DL,2                ;PL = 00 IF VECTOR, 10 IF MATRIX
          MOV       AUX0F,EDX           ;SAVE FLAG VECTOR/MATRIX
          MOV       EDX,DWORD PTR [A2]
          MOV       AUX0D,EDX           ;SAVE NUMBER OF ELEMENTS
          MOV       EDX,DWORD PTR [A6]
          MOV       AUX0E,EDX           ;SAVE DIMENSIONS
          MOVZX     EDX,CH
          MOV       AUX09,EDX           ;SAVE ELEMENT LENGTH
          MOV       AUX08,EDX

SORT0:    CALL      FIELD               ;PROCESS A FIELD SPECIFICATION INTO F5F4
          CMP       A4,0
          JE        SORT1               ;BRANCH IF NOT SPECIFIED
          TEST      SL,2
          JZ        SORT1               ;BRANCH IF NOT X PHASE
          MOV       A6,0
          MOV       EDX,AUX0A           ;GET SOURCE ADDRESS
          MOV       EAX,DWORD PTR [A5]
          ADD       EDX,EAX
          MOV       AUX0A,EDX           ;SAVE ADDRESS AT START OF FIELD
          MOVZX     EDX,A4
          MOV       AUX09,EDX           ;SAVE FIELD LENGTH

SORT1:    CALL      CCLTO               ;MUST FIND "TO"
          CALL      AVARD               ;SCAN WORK STRING
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        SORT2               ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX0D           ;GET NUMBER OF SOURCE ELEMENTS
          MOV       DWORD PTR [A0],EDX
          SHL       DWORD PTR [A0],2    ;NO. * 2 = REQUIRED NO. OF WORK BYTES
          JC        ERR_51S             ;ERROR IF OVERFLOW
          MOV       EDX,VSPTR
          MOV       ECX,DWORD PTR [EDX]
          ADD       VSPTR,4             ;GET WORK STRING LENGTH
          CMP       ECX,DWORD PTR [A0]
          JB        ERR_51S             ;ERROR IF TOO SHORT

          MOV       EDX,VSPTR           ;GET WORK ADDRESS
          MOV       EDX,DWORD PTR [EDX]
          ADD       VSPTR,4
          MOV       AUX0B,EDX           ;SAVE WORK ADDRESS
SORT2:    CMP       BYTE PTR [TEXTPTR],','   ;MUST FIND A COMMA
          JNE       ERROR_13
          INC       TEXTPTR
          CALL      ARRAY
          TEST      R0,0FH
          JNZ       ERR_28S             ;ERROR IF NOT FOUND
          TEST      R5,1
          JZ        ERR_28S             ;ERROR IF NOT ALPHA
          TEST      SL,2
          JZ        SORTH               ;BRANCH IF NOT X PHASE
          MOV       AUX0C,EDX           ;SAVE LOCATOR ADDRESS
          CMP       CH,4
          JNE       ERR_50S             ;ERROR IF NOT 2 BYTE ELEMENTS

; <<<< SHOULD WE REQUIRE TWO BYTE ELEMENTS? >>>>>>
;

          MOV       EDX,AUX0D           ;GET REQUIRED NO. OF ELEMENTS
          CMP       DWORD PTR [A2],EDX
          JB        ERR_51S             ;ERROR IF LOCATOR SIZE TOO SMALL
          MOV       R0,1                ;SET SIZE FLAG DEFAULT =
          CMP       DWORD PTR [A2],EDX
          JBE       SORTX               ;BRANCH IF SIZE = SOURCE
          MOV       R0,0                ;SET SIZE FLAG >
SORTX:    MOV       EDX,AUX0F           ;GET FLAG VECTOR/MATRIX
          MOV       AL,DL
          OR        AL,R0
          MOV       K,AL                ;SET K = FLAGS
          MOV       EDX,AUX0E           ;GET SOURCE DIMENSIONS
          MOV       DWORD PTR [A4],EDX  ;SAVE DIMENSIONS
          MOV       EDX,AUX08           ;GET SOURCE ELEMENT LENGTH
          MOV       DWORD PTR [A6],EDX  ;SAVE IT


;        REGISTERS NOW SET UP:
;               AUX 9 - FIELD LENGTH (OR ELEMENT LENGTH)
;               AUX A - ADDRESS OF SOURCE ARRAY (+FIELD)
;               AUX B - ADDRESS OF WORK STRING
;               AUX C - ADDRESS OF LOCATOR ARRAY
;               AUX D - NO. OF ELEMENTS IN SOURCE ARRAY
;
;               F5,F4 - DIMENSIONS OF SOURCE ARRAY
;               F7,F6 - ELEMENT LENGTH (F7=00)
;               K (LOW) - SIZE FLAG SET TO ZER0 IF LOC
;                   ARRAY IS > SORT ARRAY BY AT LEAST 2
;               K (HIGH) - SORT ARRAY TYPE ( VECTOR = 0,
;                          MATRIX = 1 )
;
;
;         BEGIN BY PUTTING SEQUENTIAL SUBSCRIPTS IN THE
;            LOCATOR ARRAY AND THE CORRESPONDING ELEMENT
;               ADDRESSES IN THE WORK ARRAY.
;

          MOVZX     AX,K
          SHR       AX,4
          MOV       A1,AX               ;SET INITIAL SUBSCRIPT TO 0000 IF VECTOR,

; 0100 IF ARRAY.
;

          MOV       EDX,AUX0A           ;GET SORT ARRAY ADDRESS
          MOV       DWORD PTR [A2],EDX  ;SAVE IT
          MOV       EDX,AUX0B           ;GET WORK STRING ADDRESS
          MOV       AUX08,EDX           ;SAVE IT
          MOV       EDX,AUX0C           ;GET LOCATOR ARRAY ADDRESS
          JMP       SORTA2

SORTA0:   INC       A1                  ;INCREMENT ROW SUBSCRIPT
SORTA2:   MOV       A0,0                ;INITIALIZE COLUMN SUBSCRIPT
SORTA1:   INC       A0
SORTA:    TEST      K,0F0H
          JZ        SORTAX              ;VECTOR

          MOV       AX,A1
          xchg      ah,al
          MOV       WORD PTR [EDX],AX
          ADD       EDX,2               ;PUT SUBSCRIPT IN LOCATOR ARRAY
          MOV       AX,A0
          xchg      ah,al
          MOV       WORD PTR [EDX],AX
          XCHG      EDX,AUX08           ;SAVE LOCATOR ADDRESS
          ADD       AUX08,2
          JMP       SORTAY

SORTAX:   MOV       EAX,DWORD PTR [A0]
          MOV       ECX,EAX
          SHR       ECX,16
          xchg      ch,cl
          MOV       WORD PTR [EDX],CX
          ADD       EDX,2               ;PUT SUBSCRIPT IN LOCATOR ARRAY
          xchg      ah,al
          MOV       WORD PTR [EDX],AX
          XCHG      EDX,AUX08           ;SAVE LOCATOR ADDRESS
          ADD       AUX08,2
SORTAY:   MOV       ECX,DWORD PTR [A2]
          MOV       DWORD PTR [EDX],ECX ;PUT ELEMENT ADDRESS IN WORK STRING
          ADD       EDX,4
          XCHG      EDX,AUX08           ;SAVE WORK ADDRESS
          MOV       EAX,DWORD PTR [A6]
          ADD       DWORD PTR [A2],EAX  ;ADDRESS OF NEXT SORT ELEMENT
          TEST      K,0F0H
          JZ        SORTBA              ;BRANCH IF VECTOR
          MOV       AX,A0
          CMP       AX,A4
          JB        SORTA1              ;CHECK COLUMN SUBSCRIPT
          MOV       AX,A1
          CMP       AX,A5
          JB        SORTA0              ;CHECK ROW SUBSCRIPT

SORTAA:   TEST      K,0FH
          JNZ       SORTB               ;BRANCH IF LOCATOR SIZE=SORT SIZE
          MOV       DWORD PTR [EDX],0   ;WRITE '0000' INTO LOCATOR
          ADD       EDX,4
SORTB:    MOV       EDX,AUX09           ;GET FIELD LENGTH
          MOV       A7,DX               ;SAVE IT
          MOV       DWORD PTR [A0],1
          MOV       EDX,AUX0D           ;GET NUMBER OF ELEMENTS
          CMP       DWORD PTR [A0],EDX
          JB        SORTC               ;BRANCH IF MORE THAN ONE ELEMENT
          JMP       ENDSTMTZ

SORTBA:   INC       DWORD PTR [A0]      ;INCREMENT SUBSCRIPT
          MOV       EAX,DWORD PTR [A0]
          CMP       EAX,DWORD PTR [A4]
          JBE       SORTA               ;BRANCH IF NOT DONE
          JMP       SORTAA

;
;        BEGINNING OF NEW PASS.  COMPUTE NEW DISTANCE 'D'
;             D(1)=(N+1)/2    D(I)=(D(I-1)+1)/2
;

SORTC:    INC       EDX                  ;ADD 1, BEFORE DIVIDING BY TWO
          SHR       EDX,1                ;PHPL= (D+1)/2, NEW DISTANCE
          MOV       AUX09,EDX            ;SAVE DISTANCE 'D'
          MOV       DWORD PTR [A2],EDX
          SHL       DWORD PTR [A2],2     ;DOUBLE TO GET ADDRESS INCREMENT (F3,F2)
          MOV       EDX,AUX09            ;GET CURRENT DISTANCE 'D'
          MOV       DWORD PTR [A0],EDX   ;SUBTRACT DISTANCE BETWEEN ELEMENTS FROM
          NOT       DWORD PTR [A0]

;   THE TOTAL NUMBER OF ELEMENTS TO FIND THE NUMBER
;
          MOV       EDX,AUX0D           ;OF COMPARISONS TO BE MADE
          STC
          ADC       EDX,DWORD PTR [A0]
          MOV       AUX0F,EDX           ;SAVE COMPARISON COUNT IN AUX 7
          MOV       EDX,DWORD PTR [A0]    ;SAVE NEGATIVE OF DISTANCE ( -D )
          MOV       AUX06,EDX

;
;        NOW BUBBLE 'D' LISTS OF ELEMENTS EACH A DISTANCE
;           'D' APART IN THE ORIGINAL ARRAY.
;

SORTD:    MOV       EDX,AUX0B           ;POINT TO START OF WORK STRING
          MOV       AUX0E,EDX           ;SAVE IT
          MOV       EDX,AUX0C           ;POINT TO START OF LOCATOR ARRAY
          MOV       AUX08,EDX           ;SAVE IT
          MOV       DWORD PTR [A5],0    ;CLEAR COMPARISON COUNT
          MOV       R4,0                ;CLEAR EXCHANGE FLAG

;        DO ONE BUBBLE OF EACH LIST AND THEN CHECK IF
;          ALL LISTS ARE SORTED (EXCHANGE FLAG F4 = 0)
;

SORTE:    MOV       EDX,AUX0E           ;POINT TO WORK STRING
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX             ;POINT TO ELEMENT A DISTANCE 'D' AWAY
          MOV       EDX,DWORD PTR [EDX] ;READ
                                        ;ADDRESS OF ELEMENT 2 IN PH,PL
          MOV       AUX00,EDX           ;SAVE ADDRESS
          MOVZX     EAX,A7
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,AUX0E           ;POINT TO WORK STRING
          MOV       EDX,DWORD PTR [EDX] ;GET ADDRESS OF ELEMENT 1
          CALL      COMPARE             ;CHECK IF ELEMENT 1 < ELEMENT 2
          CMP       R0,1
          JNE       SORTF               ;BRANCH IF IT IS

;        THE FOLLOWING CODE EXCHANGES THE VALUES IN THE WORK
;          BUFFER AND THOSE IN THE LOCATOR ARRAY THAT ARE
;          A DISTANCE 'D' APART (GIVEN BY THE INCREMENT IN
;          F3,F2).
;

          MOV       EDX,AUX0E           ;POINT TO WORK STRING
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX             ;FIND SECOND ADDRESS
          MOV       CX,WORD PTR [EDX]
          XCHG      EDX,AUX0E           ;READ SECOND VALUE
          MOV       AX,CX
          MOV       CX,WORD PTR [EDX]   ;READ FIRST VALUE
          MOV       WORD PTR [EDX],AX
          XCHG      EDX,AUX0E           ;WRITE SECOND INTO FIRST
          ADD       AUX0E,2
          MOV       WORD PTR [EDX],CX
          ADD       EDX,2               ;WRITE FIRST INTO SECOND
          MOV       CX,WORD PTR [EDX]
          XCHG      EDX,AUX0E           ;READ SECOND VALUE
          MOV       AX,CX
          MOV       CX,WORD PTR [EDX]   ;READ FIRST VALUE
          MOV       WORD PTR [EDX],AX
          XCHG      EDX,AUX0E           ;WRITE 2ND INTO 1ST/ RESTORE WORK PTR
          SUB       AUX0E,2
          MOV       WORD PTR [EDX],CX   ;WRITE FIRST INTO SECOND

          MOV       EDX,AUX08           ;POINT TO LOCATOR ARRAY
          MOV       EAX,DWORD PTR [A2]
          ADD       EDX,EAX             ;FIND SECOND ADDRESS
          MOV       CX,WORD PTR [EDX]
          XCHG      EDX,AUX08           ;READ SECOND VALUE
          MOV       AX,CX
          MOV       CX,WORD PTR [EDX]   ;READ FIRST VALUE
          MOV       WORD PTR [EDX],AX
          XCHG      EDX,AUX08           ;WRITE SECOND INTO FIRST
          ADD       AUX08,2
          MOV       WORD PTR [EDX],CX
          ADD       EDX,2               ;WRITE FIRST INTO SECOND
          MOV       CX,WORD PTR [EDX]
          XCHG      EDX,AUX08           ;READ SECOND VALUE
          MOV       AX,CX
          MOV       CX,WORD PTR [EDX]   ;READ FIRST VALUE
          MOV       WORD PTR [EDX],AX
          XCHG      EDX,AUX08           ;WRITE 2ND INTO 1ST/ RESTORE LOCATOR PTR
          SUB       AUX08,2
          MOV       WORD PTR [EDX],CX   ;WRITE FIRST INTO SECOND
          MOV       R4,1                ;SET EXCHANGE FLAG

SORTF:    INC       DWORD PTR [A5]      ;INCREMENT COMPARISON COUNT
          MOV       EDX,AUX0F           ;GET TOTAL NUMBER OF COMPARISONS TO BE MADE
          CMP       DWORD PTR [A5],EDX
          JB        SORTY               ;BRANCH IF NOT DONE WITH BUBBLE
          JMP       SORTG               ;END OF BUBBLE
SORTY:    MOV       EDX,AUX0E           ;POINT TO WORK STRING
          MOV       AUX0E,EDX           ;INCREMENT TO NEXT LIST
          ADD       AUX0E,4
          MOV       EDX,AUX08           ;POINT TO LOCATOR ARRAY
          MOV       AUX08,EDX           ;INCREMENT TO NEXT LIST
          ADD       AUX08,4
          JMP       SORTE               ;CONTINUE WITH THE BUBBLE

SORTG:    TEST      R4,0FH
          JZ        SORTGA              ;BRANCH IF ALL LISTS ARE SORTED
          MOV       EDX,AUX06           ;NUMBER OF COMPARISONS FOR NEXT BUBBLE IS 1 LESS
          STC
          ADC       EDX,DWORD PTR [A5]  ;FOR EACH LIST OR (COMPARISONS - D)

; SINCE THERE ARE D LISTS.
;

          JNC       SORTGA              ;DONE IF RESULT IS <= 0
          CMP       EDX,0
          JNE       SORTI               ;BRANCH IF NOT ZERO
SORTGA:   MOV       DWORD PTR [A0],1
          MOV       EDX,AUX09           ;GET NUMBER OF LISTS = 'D'
          CMP       DWORD PTR [A0],EDX
          JB        SORTC               ;BRANCH IF MORE THAN ONE LIST
SORTH:    JMP       ENDSTMTZ            ;DONE WHEN ONLY 1 LIST ALL SORTED
SORTI:    test      byte ptr fs:[cpust],10h
          jz        sortd               ;BRANCH IF < 30MS. ELAPSED
          MOV       EDX,DWORD PTR [A2]    ;SAVE IN AN AUX FOR SWITCH PARTITIONS
          MOV       AUX0A,EDX
          MOV       EAX,EDX
          MOVZX     EDX,A7               ;SAVE IN AN AUX FOR SWITCH PARTITONS
          SHL       EDX,16
          MOV       DX,AX
          MOV       AUX0E,EDX
          MOV       R0,SAV06            ;SAVE AUX 06-1F (NEED AUX 6,9,A-F)
          CALL      BREAKPTQ            ;TRANSFER CONTROL TO NEXT USER
          MOV       EDX,AUX0A
          MOV       DWORD PTR [A2],EDX  ;RESTORE
          MOV       EDX,AUX0E
          MOV       EAX,EDX
          SHR       EAX,16
          MOV       A7,AX               ;RESTORE
          JMP       SORTD

;        MAT MOVE  -  THERE ARE FOUR GENERAL CASES IN
;                        MOVING DATA ELEMENTS FROM ONE
;                        ARRAY TO ANOTHER:
;                         1)  ALPHA --> ALPHA
;                         2)  NUMERIC --> ALPHA
;                         3)  ALPHA --> NUMERIC
;                         4)  NUMERIC --> NUMERIC
;
;                     CASE 2 IMPLIES CONVERSION FROM 2600
;                         INTERNAL FORMAT TO SORT FORMAT.
;                     CASE 3 IMPLIES CONVERSION FROM SORT
;                         FORMAT TO 2600 INTERNAL FORMAT.
;
;        GENERAL FORM:
;    ****************************************************
;    *  MAT MOVE (move array),(locator array element),n *
;    *                    TO (receiver array element)   *
;    ****************************************************
;
;          WHERE:  (move array) = ALPHA OR NUMERIC
;                  (loc array element) = ALPHA SUBSCRIPTS
;                         IN THE ARRAY BEGINNING WITH THE
;                         ELEMENT SPECIFIED (OR BEGIN IF
;                         NO SUBSCRIPT GIVEN).
;                  n = NUMERIC SCALAR VARIABLE GIVING
;                         MAXIMUM # OF ELEMENTS TO MOVE
;                         AND RECEIVING #  OF ELEMENTS
;                         ACTUALLY MOVED.
;                  (rec array element) = DESTINATION OF
;                         ELEMENTS FROM MOVE ARRAY
;                         BEGINNING WITH THE ELEMENT
;                         SPECIFIED (OR BEGIN IF NO
;                         SUBSCRIPT GIVEN).
;
;          THE LOCATOR ARRAY IS OPTIONAL.  HENCE, THE
;             MOVE MAY BE DIRECT OR INDIRECT.
;
;          REGISTERS MODIFIED:   ALL, AUX 0-E
;                              (AUX 0,1,5 IN ENTRY PHASE)
;
;          EXPR BETTER NOT USE:   AUX 3-E


;     MOVE ARRAY
;


MATMOVE_S:CALL      ARRAY               ;SCAN MOVE ARRAY VARIABLE
          TEST      R0,0FH
          JNZ       ERROR_26            ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        JLMOVE1             ;BRANCH IF NOT X PHASE
          MOV       AUX09,EDX           ;SAVE ADDRESS OF MOVE ARRAY
          TEST      R5,1
          JZ        MATMOVEX            ;BRANCH IF NUMERIC ARRAY
          DEC       EDX                 ;DEDUCT 1 IF ALPHA
MATMOVEX: MOV       AUX06,EDX           ;SAVE ADDRESS OF DIMS
          SUB       AUX06,4
          XOR       EDX,EDX             ;CLEAR STATUS FLAG
          TEST      R5,1
          JZ        MATMOVEY            ;BRANCH IF NUMERIC
          OR        EDX,12H             ;SET STATUS FLAG TO ALPHA
MATMOVEY: TEST      R5,4
          JZ        MATMOVEZ            ;BRANCH IF VECTOR
          OR        EDX,40H             ;SET STATUS FLAG TO MATRIX
MATMOVEZ: MOV       AUX07,EDX           ;SAVE IT
          MOV       EDX,DWORD PTR [A2]
          MOV       AUX08,EDX           ;SAVE ELEMENT COUNT OF MOVE ARRAY
          MOV       AUX0E,0             ;FIELD POSITION = 0 FOR NUMERIC ARRAYS
          MOV       EDX,8                ;NUMERIC LENGTH = 8
          TEST      R5,1
          JZ        MATMOVEA            ;BRANCH IF NUMERIC
          MOVZX     EDX,CH               ;SAVE ELEMENT LENGTH IF ALPHA
MATMOVEA: MOV       AUX0A,EDX
          MOV       AUX0B,EDX
JLMOVE1:  TEST      R5,1
          JZ        JLMOVE2             ;BRANCH IF NUMERIC
          CALL      FIELD               ;PROCESS A FIELD SPECIFICATION IF ALPHA
          TEST      SL,2
          JZ        JLMOVE2             ;BRANCH IF NOT X PHASE
          MOV       A6,0
          MOV       EDX,DWORD PTR [A5]
          MOV       AUX0E,EDX           ;TEMPORARY SAVE OF FIELD POSITION
          CMP       A4,0
          JE        JLMOVE2             ;BRANCH IF FIELD IS NOT SPECIFIED
          MOV       EDX,AUX09           ;GET MOVE ADDRESS
          MOV       EAX,DWORD PTR [A5]
          ADD       EDX,EAX             ;ADDRESS AT START OF FIELD
          MOV       AUX09,EDX           ;SAVE IT
          MOVZX     EDX,A4
          MOV       AUX0B,EDX           ;SAVE FIELD LENGTH
JLMOVE2:  CMP       BYTE PTR [TEXTPTR],','  ;LOOK FOR A COMMA
          JNE       JLMOVE4             ;BRANCH IF NOT FOUND
          INC       TEXTPTR

;   LOCATOR ARRAY(ELEMENT)
;
          MOV       AUX20,TEXTPTR
          CALL      VARNAME             ;LOOK FOR A LOCATOR VARIABLE
          TEST      R0,0FH
          JNZ       JLMOVE3A            ;BRANCH IF NOT FOUND
          TEST      R2,1
          JNZ       JLMOVE2A            ;BRANCH IF ALPHA
          MOV       TEXTPTR,AUX20       ;RETORE TEXT POINTER IF NOT ALPHA
          JMP       JLMOVE3A

JLMOVE2A: TEST      R2,2
          JZ        ERROR_26            ;ERROR IF NOT AN ARRAY
          CALL      ARAYELRP            ;PROCESS ARRAY ELEMENT
          TEST      SL,2
          JZ        JLMOVE3             ;BRANCH IF NOT X PHASE
          CMP       CH,4
          JNE       ERR_50S             ;ERROR IF ELEMENT LENGTH NOT = 2
          MOV       AUX09,EDX           ;SAVE LOCATOR ADDRESS
          MOV       EDX,AUX07           ;GET STATUS FLAG
          OR        EDX,4               ;SET STATUS TO INDIRECT
          MOV       AUX07,EDX           ;SAVE IT
          MOV       EAX,DWORD PTR [A6]    ;FIND # OF ELEMENTS
          MOV       DWORD PTR [A0],EAX    ;FIND # OF ELEMENTS
          TEST      K,40H
          JZ        MATMOVEB            ;BRANCH IF VECTOR
          MOVZX     EAX,A0              ;MULTIPLY DIMENSIONS FOR MATRIX
          MOVZX     EBX,A1
          PUSH      EDX
          MUL       EBX
          POP       EDX
          MOV       DWORD PTR [A0],EAX
MATMOVEB: MOV       EDX,AUX02           ;GET POSITION # OF ELEMENT-1
          MOV       EAX,DWORD PTR [A0]
          SUB       EAX,EDX
          MOV       EDX,EAX               ;# OF REMAINING ELEMENTS IN PH,PL
          MOV       AUX08,EDX           ;SAVE IT
          MOV       EDX,AUX0E           ;GET FIELD POSITION
          MOV       AUX0A,EDX           ;SAVE IT
JLMOVE3:  CMP       BYTE PTR [TEXTPTR],',' ;LOOK FOR A COMMA
          JNE       JLMOVE4             ;BRANCH IF NOT FOUND
          INC       TEXTPTR

;   COUNT VARIABLE
;

JLMOVE3A: MOV       K,3                 ;SET = SCALAR, RECEIVER
          CALL      VAR                 ;SCAN NUMERIC VAR, ERROR IF NOT FOUND
          TEST      SL,2
          JZ        JLMOVE4             ;BRANCH IF NOT X PHASE
          DEC       OSPTR               ;REMOVE OS ATOM
          MOV       EDX,VSPTR           ;GET ADDRESS OF VAR
          MOV       EDX,DWORD PTR [EDX]
          ADD       VSPTR,4
          MOV       AUX0F,EDX           ;SAVE IT
          MOV       EAX,VSPTR           ;VALUE TO VS
          SUB       EAX,8
          MOV       ECX,DWORD PTR [EDX]
          MOV       DWORD PTR [EAX],ECX
          MOV       ECX,DWORD PTR [EDX+4]
          MOV       DWORD PTR [EAX+4],ECX
          MOV       VSPTR,EAX
          CMP       EAX,DWORD PTR [@@VSEND]
          JB        ERR_02

          CALL      FLTBIN              ;CONVERT TO BINARY IN F1,F0
          TEST      R2,02H
          JNZ       MATMOVEC            ;BRANCH IF 0<VAR<65535
          JMP       REM                 ;N SPECIFIES ZERO ELEMENTS MOVED
MATMOVEC: MOV       EDX,AUX08           ;GET MAXIMUM MOVE COUNT
          CMP       DWORD PTR [A0],EDX
          JB        MATMOVED            ;BRANCH IF N<MAX
          JMP       MATMOVEE
MATMOVED: MOV       EDX,DWORD PTR [A0]
          MOV       AUX08,EDX           ;SAVE MAXIMUM MOVE COUNT
MATMOVEE: MOV       EDX,AUX07           ;GET STATUS FLAG
          OR        EDX,8               ;SET STATUS = N SPECIFIED
          MOV       AUX07,EDX           ;SAVE STATUS FLAG
JLMOVE4:  CALL      CCLTO               ;MUST FIND "TO"

;   RECEIVER ARRAY [(ELEMENT)]
;

          CALL      VARNAME             ;LOOK FOR RECEIVER VARIABLE
          TEST      R0,0FH
          JNZ       ERROR_26            ;ERROR IF NOT FOUND
          TEST      R2,2
          JZ        ERROR_26            ;ERROR IF NOT AN ARRAY
          TEST      R2,1
          JZ        MATMOVEF            ;BRANCH IF NOT ALPHA
          MOV       R0,1                ;SET FLAG TO ALPHA
MATMOVEF: MOVZX     EDX,R0
          MOV       AUX05,EDX           ;SAVE FLAG FOR ALPHA/NUMERIC
          CALL      ARAYELRP            ;PROCESS ARRAY VARIABLE
          TEST      SL,2
          JZ        JLMOVE5             ;BRANCH IF NOT X PHASE
          MOV       AUX0C,EDX           ;SAVE RECEIVER ADDRESS
          MOV       EDX,AUX07           ;GET STATUS FLAG
          TEST      K,10H
          JZ        MATMOVEG            ;BRANCH IF NOT ALPHA
          OR        EDX,1
          MOV       AUX07,EDX           ;SAVE STATUS FLAG
MATMOVEG: MOV       EAX,DWORD PTR [A6]
          MOV       DWORD PTR [A0],EAX  ;GET DIMENSIONS TO F1,F0
          TEST      K,40H
          JZ        MATMOVEH            ;BRANCH IF VECTOR
          MOVZX     EAX,A0              ;FIND TOTAL # OF ELEMENTS
          MOVZX     EBX,A1
          PUSH      EDX
          MUL       EBX
          POP       EDX
          MOV       DWORD PTR [A0],EAX
MATMOVEH: MOV       EDX,AUX02           ;GET POSITION # OF THIS ELEMENT-1
          SUB       DWORD PTR [A0],EDX  ;REMAINING # OF ELEMENTS IN F1,F0
          MOV       EDX,AUX08           ;GET MAXIMUM MOVE COUNT
          CMP       DWORD PTR [A0],EDX
          JB        MATMOVEI            ;BRANCH IF RECEIVER < MAX
          JMP       MATMOVEJ
MATMOVEI: MOV       EDX,DWORD PTR [A0]
          MOV       AUX08,EDX           ;SAVE MAXIMUM MOVE COUNT
MATMOVEJ: MOV       EDX,8               ;NUMERIC LENGTH = 8
          TEST      K,10H
          JZ        MATMOVEK            ;BRANCH IF NOT ALPHA
          MOVZX     EDX,CH              ;SAVE ELEMENT LENGTH IF ALPHA
MATMOVEK: MOV       AUX0E,EDX           ;SAVE ELEMENT LENGTH
          MOV       AUX0D,EDX
JLMOVE5:  MOV       EDX,AUX05           ;GET FLAG ALPHA/NUMERIC
          MOV       AL,DL
          AND       AL,0FH
          CMP       AL,1
          JNE       JLMOVE6             ;BRANCH IF NOT ALPHA
          MOV       EDX,AUX0D
          CALL      FIELD               ;PROCESS A FIELD SPECIFICATION IF ALPHA
          CMP       A4,0
          JE        JLMOVE6             ;BRANCH IF NOT SPECIFIED135
          TEST      SL,2
          JZ        JLMOVED             ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          MOVZX     EAX,A5
          ADD       EDX,EAX             ;ADDRESS AT START OF FIELD
          MOV       AUX0C,EDX           ;SAVE IT
          MOVZX     EDX,A4
          MOV       AUX0E,EDX           ;SAVE FIELD LENGTH
JLMOVE6:  TEST      SL,2
          JZ        JLMOVED             ;BRANCH IF NOT X PHASE
          MOV       EDX,AUX08           ;GET MOVE COUNT
          MOV       AUX08,EDX           ;DECREMENT FOR START OF LOOP
          DEC       AUX08
          XCHG      EDX,AUX07           ;SET # MOVED
          MOV       K,DL                ;SAVE STATUS FLAG
          TEST      K,1
          JNZ       JLMOVE6A
          TEST      K,2
          JZ        JLMOVE6A            ;BRANCH IF NOT ALPHA ==> NUMERIC
          MOV       EDX,AUX0B
          DEC       EDX
          CMP       EDX,0
          JNE       JLMOVE6A
          JMP       ERROR_51            ;ERROR IF LENGTH OF MOVE ELEMENT = 1

;        REGISTER USE IS NOW AS FOLLOWS:
;          AUX 6 - POINTS TO DIMS OF MOVE ARRAY
;          AUX 7 - # 0F ELEMENTS MOVED
;          AUX 8 - MAXIMUM # OF ELEMENTS TO BE MOVED
;                     = SMALLEST OF # OF MOVE ELEMENTS,
;                       RECEIVER ELEMENTS, OR N FOR
;                       DIRECT MOVES
;                     = SMALLEST OF # OF LOCATOR ELEMENTS
;                       OR RECEIVER ELEMENTS OR N FOR
;                       INDIRECT MOVES
;          AUX 9 - MOVE ARRAY ADDRESS IF DIRECT MOVE OR
;                  LOCATOR ARRAY ADDRESS IF INDIRECT
;          AUX A - ELEMENT LENGTH OF MOVE ARRAY IF DIRECT
;                  MOVE (USED TO POINT TO NEXT ELEMENT)
;                  OR FIELD POSITION WITHIN MOVE ELEMENT
;                  IF INDIRECT MOVE (MAY BE = 0 IF FIELD
;                  IS NOT SPECIFIED)
;          AUX B - FIELD LENGTH OF MOVE ARRAY OR ELEMENT
;                  LENGTH IF NO FIELD SPECIFIED
;          AUX C - ADDRESS OF RECIEVER ( FIELD POSITION
;                  IS ADDED IF IT IS SPECIFIED)
;          AUX D - ELEMENT LENGTH OF RECEIVER (USED AS
;                  INCREMENT TO NEXT ELEMENT POSITIONED
;                  TO FIELD IF SPECIFIED)
;          AUX E - FIELD LENGTH OF RECEIVER OR ELEMENT
;                  LENGTH IF NO FIELD SPECIFIED
;          AUX F - ADDRESS OF NUMERIC SCALAR VARIABLE (N)
;
;          STATUS FLAG (K):
;                  BITS 1,0 - MOVE FUNCTION          (03)
;                          11 = ALPHA --> ALPHA
;                          10 = ALPHA --> NUMERIC
;                          01 = NUMERIC --> ALPHA
;                          00 = NUMERIC --> NUMERIC
;                  BIT 2 - INDIRECT MOVE = 1         (04)
;                          DIRECT MOVE = 0
;                  BIT 3 - N SPECIFIED (1=YES,0=NO)  (08)
;                  BIT 4 - SAME AS BIT 1 (FOR ARAYELP)  (10)
;                  BIT 6 - MOVE ARRAY (VECTOR=0,MATRIX=1) (40)
;                  BITS 7,5 - NOT USED
;
;

JLMOVE6A: TEST      K,4                 ;BRANCH IF INDIRECT
          JNZ       MOVEAAI

;        'DIRECT MOVE'
;

          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          JMP       MOVEAAD

MOVEAA1:  MOV       EDX,AUX0A           ;GET ADDRESS INCREMENT
          MOV       EAX,EDX
          MOV       EDX,AUX09           ;INCREMENT VALUE ADDRESS
          ADD       EDX,EAX
          MOV       AUX09,EDX           ;SAVE IT
          MOV       EDX,AUX0D           ;GET ADDRESS INCREMENT
          MOV       EAX,EDX
          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          ADD       EDX,EAX             ;INCREMENT
          MOV       AUX0C,EDX           ;SAVE
MOVEAAD:  MOV       AUX00,EDX           ;SAVE RECEIVER ADDRESS
          MOV       EDX,AUX0B           ;GET VALUE LENGTH
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX0E           ;GET RECEIVER LENGTH
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX09           ;GET VALUE ADDRESS
          MOV       AL,K
          AND       AL,03H
          CMP       AL,3                ;BRANCH IF ALPHA --> ALPHA
          JE        MOVEAAD1
          CMP       AL,1                ;BRANCH IF NUMERIC --> ALPHA
          JE        MOVECON
          CMP       AL,2                ;BRANCH IF ALPHA --> NUMERIC
          JE        MOVEDCON
MOVEAAD1: CALL      ARECEVAL            ;MOVE VALUE TO RECEIVER
          MOV       EDX,AUX08           ;GET MOVE COUNT
          MOV       AUX08,EDX           ;DECREMENT AND SAVE
          DEC       AUX08
          MOV       EAX,EDX
          DEC       EDX
          TEST      EAX,0FFH
          JNZ       MOVEAA1
          CMP       EAX,0
          JE        MOVEDONE
          test      byte ptr fs:[cpust],10h
          jz        moveaa1             ;BRANCH IF < 30MS. ELAPSED
          MOV       DL,K
          MOV       AUX05,EDX           ;SAVE K
          MOV       R0,SAV03
          CALL      BREAKPTQ            ;BREAKPOINT
          MOV       EDX,AUX05
          MOV       K,DL                ;RESTORE K
          JMP       MOVEAA1

MOVEDONE: TEST      K,8                 ;BRANCH IF N IS NOT SPECIFIED
          JZ        JLMOVED
          MOV       EDX,AUX07           ;GET # OF ELEMENTS MOVED
          MOV       DWORD PTR [A0],EDX
MOVEIDD:  CALL      $BINFLT             ;CONVERT TO FLOATING
          MOV       EDX,AUX0F           ;GET SCALAR VARIABLE ADDRESS
          MOV       EAX,VSPTR           ;PUT # MOVED INTO SCALAR VARIABLE
          MOV       ECX,DWORD PTR [EAX]
          MOV       DWORD PTR [EDX],ECX
          MOV       ECX,DWORD PTR [EAX+4]
          MOV       DWORD PTR [EDX+4],ECX
          ADD       VSPTR,8
JLMOVED:  JMP       ENDSTMTZ

;       'INDIRECT MOVE'
;

MOVEAAI0: MOV       EDX,AUX0D           ;GET ADDRESS INCREMENT
          MOV       EAX,EDX
          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          ADD       EDX,EAX
          MOV       AUX0C,EDX           ;SAVE IT
MOVEAAI:  MOV       EDX,AUX09           ;GET LOCATOR ADDRESS
          test      k,40h
          jz        moveaax
          MOV       CX,WORD PTR [EDX]   ;READ SUBSCRIPTS AND PUT IN F1,F0
          xchg      ch,cl
          ADD       EDX,2
          MOV       A1,CX
          MOV       CX,WORD PTR [EDX]
          xchg      ch,cl
          MOV       AUX09,EDX           ;INCREMENT LOCATOR ADDRESS
          ADD       AUX09,2
          MOV       A0,CX
          CMP       A0,0
          JNE       MOVEXY
          CMP       A1,0                ;DONE IF SUBSCRIPT IS '0000'
          JE        MOVEID
          jmp       movexy
moveaax:
          MOV       AX,WORD PTR [EDX]   ;READ SUBSCRIPTS AND PUT IN F1,F0
          xchg      ah,al
          ADD       EDX,2
          SHL       EAX,16
          MOV       AX,WORD PTR [EDX]
          xchg      ah,al
          MOV       AUX09,EDX           ;INCREMENT LOCATOR ADDRESS
          ADD       AUX09,2
          MOV       DWORD PTR [A0],EAX
          CMP       A0,0
          JNE       MOVEXY
          CMP       A1,0                ;DONE IF SUBSCRIPT IS '0000'
          JE        MOVEID
MOVEXY:   MOV       EDX,AUX06           ;PC'S POINT TO DIMS OF MOVE ARRAY
          CALL      ARAYELP             ;DETERMINE ADDRESS
          MOV       EBX,AUX0A           ;GET FIELD POSITION
          MOV       EAX,EBX
          ADD       EAX,EDX             ;ADD TO FIND VALUE ADDRESS
          MOV       AUX01,EAX           ;SAVE IT
          MOV       EDX,AUX0C           ;GET RECEIVER ADDRESS
          MOV       AUX00,EDX
          MOV       EDX,AUX0B           ;GET VALUE LENGTH
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX0E           ;GET RECEIVER LENGTH
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX01           ;GET VALUE ADDRESS
          MOV       AL,K
          AND       AL,03H
          CMP       AL,3                ;BRANCH IF ALPHA --> ALPHA
          JE        MOVEAAI1
          CMP       AL,1                ;BRANCH IF NUMERIC --> ALPHA
          JE        MOVECON
          CMP       AL,2                ;BRANCH IF ALPHA --> NUMERIC
          JE        MOVEDCON
MOVEAAI1: CALL      ARECEVAL            ;MOVE VALUE TO RECEIVER
          MOV       EDX,AUX08           ;GET MOVE COUNT
          MOV       AUX08,EDX           ;DECREMENT AND SAVE
          DEC       AUX08
          MOV       EAX,EDX
          DEC       EDX
          TEST      EAX,0FFH
          JNZ       MOVEAAI0
          CMP       EAX,0                ;BRANCH IF DONE
          JE        MOVEDONE
          test      byte ptr fs:[cpust],10h
          jz        moveaai0            ;BRANCH IF < 30MS. ELAPSED
          MOV       DL,K
          MOV       AUX05,EDX           ;SAVE K
          MOV       R0,SAV03
          CALL      BREAKPTQ            ;BREAKPOINT
          MOV       EDX,AUX05
          MOV       K,DL                ;RESTORE K
          JMP       MOVEAAI0

MOVEID:   TEST      K,8                 ;BRANCH IF N IS NOT SPECIFIED
          JZ        JLMOVED
          MOV       EDX,AUX07           ;GET # OF ELEMENTS EXPECTED WOULD BE MOVED
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX08           ;GET # OF ELEMENTS NOT YET MOVED FROM EXPECTED
          SUB       DWORD PTR [A0],EDX    ;CALCULATES # ACTUALLY MOVED
          DEC       DWORD PTR [A0]
          JMP       MOVEIDD

;        CONVERSION ROUTINES:
;
;        THIS SECTION 'CONVERTS' DATA ELEMENTS FROM
;           2600 INTERNAL FORMAT TO SORT FORMAT.
;
;
;        SET UP F7,F6,F5 SUCH THAT:
;          F7 (HIGH,LOW) = SIGN DIGIT, UNUSED
;          F6 (HIGH,LOW) = EXP (HIGH), EXP (LOW)
;          F5 (HIGH,LOW) = ZERO      , 1ST MANTISSA DIGIT
;

MOVECON:  MOV       R8,0
          CALL      NEWTOOLD            ;CONVERT NEW FORMAT TO OLD FORMAT
          MOV       CX,WORD PTR [EDX]   ;READ VALUE AND PUT INTO SCRATCH AREA
          XCHG      CH,CL
          MOV       AUX01,EDX           ;READ VALUE AND PUT INTO SCRATCH AREA
          ADD       AUX01,2
          MOV       R7,CH
          MOV       R5,CL
          AND       R5,0FH
          AND       CH,0FH
          AND       CL,0F0H
          OR        CL,CH
          MOV       R6,CL
          MOV       EDX,OFFSET MATHPAGE+2       ;POINT TO SCRATCH AREA
          XCHG      EDX,AUX01           ;VALUE ADDRESS TO PH,PL
          MOV       CX,WORD PTR [EDX]   ;READ DIGITS 2-5
          XCHG      CH,CL
          XCHG      EDX,AUX01
          ADD       AUX01,2
          MOV       BYTE PTR [EDX],CH
          INC       EDX
          MOV       BYTE PTR [EDX],CL
          XCHG      EDX,AUX01
          INC       AUX01
          MOV       CX,WORD PTR [EDX]   ;READ DIGITS 6-9
          XCHG      CH,CL
          XCHG      EDX,AUX01
          ADD       AUX01,2
          MOV       BYTE PTR [EDX],CH
          INC       EDX
          MOV       BYTE PTR [EDX],CL
          XCHG      EDX,AUX01
          INC       AUX01
          MOV       CX,WORD PTR [EDX]   ;READ DIGITS 10-13
          XCHG      CH,CL
          XCHG      EDX,AUX01
          ADD       AUX01,2
          MOV       BYTE PTR [EDX],CH   ;TURN OFF CARRY
          INC       EDX
          MOV       BYTE PTR [EDX],CL
          XCHG      EDX,AUX01
          INC       AUX01
          TEST      R5,0FH
          JNZ       JLMOVE7A            ;BRANCH IF NOT ZERO
          MOV       R7,80H              ;SET UP ZERO IN SORT FORMAT
JLMOVE7:  MOV       EDX,OFFSET MATHPAGE ;POINT TO SCRATCH AREA
          MOV       AL,R7               ;WRITE SIGN, EXPONENT AND FIRST DIGIT
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          MOV       AL,R6
          MOV       BYTE PTR [EDX],AL
          DEC       EDX
          CMP       R8,0
          JE        JLMOVE99
          CALL      OLDTONEW
JLMOVE99: MOV       R8,0
          TEST      K,4
          JNZ       MOVEAAI1            ;BRANCH IF INDIRECT
          JMP       MOVEAAD1


;        FIRST FIX SIGN DIGIT ( F7(HIGH) ):
;               0 --> 9
;               8 --> 8
;               9 --> 1
;               1 --> 0
;

JLMOVE7A: MOV       AL,R7
          AND       AL,0F0H
          CMP       AL,80H
          JE        JLMOVE8
          MOV       AL,R7               ;CARRY WAS TURNED OFF ABOVE
          ADD       AL,90H
          DAA
          MOV       R7,AL
          AND       AL,0F0H
          CMP       AL,80H
          JNE       JLMOVEX
          MOV       R7,10H
JLMOVEX:  TEST      R7,0F0H
          JNZ       JLMOVE9

;        TAKE 9'S COMPLEMENT OF EXPONENT IF MANTISSA
;           AND EXPONENT SIGNS DIFFER
;

JLMOVE8:  STC
          MOV       AL,0
          SBB       AL,R6
          DAS
          MOV       R6,AL
          INC       EDX
JLMOVE9:  AND       R7,0F0H
          MOV       AL,R6
          SHR       AL,4
          OR        R7,AL               ;PACK SIGN, EXPONENT, AND FIRST DIGIT INT
          SHL       R6,4
          MOV       AL,R5
          AND       AL,0FH
          OR        R6,AL               ;TWO BYTES FOR WRITE TO SCRATCH AREA.
          TEST      R7,80H
          JNZ       JLMOVE7             ;BRANCH IF MANTISSA IS POSITIVE

;
;        IF MANTISSA IS NEGATIVE, TAKE 9'S COMPLEMENT
;           OF ALL MANTISSA DIGITS.
;

MOVECOMP: STC                           ;COMPLEMENTS FIRST DIGIT
          MOV       AL,0                ;COMPLEMENTS FIRST DIGIT
          SBB       AL,R5
          DAS
          MOV       R5,AL
          INC       EDX
          AND       R6,0F0H             ;PACK INTO F6 FOR WRITE TO SCRATCH AREA
          MOV       AL,R5
          AND       AL,0FH
          OR        R6,AL
          MOV       EDX,OFFSET MATHPAGE+2        ;COMPLEMENT DIGITS 2-5
          MOV       CX,WORD PTR [EDX]
          XCHG      CH,CL
          STC
          MOV       AL,0
          SBB       AL,CH
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          STC
          MOV       AL,0
          SBB       AL,CL
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          MOV       CX,WORD PTR [EDX]   ;COMPLEMENT DIGITS 6-9
          XCHG      CH,CL
          STC
          MOV       AL,0
          SBB       AL,CH
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          STC
          MOV       AL,0
          SBB       AL,CL
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          MOV       CX,WORD PTR [EDX]   ;COMPLEMENT DIGITS 10-13
          XCHG      CH,CL
          STC
          MOV       AL,0
          SBB       AL,CH
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          STC
          MOV       AL,0
          SBB       AL,CL
          DAS
          MOV       BYTE PTR [EDX],AL
          INC       EDX
          JMP       JLMOVE7


;        THIS SECTION 'UNCONVERTS' DATA FROM SORT FORMAT
;           TO 2600 INTERNAL FORMAT.
;
;        ALPHA ELEMENTS MAY BE DEFINED TO ANY LENGTH, BUT
;           ONLY THE FIRST EIGHT BYTES WILL BE USED IN
;           'UNCONVERT'.  IF THE VALUE LENGTH IS LESS
;           THAN 8 BYTES, THE VALUE WILL BE PADDED WITH
;           ZERO'S IF THE VALUE IS POSITIVE OR NINE'S IF
;           THE VALUE IS NEGATIVE BEFORE UNCONVERSION.
;
;

MOVEDCON: MOV       R8,1
          MOV       AUX01,EDX           ;SAVE VALUE ADDRESS
          MOV       A5,08H              ;MAXIMUM OF 8 BYTES WILL BE MOVED
          MOV       AX,A0
          CMP       A5,AX
          JBE       MOVEDCONX           ;BRANCH IF LENGTH IS 8 OR MORE BYTES
          MOV       A5,AX               ;VALUE LENGTH IS LESS THAN 8
MOVEDCONX:MOV       A0,08H              ;LENGTH OF SCRATCH AREA
          MOV       EDX,OFFSET MATHPAGE ;POINTS TO SCRATCH AREA
          XCHG      EDX,AUX01           ;POINTERS READY FOR FIRST READ
          ADD       A5,DX               ;END CONDITION FOR READ/WRITE
MOVE10A:  MOV       CH,BYTE PTR [EDX]
          XCHG      EDX,AUX01
          INC       AUX01
          MOV       BYTE PTR [EDX],CH
          XCHG      EDX,AUX01
          INC       AUX01
          CMP       A5,DX
          JNE       MOVE10A             ;BRANCH IF NOT DONE
          MOV       EDX,OFFSET MATHPAGE ;POINT TO SIGN,EXPONENT, AND FIRST DIGIT
          MOV       CX,WORD PTR [EDX]
          XCHG      CH,CL
          MOV       R7,CH               ;SETUP F7F6F5 AS SHOWN ABOVE FOR'CONVERT'
          MOV       R5,CL
          AND       R5,0FH
          AND       CH,0FH
          AND       CL,0F0H
          OR        CL,CH
          MOV       R6,CL
          MOV       EDX,AUX01           ;GET NEXT POSITION IN SCRATCH AREA
          TEST      R7,80H
          JZ        MOVE10B             ;BRANCH IF NUMBER IS NEGATIVE
          JMP       MOVEDCONY
MOVEXX:   MOV       BYTE PTR [EDX],0    ;PAD SCRATCH AREA WITH ZERO'S
          INC       EDX
MOVEDCONY:CMP       EDX,OFFSET MATHPAGE+8  ;LOOP IF NOT END OF SCRATCH AREA
          JNE       MOVEXX
          TEST      R5,0FH
          JNZ       MOVE10              ;BRANCH IF VALUE IS NOT ZERO
          MOV       R7,0                ;INTERNAL FORMAT ZERO
          JMP       JLMOVE7
MOVE10B:  MOV       R4,99H              ;PAD SCRATCH AREA WITH NINE'S
          JMP       MOVEDCONZ
MOVEYY:   MOV       AL,R4
          MOV       BYTE PTR [EDX],AL
          INC       EDX
MOVEDCONZ:CMP       EDX,OFFSET MATHPAGE+8
          JNE       MOVEYY

;        FIRST FIX SIGN DIGIT ( F7(HIGH) ):
;               9 --> 0
;               8 --> 8
;               1 --> 9
;               0 --> 1
;

MOVE10:   INC       EDX                 ;TURN OFF CARRY
          MOV       AL,R7
          AND       AL,0F0H
          CMP       AL,80H
          JE        JLMOVE11
          MOV       AL,R7
          ADD       AL,10H
          DAA
          MOV       R7,AL
          AND       AL,0F0H
          CMP       AL,20H
          JNE       MOVEZZ
          MOV       R7,90H
MOVEZZ:   MOV       AL,R7
          AND       AL,0F0H
          CMP       AL,10H
          JNE       JLMOVE12

;  IF SIGNS OF MANTISSA AND EXPONENT DO NOT DIFFER, DON'T
;    COMPLEMENT EXPONENT.

JLMOVE11: STC
          MOV       AL,0
          SBB       AL,R6
          DAS
          MOV       R6,AL
          INC       EDX
JLMOVE12: AND       R7,0F0H
          MOV       AL,R6
          SHR       AL,4
          OR        R7,AL               ;PACK SIGN, EXPONENT, AND FIRST DIGIT INT
          SHL       R6,4
          MOV       AL,R5
          AND       AL,0FH
          OR        R6,AL               ;TWO BYTES FOR WRITE TO SCRATCH AREA.
          TEST      R7,10H
          JZ        JLMOVE7             ;BRANCH IF MANTISSA IS POSITIVE
          JMP       MOVECOMP

;        MAT SEARCH  -  SEARCH A DESIGNATED STRING FOR
;                        CHARACTER STRINGS WHICH SATISFY
;                         A GIVEN RELATION AND PUT THE
;                          TWO BYTE POSITION INTO A
;                           LOCATOR STRING.
;
;        GENERAL FORM:
;        ***********************************************
;        * MAT SEARCH (search),(relation) (compare)    *
;        *            TO (locator) STEP (expression)   *
;        ***********************************************
;
;               WHERE:  (search) = STRING TO BE SEARCHED (no test for null string)
;                       (relation) = RELATION TESTED
;                       (compare) = COMPARISON STIRNG (no test for null string)
;                       (locator) = STRING OF POSITIONS
;                       (expression) = STEP PARAMETER
;
;        MODIFIED REGISTERS:  ALL, AUX 0-A,E,F
;
;        EXPR BETTER NOT USE:   AUX 3-A,E,F
;

SEARCH_S: CALL      ALPHINL             ;SCAN SEARCH VARIABLE OR LITERAL (RELEASE 1.5
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        SEARCH1             ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET SEARCH ADDRESS & LENGTH
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX03,EDX           ;SAVE INITIAL SEARCH STRING ADRESS
          MOV       AUX09,EDX           ;AUX9 = ADR(SEARCH STRING)-1
          DEC       AUX09
          ADD       EDX,EAX             ;PC's = ADR(END OF SEARCH STRING)+1
          MOV       AUX0A,EDX           ;AUXA = ADR(END OF SEARCH STRING)
          DEC       AUX0A
          MOV       EDX,DWORD PTR [A0]  ;AUX4 = INITIAL LENGTH TO SEARCH
          MOV       AUX04,EDX

SEARCH1:  CMP       BYTE PTR [TEXTPTR],','  ;MUST FIND COMMA
          JNE       ERROR_13
          INC       TEXTPTR
          CALL      RELOP               ;PUTS RELATION OPERATOR IN F1
          TEST      SL,2
          JZ        SEARCHX             ;BRANCH IF NOT X PHASE
          MOV       DL,R1
          OR        DL,0F0H             ;PL = RELOP AND FORWARD SEARCH DIRECTION
          MOV       AUX08,EDX           ;SAVE RELOP
SEARCHX:  MOV       K,4                 ;(CURRENT LENGTH))
          CALL      ALPHA1              ;SCAN ALPHA VALUE ARGUMENT
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        SEARCH2             ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET VALUE ADDRESS & LENGTH
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX05,EDX           ;SAVE ADDRESS
          MOV       EDX,DWORD PTR [A0]
          MOV       AUX06,EDX           ;SAVE LENGTH

SEARCH2:  CALL      CCLTO               ;MUST FIND "TO"
          CALL      AVARD               ;SCAN POSITION VARIABLE
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        SEARCH2B            ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET LOCATOR ADDRESS & LENGTH
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX0E,EDX           ;SAVE ADDRESS
          AND       A0,0FFFCH           ;MAKE LENGTH EVEN
          CMP       A0,0
          JNE       SEARCHY
          CMP       A1,0
          JE        ERR_51S             ;ERROR IF LENGTH WAS < 2 BYTES
SEARCHY:  ADD       EDX,DWORD PTR [A0]
          MOV       AUX0F,EDX           ;SAVE LAST ADDRESS
SEARCH2B: MOV       R0,@STEP
          CALL      CCL                 ;LOOK FOR "STEP"
          TEST      R0,0FH
          JZ        SEARCHS             ;BRANCH IF FOUND
          TEST      SL,2
          JZ        SEARCH6             ;BRANCH IF NOT X PHASE
          MOV       EDX,1               ;SET STEP VALUE TO ONE
SEARCH2A: MOV       AUX07,EDX           ;SAVE STEP VALUE

;       REGISTERS NOW SET UP AS FOLLOWS:
;         AUX3 = CURRENT ADR WITHIN SEARCH STRING
;         AUX4 = REMAINING LENGTH OF SEARCH STRING
;         AUX5 = ADR(STRING TO SEARCH FOR)
;         AUX6 = LENGTH OF STRING TO SEARCH FOR
;         AUX7 = STEP SIZE
;         AUX8 = COMPARISON OPERATOR
;         AUX9 = ADR(SEARCH STRING)-1
;         AUXA = ADR(END OF SEARCH STRING)
;         AUXE = ADR(LOCATOR STRING)
;         AUXF = LOCATOR LAST ADR

SEARCH3:  CALL      SRCHALPH            ;SEARCH FOR NEXT OCCURANCE
          TEST      R7,0FH
          JZ        SEARCHD             ;BRANCH IF NO MATCH; EXIT MAT SEARCH
          MOV       DWORD PTR [A2],EDX  ;F3F2 = ADR(MATCHED STRING)
          MOV       EDX,AUX09           ;PC's = ADR(SEARCH STRING)-1
          MOV       EAX,DWORD PTR [A2]
          SUB       EAX,EDX             ;F1F0 = OFFSET OF FOUND STRING IN SEARCH STRING
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,AUX07           ;PC's = STEP SIZE
          MOV       EBX,EDX
          MOV       EAX,DWORD PTR [A2]
          ADD       EAX,EBX             ;F5F4 = ADR(NEXT STRING TO TEST)
          MOV       DWORD PTR [A4],EAX
          JC        SEARCH4             ;BRANCH IF 16 BIT OVERFLOW
          MOV       EDX,AUX0A           ;PC's = ADR(END OF SEARCH STRING)
          CMP       DWORD PTR [A4],EDX
          JBE       SEARCH7             ;BRANCH IF STILL WITHIN SEARCH STRING

;      LAST OCCURANCE FOUND
;

SEARCH4:  CALL      POSARR1             ;PUT LAST OCCURANCE IN LOCATOR ARRAY
          CMP       R1,0
          JE        SEARCH6             ;BRANCH IF LOCATOR STRING FULL

;      SIGNAL END OF LOCATOR ARRAY
;

SEARCHD:  MOV       DWORD PTR [A0],0    ;F1F0 = ZERO
          CALL      POSARR1             ;PUT IN LOCATOR STRING
SEARCH6:  JMP       ENDSTMTZ            ;EXIT STATEMENT

;      LOOP AND SEARCH AGAIN
;

SEARCH7:  MOV       DWORD PTR [A2],EDX  ;F3F2 = ADR(END OF SEARCH STRING)+1
          INC       DWORD PTR [A2]
          MOV       EDX,DWORD PTR [A2]  ;PC's = REMAINING LENGTH TO SEARCH
          SUB       EDX,DWORD PTR [A4]
          MOV       AUX04,EDX           ;AUX4 = REMAINING LENGTH TO SEARCH
          MOV       EDX,DWORD PTR [A4]  ;AUX3 = ADR(NEXT STRING TO TEST)
          MOV       AUX03,EDX
          CALL      POSARR1             ;PUT OCCURANCE OF STRING IN LOCATOR STRING
          CMP       R1,0
          JNE       SEARCH3             ;LOOP IF LOCATOR STRING NOT FULL
          JMP       SEARCH6             ;EXIT STATEMENT IF LOCATOR STRING FULL

;
;        PROCESS STEP FUNCTION
;

SEARCHS:  CALL      EXPR                ;EVALUATE EXPRESSION
          TEST      SL,2
          JZ        SEARCH6             ;BRANCH IF NOT X PHASE
          CALL      FLTBIN              ;PUT STEP INTO F1,F0
          TEST      R2,2
          JZ        QERR34              ;ERROR IF NOT 0<X<2^32
          MOV       EDX,DWORD PTR [A0]
          JMP       SEARCH2A
QERR34:   JMP       ERR34


;        MAT MERGE  -  MERGE ROWS OF SORTED DATA
;                         INTO A SINGLE OUTPUT BUFFER
;                         IN ASCENDING ORDER.
;
;        GENERAL FORM:
;        ************************************************
;        *  MAT MERGE (merge array) TO (control area),  *
;        *               (work area),(locator array)    *
;        ************************************************
;
;               WHERE:   (merge array) - ROWS OF SORTED
;                              DATA WITH NOT MORE THAN
;                              254 ELEMENTS PER ROW.
;
;                        (control area) - USED TO STORE
;                              THE 1-BYTE COLUMN POSITION
;                              OF THE NEXT ELEMENT IN
;                              EACH ROW.  THE LAST BYTE
;                              IS USED AS A CONTROL FLAG
;                              SIGNALING TO THE USER IF A
;                              ROW IS EMPTY OR IF THE
;                              LOCATOR ARRAY IS FULL.
;
;                              REQUIRES: N+1 BYTES WHERE
;                                N = (no. of rows)  IN
;                                THE MERGE ARRAY.
;
;                        (work area) - USED TO STORE THE
;                              ADDRESS OF THE ELEMENT
;                              SPECIFIED BY THE COLUMN
;                              POSITION IN THE RESPECTIVE
;                              ROW OF THE CONTROL AREA.
;
;                              REQUIRES: N*2 BYTES WHERE
;                                N = (no. of rows) IN THE
;                                MERGE ARRAY.
;
;                        (locator array) - AN ARRAY WHICH
;                              RECEIVES THE 2-BYTE
;                              SUBSCRIPTS OF ELEMENTS IN
;                              THE MERGE ARRAY AS THEY
;                              ARE ARRANGED IN ASCENDING
;                              ORDER.
;
;
;        MODIFIED REGISTERS:  ALL, AUX 0-F
;                             (AUX 0,1 IN ENTRY PHASE)
;
;        EXPR BETTER NOT USE:   AUX 3,4,5,7,8,9,A,B,C,D
;

MERGE_S:  CALL      ARRAY               ;SCAN MERGE ARRAY VARIABLE
          TEST      R0,0FH
          JNZ       ERR_28S             ;ERROR IF NOT FOUND
          TEST      R5,1
          JZ        ERR_28S             ;ERROR IF NOT ALPHA
          TEST      SL,2
          JZ        MERGE0              ;BRANCH IF NOT X PHASE
          MOV       AL,R5
;         AND       AL,0FH              ;05/12/92
          AND       AL,07H
          CMP       AL,7
          JNE       ERR_50S             ;ERROR IF NOT ALPHA MATRIX
          MOV       AUX0B,EDX           ;SAVE ADDRESS OF DIMS
          SUB       AUX0B,5
          MOV       EDX,DWORD PTR [A6]
          CMP       DX,0FFFFH
          JAE       ERR_50S             ;ERROR IF MORE THAN 254 ELEMENTS/ROW
          MOV       AUX0C,EDX           ;SAVE MERGE DIMS
          MOVZX     EDX,CH
          MOV       AUX0D,EDX           ;SAVE ELEMENT LENGTH
          MOV       AUX03,EDX           ;SAVE AS DEFAULT FIELD LENGTH
MERGE0:   CALL      FIELD               ;SCAN AND PROCESS A FIELD SPECIFICATION
          TEST      SL,2
          JZ        MERGE1              ;BRANCH IF NOT X PHASE
          MOV       A6,0
          MOV       EDX,DWORD PTR [A5]  ;SAVE INCREMENT TO START OF FIELD
          MOV       AUX0A,EDX           ;SAVE IT
          CMP       A4,0
          JE        MERGE1              ;BRANCH IF FIELD NOT SPECIFIED
          MOVZX     EDX,A4
          MOV       AUX03,EDX           ;SAVE FIELD LENGTH
MERGE1:   CALL      CCLTO               ;MUST FIND "TO"

; CONTROL AREA
;
          CALL      AVARD               ;SCAN CONTROL AREA VARIABLE
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        MERGE2              ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET ADDRESS LENGTH OF CONTROL AREA
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX08,EDX           ;SAVE ADDRESS
          ADD       EDX,EAX
          MOV       AUX09,EDX           ;SAVE ADDRESS OF CONTROL FLAG (LAST WORD)
          SUB       AUX09,2
          MOV       EDX,AUX0C           ;GET MERGE DIMS
          SHR       EDX,16              ;NO. OF ROWS PUT IN PH,PL
          MOV       AUX07,EDX           ;TEMPORARY SAVE # OF ROWS
          SHL       EDX,1
          ADD       EDX,2
          CMP       DWORD PTR [A0],EDX
          JB        ERR_51S             ;ERROR IF NOT LARGE ENOUGH
MERGE2:   CMP       BYTE PTR [TEXTPTR],','  ;MUST FIND COMMA
          JNE       ERROR_13
          INC       TEXTPTR

; WORK VARIABLE
;
          CALL      AVARD               ;SCAN WORK AREA VARIALBLE
          TEST      R0,0FH
          JNZ       ERR_29I             ;ERROR IF NOT FOUND
          TEST      SL,2
          JZ        MERGE3              ;BRANCH IF NOT X PHASE
          MOV       EDX,VSPTR           ;GET ADDRESS & LENGTH OF BUFFER
          MOV       EAX,DWORD PTR [EDX]
          MOV       DWORD PTR [A0],EAX
          MOV       EDX,DWORD PTR [EDX+4]
          ADD       VSPTR,8
          MOV       AUX06,EDX           ;SAVE ADDRESS
          MOV       EDX,AUX07           ;GET NO. OF ROWS IN MERGE ARRAY
          MOV       DWORD PTR [A2],EDX
          SHL       EDX,2                ;DOUBLE TO FIND REQUIRED #  OF WORK BYTES
          CMP       DWORD PTR [A0],EDX
          JB        ERR_51S             ;ERROR IF WORK AREA ISN'T LARGE ENOUGH
MERGE3:
          CMP       BYTE PTR [TEXTPTR],',' ;MUST FIND A COMMA
          JNE       ERROR_13
          INC       TEXTPTR

; LOCATOR ARRAY
;
          CALL      ARRAY               ;SCAN LOCATOR ARRAY
          TEST      R0,0FH
          JNZ       ERR_28S             ;ERROR IF NOT FOUND
          TEST      R5,1
          JZ        ERR_28S             ;ERROR IF NOT ALPHA
          TEST      SL,2
          JZ        MERGE14             ;BRANCH IF NOT X PHASE
          CMP       CH,4
          JNE       ERR_50S             ;ERROR IF ELEMENT LENGTH IS NOT 4
          MOV       AUX0E,EDX           ;SAVE LOCATOR ADDRESS
          MOV       EAX,DWORD PTR [A2]
          SHL       EAX,2
          ADD       EDX,EAX
          MOV       AUX0F,EDX           ;SAVE LAST ADDRESS + 2

;        REGISTER USE IS AS FOLLOWS:
;               AUX 3 - FIELD LENGTH OF MERGE ARRAY
;                       ELEMENT ( = ELEMENT LENGTH IF
;                       FIELD IS NOT SPECIFIED)
;               AUX 6 - ADDRESS OF WORK AREA
;               AUX 8 - ADDRESS OF CONTROL AREA
;               AUX 9 - ADDRESS OF CONTROL FLAG ( LAST
;                       BYTE OF CONTROL AREA)
;               AUX A - INCREMENT TO START OF FIELD
;               AUX B - ADDRESS OF DIMS OF MERGE ARRAY
;               AUX C - DIMENSIONS OF MERGE ARRAY
;               AUX D - ELEMENT LENGTH OF MERGE ARRAY
;               AUX E - ADDRESS OF LOCATOR ARRAY
;               AUX F - LAST ADDRESS OF LOCATOR ARRAY
;
;
;        READ STARTING POSITIONS FOR EACH ROW FROM
;        CONTROL AREA AND PUT CORRESPONDING ELEMENT
;        ADDRESSES INTO WORK BUFFER.  IF POSITION IS 'FF'
;        THEN IGNORE ROW BY PUTTING '0000' INTO WORK
;        BUFFER.  SAVE ROW # AND ELEMENT ADDRESS OF THE
;        FIRST ROW NOT IGNORED.
;
;        K IS USED AS A FLAG FOR ARAYELP INDICATING
;           ALPHA MATRIX ( HIGH 8 BITS = 5 ) AND AS A
;              FLAG INDICATING WHETHER ALL ROWS ARE
;                 TO BE IGNORED ( LOW 8 BITS, 0=YES,1=NO)
;

          MOV       K,50H
          MOV       AUX05,0             ;INIT ROW POSITION (0001)
          INC       AUX05
          MOV       EDX,AUX0B
          MOV       CX,WORD PTR [EDX]
          MOVZX     EDX,CX
          DEC       EDX
          PUSH      EDX                 ;SAVE NO. OF ROWS IN MERGE ARRAY -1
          MOV       EDX,AUX06           ;SAVE WORK AREA ADDRESS
          MOV       AUX07,EDX
          MOV       EDX,AUX08           ;ADDRESS OF CONTROL AREA
MERGE4:   MOV       CX,WORD PTR [EDX]   ;READ COLUMN POSITION IN ROW
          XCHG      CH,CL
          MOV       AUX04,EDX           ;READ COLUMN POSITION IN ROW
          ADD       AUX04,2
          MOV       EDX,AUX05
          INC       AUX05               ;INCREMENT TO NEXT ROW AND SAVE
          CMP       CX,0FFFFH
          JNE       MERGE4A             ;BRANCH IF NOT IGNORE ROW
          MOV       DWORD PTR [A0],0    ;PUT '0000' INTO WORK AREA
          JMP       MERGE45

MERGE4A:  MOV       AX,CX               ;PUT ELEMENT SUBSCRIPT INTO F1,F0
          MOV       A0,AX
          MOV       A1,DX
          MOV       AL,K
          AND       AL,0FH
          CMP       AL,1
          JE        MERGE4B             ;BRANCH IF NOT FIRST GOOD ROW
          MOV       K,51H               ;FLAG = FIRST GOOD ROW FOUND
          MOV       EAX,EDX
          SHL       EDX,16
          MOV       DX,AX
          MOV       AUX05,EDX           ;SAVE ROW NUMBER OF FIRST GOOD ROW
          INC       AUX05
          MOV       EDX,AUX07           ;SAVE WORK AREA ADDRESS OF FIRST GOOD ROW
          MOV       AUX01,EDX
MERGE4B:  MOV       EDX,AUX0B           ;GET MERGE DIM ADDRESS
          CALL      ARAYELP             ;GET ELEMENT ADDRESS TO PC'S

MERGE4C:  MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX0A           ;GET INCREMENT TO START OF FIELD
          MOV       EAX,EDX             ;F1,0 = ADDRESS OF START OF FIELD
          ADD       DWORD PTR [A0],EAX
MERGE45:  MOV       EDX,AUX07           ;GET ADDRESS IN WORK AREA
          MOV       ECX,DWORD PTR [A0]
          MOV       DWORD PTR [EDX],ECX ;WRITE ELEMENT ADDRESS INTO WORK AREA
          ADD       EDX,4
          MOV       AUX07,EDX
          POP       EDX                 ;GET ROW COUNT
          MOV       EAX,EDX
          DEC       EDX
          CMP       EAX,0
          JE        MERGE5              ;BRANCH IF LAST ROW DONE
          PUSH      EDX                 ;DECREMENT ROW COUNT
          MOV       EDX,AUX04           ;GET ADDRESS OF CONTROL AREA
          JMP       MERGE4
MERGE5:   TEST      K,0FH
          JNZ       MERGE_S5            ;BRANCH IF ALL ROWS IGNORED

          MOV       DWORD PTR [A0],0    ;F1F0 = ZERO
          CALL      POSARR1             ;PUT IN LOCATOR STRING
          JMP       ENDSTMTZ            ;EXIT STATEMENT


;        NOW PERFORM POLYPHASE MERGE:
;          1) ONE EACH PASS, COMPARE NEXT ELEMENT OF
;             EACH ROW, AS GIVEN BY ADDRESS IN WORK AREA.
;          2) PUT SUBSCRIPT OF LOWEST ELEMENT FOUND IN
;             PASS INTO THE LOCATOR ARRAY.
;          3) INCREMENT THE COLUMN SUBSCRIPT IN THE
;             CONTROL AREA, AND PUT THE ADDRESS OF THIS
;             NEXT ELEMENT INTO THE WORK AREA.
;          4) REPEAT STEPS 1 - 3 UNTIL THE END OF A ROW
;             IS REACHED OR UNTIL THE LOCATOR ARRAY IS
;             FILLED.
;
;        REGISTERS NOW SET UP AS FOLLOWS:
;               AUX 1 - WORK AREA ADDRESS OF FIRST GOOD
;                       ROW
;               AUX 5 (HIGH 8 BITS) - ROW NUMBER OF FIRST
;                       GOOD ROW.
;
;        REGISTER USE WILL BE AS FOLLOWS:
;               AUX 2 (LOW 8 BITS) - ROW POSITION OF
;                       LOWEST ELEMENT IN THE PASS
;               AUX 3 - ADDRESS OF LOWEST ELEMENT IN THE
;                       PASS
;               AUX 4 - CURRENT ADDRESS IN THE WORK AREA
;               AUX 5 (LOW 8 BITS) - CURRENT ROW POSITION
;               F7,F6 - DIMENSIONS OF THE MERGE ARRAY
;               F5,F4 - FIELD LENGTH OF THE MERGE ARRAY
;

MERGE_S5:
          MOV       EDX,AUX0C           ;GET DIMENSIONS OF MERGE ARRAY
          MOV       DWORD PTR [A6],EDX
          MOV       EDX,AUX03           ;GET FIELD LENGTH OF MERGE ARRAY
          MOV       DWORD PTR [A4],EDX
MERGE6:   MOV       EDX,AUX01           ;GET FIRST GOOD WORK ADDRESS
          MOV       ECX,DWORD PTR [EDX]
          ADD       EDX,4               ;READ ELEMENT ADDRESS
          MOV       AUX04,EDX
          MOV       AUX03,ECX           ;INIT FIRST GOOD ROW AS LOWEST
          MOV       EDX,AUX05           ;GET FIRST GOOD ROW POSITION
          MOV       EAX,EDX
          SHR       EAX,16
          MOV       DX,AX
          MOV       AUX05,EDX           ;START COMPARE AT NEXT ROW
          INC       AUX05
          AND       EDX,0FFFFH
          MOV       AUX02,EDX           ;INIT FIRST ROW AS LOWEST
          DEC       AUX02

MERGE7:   CMP       A7,DX
          JE        MERGE9              ;BRANCH IF LAST ROW
          MOV       EDX,AUX04           ;GET NEXT WORK ADDRESS
          MOV       ECX,DWORD PTR [EDX]
          ADD       EDX,4               ;READ ELEMENT ADDRESS
          MOV       AUX04,EDX
          CMP       ECX,0
          JE        MERGE8              ;BRANCH IF IGNORED '0000'
          MOV       AUX0C,ECX           ;SAVE ELEMENT ADDRESS
          MOV       AUX00,ECX
          MOV       EDX,AUX03           ;ADDRESS OF CURRENT LOWEST ELEMENT
          MOV       EAX,DWORD PTR [A4]
          MOV       DWORD PTR [A0],EAX  ;SET UP FIELD LENGTH
          CALL      COMPARE             ;COMPARE NEXT ELEMENT WITH LOWEST FOR PASS
          CMP       R0,1
          JNE       MERGE8              ;BRANCH IF STILL THE LOWEST
          MOV       EDX,AUX0C           ;GET ELEMENT ADDRESS OF NEW ELEMENT
          MOV       AUX03,EDX           ;SAVE AS THE NEW LOWEST FOR THE PASS
          MOV       EDX,AUX05           ;GET CURRENT ROW POSITION
          MOV       AUX05,EDX           ;INCREMENT TO THE NEXT ROW
          INC       AUX05
          AND       EDX,0FFFFH
          MOV       AUX02,EDX           ;SAVE AS THE NEW LOWEST FOR THE PASS
          DEC       AUX02
          JMP       MERGE7

MERGE8:   MOV       EDX,AUX05           ;INCREMENT TO THE NEXT ROW
          MOV       AUX05,EDX
          INC       AUX05
          JMP       MERGE7

;        END OF PASS:  NOW PUT SUBSCRIPTS OF LOWEST
;              ELEMENT FOUND FOR THE PASS INTO THE
;              LOCATOR ARRAY AND TEST FOR END OF MERGE
;              CONDITIONS.
;

MERGE9:   MOV       EDX,AUX02           ;GET ROW POSITON OF LOWEST ELEMENT
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX08           ;GET ADDRESS OF CONTROL AREA
          MOV       EAX,DWORD PTR [A0]
          SHL       EAX,1
          ADD       EDX,EAX             ;FIND ADDRESS OF ROW IN CONTROL AREA
          MOV       CX,WORD PTR [EDX]
          XCHG      CH,CL
          MOV       AUX00,EDX           ;SAVE AND READ
          MOV       AX,A0
          INC       AX
          MOV       A1,AX               ;PUT SUBSCRIPT INTO F1,F0
          MOV       AX,CX
          MOV       A0,AX
          CALL      POSARR1             ;PUT SUBSCRIPT INTO LOCATOR ARRAY
          MOV       AX,A0
          CMP       AX,A6
          JE        MERGE10             ;BRANCH IF ROW IS EXHAUSTED
          MOV       AL,R1
          MOV       K,AL                ;SAVE LOCATOR STATUS
          INC       A0                  ;INCREMENT COLUMN POSITION
          MOV       EDX,AUX00           ;GET CONTROL AREA ADDRESS
          MOV       AX,A0
          XCHG      AH,AL
          MOV       WORD PTR [EDX],AX   ;WRITE COLUMN SUBSCRIPT INTO CONTROL AREA
          MOV       EDX,AUX03           ;ADDRESS OF LOWEST ELEMENT
          MOV       DWORD PTR [A0],EDX
          MOV       EDX,AUX0D           ;GET ELEMENT LENGTH OF MERGE ARRAY
          MOV       EAX,EDX
          ADD       DWORD PTR [A0],EAX  ;FIND ADDRESS OF NEXT ELEMENT IN ROW
          MOV       EDX,AUX02           ;GET ROW POSITION OF LOWEST
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX06           ;GET ADDRESS OF WORK AREA
          MOV       EAX,DWORD PTR [A2]
          SHL       EAX,2
          ADD       EDX,EAX             ;FIND ADDRESS OF ROW IN WORK AREA
          MOV       EAX,DWORD PTR [A0]
          MOV       DWORD PTR [EDX],EAX ;WRITE IN NEXT ELEMENT ADDRESS
          ADD       EDX,4
          CMP       K,0
          JNE       MERGE6              ;BRANCH IF LOCATOR NOT FULL

;
;        THIS SECTION CODES FOR LOCATOR ARRAY FULL
;

MERGE12:  MOV       EDX,AUX09           ;GET ADDRESS OF CONTROL FLAG
MERGE13:  MOV       WORD PTR [EDX],0    ;'00' SIGNALS LOCATOR FULL
MERGE14:  JMP       ENDSTMTZ

;
;        THIS SECTION CODES FOR ROW EMPTY
;

MERGE10:  MOV       EDX,AUX02           ;GET ROW POSITION
          MOV       A0,DX
          INC       A0
          MOV       EDX,AUX09           ;GET ADDRESS OF CONTROL FLAG
          MOV       AX,A0
          XCHG      AH,AL
          MOV       WORD PTR [EDX],AX   ;SIGNAL THAT ROW IS EMPTY
          MOV       EDX,AUX00           ;GET CONTROL AREA ADDRESS
          MOV       WORD PTR [EDX],0FFFFH ;SIGNAL TO IGNORE ROW
          TEST      R1,0FH
          JZ        MERGE14             ;BRANCH IF LOCATOR IS FULL
          MOV       EDX,AUX0E           ;GET LOCATOR ADDRESS
          MOV       WORD PTR [EDX],0    ;WRITE '0000' AS END OF BUFFER
          ADD       EDX,2
          JMP       MERGE13

;
;
;        CALLS TO ERROR ROUTINES
;
;

ERROR_26: MOV       R0,26H
          JMP       ERROR               ;MISSING ARRAY
ERR_51S:  JMP       ERROR_51            ;VARIABLE TOO SMALL
ERR_28S:  JMP       ERROR_28            ;MISSING ALPHA ARRAY
ERR_50S:  JMP       ERROR_50            ;ILLEGAL DIM'S OR LENGTH
ERR_29I:  JMP       ERROR_29            ;MISSING ALPHA VARIABLE
ERR_55S:  JMP       ERROR_55            ;UNDEFINED ARRAY


;        POSARR1   -  PUTS THE TWO BYTE VALUE STORED IN
;                     F1F0 INTO THE ADDRESS GIVEN BY
;                     AUX E, AND THEN INCREMENTS THE
;                     ADDRESS AND CHECKS IF DONE.
;
;        ENTRY:  AUX E = RECEIVER ADDRESS
;                AUX F = LAST ADDRESS + 2
;
;        RETURN: F1 = 0 IF DONE
;                F1 <> 0 IF NOT DONE
;
;        MODIFIED REGISTERS:  PH,PL,F0-F3,AUX E,F
;

POSARR1:  MOV       EDX,AUX0F           ;GET LAST ADDRESS
          MOV       DWORD PTR [A2],EDX
          MOV       EDX,AUX0E           ;GET CURRENT ADDRESS
          MOV       EAX,DWORD PTR [A0]
          PUSH      ECX
          MOV       ECX,EAX
          SHR       ECX,16
          xchg      ch,cl
          MOV       WORD PTR [EDX],CX   ;WRITE VALUE IN
          POP       ECX
          xchg      ah,al
          MOV       WORD PTR [EDX+2],AX
          ADD       EDX,4
          MOV       AUX0E,EDX           ;SAVE NEW CURRENT ADDRESS
          CMP       DWORD PTR [A2],EDX
          JBE       POSA0               ;BRANCH IF DONE
          MOV       R1,1
          RET

POSA0:    MOV       R1,0
          RET

CODE      ENDS
          END
