;===========================================================================
;
;
;
;===========================================================================
;           This module was originally M60
;===========================================================================
            .XLIST
            INCLUDE PREAMBLE
            .LIST

PUBLIC    FOR
PUBLIC    NEXT
PUBLIC    ERROR_40

EXTRN     CCL     :NEAR
EXTRN     CCLTO   :NEAR
EXTRN     ENDSTMT :NEAR
EXTRN     ENDSTMTS:NEAR
EXTRN     ERROR   :NEAR
EXTRN     ERR_02  :NEAR
EXTRN     EXPR    :NEAR
EXTRN     MATHADD1:NEAR
EXTRN     MATHCMP1:NEAR
EXTRN     MATHSUB1:NEAR
EXTRN     MVSCRTCH:NEAR
EXTRN     PRNTCR  :NEAR
EXTRN     RECNUM  :NEAR
EXTRN     TRACDSBL:NEAR
EXTRN     TRACELNB:NEAR
EXTRN     TRACENBL:NEAR
EXTRN     TWRIT2SP:NEAR
EXTRN     VAR     :NEAR
EXTRN     VVFMTA  :NEAR
EXTRN     WRTBLK  :NEAR
EXTRN     WRTWORD :NEAR
EXTRN     _END    :NEAR
EXTRN     _NEXT   :NEAR

EXTRN     @@OSEND :DWORD
EXTRN     @@VSEND :DWORD


;-------------------------------------------------------------------------------
;                                                                              ;
;      ***********************************                                     ;
;      *                                 *                                     ;
;      *  MODULE-60: GENERAL STATEMENTS  *                                     ;
;      *                                 *                                     ;
;      ***********************************                                     ;
;                                                                              ;
;                                                                              ;
;      DESCRIPTION: GENERAL BASIC STATEMENTS                                   ;
;                     FOR      NEXT                                            ;
;                                                                              ;
;      ENTRY: BRANCH TO APPROPRIATE STATEMENT ROUTINE.                         ;
;             TEXTPTR POINTS AFTER VERB.                                       ;
;             TEXTLPTR POINTS TO CURRENT LINE.                                 ;
;                                                                              ;
;      EXIT: BRANCH TO 'ENDSTMT','ENDSTIO',OR 'ERROR'                          ;
;-------------------------------------------------------------------------------




;      FOR: PROCESS A 'FOR' STATEMENT
;
;           STMT FORMAT:  FOR v = s TO l [STEP i]
;                WHERE: v = scalar numeric variable
;                           (index variable)
;                       s,l,i = expressions
;                       s = starting value of index var.
;                       l = limit of index variable
;                       i = index increment
;
;           ENTRY: TEXTLPTR POINTS TO CURRENT LINE
;                  TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;           DESCRIPTION: A BASIC PROGRAM LOOP IS INITIATED
;                BY A FOR STATEMENT AND TERMINATED BY A
;                NEXT STATEMENT.  WHEN THE FOR LOOP IS
;                ENCOUNTERED, THE FOLLOWING ACTION OCCURS:
;                   1. INDEX VARIABLE SET = STARTING VALUE
;                   2. ADDRESS OF INDEX VARIABLE --> VS
;                   3. LIMIT VALUE --> VS
;                   4. STEP VALUE --> VS
;                   5. STMTCNT --> VS
;                   6. TEXTPTR (POINTS TO AFTER FOR STMT)
;                      --> VS
;                   7. TEXTLPTR --> VS
;                   8. FOR ATOM --> OS
;                THE 24 BYTES OF INFORMATION IN THE VALUE
;                STACK (VS) IS KNOWN AS THE FOR-GROUP. THE
;                FOR-GROUP IS USED DURING 'NEXT' STATEMENT
;                PROCESSING. THE FOR-GROUP IS DELETED IF
;                   1. THE LOOP COMPLETES (IE, THE INDEX
;                      VALUE EXCEEDS THE LIMIT IN 'NEXT'
;                      STATEMENT PROCESSING)
;                   2. A 'NEXT' STATEMENT WITH A DIFFERENT
;                      INDEX VARIABLE, WHOSE CORRESPONDING
;                      'FOR' STATEMENT WAS EXECUTED BEFORE
;                      THE 'FOR' STATEMENT THAT GENERATED
;                      THE FOR-GROUP LAST IN THE STACK, IS
;                      ENCOUNTERED.
;                   3. A 'RETURN' STATEMENT IS ENCOUNTERED
;                      WHOSE CORRESPONDING 'GOSUB' STATE-
;                      MENT WAS EXECUTED BEFORE THE 'FOR'
;                      STATEMENT THAT GENERATED THE FOR-
;                      GROUP LAST IN THE STACK.
;
;           FOR-GROUP: VALUE STACK (VS)    32 BYTES
;                ---------------------------------------
;                ! TEXTLPTR                  (4 BYTES) !
;                ! TEXTPTR                   (4 BYTES) !
;                ! STMTCNT                   (4 BYTES) !
;                ! STEP VALUE                (8 BYTES) !
;                ! LIMIT VALUE               (8 BYTES) !
;                ! ADDRESS OF INDEX VARIABLE (4 BYTES) !
;                ---------------------------------------
;
;           MODIFIED REGISTERS: ALL,AUX0,1,2


FOR:      CALL    TRACENBL                ;ENABLE TRACE IF TRACE IS ON
          MOV     K,3                     ;FLAG = SCALAR RECEIVER VARIABLE
          CALL    VAR                     ;PROCESS INDEX VARIABLE
          TEST    SL,2                    ;BRANCH IF NOT XPASE
          JZ      FOR4                    ;
          MOV     EDX,VSPTR               ;
          MOV     ECX,DWORD PTR [EDX]     ;
          SUB     EDX,4
          MOV     DWORD PTR [EDX],ECX     ;SAVE ADDRESS OF INDEX VARIABLE IN VS
          MOV     VSPTR,EDX
          CMP     DWORD PTR [@@VSEND],EDX ;CHECK VS STACK OVERFLOW
          JA      ERR_02

FOR4:     MOV     EDX,TEXTPTR             ;LOOK FOR '='
          MOV     CH,BYTE PTR [EDX]
          CMP     CH,'='
          JE      FOR30
          MOV     R0,12H
          JMP     ERROR
FOR30:    INC     TEXTPTR
          CALL    RECNUM                  ;INDEX VARIABL = STARTING VALUE
          CALL    CCLTO                   ;LOOK FOR 'TO'
          CALL    EXPR                    ;SCAN AND EVALUATE LIMIT VALUE
          MOV     R0,@STEP                ;
          CALL    CCL                     ;LOOK FOR 'STEP'
          TEST    R0,0FH                  ;BRANCH IF ?
          JZ      FOR1                    ;
          TEST    SL,2                    ;BRANCH IF NOT XPHASE
          JZ      FOR2
          MOV     EDX,VSPTR
          MOV     WORD PTR [EDX-8],0      ;PUT 1 INTO VS
          MOV     WORD PTR [EDX-6],1      ;
          MOV     DWORD PTR [EDX-4],0     ;
          SUB     EDX,8
          MOV     VSPTR,EDX
          CMP     DWORD PTR [@@VSEND],EDX ;CHECK VS STACK OVERFLOW
          JA      ERR_02                  ;

FOR3:     MOV     EDX,VSPTR                     ;
          MOV     EAX,STMTCNT
          MOV     DWORD PTR [EDX-4],EAX     ;STMTCNT ==>  VS
          MOV     DWORD PTR [EDX-8],TEXTPTR ;TEXTPTR ==>  VS
          MOV     EAX,TEXTLPTR
          MOV     DWORD PTR [EDX-12],EAX    ;TEXTLPTR ==> VS
          SUB     EDX,12
          MOV     VSPTR,EDX
          CMP     DWORD PTR [@@VSEND],EDX ;CHECK VS STACK OVERFLOW
          JA      ERR_02

          MOV     EDX,OSPTR
          INC     EDX
          CMP     EDX,DWORD PTR [@@OSEND] ;CHECK OS OVERFLOW
          JNE     FOR40
          MOV     R0,4
          JMP     ERROR                   ;OS OVERFLOW
FOR40:    MOV     BYTE PTR [EDX],@FOR     ;FOR ATOM -> OS
          MOV     OSPTR,EDX

FOR50:    CALL    TRACDSBL                ;DISABLE TRACE
          JMP     ENDSTMTS                ;SAVE STACKS, EXIT

FOR1:     CALL    EXPR                    ;SCAN AND EVALUATE INCREMENT VALUE
          TEST    SL,2                    ;BRANCH IF XPHASE
          JNZ     FOR3                    ;

FOR2:     JMP     ENDSTMT                 ;EXIT



;      NEXT: PROCESS A 'NEXT' STATEMENT
;
;           STMT FORMAT:  NEXT v [,v ...]
;                WHERE: v = numeric scalar variable
;                       (index variable)
;
;                NOTE: 'NEXT X,Y,Z' IS EQUIVALENT TO
;                      'NEXT X: NEXT Y: NEXT Z'
;
;           ENTRY: TEXTLPTR POINTS TO CURRENT LINE
;                  TEXTPTR POINTS TO NEXT TEXT CHARACTER
;
;           DESCRIPTION: WHEN A 'NEXT' STATEMENT IS
;                ENCOUNTERED, THE FOLLOWING ACTION OCCURS:
;                   1. THE LAST OS ATOM IS CHECKED. IF IT
;                      IS NOT A FOR ATOM, ERR27 RESULTS.
;                   2. IF THE 'FOR' VARIABLE IN THE LAST
;                      FOR-GROUP <> 'NEXT' VARIABLE, THE
;                      FOR-GROUP IS FLUSHED AND 1. IS DONE
;                      AGAIN.
;                   3. INDEX VARIABLE = INDEX VAR. + STEP
;                   4. INDEX VALUE IS COMPARED TO LIMIT.
;                   5. IF SIGN OF STEP IS IN WRONG
;                      DIRECTION OR STEP = 0, FOR-GROUP
;                      IS FLUSHED AND NEXT SEQUENTIAL
;                      STATEMENT IS EXECUTED.
;                   6. A TRANSFER IS MADE TO THE STATMENT
;                      FOLLOWING THE CORRESPONDING FOR
;                      STATEMENT.
;
;           MODIFIED REG: ALL,AUX0,1,2

NEXT11:   CALL    MATHSUB1                ;INDEX = INDEX - STEP
          CMP     R0,0                    ;BRANCH IF MATH ERROR
          JNE     NEXT12
NEXT1:    DEC     OSPTR                   ;REMOVE FOR ATOM FROM OS
          ADD     VSPTR,20H
          TEST    MODE,00000004H
          JZ      NEXT2                   ;
          MOV     EDX,OFFSET _NEXT+6      ;
          MOV     CH,BYTE PTR [EDX]       ;
          CALL    WRTWORD                 ;OUTPUT 'NEXT'
          MOV     EDX,OFFSET _END+4       ;
          MOV     CH,BYTE PTR [EDX]       ;
          CALL    WRTWORD                 ;OUTPUT 'END'
          CALL    PRNTCR                  ;PRINT CR
NEXT2:    MOV     EDX,TEXTPTR
          MOV     CH,BYTE PTR [EDX]
          CMP     CH,','
          JE      NEXT30
          JMP     NEXT3
NEXT30:   INC     TEXTPTR

NEXT:     CALL    TRACENBL                ;ENABLE TRACE IF TRACE ON
          MOV     K,3                     ;FLAG = SCALAR RECEIVER VARIABLE
          CALL    VAR                     ;PROCESS INDEX VARIABLE
          TEST    SL,2                    ;BRANCH IF NOT XPHASE
          JZ      NEXT2                   ;
          MOV     EDX,VSPTR               ;
          MOV     ECX,DWORD PTR [EDX]     ;GET 'NEXT' VARIABLE ADDRESS FROM VS
          ADD     VSPTR,4
          MOV     AUX01,ECX               ;AUX1 = ADDRESS OF 'NEXT' VARIABLE
NEXT10:   MOV     EDX,OSPTR               ;
          DEC     EDX                     ;REMOVE LAST OS ATOM FROM OS
          MOV     CH,BYTE PTR [EDX]       ;
          MOV     OSPTR,EDX               ;CH = NEXT OS ATOM
          CMP     CH,@FOR                 ;
          JE      NEXT4                   ;BRANCH IF 'FOR' ATOM
ERROR_40: MOV     R0,40H                  ;ILLEGAL FOR/NEXT USAGE
          JMP     ERROR                   ;
NEXT4:    MOV     EDX,VSPTR               ;
          ADD     EDX,1CH                 ;PC'S POINT TO 'FOR' VARIABLE ADDRESS
          MOV     ECX,DWORD PTR [EDX]     ;ECX = ADDRESS OF 'FOR' VARIABLE
          CMP     ECX,AUX01
          JNE     NEXT5
          SUB     EDX,10H
          CALL    MATHADD1                ;INDEX VARIABLE = INDEX VALUE + STEP
          CMP     R0,0
          JE      NEXT20                  ;BRANCH IF NO MATH ERROR
NEXT12:   JMP     ERROR                   ;
NEXT20:   MOV     EDX,VSPTR               ;
          ADD     EDX,14H                 ;PC'S = ADDRESS OF LIMIT VALUE
          CALL    MATHCMP1                ;COMPARE INDEX VALUE TO LIMIT VALUE
          MOV     EDX,VSPTR               ;
          ADD     EDX,0CH                 ;PC'S POINT TO STEP VALUE
          MOV     CX,WORD PTR [EDX+6]     ;CX  = 1ST WORD OF STEP VALUE
          TEST    R7,01H                  ;BRANCH IF INDEX < LIMIT
          JNZ     NEXT6                   ;
          TEST    R7,02H                  ;BRANCH IF INDEX = LIMIT
          JNZ     NEXT7                   ;
          OR      CX,CX                   ;BRANCH IF INDEX > LIMIT  STEP >= 0
          JNS     NEXT11                  ;
          JMP     NEXT8
NEXT6:    OR      CX,CX                   ;BRANCH IF INDEX < LIMIT & STEP < 0
          JS      NEXT11
NEXT7:    MOV     AX,WORD PTR [EDX+2]
          OR      AX,AX
          JNZ     NEXT8
          MOV     AX,WORD PTR [EDX+4]
          OR      AX,CX
          JZ      NEXT1                   ;BRANCH IF STEP = 0
NEXT8:    TEST    MODE,00000004H          ;BRANCHI IF TRACE OFF
          JZ      NEXT9                   ;
          MOV     EDX,AUX01               ;
          CALL    VVFMTA                  ;FORMAT VALUE FOR PRINTING
          CALL    MVSCRTCH                ;MOVE VALUE TO 'IOBUFFER'
          CALL    WRTBLK                  ;PRINT VALUE
          CALL    TWRIT2SP                ;OUTPUT 2 SPACES
NEXT9:    MOV     EDX,VSPTR
          MOV     EAX,DWORD PTR [EDX]
          MOV     TEXTLPTR,EAX              ;SET TEXTLPTR TO LINE TO TRANSFER TO
          MOV     TEXTPTR,DWORD PTR [EDX+4] ;SET TEXTPTR TO LINE TO TRANSFER TO
          MOV     EAX,DWORD PTR [EDX+8]
          MOV     STMTCNT,EAX               ;SET STMTCNT TO STATEMENT TO TRANSFER TO
          ADD     EDX,12
          CALL    TRACELNB                ;IF TRACE, PRINT 'TRANSFER TO XXXX' MESSAGE
NEXT3:    CALL    TRACDSBL                ;DISABLE TRACE
          JMP     ENDSTMTS                ;SAVE STACK PTRS, EXIT
NEXT5:    ADD     VSPTR,20H               ;REMOVE FOR-GROUP FROM VS
          JMP     NEXT10                  ;

CODE      ENDS
          END
