;===========================================================================
;
;
;
;===========================================================================
;           This module was originally M91
;===========================================================================
	    .XLIST
	    INCLUDE PREAMBLE
	    .LIST

PUBLIC    BUFSTRCM
PUBLIC    ERRORMSG
PUBLIC    ENDMSG
PUBLIC    LDCMSG7
PUBLIC    LDCMSG14
PUBLIC    LDCMSG15
PUBLIC    LDCMSG16
PUBLIC    LDTMSG1
PUBLIC    LDTMSG2
PUBLIC    LDTMSG3
PUBLIC    LDTMSG4
PUBLIC    LDTMSG5
PUBLIC    LDTMSG6
PUBLIC    LDTMSG17
PUBLIC    MSGBASE
PUBLIC    MSGATOMS
PUBLIC    READYMSG
PUBLIC    SYSMSGS
PUBLIC    SUPERSTR
PUBLIC    TRNSFMSG
PUBLIC    VRFYMSG
PUBLIC    WRTSTRCM
PUBLIC    WRTERRSTRCM

EXTRN     CPBASE  :DWORD
EXTRN     OSMOD   :WORD
EXTRN     TWRITE  :NEAR

;------------------------------------------------------------------------------;
;                                                                              ;
;           ******************************                                     ;
;           *                            *                                     ;
;           *       MOD91 MESSAGES       *                                     ;
;           *                            *                                     ;
;           ******************************                                     ;
;                                                                              ;
;    This module contains the system messages stored in control memory.        ;
;    This includes all the error messages.  The messages are stored as         ;
;    an atomized superstring, where each message is a substring thereof.       ;
;    A vector contains pointers to the messages.                               ;
;                                                                              ;
; The following BASIC-2 files are used to create the message source code files:;
;                                                                              ;
;             MSG.COMP -- message compiler                                     ;
;             SUPERST! -- message compiler overlay                             ;
;             MSG.SRC  -- message compiler overlay                             ;
;                                                                              ;
;             B2MSGS   -- BASIC-2 messages                                     ;
;                                                                              ;
;             PRNTMSGS -- display/print messages                               ;
;                                                                              ;
;    To run the compiler, LOAD RUN "MSG.COMP".  The compiler uses the          ;
;    B2MSGS file for input and produces 2600 source files as output.           ;
;    EJECTs and ORGs must be edited in by hand after the compilation.          ;
;    Also, change MSGBASE to 0 and add in SUPERSTR EQU.                        ;
;                                                                              ;
;    The module also contains routines used to display messages.               ;
;                                                                              ;
;******************************************************************************;
;                                                                              ;
;    [The BASIC-3 logic was changed so that the CM address in the text-vector  ;
;    element is now relative to the beginning of the superstring.  This allows ;
;    the superstring to be located anywhere in CM, not just in the 1st 4K.     ;
;------------------------------------------------------------------------------;


;WRTSTRCM: output (OBS) a message stored in control memory, using
;          TWRITE logic.
;
;BUFSTRCM: copy the message stored in control memory to RIOBUFFR in data memory.
;
;        The message is a string of characters stored in consecutive locations in
;        control memory.  The 1st byte of the string can start in any byte of the
;        control memory word.  Normal message characters are restricted to 01-7F.
;        The byte value 00 is a word atom indicating that
;        the next byte represents an atomized word.  The atomized word is also
;        stored in control memory and follows the same rules as the message
;        itself.
;
;        The message is defined by a control memory word, called the text-vector
;        element.  The text-vector element has the following format:
;
;                  K,bit 80 = parity for control memory word
;                   ,bit 7F = message length (1-127)
;                  PH,bit 80 = 1 iff message starts in byte 1 of CM word
;                    ,bit 40 = 1 iff message starts in byte 2 of CM word
;                    ,bit C0 = 0 iff message starts in byte 3 of CM word
;                  PHPL,bit 0FFF = address in CM of message (now relative to superstring)
;
;        Likewise, a text-vector defines the atomized words in the messages.
;        This vector is MSGWORD, where:
;
;                  1st element defines atomized word 00
;                  2nd element defines atomized word 01
;                  3rd element defines atomized word 02
;                   .
;                   .
;                   .
;
;     Register use:
;        F2, bit 02 = 1 if message is to be buffered in RIOBUFFR
;                     0 if message is to output using TWRITE logic
;        AUX0 = buffer ptr for BUFSTRCM
;
;     Entry: device enabled.
;            PC's = address of the text-vector element that defines the limits
;                   of the message.  REF: "text super-string".
;
;        Return:   (BUFSTRCM)
;                  RIOBUFFR contains message
;                  AUX0 points to byte after message
;
;        Modified registers: PC's,F0-2,K
;                            AUX0 (BUFSTRCM only)

;------------------ setup ------------------------------------------------------

BUFSTRCM: MOV      R2,2                     ;flag = buffer characters of message
	  MOV      EAX,DWORD PTR [CPBASE]
	  LEA      EAX,[EAX].@IOBUFFR
	  MOV      AUX00,EAX                ;iobuffer ptr
	  JMP      WRTSTRC0

WRTERRSTRCM:
	  MOV      R2,0                     ;flag = output character of message
	  mov      eax,dword ptr [cpbase]
	  cmp      byte ptr [eax].@errf,5
	  jne      wrtstrc0
	  mov      eax,dword ptr [cpbase]
	  mov      eax,dword ptr [eax].@spreg1
	  mov      ch,byte ptr [eax]
	  shr      ch,4
	  or       ch,30h
	  mov      k,ch
	  call     wrtstrch
	  mov      eax,dword ptr [cpbase]
	  mov      eax,dword ptr [eax].@spreg1
	  mov      ch,byte ptr [eax]
	  and      ch,0fh
	  or       ch,30h
	  mov      k,ch
	  call     wrtstrch
	  mov      eax,dword ptr [cpbase]
	  mov      eax,dword ptr [eax].@spreg1
	  mov      ch,byte ptr [eax+1]
	  shr      ch,4
	  or       ch,30h
	  mov      k,ch
	  call     wrtstrch
	  mov      eax,dword ptr [cpbase]
	  mov      eax,dword ptr [eax].@spreg1
	  mov      ch,byte ptr [eax+1]
	  and      ch,0fh
	  or       ch,30h
	  mov      k,ch
	  call     wrtstrch
	  mov      k,20h
	  call     wrtstrch
	  jmp      wrtstrc0
WRTSTRCM: MOV      R2,0                     ;flag = output character of message
WRTSTRC0: MOV      AL,BYTE PTR [EDX]        ;read the message length
	  MOV      R0,AL
	  MOV      EDX,DWORD PTR [EDX+1]    ;read the message address
WRTSTRC1: MOV      AL,BYTE PTR [EDX]
	  MOV      K,AL
	  CALL     WRTSTRCH                 ;output character
	  DEC      R0
	  JZ       WRTSTRC2                 ;end of string
	  INC      EDX
	  JMP      WRTSTRC1                 ;output another character
WRTSTRC2: RET


;WRTSTRCH:  subroutine used by WRTSTRCM to process the next byte of the message.
;
;        If the next byte of the message is a normal message character, that
;        character is output.  If the character is 00, a flag is set indicating
;        that the next byte of message is a word atom.  The next time WRTSTRCH
;        is called, the specified word in output by calling WRTSTRCM. (Note,
;        the recursion.  4 additional stack levels are used.)
;
;        Entry:  K = next byte of message
;                F2,bit 01 = 1 iff last character was a message atom indicator (hex00)
;                  ,bit 02 = 1 iff character to be buffered, 0 if to be output
;
;        Return:  F2,bit 01 = 1 iff next byte of message is a word atom
;        Modified registers:  F2, K, F1

;------------------ byte = 00 --------------------------------------------------

WRTSTRCH: TEST      R2,1
	  JNZ       WRTSTRH2            ;branch if next byte is a word atom
	  CMP       K,0
	  JNE       WRTSTRH1            ;branch if byte not 00
	  OR        R2,1                ;flag = 00 found
	  RET

;------------------- normal character -------------------------------------------

WRTSTRH1: MOV       CX,WORD PTR [OSMOD] ;Read The First Byte Of OSMOD
	  XCHG      CH,CL
	  TEST      CH,4
	  JZ        WSTWR               ;If Bit 4 is Low - Don't Xlate
	  MOV       AL,K
	  AND       AL,0F0H
	  CMP       AL,70H
	  JE        WST7X               ;If 7X, Check Further
	  CMP       AL,60H
	  JNE       WSTWR               ;If < 60, Don't Xlate
	  TEST      K,0FH
	  JZ        WSTWR               ;If 60, Don't Xlate
	  JMP       WSTXL               ;Go Translate

WST7X:    TEST      K,8
	  JZ        WSTXL               ;If Less Than 78, Xlate
	  TEST      K,4
	  JNZ       WSTWR               ;If >=7C, Don't Xlate
	  MOV       AL,K
	  AND       AL,3
	  CMP       AL,3
	  JE        WSTWR               ;IF >7A, Don't Xlate

WSTXL:    XOR       K,20H               ;Make Into Upper Case

WSTWR:    TEST      R2,2
	  JNZ       WRTSTRH3            ;*!*!* RMK FHU
	  JMP       TWRITE              ;output message character

WRTSTRH3: XCHG      AUX00,EDX
	  MOV       AL,K
	  MOV       BYTE PTR [EDX],AL   ;buffer message character
	  XCHG      AUX00,EDX
	  INC       AUX00
	  RET

;------------------ word atom --------------------------------------------------

WRTSTRH2: PUSH      EDX                 ;save PC's
	  MOV       DL,R0
	  MOV       DH,R2
	  PUSH      EDX                 ;save F0,2

	  MOV       EDX,offset MSGATOMS
	  MOV       EAX,0
	  MOV       AL,K
	  ADD       EDX,EAX
	  SHL       EAX,2               ;multiply 5  for each message atom
	  ADD       EDX,EAX             ;point at text-vector element for word
	  AND       R2,2
	  CALL      WRTSTRC0            ;output atomized word

	  POP       EDX
	  MOV       R2,DH
	  AND       R2,0FEH             ;restore F2 (but turn off 00 flag)
	  MOV       R0,DL               ;restore F0
	  POP       EDX                 ;restore PC's
	  RET

;  Superstring

SUPERSTR LABEL     NEAR
MSGBASE  LABEL     NEAR

	 DB     53H,79H,73H
	 DB     74H,65H,6DH
	 DB     00H,09H,4CH
	 DB     69H,6EH,65H
	 DB     00H,0FH,20H
	 DB     4CH,6FH,6EH
	 DB     67H,00H,02H
	 DB     20H,69H,6EH
	 DB     20H,49H,6DH
	 DB     6DH,65H,64H
	 DB     69H,61H,74H
	 DB     65H,20H,4DH
	 DB     6FH,64H,65H
	 DB     53H,74H,61H
	 DB     74H,65H,6DH
	 DB     65H,6EH,74H
	 DB     20H,00H,02H
	 DB     20H,48H,65H
	 DB     72H,65H,00H
	 DB     00H,00H,0EH
	 DB     20H,52H,65H
	 DB     73H,6FH,6CH
	 DB     76H,65H,64H
	 DB     00H,01H,20H
	 DB     27H,28H,27H
	 DB     00H,01H,20H
	 DB     27H,29H,27H
	 DB     00H,01H,20H
	 DB     27H,3DH,27H
	 DB     00H,01H,20H
	 DB     27H,2CH,27H
	 DB     00H,01H,20H
	 DB     27H,2AH,27H
	 DB     00H,01H,20H
	 DB     27H,3EH,27H
	 DB     00H,01H,20H
	 DB     4CH,65H,74H
	 DB     74H,65H,72H
	 DB     00H,01H,20H
	 DB     48H,65H,78H
	 DB     20H,44H,69H
	 DB     67H,69H,74H
	 DB     00H,01H,20H
	 DB     52H,65H,6CH
	 DB     61H,74H,69H
	 DB     6FH,6EH,61H
	 DB     6CH,20H,4FH
	 DB     70H,65H,72H
	 DB     61H,74H,6FH
	 DB     72H,20H,53H
	 DB     74H,61H,63H
	 DB     6BH,20H,4FH
	 DB     76H,65H,72H
	 DB     66H,6CH,6FH
	 DB     77H,00H,01H
	 DB     20H,57H,6FH
	 DB     72H,64H,45H
	 DB     6EH,64H,20H
	 DB     6FH,66H,20H
	 DB     56H,61H,6CH
	 DB     69H,64H,20H
	 DB     53H,79H,6EH
	 DB     74H,61H,78H
	 DB     00H,01H,20H
	 DB     4CH,69H,6EH
	 DB     65H,00H,05H
	 DB     20H,43H,6FH
	 DB     6EH,66H,6CH
	 DB     69H,63H,74H
	 DB     00H,02H,20H
	 DB     50H,4CH,4FH
	 DB     54H,20H,00H
	 DB     0BH,00H,0FH
	 DB     20H,4CH,61H
	 DB     72H,67H,65H
	 DB     00H,03H,20H
	 DB     4CH,69H,74H
	 DB     65H,72H,61H
	 DB     6CH,00H,02H
	 DB     20H,45H,78H
	 DB     70H,72H,65H
	 DB     73H,73H,69H
	 DB     6FH,6EH,20H
	 DB     6FH,72H,20H
	 DB     00H,01H,20H
	 DB     00H,04H,20H
	 DB     6FH,72H,00H
	 DB     08H,00H,0FH
	 DB     20H,4CH,6FH
	 DB     6EH,67H,00H
	 DB     01H,20H,4EH
	 DB     75H,6DH,65H
	 DB     72H,69H,63H
	 DB     20H,53H,63H
	 DB     61H,6CH,61H
	 DB     72H,20H,00H
	 DB     04H,20H,6FH
	 DB     72H,00H,08H
	 DB     00H,0FH,20H
	 DB     53H,68H,6FH
	 DB     72H,74H,00H
	 DB     01H,20H,41H
	 DB     72H,72H,61H
	 DB     79H,20H,00H
	 DB     04H,00H,01H
	 DB     20H,4EH,75H
	 DB     6DH,65H,72H
	 DB     69H,63H,20H
	 DB     41H,72H,72H
	 DB     61H,79H,00H
	 DB     01H,20H,41H
	 DB     6CH,70H,68H
	 DB     61H,20H,41H
	 DB     72H,72H,61H
	 DB     79H,00H,01H
	 DB     20H,41H,6CH
	 DB     70H,68H,61H
	 DB     20H,00H,04H
	 DB     44H,4FH,00H
	 DB     0EH,20H,4DH
	 DB     61H,74H,63H
	 DB     68H,65H,64H
	 DB     20H,77H,69H
	 DB     74H,68H,20H
	 DB     45H,4EH,44H
	 DB     44H,4FH,53H
	 DB     74H,61H,72H
	 DB     74H,20H,3EH
	 DB     20H,45H,6EH
	 DB     64H,2DH,6FH
	 DB     66H,2DH,46H
	 DB     69H,6CH,65H
	 DB     00H,02H,00H
	 DB     08H,4EH,6FH
	 DB     20H,00H,00H
	 DB     20H,50H,72H
	 DB     6FH,74H,65H
	 DB     63H,74H,65H
	 DB     64H,00H,07H
	 DB     20H,4CH,69H
	 DB     6EH,65H,00H
	 DB     05H,20H,6FH
	 DB     72H,20H,43H
	 DB     4FH,4EH,54H
	 DB     49H,4EH,55H
	 DB     45H,20H,00H
	 DB     02H,20H,53H
	 DB     65H,63H,74H
	 DB     6FH,72H,20H
	 DB     41H,64H,64H
	 DB     72H,65H,73H
	 DB     73H,20H,6FH
	 DB     72H,20H,4EH
	 DB     6FH,20H,50H
	 DB     6CH,61H,74H
	 DB     74H,65H,72H
	 DB     00H,07H,20H
	 DB     44H,45H,46H
	 DB     46H,4EH,27H
	 DB     20H,53H,75H
	 DB     62H,72H,6FH
	 DB     75H,74H,69H
	 DB     6EH,65H,00H
	 DB     07H,20H,46H
	 DB     4EH,27H,73H
	 DB     20H,4EH,65H
	 DB     73H,74H,65H
	 DB     64H,00H,0FH
	 DB     20H,44H,65H
	 DB     65H,70H,6CH
	 DB     79H,4EH,45H
	 DB     58H,54H,20H
	 DB     77H,69H,74H
	 DB     68H,6FH,75H
	 DB     74H,20H,46H
	 DB     4FH,52H,45H
	 DB     41H,44H,59H
	 DB     20H,28H,42H
	 DB     41H,53H,49H
	 DB     43H,2DH,32H
	 DB     29H,20H,50H
	 DB     41H,52H,54H
	 DB     49H,54H,49H
	 DB     4FH,4EH,6FH
	 DB     74H,20H,45H
	 DB     6EH,6FH,75H
	 DB     67H,68H,20H
	 DB     4DH,65H,6DH
	 DB     6FH,72H,79H
	 DB     52H,45H,54H
	 DB     55H,52H,4EH
	 DB     20H,77H,69H
	 DB     74H,68H,6FH
	 DB     75H,74H,20H
	 DB     47H,4FH,53H
	 DB     55H,42H,75H
	 DB     66H,66H,65H
	 DB     72H,20H,45H
	 DB     78H,63H,65H
	 DB     65H,64H,65H
	 DB     64H,00H,02H
	 DB     20H,49H,6DH
	 DB     61H,67H,65H
	 DB     00H,02H,20H
	 DB     00H,11H,20H
	 DB     4FH,70H,65H
	 DB     72H,61H,6EH
	 DB     64H,49H,6EH
	 DB     63H,6FH,6DH
	 DB     70H,61H,74H
	 DB     69H,62H,6CH
	 DB     65H,20H,4FH
	 DB     70H,65H,72H
	 DB     61H,6EH,64H
	 DB     20H,44H,00H
	 DB     10H,73H,00H
	 DB     02H,20H,4DH
	 DB     69H,63H,72H
	 DB     6FH,63H,6FH
	 DB     6DH,6DH,61H
	 DB     6EH,64H,00H
	 DB     01H,20H,42H
	 DB     75H,66H,66H
	 DB     65H,72H,20H
	 DB     00H,04H,00H
	 DB     02H,20H,44H
	 DB     65H,76H,69H
	 DB     63H,65H,00H
	 DB     12H,49H,6EH
	 DB     74H,65H,72H
	 DB     72H,75H,70H
	 DB     74H,20H,54H
	 DB     61H,62H,6CH
	 DB     65H,20H,46H
	 DB     75H,6CH,6CH
	 DB     00H,02H,20H
	 DB     41H,72H,72H
	 DB     61H,79H,20H
	 DB     44H,00H,10H
	 DB     73H,20H,6FH
	 DB     72H,20H,00H
	 DB     04H,20H,4CH
	 DB     65H,6EH,67H
	 DB     74H,68H,4EH
	 DB     6FH,6EH,63H
	 DB     6FH,6DH,6DH
	 DB     6FH,6EH,20H
	 DB     00H,04H,73H
	 DB     20H,41H,6CH
	 DB     72H,65H,61H
	 DB     64H,79H,20H
	 DB     44H,65H,66H
	 DB     69H,6EH,65H
	 DB     64H,43H,6FH
	 DB     6DH,6DH,6FH
	 DB     6EH,20H,00H
	 DB     04H,20H,52H
	 DB     65H,71H,75H
	 DB     69H,72H,65H
	 DB     64H,00H,07H
	 DB     20H,00H,04H
	 DB     53H,75H,62H
	 DB     73H,63H,72H
	 DB     69H,70H,74H
	 DB     20H,4FH,75H
	 DB     74H,20H,6FH
	 DB     66H,20H,52H
	 DB     61H,6EH,67H
	 DB     65H,00H,02H
	 DB     20H,53H,54H
	 DB     52H,20H,00H
	 DB     0BH,00H,02H
	 DB     20H,46H,69H
	 DB     65H,6CH,64H
	 DB     2FH,44H,65H
	 DB     6CH,69H,6DH
	 DB     69H,74H,65H
	 DB     72H,00H,12H
	 DB     00H,02H,20H
	 DB     52H,65H,64H
	 DB     00H,10H,55H
	 DB     6EH,64H,65H
	 DB     72H,66H,6CH
	 DB     6FH,77H,44H
	 DB     69H,76H,69H
	 DB     73H,69H,6FH
	 DB     6EH,20H,62H
	 DB     79H,20H,00H
	 DB     0CH,20H,52H
	 DB     61H,69H,73H
	 DB     65H,64H,20H
	 DB     74H,6FH,20H
	 DB     00H,0DH,20H
	 DB     50H,6FH,77H
	 DB     65H,72H,00H
	 DB     0DH,00H,05H
	 DB     20H,52H,61H
	 DB     69H,73H,65H
	 DB     64H,20H,74H
	 DB     6FH,20H,4EH
	 DB     6FH,6EH,2DH
	 DB     49H,6EH,74H
	 DB     65H,67H,65H
	 DB     72H,20H,50H
	 DB     6FH,77H,65H
	 DB     72H,4CH,4FH
	 DB     47H,20H,6FH
	 DB     66H,20H,00H
	 DB     0CH,20H,2FH
	 DB     20H,6FH,72H
	 DB     20H,5EH,20H
	 DB     5AH,65H,72H
	 DB     6FH,4CH,4FH
	 DB     47H,20H,6FH
	 DB     66H,20H,00H
	 DB     0DH,00H,08H
	 DB     49H,6EH,73H
	 DB     75H,66H,66H
	 DB     69H,63H,69H
	 DB     65H,6EH,74H
	 DB     20H,44H,61H
	 DB     74H,61H,00H
	 DB     09H,53H,69H
	 DB     6EH,67H,75H
	 DB     6CH,61H,72H
	 DB     20H,00H,11H
	 DB     00H,0EH,20H
	 DB     53H,71H,75H
	 DB     61H,72H,65H
	 DB     20H,52H,6FH
	 DB     6FH,74H,20H
	 DB     6FH,66H,20H
	 DB     00H,0DH,00H
	 DB     08H,00H,02H
	 DB     20H,49H,4EH
	 DB     50H,55H,54H
	 DB     20H,44H,61H
	 DB     74H,61H,57H
	 DB     72H,6FH,6EH
	 DB     67H,20H,00H
	 DB     04H,20H,54H
	 DB     79H,70H,65H
	 DB     00H,02H,00H
	 DB     05H,00H,03H
	 DB     20H,50H,61H
	 DB     72H,74H,69H
	 DB     74H,69H,6FH
	 DB     6EH,20H,52H
	 DB     65H,66H,65H
	 DB     72H,65H,6EH
	 DB     63H,65H,50H
	 DB     72H,69H,6EH
	 DB     74H,20H,44H
	 DB     72H,69H,76H
	 DB     65H,72H,00H
	 DB     09H,00H,03H
	 DB     20H,50H,61H
	 DB     73H,73H,77H
	 DB     6FH,72H,64H
	 DB     00H,0AH,00H
	 DB     0EH,20H,4FH
	 DB     70H,65H,6EH
	 DB     00H,0AH,20H
	 DB     46H,75H,6CH
	 DB     6CH,00H,0AH
	 DB     00H,0EH,20H
	 DB     46H,6FH,75H
	 DB     6EH,64H,49H
	 DB     6EH,64H,65H
	 DB     78H,20H,46H
	 DB     75H,6CH,6CH
	 DB     43H,61H,74H
	 DB     61H,6CH,6FH
	 DB     67H,20H,45H
	 DB     6EH,64H,00H
	 DB     09H,4EH,6FH
	 DB     20H,45H,6EH
	 DB     64H,2DH,6FH
	 DB     66H,2DH,00H
	 DB     0AH,00H,0EH
	 DB     20H,53H,63H
	 DB     72H,61H,74H
	 DB     63H,68H,65H
	 DB     64H,57H,72H
	 DB     6FH,6EH,67H
	 DB     20H,52H,65H
	 DB     63H,6FH,72H
	 DB     64H,20H,54H
	 DB     79H,70H,65H
	 DB     53H,65H,63H
	 DB     74H,6FH,72H
	 DB     20H,41H,64H
	 DB     64H,72H,65H
	 DB     73H,73H,20H
	 DB     42H,65H,79H
	 DB     6FH,6EH,64H
	 DB     20H,45H,6EH
	 DB     64H,2DH,6FH
	 DB     66H,2DH,00H
	 DB     0AH,20H,41H
	 DB     6CH,72H,65H
	 DB     61H,64H,79H
	 DB     20H,45H,78H
	 DB     69H,73H,74H
	 DB     73H,44H,69H
	 DB     73H,6BH,20H
	 DB     44H,72H,69H
	 DB     76H,65H,00H
	 DB     0EH,20H,52H
	 DB     65H,61H,64H
	 DB     79H,54H,69H
	 DB     6DH,65H,6FH
	 DB     75H,74H,00H
	 DB     09H,44H,69H
	 DB     73H,6BH,20H
	 DB     43H,6FH,6EH
	 DB     74H,72H,6FH
	 DB     6CH,6CH,65H
	 DB     72H,00H,09H
	 DB     44H,65H,76H
	 DB     69H,63H,65H
	 DB     00H,09H,4CH
	 DB     52H,43H,00H
	 DB     09H,20H,20H
	 DB     20H,77H,69H
	 DB     64H,74H,68H
	 DB     4DH,44H,54H
	 DB     3AH,50H,44H
	 DB     54H,3AH,0DH
	 DB     45H,4EH,44H
	 DB     20H,50H,52H
	 DB     4FH,47H,52H
	 DB     41H,4DH,0DH
	 DB     46H,52H,45H
	 DB     45H,20H,53H
	 DB     50H,41H,43H
	 DB     45H,3DH,54H
	 DB     52H,41H,4EH
	 DB     53H,46H,45H
	 DB     52H,20H,54H
	 DB     4FH,45H,52H
	 DB     52H,4FH,52H
	 DB     20H,49H,4EH
	 DB     20H,53H,45H
	 DB     43H,54H,4FH
	 DB     52H,65H,61H
	 DB     64H,2DH,41H
	 DB     66H,74H,65H
	 DB     72H,2DH,57H
	 DB     72H,69H,74H
	 DB     65H,00H,09H
	 DB     0DH,0DH,4EH
	 DB     41H,4DH,45H
	 DB     20H,20H,20H
	 DB     20H,20H,54H
	 DB     59H,50H,45H
	 DB     20H,20H,53H
	 DB     54H,41H,52H
	 DB     54H,20H,20H
	 DB     20H,45H,4EH
	 DB     44H,20H,20H
	 DB     20H,55H,53H
	 DB     45H,44H,20H
	 DB     20H,20H,46H
	 DB     52H,45H,45H
	 DB     0DH,43H,55H
	 DB     52H,52H,45H
	 DB     4EH,54H,20H
	 DB     45H,4EH,44H
	 DB     20H,20H,20H
	 DB     3DH,0DH,49H
	 DB     4EH,44H,45H
	 DB     58H,20H,53H
	 DB     45H,43H,54H
	 DB     4FH,52H,53H
	 DB     20H,3DH,0DH
	 DB     45H,4EH,44H
	 DB     20H,43H,41H
	 DB     54H,2EH,20H
	 DB     41H,52H,45H
	 DB     41H,20H,3DH
	 DB     50H,72H,6FH
	 DB     67H,72H,61H
	 DB     6DH,4DH,69H
	 DB     73H,73H,69H
	 DB     6EH,67H,49H
	 DB     6CH,6CH,65H
	 DB     67H,61H,6CH
	 DB     49H,6EH,76H
	 DB     61H,6CH,69H
	 DB     64H,56H,61H
	 DB     72H,69H,61H
	 DB     62H,6CH,65H
	 DB     20H,4EH,75H
	 DB     6DH,62H,65H
	 DB     72H,44H,75H
	 DB     6DH,6DH,79H
	 DB     20H,66H,6FH
	 DB     72H,20H,61H
	 DB     74H,6FH,6DH
	 DB     20H,30H,30H
	 DB     30H,36H,55H
	 DB     6EH,64H,65H
	 DB     66H,69H,6EH
	 DB     65H,64H,20H
	 DB     56H,61H,6CH
	 DB     75H,65H,20H
	 DB     45H,78H,63H
	 DB     65H,65H,64H
	 DB     73H,20H,00H
	 DB     14H,00H,09H
	 DB     20H,45H,72H
	 DB     72H,6FH,72H
	 DB     41H,72H,67H
	 DB     75H,6DH,65H
	 DB     6EH,74H,4EH
	 DB     65H,67H,61H
	 DB     74H,69H,76H
	 DB     65H,20H,4EH
	 DB     6FH,74H,20H
	 DB     54H,6FH,6FH
	 DB     69H,6DH,65H
	 DB     6EH,73H,69H
	 DB     6FH,6EH,4DH
	 DB     61H,74H,72H
	 DB     69H,78H,20H
	 DB     53H,70H,65H
	 DB     63H,69H,66H
	 DB     69H,63H,61H
	 DB     74H,69H,6FH
	 DB     6EH,46H,6FH
	 DB     72H,6DH,61H
	 DB     74H,2EH,2EH
MSGADD   DB     "PHT:"
MSGADD1  DB     0DH,0DH,"NAME",5 DUP (20H)
	 DB     "TYPE",2 DUP (20H)
	 DB     "START",5 DUP (20H)
	 DB     "END",7 DUP (20H)
	 DB     "USED",6 DUP (20H)
	 DB     "FREE",0DH

;     Error messages

ERRORMSG LABEL     NEAR                  ;error messages

err_0     DB        08H
	  DD   OFFSET MSGBASE+0+3*000H   ;System Error
err_001   DB        11H
	  DD   OFFSET MSGBASE+1+3*0B7H   ;Not Enough Memory
err_002   DB        11H
	  DD   OFFSET MSGBASE+1+3*0B7H   ;Not Enough Memory
err_003   DB        11H
	  DD   OFFSET MSGBASE+1+3*0B7H   ;Not Enough Memory
err_004   DB        17H
	  DD   OFFSET MSGBASE+2+3*02EH   ;Operator Stack Overflow
err_005   DB        0BH
	  DD   OFFSET MSGBASE+2+3*002H   ;Line Too Long
err_006   DB        0CH
	  DD   OFFSET MSGBASE+1+3*086H   ;Program Protected
err_007   DB        14H
	  DD   OFFSET MSGBASE+1+3*006H   ;Illegal in Immediate Mode
err_008   DB        11H
	  DD   OFFSET MSGBASE+0+3*00DH   ;Statement Illegal Here
err_009   DB        0DH
	  DD   OFFSET MSGBASE+2+3*012H   ;Program Not Resolved
err_010   DB        06H
	  DD   OFFSET MSGBASE+0+3*017H   ;Missing '('
err_11    DB        06H
	  DD   OFFSET MSGBASE+0+3*019H   ;Missing ')'
err_12    DB        06H
	  DD   OFFSET MSGBASE+0+3*01BH   ;Missing '='
err_13    DB        06H
	  DD   OFFSET MSGBASE+0+3*01DH   ;Missing ','
err_14    DB        06H
	  DD   OFFSET MSGBASE+0+3*01FH   ;Missing '*'
err_15    DB        06H
	  DD   OFFSET MSGBASE+0+3*021H   ;Missing '>'
err_16    DB        09H
	  DD   OFFSET MSGBASE+0+3*023H   ;Missing Letter
err_17    DB        0CH
	  DD   OFFSET MSGBASE+0+3*026H   ;Missing Hex Digit
err_18    DB        16H
	  DD   OFFSET MSGBASE+0+3*02AH   ;Missing Relational Operator
err_19    DB        07H
	  DD   OFFSET MSGBASE+1+3*036H   ;Missing Word
err_20    DB        13H
	  DD   OFFSET MSGBASE+2+3*038H   ;End of Valid Syntax
err_21    DB        09H
	  DD   OFFSET MSGBASE+0+3*03FH   ;Missing Line Number
err_22    DB        0AH
	  DD   OFFSET MSGBASE+0+3*045H   ;Illegal PLOT Argument
err_23    DB        0AH
	  DD   OFFSET MSGBASE+0+3*04BH   ;Invalid Literal
err_24    DB        16H
	  DD   OFFSET MSGBASE+1+3*04EH   ;Illegal Expression or Missing Variable
err_25    DB        14H
	  DD   OFFSET MSGBASE+2+3*059H   ;Missing Numeric Scalar Variable
err_26    DB        0BH
	  DD   OFFSET MSGBASE+2+3*064H   ;Missing Array Variable
err_27    DB        10H
	  DD   OFFSET MSGBASE+1+3*068H   ;Missing Numeric Array
err_28    DB        0EH
	  DD   OFFSET MSGBASE+2+3*06DH   ;Missing Alpha Array
err_29    DB        0BH
	  DD   OFFSET MSGBASE+1+3*072H   ;Missing Alpha Variable
err_30    DB        01H
	  DD   OFFSET MSGBASE+2+3*004H   ;
err_31    DB        17H
	  DD   OFFSET MSGBASE+0+3*076H   ;DO Not Matched with ENDDO
err_32    DB        0BH
	  DD   OFFSET MSGBASE+2+3*07DH   ;Start > End
err_33    DB        0FH
	  DD   OFFSET MSGBASE+0+3*040H   ;Line Number Conflict
err_34    DB        04H
	  DD   OFFSET MSGBASE+0+3*084H   ;Illegal Value
err_35    DB        05H
	  DD   OFFSET MSGBASE+1+3*085H   ;No Program
err_36    DB        18H
	  DD   OFFSET MSGBASE+1+3*08AH   ;Undefined Line Number or CONTINUE Illegal
err_37    DB        14H
	  DD   OFFSET MSGBASE+0+3*09CH   ;Undefined DEFFN' Subroutine
err_38    DB        05H
	  DD   OFFSET MSGBASE+2+3*0A2H   ;Undefined FN
err_39    DB        14H
	  DD   OFFSET MSGBASE+2+3*0A3H   ;FN's Nested Too Deeply
err_40    DB        10H
	  DD   OFFSET MSGBASE+1+3*0AAH   ;NEXT without FOR
err_41    DB        14H
	  DD   OFFSET MSGBASE+0+3*0BDH   ;RETURN without GOSUB
err_42    DB        08H
	  DD   OFFSET MSGBASE+1+3*0C8H   ;Illegal Image
err_43    DB        0DH
	  DD   OFFSET MSGBASE+0+3*0CBH   ;Illegal Matrix Operand
err_44    DB        0BH
	  DD   OFFSET MSGBASE+1+3*145H   ;Matrix Not Square
err_45    DB        19H
	  DD   OFFSET MSGBASE+1+3*0CFH   ;Incompatible Operand Dimensions
err_46    DB        0FH
	  DD   OFFSET MSGBASE+2+3*0D7H   ;Illegal Microcommand
err_47    DB        0CH
	  DD   OFFSET MSGBASE+2+3*0DCH   ;Missing Buffer Variable
err_48    DB        0BH
	  DD   OFFSET MSGBASE+2+3*0E0H   ;Illegal Device Specification
err_49    DB        14H
	  DD   OFFSET MSGBASE+1+3*0E4H   ;Interrupt Table Full
err_50    DB        1AH
	  DD   OFFSET MSGBASE+0+3*0EBH   ;Illegal Array Dimensions or Variable Length
err_51    DB        0FH
	  DD   OFFSET MSGBASE+2+3*05FH   ;Variable or Value Too Short
err_52    DB        0EH
	  DD   OFFSET MSGBASE+0+3*055H   ;Variable or Value Too Long
err_53    DB        1DH
	  DD   OFFSET MSGBASE+2+3*0F3H   ;Noncommon Variables Already Defined
err_54    DB        12H
	  DD   OFFSET MSGBASE+1+3*0FDH   ;Common Variable Required
err_55    DB        05H
	  DD   OFFSET MSGBASE+1+3*103H   ;Undefined Variable
err_56    DB        16H
	  DD   OFFSET MSGBASE+0+3*105H   ;Subscript Out of Range
err_57    DB        09H
	  DD   OFFSET MSGBASE+1+3*10CH   ;Illegal STR Argument
err_58    DB        14H
	  DD   OFFSET MSGBASE+1+3*10FH   ;Illegal Field/Delimiter Specification
err_59    DB        08H
	  DD   OFFSET MSGBASE+0+3*116H   ;Illegal Redimension
err_60    DB        09H
	  DD   OFFSET MSGBASE+2+3*118H   ;Underflow
err_61    DB        08H
	  DD   OFFSET MSGBASE+2+3*033H   ;Overflow
err_62    DB        0EH
	  DD   OFFSET MSGBASE+2+3*11BH   ;Division by Zero
err_63    DB        0EH
	  DD   OFFSET MSGBASE+2+3*133H   ;Zero / or ^ Zero
err_64    DB        15H
	  DD   OFFSET MSGBASE+2+3*11FH   ;Zero Raised to Negative Power
err_65    DB        20H
	  DD   OFFSET MSGBASE+2+3*126H   ;Negative Number Raised to Non-Integer Power
err_66    DB        13H
	  DD   OFFSET MSGBASE+0+3*147H   ;Square Root of Negative Value
err_67    DB        09H
	  DD   OFFSET MSGBASE+1+3*131H   ;LOG of Zero
err_68    DB        0BH
	  DD   OFFSET MSGBASE+1+3*138H   ;LOG of Negative Value
err_69    DB        0AH
	  DD   OFFSET MSGBASE+2+3*047H   ;Argument Too Large
err_70    DB        11H
	  DD   OFFSET MSGBASE+0+3*13CH   ;Insufficient Data
err_71    DB        10H
	  DD   OFFSET MSGBASE+0+3*1F9H   ;Value Exceeds Format
err_72    DB        0BH
	  DD   OFFSET MSGBASE+1+3*142H   ;Singular Matrix
err_73    DB        0DH
	  DD   OFFSET MSGBASE+1+3*14DH   ;Illegal INPUT Data
err_74    DB        0DH
	  DD   OFFSET MSGBASE+2+3*151H   ;Wrong Variable Type
err_75    DB        04H
	  DD   OFFSET MSGBASE+0+3*156H   ;Illegal Number
err_76    DB        0FH
	  DD   OFFSET MSGBASE+1+3*0C3H   ;Buffer Exceeded
err_77    DB        16H
	  DD   OFFSET MSGBASE+1+3*157H   ;Invalid Partition Reference
err_78    DB        0EH
	  DD   OFFSET MSGBASE+2+3*15EH   ;Print Driver Error
err_79    DB        0BH
	  DD   OFFSET MSGBASE+1+3*163H   ;Invalid Password
err_80    DB        09H
	  DD   OFFSET MSGBASE+0+3*167H   ;File Not Open
err_81    DB        07H
	  DD   OFFSET MSGBASE+0+3*16AH   ;File Full
err_82    DB        0AH
	  DD   OFFSET MSGBASE+1+3*16CH   ;File Not Found
err_83    DB        11H
	  DD   OFFSET MSGBASE+2+3*18EH   ;File Already Exists
err_84    DB        0EH
	  DD   OFFSET MSGBASE+2+3*17AH   ;File Not Scratched
err_85    DB        0AH
	  DD   OFFSET MSGBASE+2+3*16FH   ;Index Full
err_86    DB        0DH
	  DD   OFFSET MSGBASE+0+3*173H   ;Catalog End Error
err_87    DB        0CH
	  DD   OFFSET MSGBASE+1+3*177H   ;No End-of-File
err_88    DB        11H
	  DD   OFFSET MSGBASE+1+3*17FH   ;Wrong Record Type
err_89    DB        1FH
	  DD   OFFSET MSGBASE+0+3*185H   ;Sector Address Beyond End-of-File
err_90    DB        11H
	  DD   OFFSET MSGBASE+1+3*19DH   ;Disk Controller Error
err_91    DB        12H
	  DD   OFFSET MSGBASE+1+3*194H   ;Disk Drive Not Ready
err_92    DB        09H
	  DD   OFFSET MSGBASE+1+3*19AH   ;Timeout Error
err_93    DB        04H
	  DD   OFFSET MSGBASE+2+3*1FDH   ;Format Error
err_94    DB        11H
	  DD   OFFSET MSGBASE+1+3*19DH   ;Disk Controller Error
err_95    DB        08H
	  DD   OFFSET MSGBASE+0+3*1A3H   ;Device Error
err_96    DB        06H
	  DD   OFFSET MSGBASE+1+3*140H   ;Data Error
err_97    DB        05H
	  DD   OFFSET MSGBASE+2+3*1A5H   ;LRC Error
err_98    DB        1FH
	  DD   OFFSET MSGBASE+2+3*091H   ;Illegal Sector Address or No Platter
err_99    DB        12H
	  DD   OFFSET MSGBASE+0+3*1BDH   ;Read-After-Write Error


;     System messages

SYSMSGS   LABEL NEAR

LDTMSG1   DB        08H
	  DD   OFFSET MSGBASE+1+3*1A7H   ;width
LDTMSG2   DB        03H
	  DD   OFFSET MSGBASE+1+3*1F2H   ;at
LDTMSG3   DB        03H
	  DD   OFFSET MSGBASE+0+3*007H   ;in
LDTMSG4   DB        03H
	  DD   OFFSET MSGBASE+2+3*122H   ;to
LDTMSG5   DB        04H
	  DD   OFFSET MSGBASE+0+3*1AAH   ;MDT:
LDTMSG6   DB        04H
	  DD   OFFSET MSGBASE+1+3*1ABH   ;PDT:
READYMSG  DB        19H
	  DD   OFFSET MSGBASE+1+3*0AFH   ;READY (BASIC-2) PARTITION
ENDMSG    DB        18H
	  DD   OFFSET MSGBASE+2+3*1ACH   ;.END PROGRAM.FREE SPACE=
TRNSFMSG  DB        0BH
	  DD   OFFSET MSGBASE+2+3*1B4H   ;TRANSFER TO
VRFYMSG   DB        0FH
	  DD   OFFSET MSGBASE+1+3*1B8H   ;ERROR IN SECTOR
LDCMSG7   DB        34H
	  DD   OFFSET MSGADD1            ;..NAME     TYPE  START   END   USED   FREE.
LDCMSG14  DB        10H
	  DD   OFFSET MSGBASE+1+3*1D6H   ;.INDEX SECTORS =
LDCMSG15  DB        10H
	  DD   OFFSET MSGBASE+2+3*1DBH   ;.END CAT. AREA =
LDCMSG16  DB        10H
	  DD   OFFSET MSGBASE+0+3*1D1H   ;.CURRENT END   =
LDTMSG17  DB        04H
	  DD   OFFSET MSGADD             ;PHT:

;     Substrings to atomize

MSGATOMS  LABEL NEAR

	  DB        07H
	  DD   OFFSET MSGBASE+0+3*1E1H   ;Program
	  DB        07H
	  DD   OFFSET MSGBASE+1+3*1E3H   ;Missing
	  DB        07H
	  DD   OFFSET MSGBASE+2+3*1E5H   ;Illegal
	  DB        07H
	  DD   OFFSET MSGBASE+0+3*1E8H   ;Invalid
	  DB        08H
	  DD   OFFSET MSGBASE+1+3*1EAH   ;Variable
	  DB        07H
	  DD   OFFSET MSGBASE+0+3*1EDH   ;Number
	  DB        13H
	  DD   OFFSET MSGBASE+1+3*1EFH   ;Dummy for atom 0006
	  DB        09H
	  DD   OFFSET MSGBASE+2+3*1F5H   ;Undefined
	  DB        06H
	  DD   OFFSET MSGBASE+2+3*1F8H   ;Value
	  DB        06H
	  DD   OFFSET MSGBASE+0+3*1FFH   ;Error
	  DB        04H
	  DD   OFFSET MSGBASE+2+3*082H   ;File
	  DB        08H
	  DD   OFFSET MSGBASE+0+3*201H   ;Argument
	  DB        04H
	  DD   OFFSET MSGBASE+0+3*137H   ;Zero
	  DB        08H
	  DD   OFFSET MSGBASE+2+3*203H   ;Negative
	  DB        04H
	  DD   OFFSET MSGBASE+1+3*206H   ;Not
	  DB        04H
	  DD   OFFSET MSGBASE+2+3*207H   ;Too
	  DB        08H
	  DD   OFFSET MSGBASE+0+3*209H   ;imension
	  DB        06H
	  DD   OFFSET MSGBASE+2+3*20BH   ;Matrix
	  DB        0EH
	  DD   OFFSET MSGBASE+2+3*20DH   ;Specification
	  DB        0CH
	  DD   OFFSET MSGBASE+0+3*080H   ;End-of-File
	  DB        06H
	  DD   OFFSET MSGBASE+1+3*212H   ;Format

CODE      ENDS
	  END
