;===========================================================================
;
;           P L O T   S T A T E M E N T
;
;===========================================================================
;           This module was originally M28
;===========================================================================
	    .XLIST
	    INCLUDE PREAMBLE
	    .LIST

PUBLIC    PLOT

EXTRN     CPBASE  :DWORD
EXTRN     ENBLDEVG:NEAR
EXTRN     ENDSTIOZ:NEAR
EXTRN     OBSTROBE:NEAR
EXTRN     ERROR_13:NEAR
EXTRN     ERROR_34:NEAR
EXTRN     MATHABS:NEAR
EXTRN     ALPHA1:NEAR
EXTRN     FLTBIN:NEAR
EXTRN     ERROR:NEAR
EXTRN     EXPR:NEAR

;    PLOT STATEMENT
;        GENERAL FORM
;          PLOT  ELEMENT [,ELEMENT ... ]
;   ELEMENT ::==[EXPR 0] < [EXPR 1], [EXPR 2], [TERM]>
;   TERM    ::==  LITERAL/
;                 ALPHA VARIABLE/
;                 D/U/R/S/C
;
;   EXPRESSION 0 - REPETITION FACTOR 1 TO 999
;   EXPRESSION 1 - SIGNED HORIZONTAL INCREMENT, DEFAULT 0
;   EXPRESSION 2 - SIGNED VERTICAL INCREMENT, DEFAULT 0
;
;   DEVICE USED IS WHATEVER WAS SELECTED BY THE
;                         SELECT PLOT STATEMENT.
;   DEVICE TYPE MAY BE EITHER 4 OR C:
;      IF DEVICE TYPE IS 4, THE 2200 'COMPATIBLE' STROBE
;      SEQUENCE IS USED, WITH UP TO 999 MOTION STROBES
;      SENT OUT.  MAXIMUM VALUE FOR EXPRESSIONS 1 AND 2
;      IS  -999 TO +999.
;
;      IF DEVICE TYPE IS C, THE 2600 VECTOR SEQUENCE IS
;      USED, WITH A 5 BYTE MOTION VECTOR SENT TO THE
;      PLOTTER.  THE FIRST BYTE IS AN ATOM, FOLLOWED BY
;      TWO 2-BYTE SIGNED (TWO'S COMPLEMENT) COUNTS.
;      THE MAXIMUM VALUE IS -32K TO +32K.
;  IF EXPRESSION 0 IS GREATER THAN 1, THE ENTIRE SEQUENCE
;      REPRESENTED BY THE ELEMENT IS REPEATEDLY STROBED.

PLOT:     TEST      SL,2               ;SKIP IF NOT X PHASE
	  JZ        PLOT1
	  MOV       EDX,DWORD PTR [CPBASE]
	  LEA       EDX,[EDX].@DEVPLOT
	  MOV       CX,WORD PTR [EDX]
	  XCHG      CH,CL
	  MOV       DX,CX
	  CALL      ENBLDEVG           ;ENABLE PLOT DEVICE

PLOT1:    MOV       EDX,1
	  CMP       BYTE PTR [TEXTPTR],'<'
	  JNE       PLOT1X
	  INC       TEXTPTR
	  JMP       PLOT1?pd8          ;< FOUND,  SKIP
PLOT1X:
	  CALL      EXPR               ;EVALUATE EXPRESSION 0 (REP COUNT)
	  CMP       BYTE PTR [TEXTPTR],'<'
	  JNE       PLOTERR
	  INC       TEXTPTR
	  TEST      SL,2               ;SKIP IF NOT X PHASE
	  JZ        PLOT2
	  CALL      FLTBIN             ;CONVERT TO TWO BYTE BINARY COUNT
	  MOV       AL,R2
	  AND       AL,0FH
	  CMP       AL,02H             ;ERROR IF ZERO
	  JNE       PLOTERR
	  CMP       DWORD PTR [A0],0FFFFH
	  JA        PLOTERR
	  MOV       EDX,DWORD PTR A0

PLOT1?pd8:
	  TEST      SL,2               ;SKIP IF NOT X PHASE
	  JZ        PLOT2
	  MOV       AUX07,EDX          ;AUX 7 IS REPETITION COUNT
	  MOV       EDX,0
	  MOV       AUX08,EDX          ;AUX 8 IS HORIZONTAL VECTOR (DEFAULT = 0000)
	  MOV       AUX09,EDX          ;AUX 9 IS VERTICAL VECTOR (DEFAULT = 0000)
	  MOV       AUX0A,EDX          ;AUX A IS ADDRESS OF DATA
	  MOV       AUX0B,EDX          ;AUX B IS LENGTH OF DATA
PLOT2:    CMP       BYTE PTR [TEXTPTR],','
	  JNE       PLOT2X
	  INC       TEXTPTR            ;F, SKIP
	  JMP       PLOT3
PLOT2X:
	  CALL      EXPR               ;EVALUATE EXPRESSION 1
	  CMP       BYTE PTR [TEXTPTR],','
	  JNE       ERROR_13           ;BRANCH IF NOT MATCH
	  INC       TEXTPTR
	  TEST      SL,2               ;SKIP IF NOT X PHASE
	  JZ        PLOT3
	  CALL      FLTBINS            ;CONVERT TO TWO BYTE BINARY
	  MOVZX     EDX,WORD PTR R0
	  MOV       AUX08,EDX          ;SAVE HORIZONTAL COMPONENT
PLOT3:    CMP       BYTE PTR [TEXTPTR],','
	  JNE       PLOT3X
	  INC       TEXTPTR            ;F, SKIP
	  JMP       PLOT4
PLOT3X:
	  CALL      EXPR               ;EVALUATE EXPRESSION 2
	  CMP       BYTE PTR [TEXTPTR],','
	  JNE       ERROR_13           ;BRANCH IF NOT MATCH
	  INC       TEXTPTR
	  TEST      SL,2               ;SKIP IF NOT X PHASE
	  JZ        PLOT4
	  CALL      FLTBINS            ;CONVERT TO TWO BYTE BINARY
	  MOVZX     EDX,WORD PTR R0
	  MOV       AUX09,EDX          ;SAVE VERTICAL COMPONENT
PLOT4:    MOV       R7,20H             ;FLAG POSSIBLE NULL OR ALPHA PARAMETER
	  CMP       BYTE PTR [TEXTPTR],'>'
	  JNE       PLOT4X
	  INC       TEXTPTR            ;SKIP IF FOUND (NULL PARAMETER)
	  JMP       PLOT6
PLOT4X:
	  MOV       K,4                ;(CURRENT LENGTH)
	  CALL      ALPHA1             ;LOOK FOR ALPHA VARIABLE OR LITERAL
	  TEST      R0,0FH             ;NF, SKIP
	  JNZ       PLOT5
	  TEST      SL,2               ;SKIP IF NOT X PHASE
	  JZ        PLOT5?pd9
	  mov       edx,vsptr          ;POP OUT ADDRESS & LENGTH
	  mov       eax,dword ptr [edx]
	  mov       dword ptr [a0],eax
	  mov       edx,dword ptr [edx+4]
	  add       vsptr,8
	  MOV       AUX0A,EDX          ;AUX A = ADDRESS
	  MOV       EDX,DWORD PTR A0
	  MOV       AUX0B,EDX          ;AUX B = LENGTH
	  MOV       R7,20H
	  JMP       PLOT5?pd9

PLOT5:    MOV       EDX,TEXTPTR        ;GET NEXT CHARACTER (U,D,S,C,R)
	  MOV       CH,BYTE PTR [EDX]
	  INC       TEXTPTR
	  MOV       AH,CH
	  AND       AH,0F0H
	  CMP       AH,40H             ;SKIP IF NOT 4X (A-O)
	  JNE       PLOT5?pd2
	  MOV       R7,0DH
	  CMP       CH,43H             ;CODE IS 0D FOR 'C'
	  JE        PLOT5?pd9
	  MOV       R7,0FH
	  CMP       CH,44H
	  JE        PLOT5?pd9          ;CODE IF 0F FOR 'D'

PLOTERR:  MOV       R0,22H
	  JMP       ERROR              ;ILLEGAL PLOT ARGUMENT

PLOT5?pd2:
	  CMP       AH,50H             ;ERROR IF NOT 5X (P-Z)
	  JNE       PLOTERR
	  MOV       R7,0AH
	  CMP       CH,53H             ;CODE IS  0A FOR 'S'
	  JE        PLOT5?pd9
	  MOV       R7,0EH
	  CMP       CH,55H             ;CODE IS  0E FOR 'U'
	  JE        PLOT5?pd9
	  CMP       CH,52H             ;ANYTHING OTHER THAN R IS AN ERROR
	  JNE       PLOTERR
	  MOV       R7,0FAH            ;CODE IS  FA FOR 'R'
PLOT5?pd9:
	  CMP       BYTE PTR [TEXTPTR],'>'
	  JNE       PLOTERR            ;ERROR IF MISSING
	  INC       TEXTPTR
PLOT6:    TEST      SL,2               ;DONE WITH THIS ELEMENT IF NOT X PHASE
	  JZ        PLOT91
PLOT6?pd08:
	  MOV       K,0F0H
	  CALL      OBSTROBE           ;STROBE OUT THE PLOT MODE CHARACTER
	  CALL      VECTOR             ;SEND OUT THE VECTOR, IF ANY
	  MOV       AH,R7
	  AND       AH,0F0H
	  CMP       AH,20H             ;SKIP IF U,D,S,C, OR R PARAMETER
	  JNE       PLOT7
	  MOV       K,0FEH
	  CALL      OBSTROBE           ;STROBE OUT THE OTHER PLOT MODE CHARACTER
	  MOV       EDX,AUX0B
	  MOV       DWORD PTR [A0],EDX     ;F1F0 = LENGTH OF ALPHA STRING
	  CMP       DWORD PTR [A0],0    ;SKIP IF NULL STRING
	  JE        PLOT6?pd3
	  CMP       DWORD PTR [A0],01H
	  JE        PLOT6?pd2          ;SKIP IF EXACTLY 1 BYTE
PLOT6?pd1:
	  MOV       K,0F5H
	  CALL      OBSTROBE           ;STROBE OUT A PRINT MODE CHAR (LEN>1)
PLOT6?pd2:
	  MOV       EDX,AUX0A          ;POINT AT FIRST CHARACTER OF STRING
	  MOV       DWORD PTR [A0],EDX   ;F1F0 = ADDRESS OF LAST+1
PLOT6?pd25:
	  MOV       CH,[EDX]
	  MOV       K,CH
	  INC       EDX
	  CALL      OBSTROBE           ;STROBE OUT NEXT CHARACTER
	  CMP       DWORD PTR [A0],EDX
	  JNE       PLOT6?pd25         ;LOOP TILL END OF STRING
	  JMP       PLOT6?pd3

PLOT7:    MOV       AL,R7
	  MOV       K,AL
	  CALL      OBSTROBE           ;STROBE OUT CODED PARAMETER

PLOT6?pd3:
	  DEC       AUX07
	  MOV       EDX,AUX07          ;DECREMENT REPETITION COUNT
	  CMP       EDX,0               ;LOOP BACK IF REPETITION COUNT NON-ZERO
	  JNE       PLOT6?pd08

PLOT9:    MOV       K,0F5H
	  CALL      OBSTROBE           ;STROBE OUT PRINT MODE CHARACTER
PLOT91:   CMP       BYTE PTR [TEXTPTR],','
	  JNE       ENDSTIOZ           ;END OF STATEMENT
	  INC       TEXTPTR            ;LOOP IF FOUND
	  JMP       PLOT1

;  VECTOR:  OUTPUT THE MOTION VECTOR, IN ONE OF TWO
;     POSSIBLE FORMATS.
;

VECTOR:   MOV       EDX,AUX08
	  MOV       WORD PTR R4,DX     ;F5F4 = HORIZONTAL COUNT
	  MOV       EDX,AUX09
	  MOV       WORD PTR R2,DX     ;F3F2 = VERTICAL COUNT
	  MOV       R1,88H             ;FLAG = BOTH NONZERO
	  CMP       WORD PTR R4,0
	  JNE       M28_01             ;SKIP IF HORIZ COUNT > 0
	  AND       R1,0FH             ;HORIZONTAL COUNT = 0
M28_01:
	  CMP       WORD PTR R2,0
	  JNE       M28_02             ;SKIP IF VERT COUNT >0
	  AND       R1,0F0H            ;VERTICAL COUNT  = 0
M28_02:
	  CMP       R1,0               ;RETURN IF BOTH ARE NOT ZERO
	  JE        VECTORR            ;B IF BOTH ARE ZERO
	  MOV       EDX,ACTDEV         ;CHECK ACTIVE DEVICE TYPE
	  MOV       AL,DH
	  AND       AL,0FH
	  CMP       AL,4               ;SKIP IF DEVICE TYPE 4 (2200 FORMAT)
	  JE        VECTOR2

;  2600 VECTOR FORMAT
;

	  MOV       K,0E0H
	  CALL      OBSTROBE           ;STROBE E0, THE VECTOR ATOM
	  MOV       AL,R5
	  MOV       K,AL
	  CALL      OBSTROBE  
	  MOV       AL,R4
	  MOV       K,AL
	  CALL      OBSTROBE           ;STROBE OUT HORIZONTAL COUNT
	  MOV       AL,R3
	  MOV       K,AL
	  CALL      OBSTROBE  
	  MOV       AL,R2
	  MOV       K,AL
	  JMP       OBSTROBE          ;STROBE OUT VERTICAL COUNT

VECTOR2:  TEST      R5,80H            ;SKIP IF X IS NEGATIVE
	  JNZ       VECTOR21
	  XOR       R1,90H            ;FLAG X >= 0
	  NEG       WORD PTR R4       ;COMPLEMENT X
VECTOR21: TEST      R3,80H            ;SKIP IF Y IS NEGATIVE
	  JNZ       VECTOR22
	  XOR       R1,09             ;FLAG Y >= 0
	  NEG       WORD PTR R2       ;COMPLEMENT Y

;  EACH COMPONENT IS LIMITED TO + OR - 999 (DECIMAL)
;

VECTOR22: MOV       DX,-999          ;(BINARY -999)
	  MOV       AL,R1
	  AND       AL,0F0H
	  CMP       AL,90H
	  JE        M28_04
	  CMP       WORD PTR R4,DX
	  JB        PLOTERRA         ;ERROR IF HORIZONTAL ARGUMENT > 999
M28_04:
	  MOV       AL,R1
	  AND       AL,0FH
	  CMP       AL,9             ;SKIP IF PURELY VERTICAL MOVEMENT
	  JE        VECTOR5
	  CMP       WORD PTR R2,DX
	  JB        PLOTERRA         ;ERROR IF VERTICAL ARGUMENT > 999

;  NOW F5F4 REPRESENTS THE TWO'S COMPLEMENT OF THE
;      HORIZONTAL DISTANCE TO BE MOVED
;  F3F2  REPRESENTS THE TWO'S COMPLEMENT OF THE VERTICAL
;      DISTANCE TO BE MOVED
;  THE TWO DIGITS OF F1 REPRESENT THE SIGNS OF THE TWO
;      DISTANCES, AS FOLLOWS (HIGH DIGIT = HORIZONTAL)
;           9 - ZERO MOTION
;           8 - NEGATIVE DIRECTION
;           1 - POSITIVE DIRECTION

	  MOV       AL,R1
	  AND       AL,0F0H
	  CMP       AL,90H
	  JE        VECTOR61         ;SKIP IF PURELY VERTICAL MOVEMENT

;  MOTION HAS TWO COMPONENTS;  TRAVEL AT 45 DEGREES
;

	  TEST      R1,80H           ;SKIP IF LEFTWARD MOTION (X IS NEG)
	  JNZ       VECTOR3
	  MOV       K,0F4H           ;F4 = + +
	  TEST      R1,08H
	  JZ        VECTOR4
	  MOV       K,0F2H           ;F2 = + -
	  JMP       VECTOR4   

VECTOR3:  MOV       K,0F3H           ;F3 = - +
	  TEST      R1,8
	  JZ        VECTOR4
	  MOV       K,0F1H          ;F1 = - -
VECTOR4:  CALL      OBSTROBE        ;STROBE OUT ONE 45 DEGREE MOVEMENT
	  ADD       WORD PTR R4,1   ;INCREMENT THE HORIZONTAL COUNT
	  SETC      AH
	  ADD       WORD PTR R2,1   ;INCREMENT THE VERTICAL COUNT
	  SETC      AL
	  TEST      AH,01H
	  JNZ       VECTOR6         ;SKIP IF HORIZ COUNT = 0
	  TEST      AL,01H
	  JZ        VECTOR4         ;LOOP IF BOTH COUNTS NONZERO

;  VERTICAL COUNT = 0; HORIZONTAL IS NONZERO
;

VECTOR5:  MOV       AX,WORD PTR R4
	  MOV       WORD PTR R2,AX   ;MOVE HORIZONTAL COUNT TO VERTICAL LOC
	  MOV       K,0F9H           ;F9 = + 0
	  TEST      R1,80H
	  JZ        VECTOR66
	  MOV       K,0F8H           ;F8 = - 0
	  JMP       VECTOR66  

VECTOR6:  TEST      AL,01H           ;INCREMENT THE VERTICAL COUNT
	  JNZ       VECTORR          ;EXIT IF BOTH COUNTS ZERO

;  HORIZONTAL COUNT = 0;  VERTICAL IS NONZERO
;

VECTOR61: MOV       K,0F7H           ;F7 = 0 +
	  TEST      R1,08H
	  JZ        VECTOR66
	  MOV       K,0F6H           ;F6 = 0 -
VECTOR66: CALL      OBSTROBE  
	  ADD       WORD PTR R2,1    ;INCREMENT COUNTER
	  JNC       VECTOR66         ;LOOP TILL ZERO IS REACHED

VECTORR:  RET                        ;RETURN

PLOTERRA: JMP       PLOTERR


;  FLTBINS:  CONVERT FLT PT NUMBER TO TWO BYTE SIGNED
;     TWO'S COMPLEMENT BINARY VALUE.
;

FLTBINS:  MOV       EDX,VSPTR
	  MOV       AX,WORD PTR [EDX+6]
	  MOV       R7,00H
	  TEST      AX,8000H
	  JZ        FLTBINS_X
	  MOV       R7,10H           ;SAVE SIGN BIT
FLTBINS_X:
	  CALL      MATHABS          ;TAKE ABSOLUTE VALUE OF NUMBER
	  CALL      FLTBIN           ;CONVERT TO BINARY IN F1F0
	  CMP       DWORD PTR [A0],8000H
	  JA        FLTBINS2         ;ERROR IF >32K
	  MOV       EAX,DWORD PTR [A0]
	  MOV       WORD PTR [R0],AX
	  TEST      R7,10H
	  JZ        M28_06
	  NEG       WORD PTR R0      ;TAKE TWO'S COMPLEMENT IF NEGATIVE
M28_06:
	  RET

FLTBINS2: JMP       ERROR_34         ;ERROR,  SIGNED VALUE > 32K

CODE      ENDS
	  END
