;===========================================================================
;
;     R U N   /   C O N T I N U E  /  C L E A R   /   S T E P
;
;===========================================================================
;           This module was originally M03
;===========================================================================
            .XLIST
            INCLUDE PREAMBLE
            .LIST

PUBLIC    CLEAR
PUBLIC    CONTINUE
PUBLIC    ERRB_35
PUBLIC    ERRC_34
PUBLIC    RUN
PUBLIC    RUNB
PUBLIC    STEP

EXTRN     ACKDISCT  :NEAR
EXTRN     CLEARN    :NEAR
EXTRN     ENBLCO    :NEAR
EXTRN     ENBLNULL  :NEAR
EXTRN     ENDSTMTB  :NEAR
EXTRN     ENDSTMTC  :NEAR
EXTRN     ENDSTM10  :NEAR
EXTRN     ENTERF    :NEAR
EXTRN     ERRB_36   :NEAR
EXTRN     ERRC_20   :NEAR
EXTRN     ERRMSG    :NEAR
EXTRN     ERRORC    :NEAR
EXTRN     ERROR_08  :NEAR
EXTRN     ERROR_20  :NEAR
EXTRN     FINDFNC   :NEAR
EXTRN     FINDSTI   :NEAR
EXTRN     GR_THRDS  :NEAR
EXTRN     ICLEAR    :NEAR
EXTRN     INITSTKA  :NEAR
EXTRN     INITSTKS  :NEAR
EXTRN     INITTEXT  :NEAR
EXTRN     INITVAR   :NEAR
EXTRN     INTEGER   :NEAR
EXTRN     LINE_REF  :NEAR
EXTRN     LISTLINS  :NEAR
EXTRN     LN_RANGE  :NEAR
EXTRN     PMODE     :NEAR
EXTRN     PRNTCR    :NEAR
EXTRN     RDEFNADR  :NEAR
EXTRN     RDOFLAG   :NEAR
EXTRN     RESCHK    :NEAR
EXTRN     RFILE_    :NEAR
EXTRN     SCNCHSS   :NEAR
EXTRN     SLCT@OFF  :NEAR
EXTRN     START     :NEAR
EXTRN     TAPIMODE  :NEAR
EXTRN     TWRITE    :NEAR
EXTRN     TWRITESP  :NEAR
EXTRN     UNRESLVS  :NEAR
EXTRN     UNTHREAD  :NEAR
EXTRN     VJUMP     :NEAR
EXTRN     WRT@      :NEAR
EXTRN     WRTLINE_  :NEAR

EXTRN     @@VSEND   :DWORD
EXTRN     @@OSEND   :DWORD
EXTRN     CPBASE    :DWORD
EXTRN     PART_     :BYTE

;------------------------------------------------------------------------------;
;                                                                              ;
;     *************************                                                ;
;     *                       *                                                ;
;     *  MODULE-03: COMMANDS  *                                                ;
;     *                       *                                                ;
;     *************************                                                ;
;                                                                              ;
;                                                                              ;
;     DESCRIPTION: ROUTINES FOR PROCESSING COMMANDS                            ;
;                                                                              ;
;     ENTRY: TEXTPTR POINTS JUST AFTER VERB                                    ;
;            TEXTLPTR POINTS TO THREAD OF COMMAND LINE                         ;
;                                                                              ;
;------------------------------------------------------------------------------;


;     CONTINUE: PROCESS A 'CONTINUE' STATEMENT
;
;          STATEMENT FORMAT:  CONTINUE
;
;          ENTRY: CONTLN# = LINE NO. TO CONTINUE AT
;                 CONTINUE STMTCNT = STATEMENT ON LINE
;                                    TO CONTINUE AT
;
;     NOTE: CONTINUE & STEP ARE DISABLED WHEN PROGRAM IS
;           TERMINATED BY NORMAL COMPLETION, ERROR, END,
;           OR RESET (DONE IN 'ENDSTMT','INITSTKS')
;

CONTINUE: TEST      MODE,00000002H
          JNZ       CONTINUE1
          JMP       ERROR_08    ;STMT LEGAL ONLY IN IMMEDIATE EXECUTION
CONTINUE1:                      ;CR REQUIRED
          CMP       BYTE PTR [TEXTPTR],0DH
          JNE       ERROR_20    ;ERROR IF CR NOT FOUND
          INC       TEXTPTR
          TEST      SL,2
          JNZ       STEP1       ;BRANCH IF EXECUTION PHASE
          JMP       ENDSTMTC    ;E,R PHASE



;     STEP: PROCESS A 'STEP' COMMAND.  DISPLAYS THE NEXT
;           STATEMENT TO BE EXECUTED IF PROGRAM IS NOT
;           PROTECTED, AND THEN STARTS EXECUTION AT THAT
;           POINT.  IN MULTI-STATEMENT LINES, STATEMENTS
;           THAT HAVE ALREADY BEEN EXECUTED ARE NOT DIS-
;           PLAYED, BUT COLONS SEPARATING STATEMENTS ARE
;           DISPLAYED.
;
;           ENTRY: 'STEP' ROUTINE IS ENTERED BY 'HALT/STEP
;                  KEY' BEING PRESSED DURING KEYBOARD
;                  ENTRY ('READCH' ROUTINE).
;

STEP:                            ;CRB = 1 (DISABLE INPUT)
                                 ;DELAY 10 MICRO-SECONDS
          CALL      RESCHK       ;ERROR IF PROGRAM NOT RESOLVED
          CALL      ENBLCO       ;ENABLE CONSOLE OUTPUT DEVICE
          CALL      PRNTCR       ;PRINT 'CR'
          MOV       EDX,CONTLN_  ;GET LINE NO. TO CONTINUE AT
          MOV       WORD PTR [R2],DX    ;F3,2 = LINE NO.
          CALL      FINDSTI      ;FIND LINE TO CONTINUE AT (GLOBAL OR LOCAL)
          TEST      R0,0FH
          JZ        STEP6        ;F, BRANCH
STEP3:    MOV       EDX,TEXTENDS ;FLUSH TEXT BUFFER
          MOV       TEXTEND,EDX
          JMP       ERRB_36      ;NF, ERROR
STEP6:    INC       EDX
          MOV       TEXTPTR,EDX  ;TEXTPTR POINT AFTER LINE NO.
          ADD       TEXTPTR,5
          CALL      WRT@         ;OUTPUT '@' IF GLOBAL TEXT LINE
          CALL      WRTLINE_     ;DISPLAY LINE NO.
          MOV       EDX,STMTCNT  ;GET CONTINUE STMT COUNT
          MOV       R7,DH        ;F7 = COUNT
          JMP       STEP2

STEP4:    CALL      SCNCHSS      ;SCAN TO NEXT STATEMENT ON LINE
          TEST      R0,0FH
          JNZ       STEP3        ;NF, ERROR
          MOV       K,':'
          CALL      TWRITE       ;PRINT ':'
          DEC       R7           ;DECREMENT COUNT
STEP2:    CMP       R7,00H
          JNE       STEP4        ;BRANCH IF COUNT NOT 0

          CALL      TWRITESP     ;OUTPUT SPACE

; ------------------------------------------------------------------------------
                                 ;IBASE logic added rel 3.0
          MOV       EDX,MODE
          MOV       EAX,DWORD PTR [CPBASE]  ;GET MODE OF IBASE PARTITION
          MOV       EBX,DWORD PTR [EAX].@IBASE
          CMP       EAX,EBX      ;CPBASE = IBASE  THEN USE CPBASE CURRENT MODE
          JE        STEP7
          MOV       EDX,DWORD PTR [EBX].@REGMODE   ;GET IBASE REGSAVE MODE
STEP7:    TEST      DL,40H
          JNZ       STEP5        ;BRANCH IF IO CLASS
; ------------------------------------------------------------------------------
          MOV       AUX00,TEXTPTR;AUX0 POINT TO POSITION IN LINE START PRINTING
          CALL      LISTLINS     ;DISPLAY REST OF TEXT LINE
STEP1:    CALL      RESCHK       ;ERROR IF PROGRAM NOT RESOLVED
          MOV       EDX,TEXTENDS
          MOV       TEXTEND,EDX      ;FLUSH I/O BUFFER
          AND       MODE,0FFFCH      ;EXECUTION MODE = 'RUN AT'
          CALL      ENBLNULL         ;DISABLE CO (IE, ENABLE NULL)
          AND       SL,0FFH-PHASE
          OR        SL,XPHASE        ;SET EXECUTION PHASE
          JMP       ENDSTM10

STEP5:    MOV       R7,06H
          CALL      ERRMSG
          JMP       STEP1


;     CLEAR: PROCESS A 'CLEAR' COMMAND
;
;          COMMAND FORMAT:
;               (1) CLEAR -- CLEAR ALL TEXT & VARIABLES
;               (2) CLEAR V -- CLEAR ALL VARIABLES
;               (3) CLEAR N -- CLEAR NONCOMMON VARIABLES
;               (4) CLEAR P [LINE-NO.] [,] [LINE-NO.]
;
;          ENTRY: TEXTPTR POINTS AFTER 'CLEAR'
;                 TEXLPTR POINTS TO CLEAR LINE
;
;

                                   ;Clear (program) now inhibits recall
CLEAR:    CALL      UNRESLVS       ;INIT. STACKS, CLEAR INTERRUPT TABLE
          MOV       EDX,TEXTPTR
          MOV       CH,BYTE PTR [TEXTPTR]
          INC       TEXTPTR        ;GET NEXT TEXT CHARACTER
          CMP       CH,0DH
          JNE       CLEAR0         ;BRANCH IF NOT CR
;   CLEAR  command
          AND       MODE,0FBFFH    ;turn off immediate mode bit (MODEIMM=0400)
          test      sl,2
          jz        clear_x
          mov       eax,dword ptr cpbase
          mov       byte ptr [eax].@type,0
clear_x:
          JMP       START          ;BRANCH IF 'CR'   (CLEAR)

CLEAR0:   CMP       CH,'P'
          JE        CLEAR1         ;BRANCH IF 'P'
          CMP       CH,'V'
          JE        CLEAR2         ;BRANCH IF 'V'
          CMP       CH,'N'
          JE        CLEAR3         ;BRANCH IF 'N'

; Error Syntax

          MOV       TEXTPTR,EDX     ;DECREMENT TEXTPTR
          JMP       ERRC_20

; CLEAR V  command

CLEAR2:                            ;LOOK FOR 'CR'
          CMP       BYTE PTR [TEXTPTR],0DH
          JNE       ERRC_20        ;BRANCH IF NOT MATCH
          INC       TEXTPTR
          CALL      INITVAR        ;CLEAR VARIABLES  (CLEAR V)
CLEAR5:   CALL      INITSTKS       ;INITIALIZE STACKS
          MOV       EAX,TEXTEND
          ADD       EAX,MINBUF+MINVS+0AH
          MOV       DWORD PTR [@@VSEND],EAX
          JMP       ENTERF         ;EXIT

;  CLEAR N command

CLEAR3:                            ;LOOK FOR 'CR'
          CMP       BYTE PTR [TEXTPTR],0DH
          JNE       ERRC_20        ;BRANCH IF NOT MATCH
          INC       TEXTPTR
          CALL      CLEARN         ;CLEAR NONCOMMON VARIABLES
          JMP       CLEAR5

;     CLEAR P command

CLEAR1:   CMP       BYTE PTR [TEXTPTR],0DH
          JNE       CLEAR4         ;NF, BRANCH
          INC       TEXTPTR
          AND       MODE,0FBFFH    ;turn off immediate mode bit (MODEIMM=0400)
          CALL      INITTEXT       ;INITIALIZE TEXT AREA
          MOV       EAX,TEXTEND
          ADD       EAX,MINBUF+MINVS+0AH
          MOV       DWORD PTR [@@VSEND],EAX
          JMP       ENTERF         ;EXIT

;  CLEAR P #### [,####]

CLEAR4:   CALL      LN_RANGE       ;LOOK FOR START & END LINE NUMBERS
          MOV       AX,WORD PTR [R6]  ;F3,2 = START LINE NO. FOR DELETE
          MOV       WORD PTR [R2],AX
          CMP       BYTE PTR [TEXTPTR],0DH
          JNE       ERRC_20        ;BRANCH IF NOT MATCH
          INC       TEXTPTR
          CALL      UNTHREAD       ;DELETE TEXT FROM LOW LINE NO. THRU HIGH LINE NO.
          MOV       EDX,DWORD PTR [CPBASE]
          MOV       DWORD PTR [EDX].@DBASE,EDX   ;@DBASE = CPBASE
          MOV       ECX,DWORD PTR [EDX].@TEXTBEG ;GET TEXTBEG POINTED TO BY CPBASE
          MOV       DWORD PTR [EDX].@DATAPTR,0   ;DATAPTR = 0
          MOV       DWORD PTR [EDX].@DATALPT,ECX ;DATALPT = TEXTBEG
          MOV       EAX,TEXTEND
          ADD       EAX,MINBUF+MINVS+0AH
          MOV       DWORD PTR [@@VSEND],EAX
          JMP       ENTERF         ;EXIT


;     RUN: PROCESS A 'RUN' COMMAND.
;          INITIALIZE VS & OS STACKS, DATA POINTERS,
;          STATEMENT COUNTS, END-OF-FILE INDICATOR. IF NO
;          LINE NO. SPECIFIED, CLEAR NONCOMMON VARIABLES.
;          INITIATE RESOLUTION PHASE.
;
;          ENTRY: TEXTPTR POINTS AFTER 'RUN'
;
;          COMMAND FORMAT:  RUN 'LINE-NO.'
;
;          EXITS: BRANCH TO VJUMP
;
;

RUN:     MOV       STMTCNT,0          ;STATEMENT COUNT, CONTINUE STMT COUNT = 0
         MOV       EDX,DWORD PTR [CPBASE]        ;GET CPBASE
         MOV       BYTE PTR [EDX].@ESCFLAG,0     ;ESC FLAG = 0
         MOV       BYTE PTR [EDX].@ERRF,0        ;ERR FUNCTION = 0
         MOV       DWORD PTR [EDX].@DBASE,EDX    ;@DBASE = CPBASE
         MOV       DWORD PTR [EDX].@DATAPTR,0    ;@DATAPTR = 0
         MOV       EAX,DWORD PTR [EDX].@TEXTBEG  ;GET TEXTBEG
         MOV       DWORD PTR [EDX].@DATALPT,EAX  ;DATALPTR = TEXTBEG

; ---- INITIALIZE STACKS (VS,OS) ----

         LEA       EAX,[EDX].@DVTBL
         MOV       DWORD PTR [EDX].@DVTPTR,EAX   ;DVTPTR = DVTBL BEGIN
         LEA       EAX,[EDX].@OSBEG
         MOV       OSPTR,EAX                     ;OSPTR = OSBEG BEGIN OF OS = 0
         MOV       BYTE PTR [EAX],0
         MOV       EAX,VARPTR
         MOV       VSPTR,EAX
         MOV       DWORD PTR [EDX].@SAVEVS,EAX   ;VSPTR ==> SAVEVS
         MOV       EAX,OSPTR
         MOV       DWORD PTR [EDX].@SAVEOS,EAX   ;OSPTR ==> SAVEOS
         MOV       DH,0FFH
         MOV       CONTLN_,EDX                   ;MAKE CONTINUE & STEP ILLEGAL

         AND       MODE,0FF7CH
         CALL      LINE_REF           ;LOOK FOR LINE NO.
         TEST      R0,0FH
         JNZ       RUN1               ;NF, BRANCH
         MOV       DX,WORD PTR [R2]
         MOV       CONTLN_,EDX        ;CONTLN# = LINE NO.
         CMP       BYTE PTR [TEXTPTR],','
         JNE       RUN3               ;NF, BRANCH
         INC       TEXTPTR
         CALL      INTEGER            ;LOOK FOR INTEGER
         TEST      R0,0FH
         JNZ       ERRC_34            ;NF, ERROR
         CMP       R2,0
         JNE       ERRC_34            ;ERROR IF > 255
         CMP       R1,0
         JE        ERRC_34            ;ERROR IF = 0
         MOV       EDX,STMTCNT
         MOV       DH,R1
         DEC       DH                 ;SET CONTINUE STMT COUNT
         MOV       STMTCNT,EDX
RUN3:    OR        MODE,0001H
         JMP       RUN2

RUN1:    CALL      CLEARN                      ;REMOVE NONCOMMON VARIABLES
                                               ;INITIALIZE STACKS (OS,VS,DVT)
         MOV       EDX,DWORD PTR [CPBASE]
         LEA       EAX,[EDX].@DVTBL
         MOV       DWORD PTR [EDX].@DVTPTR,EAX
         LEA       EAX,[EDX].@OSBEG
         MOV       OSPTR,EAX                    ;OSPTR = OSBEG
         MOV       BYTE PTR [EAX],0             ;BEGIN OF OS = 0
         MOV       EAX,VARPTR
         MOV       VSPTR,EAX
         MOV       DWORD PTR [EDX].@SAVEVS,EAX  ;VSPTR = SAVEVS
         MOV       EAX,OSPTR
         MOV       DWORD PTR [EDX].@SAVEOS,EAX  ;OSPTR = SAVEOS

                                                ;INITIALIZE INTERRUPT TABLE
         MOV       BYTE PTR [EDX].@ISTATUS,0    ;ISTATUS = 0
         MOV       BYTE PTR [EDX].@PRNTCNT,0    ;PRNTCNT = 0
         LEA       EDX,[EDX].@ITABLE            ;ITABLE = 0
         MOV       DWORD PTR [EDX],0
         MOV       DWORD PTR [EDX+4],0
         MOV       DWORD PTR [EDX+8],0

RUN2:    CMP       BYTE PTR [TEXTPTR],0DH
         JNE       ERRC_20
         INC       TEXTPTR
                                            ;Fetch boot and disconnect flags
RUNB:    MOV       EDX,0
         MOV       DL,PART_
         ADD       EDX,OFFSET PMODE
         MOV       AL,BYTE PTR [EDX]
         AND       BYTE PTR [EDX],0FFH-PMODEB-PMODED ;Turn off boot and disconnect flags
         TEST      AL,PMODED
         JZ        RUN5                    ;Branch if not processing disconnect
         CALL      ACKDISCT                ;Acknowledge disconnect
RUN5:    AND       SL,0FFH-PHASE           ;Mask off phase bits
         OR        SL,RPHASE               ;Turn on resolution phase

;Following logic added Rel 1.9 -- speed up of resolution of DEFFNX

         MOV       EDX,DWORD PTR [CPBASE]
         LEA       EDX,[EDX].@TEXTBEG
         MOV       ECX,DWORD PTR [EDX];point to begin of text
         MOV       R7,00H             ;search for 1st DEFFNx
         CALL      FINDFNC            ;search for DEFFN
         MOV       DWORD PTR [A0],0    ;F1F0 = 0
         CMP       CH,0
         JE        RUN9               ;branch if not found
         MOV       EDX,TEXTLPTR
         MOV       DWORD PTR [A0],EDX ;F1F0 = TEXTLPTR of DEFFNx
RUN9:    CALL      RDEFNADR
         MOV       EAX,DWORD PTR [A0]
         MOV       DWORD PTR [EDX],EAX ;SET DEFNADR PTR

;End of added logic

         MOV       EDX,DWORD PTR [CPBASE]
         MOV       ECX,DWORD PTR [EDX].@TEXTBEG   ;GET STARTING PROGRAM THREAD
         CMP       ECX,0
         JNE       RUN4               ;BRANCH IF THREAD <> 0

ERRB_35: MOV       R0,35H             ;NO PROGRAM LINES
         JMP       ERRORC

ERRC_34: MOV       R0,34H             ;ILLEGAL VALUE
         JMP       ERRORC

RUN4:    MOV       EDX,ECX
         MOV       TEXTLPTR,EDX
         ADD       TEXTLPTR,4
         XCHG      EDX,TEXTLPTR       ;TEXTLPTR POINT THREAD OF LINE
         MOV       TEXTPTR,EDX        ;TEXTPTR POINT TO AFTER LINE NO.
         ADD       TEXTPTR,2
         MOV       EDX,TEXTENDS       ;FLUSH RUN COMMAND
         MOV       TEXTEND,EDX
         MOV       EDX,DWORD PTR [CPBASE]  ;DEFFN' TABLE ENTRY COUNT = 0
         MOV       BYTE PTR [EDX].@DFN?QUTBL,0
         LEA       EAX,[EDX].@OSBEG
         ADD       EAX,64
         MOV       DWORD PTR [@@OSEND],EAX
         MOV       EAX,TEXTEND
         ADD       EAX,MINBUF+MINVS+0AH
         MOV       DWORD PTR [@@VSEND],EAX
         CALL      GR_THRDS           ;SET UP THREADS TO TEXT-LINE GROUPS
         MOV       EDX,DWORD PTR [CPBASE]
         LEA       EDX,[EDX].@ONERRPT
         MOV       DWORD PTR [EDX],0  ;initialize ON-ERR-GOTO PTR
         CALL      SLCT@OFF           ;IBASE = CPBASE
         MOV       EDX,DWORD PTR [CPBASE]
         MOV       BYTE PTR [EDX].@RESFLG,0     ;FLAG = DEFFN@ NOT YET ENCOUNTERED
                                                ;DO flag added in rel 3.0
         MOV       BYTE PTR [EDX].@DOFLAG,0     ;clear DOFLAG
         AND       WORD PTR [EDX].@FILE_,00FFH  ;ZERO FIGHEST FIELD# FOUND
         JMP       VJUMP

CODE     ENDS
         END
